/*
 * Copyright 2009, 2010
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h> // for snprintf
#include <string.h>

#include "array.h"
#include "pqlcontext.h"

#define DATATYPE_INLINE
#include "datatype.h"


struct column {
   char *name;			/* printable name, maybe user-supplied */
   unsigned id;			/* unique ID number */
   unsigned refcount;		/* usage count */
   unsigned typerefcount;	/* additional usage count from types */
};

enum datatyperepresentations {
   DT_BASE,			/* one of the base types in the manager */
   DT_SET,			/* set(MEMBER) */
   DT_SEQUENCE,			/* sequence(MEMBER) */
   DT_TUPLE,			/* (T1, T2, ...), arity at least 2 */
};
struct datatype {
   struct datatypemanager *dtm;
   char *name;

   enum datatyperepresentations rep;
   union {
      /* base - nothing */
      struct {
	 struct datatype *member;
      } set;
      struct {
	 struct datatype *member;
      } sequence;
      struct {
	 struct datatypearray members;
	 struct columnarray headings;
      } tuple;
   };

   /* Backreferences to types derived from this type T */
   struct datatype *set_of;		/* set(T) */
   struct datatype *sequence_of;	/* sequence(T) */
   struct datatypearray tuples;		/* all (T, RIGHT) */
};

struct datatypemanager {
   struct pqlcontext *pql;

   struct datatype type_unit;
   struct datatype type_bool;
   struct datatype type_int;
   struct datatype type_double;
   struct datatype type_string;
   struct datatype type_struct;
   struct datatype type_pathelement;
   struct datatype type_distinguisher;

   struct datatype type_absdbedge;
   struct datatype type_absnumber;
   struct datatype type_absatom;
   struct datatype type_absdbobj;
   struct datatype type_abstop;
   struct datatype type_absbottom;
};

////////////////////////////////////////////////////////////

struct column *column_create(struct pqlcontext *pql, const char *name) {
   struct column *col;

   col = domalloc(pql, sizeof(*col));

   /* XXX: should we share the columnid if copying a columnvar? */
   /* XXX: and if not should we be sharing the same ID counter? */
   col->id = pql->nextcolumnid++;
   col->refcount = 1;
   col->typerefcount = 0;

   if (name != NULL) {
      col->name = dostrdup(pql, name);
   }
   else {
      // XXX this needs to match what's in tcalc.c for the dumps
      char buf[32];
      snprintf(buf, sizeof(buf), ".C%u", col->id);
      col->name = dostrdup(pql, buf);
   }

   return col;
}

void column_incref(struct column *col) {
   // XXX temporary (or maybe not...)
   if (col == NULL) {
      return;
   }

   PQLASSERT(col != NULL);
   col->refcount++;
}

static void column_dofree(struct pqlcontext *pql, struct column *col) {
   dostrfree(pql, col->name);
   dofree(pql, col, sizeof(*col));
}

void column_decref(struct pqlcontext *pql, struct column *col) {
   // XXX temporary (or maybe not...)
   if (col == NULL) {
      return;
   }

   PQLASSERT(col != NULL);

   /* this will fail if COL points at free memory that's been deadbeef'd */
   PQLASSERT(col->id < pql->nextcolumnid);

   PQLASSERT(col->refcount > 0);

   col->refcount--;
   if (col->refcount + col->typerefcount == 0) {
      column_dofree(pql, col);
   }
}

static void column_typeincref(struct column *col) {
   // XXX temporary (or maybe not...)
   if (col == NULL) {
      return;
   }

   PQLASSERT(col != NULL);
   col->typerefcount++;
}

static void column_typedecref(struct pqlcontext *pql, struct column *col) {
   // XXX temporary (or maybe not...)
   if (col == NULL) {
      return;
   }

   PQLASSERT(col != NULL);

   /* this will fail if COL points at free memory that's been deadbeef'd */
   PQLASSERT(col->id < pql->nextcolumnid);

   PQLASSERT(col->typerefcount > 0);

   col->typerefcount--;
   if (col->refcount + col->typerefcount == 0) {
      column_dofree(pql, col);
   }
}

const char *column_getname(struct column *col) {
   PQLASSERT(col != NULL);
   return col->name;
}

static const char *column_getnamex(struct column *col) {
   if (col == NULL) {
      return "(NULL)";
   }
   return col->name;
}

static bool columnarray_contains(const struct columnarray *cs,
				 struct column *c) {
   unsigned i, num;

   num = columnarray_num(cs);
   for (i=0; i<num; i++) {
      if (columnarray_get(cs, i) == c) {
	 return true;
      }
   }
   return false;
}

////////////////////////////////////////////////////////////

static void datatype_destroy(struct datatype *t);

static void datatype_init(struct datatypemanager *dtm,
			  struct datatype *t,
			  const char *name,
			  enum datatyperepresentations rep) {
   t->dtm = dtm;
   t->name = dostrdup(dtm->pql, name);
   t->rep = rep;
   t->set_of = t->sequence_of = NULL;
   datatypearray_init(&t->tuples);
}

static void datatype_cleanup(struct datatype *t) {
   unsigned i, num;

   dostrfree(t->dtm->pql, t->name);
   t->name = NULL;

   switch (t->rep) {
    case DT_BASE:
    case DT_SET:
    case DT_SEQUENCE:
     /* nothing */
     break;
    case DT_TUPLE:
     datatypearray_setsize(t->dtm->pql, &t->tuple.members, 0);
     datatypearray_cleanup(t->dtm->pql, &t->tuple.members);
     {
	unsigned j, num;

	num = columnarray_num(&t->tuple.headings);
	for (j=0; j<num; j++) {
	   column_typedecref(t->dtm->pql,
			     columnarray_get(&t->tuple.headings, j));
	}
     }
     columnarray_setsize(t->dtm->pql, &t->tuple.headings, 0);
     columnarray_cleanup(t->dtm->pql, &t->tuple.headings);
     break;
   }

   if (t->set_of != NULL) {
      datatype_destroy(t->set_of);
      t->set_of = NULL;
   }

   if (t->sequence_of != NULL) {
      datatype_destroy(t->sequence_of);
      t->sequence_of = NULL;
   }

   num = datatypearray_num(&t->tuples);
   for (i=0; i<num; i++) {
      datatype_destroy(datatypearray_get(&t->tuples, i));
   }
   datatypearray_setsize(t->dtm->pql, &t->tuples, 0);
   datatypearray_cleanup(t->dtm->pql, &t->tuples);
}

static struct datatype *datatype_create(struct pqlcontext *pql,
					const char *name,
					enum datatyperepresentations rep) {
   struct datatype *t;

   t = domalloc(pql, sizeof(*t));
   datatype_init(pql->dtm, t, name, rep);
   return t;
}

static void datatype_destroy(struct datatype *t) {
   datatype_cleanup(t);
   dofree(t->dtm->pql, t, sizeof(*t));
}

////////////////////////////////////////////////////////////

struct datatypemanager *datatypemanager_create(struct pqlcontext *pql) {
   struct datatypemanager *dtm;

   dtm = domalloc(pql, sizeof(*dtm));

   dtm->pql = pql;

   datatype_init(dtm, &dtm->type_unit, "unit", DT_BASE);
   datatype_init(dtm, &dtm->type_bool, "bool", DT_BASE);
   datatype_init(dtm, &dtm->type_int, "int", DT_BASE);
   datatype_init(dtm, &dtm->type_double, "double", DT_BASE);
   datatype_init(dtm, &dtm->type_string, "string", DT_BASE);
   datatype_init(dtm, &dtm->type_struct, "struct", DT_BASE);
   datatype_init(dtm, &dtm->type_pathelement, "pathelement", DT_BASE);
   datatype_init(dtm, &dtm->type_distinguisher, "distinguisher", DT_BASE);

   datatype_init(dtm, &dtm->type_absdbedge, "dbedge", DT_BASE);
   datatype_init(dtm, &dtm->type_absnumber, "number", DT_BASE);
   datatype_init(dtm, &dtm->type_absatom, "atom", DT_BASE);
   datatype_init(dtm, &dtm->type_absdbobj, "dbobj", DT_BASE);
   datatype_init(dtm, &dtm->type_abstop, "top", DT_BASE);
   datatype_init(dtm, &dtm->type_absbottom, "bottom", DT_BASE);

   return dtm;
}

void datatypemanager_destroy(struct datatypemanager *dtm) {

   datatype_cleanup(&dtm->type_unit);
   datatype_cleanup(&dtm->type_bool);
   datatype_cleanup(&dtm->type_int);
   datatype_cleanup(&dtm->type_double);
   datatype_cleanup(&dtm->type_string);
   datatype_cleanup(&dtm->type_struct);
   datatype_cleanup(&dtm->type_pathelement);
   datatype_cleanup(&dtm->type_distinguisher);

   datatype_cleanup(&dtm->type_absdbedge);
   datatype_cleanup(&dtm->type_absnumber);
   datatype_cleanup(&dtm->type_absatom);
   datatype_cleanup(&dtm->type_absdbobj);
   datatype_cleanup(&dtm->type_abstop);
   datatype_cleanup(&dtm->type_absbottom);

   dofree(dtm->pql, dtm, sizeof(*dtm));
}

////////////////////////////////////////////////////////////

/*
 * Base types (kind "*")
 */

/* concrete */

struct datatype *datatype_unit(struct pqlcontext *pql) {
   return &pql->dtm->type_unit;
}

struct datatype *datatype_bool(struct pqlcontext *pql) {
   return &pql->dtm->type_bool;
}

struct datatype *datatype_int(struct pqlcontext *pql) {
   return &pql->dtm->type_int;
}

struct datatype *datatype_double(struct pqlcontext *pql) {
   return &pql->dtm->type_double;
}

struct datatype *datatype_string(struct pqlcontext *pql) {
   return &pql->dtm->type_string;
}

struct datatype *datatype_struct(struct pqlcontext *pql) {
   return &pql->dtm->type_struct;
}

struct datatype *datatype_pathelement(struct pqlcontext *pql) {
   return &pql->dtm->type_pathelement;
}

struct datatype *datatype_distinguisher(struct pqlcontext *pql) {
   return &pql->dtm->type_distinguisher;
}

/* abstract */

struct datatype *datatype_absdbedge(struct pqlcontext *pql) {
   return &pql->dtm->type_absdbedge;
}

struct datatype *datatype_absnumber(struct pqlcontext *pql) {
   return &pql->dtm->type_absnumber;
}

struct datatype *datatype_absatom(struct pqlcontext *pql) {
   return &pql->dtm->type_absatom;
}

struct datatype *datatype_absdbobj(struct pqlcontext *pql) {
   return &pql->dtm->type_absdbobj;
}

struct datatype *datatype_abstop(struct pqlcontext *pql) {
   return &pql->dtm->type_abstop;
}

struct datatype *datatype_absbottom(struct pqlcontext *pql) {
   return &pql->dtm->type_absbottom;
}

// XXX these two should be removed
struct datatype *datatype_dbobj(struct pqlcontext *pql) {
   return datatype_absdbobj(pql);
}

struct datatype *datatype_dbedge(struct pqlcontext *pql) {
   return datatype_absdbedge(pql);
}

////////////////////////////////////////////////////////////

/*
 * Collection types (kind "* -> *")
 */

struct datatype *datatype_set(struct pqlcontext *pql, struct datatype *t) {
   PQLASSERT(t->dtm == pql->dtm);
   if (t->set_of == NULL) {
      char buf[1024];

      snprintf(buf, sizeof(buf), "set(%s)", t->name);
      t->set_of = datatype_create(pql, buf, DT_SET);
      t->set_of->set.member = t;
   }
   return t->set_of;
}

struct datatype *datatype_sequence(struct pqlcontext *pql, struct datatype *t){
   PQLASSERT(t->dtm == pql->dtm);
   if (t->sequence_of == NULL) {
      char buf[1024];

      snprintf(buf, sizeof(buf), "seq(%s)", t->name);
      t->sequence_of = datatype_create(pql, buf, DT_SEQUENCE);
      t->sequence_of->sequence.member = t;
   }
   return t->sequence_of;
}

////////////////////////////////////////////////////////////

/*
 * Tuples
 *
 * Tuples are encoded as arrays of T (kind "(*, *, ...) -> *").
 */

/*
 * Create the monople type (t0) from t0.
 */
struct datatype *datatype_tuple_monople(struct pqlcontext *pql,
					struct datatype *t0,
					struct column *c0) {
   unsigned i, numtypes, arity;
   struct datatype *t;
   char buf[1024];

   (void)c0;

   /* must all belong together */
   PQLASSERT(t0->dtm == pql->dtm);

   /* currently, tuples containing tuples are not allowed */
   PQLASSERT(t0->rep != DT_TUPLE);

   /* check if the type already exists */
   numtypes = datatypearray_num(&t0->tuples);
   for (i=0; i<numtypes; i++) {
      t = datatypearray_get(&t0->tuples, i);
      PQLASSERT(t->rep == DT_TUPLE);
      arity = datatypearray_num(&t->tuple.members);
      PQLASSERT(arity >= 1);
      PQLASSERT(datatypearray_get(&t->tuple.members, 0) == t0);

      if (arity == 1 && columnarray_get(&t->tuple.headings, 0) == c0) {
	 /* found it */
	 return t;
      }
   }

   /* not found, make one */
   snprintf(buf, sizeof(buf), "(%s : %s)", t0->name, column_getnamex(c0));
   t = datatype_create(pql, buf, DT_TUPLE);
   datatypearray_init(&t->tuple.members);
   datatypearray_setsize(pql, &t->tuple.members, 1);
   datatypearray_set(&t->tuple.members, 0, t0);
   column_typeincref(c0);
   columnarray_init(&t->tuple.headings);
   columnarray_setsize(pql, &t->tuple.headings, 1);
   columnarray_set(&t->tuple.headings, 0, c0);

   /* store it */
   datatypearray_add(t0->dtm->pql, &t0->tuples, t, NULL);

   return t;
}

/*
 * Create the pair (t0, t1) from t0 and t1.
 */
struct datatype *datatype_tuple_pair(struct pqlcontext *pql,
				     struct datatype *t0,
				     struct datatype *t1,
				     struct column *c0,
				     struct column *c1) {
   unsigned i, numtypes, arity;
   struct datatype *t;
   char buf[1024];

   /* must all belong together */
   PQLASSERT(t0->dtm == pql->dtm);
   PQLASSERT(t1->dtm == pql->dtm);

   /* currently, tuples containing tuples are not allowed */
   PQLASSERT(t0->rep != DT_TUPLE);
   PQLASSERT(t1->rep != DT_TUPLE);

   /* check if the type already exists */
   numtypes = datatypearray_num(&t0->tuples);
   for (i=0; i<numtypes; i++) {
      t = datatypearray_get(&t0->tuples, i);
      PQLASSERT(t->rep == DT_TUPLE);
      arity = datatypearray_num(&t->tuple.members);
      PQLASSERT(arity >= 1);
      PQLASSERT(datatypearray_get(&t->tuple.members, 0) == t0);

      if (arity == 2 && datatypearray_get(&t->tuple.members, 1) == t1 &&
	  columnarray_get(&t->tuple.headings, 0) == c0 &&
	  columnarray_get(&t->tuple.headings, 1) == c1) {
	 /* found it */
	 return t;
      }
   }

   /* not found, make one */
   snprintf(buf, sizeof(buf), "(%s : %s, %s : %s)",
	    t0->name, column_getnamex(c0),
	    t1->name, column_getnamex(c1));
   t = datatype_create(pql, buf, DT_TUPLE);
   datatypearray_init(&t->tuple.members);
   datatypearray_setsize(pql, &t->tuple.members, 2);
   datatypearray_set(&t->tuple.members, 0, t0);
   datatypearray_set(&t->tuple.members, 1, t1);
   column_typeincref(c0);
   column_typeincref(c1);
   columnarray_init(&t->tuple.headings);
   columnarray_setsize(pql, &t->tuple.headings, 2);
   columnarray_set(&t->tuple.headings, 0, c0);
   columnarray_set(&t->tuple.headings, 1, c1);

   /* store it */
   datatypearray_add(t0->dtm->pql, &t0->tuples, t, NULL);

   return t;
}

/*
 * Create (t0, t1, t2).
 * XXX too much code duplication
 */
struct datatype *datatype_tuple_triple(struct pqlcontext *pql,
				       struct datatype *t0,
				       struct datatype *t1,
				       struct datatype *t2,
				       struct column *c0,
				       struct column *c1,
				       struct column *c2) {
   unsigned i, numtypes, arity;
   struct datatype *t;
   char buf[1024];

   (void)c0;
   (void)c1;
   (void)c2;

   /* must all belong together */
   PQLASSERT(t0->dtm == pql->dtm);
   PQLASSERT(t1->dtm == pql->dtm);
   PQLASSERT(t2->dtm == pql->dtm);

   /* currently, tuples containing tuples are not allowed */
   PQLASSERT(t0->rep != DT_TUPLE);
   PQLASSERT(t1->rep != DT_TUPLE);
   PQLASSERT(t2->rep != DT_TUPLE);

   /* check if the type already exists */
   numtypes = datatypearray_num(&t0->tuples);
   for (i=0; i<numtypes; i++) {
      t = datatypearray_get(&t0->tuples, i);
      PQLASSERT(t->rep == DT_TUPLE);
      arity = datatypearray_num(&t->tuple.members);
      PQLASSERT(arity >= 1);
      PQLASSERT(datatypearray_get(&t->tuple.members, 0) == t0);

      if (arity == 3
	  && datatypearray_get(&t->tuple.members, 1) == t1
	  && datatypearray_get(&t->tuple.members, 2) == t2
	  && columnarray_get(&t->tuple.headings, 0) == c0
	  && columnarray_get(&t->tuple.headings, 1) == c1
	  && columnarray_get(&t->tuple.headings, 2) == c2) {
	 /* found it */
	 return t;
      }
   }

   /* not found, make one */
   snprintf(buf, sizeof(buf), "(%s : %s, %s : %s, %s : %s)", 
	    t0->name, column_getnamex(c0),
	    t1->name, column_getnamex(c1),
	    t2->name, column_getnamex(c2));
   t = datatype_create(pql, buf, DT_TUPLE);
   datatypearray_init(&t->tuple.members);
   datatypearray_setsize(pql, &t->tuple.members, 3);
   datatypearray_set(&t->tuple.members, 0, t0);
   datatypearray_set(&t->tuple.members, 1, t1);
   datatypearray_set(&t->tuple.members, 2, t2);
   column_typeincref(c0);
   column_typeincref(c1);
   column_typeincref(c2);
   columnarray_init(&t->tuple.headings);
   columnarray_setsize(pql, &t->tuple.headings, 3);
   columnarray_set(&t->tuple.headings, 0, c0);
   columnarray_set(&t->tuple.headings, 1, c1);
   columnarray_set(&t->tuple.headings, 2, c2);

   /* store it */
   datatypearray_add(t0->dtm->pql, &t0->tuples, t, NULL);

   return t;
}

/*
 * Create (t0, t1, ..., tn) where ti are given by members[].
 */
struct datatype *datatype_tuple_specific(struct pqlcontext *pql,
					 const struct datatypearray *members,
					 const struct columnarray *headings) {
   unsigned i, j, numtypes, arity, newarity;
   struct datatype *t, *t0;
   struct column *col;
   char buf[1024];

   newarity = datatypearray_num(members);

   if (newarity == 0) {
      return datatype_unit(pql);
   }

   /* check if it already exists */
   t0 = datatypearray_get(members, 0);
   numtypes = datatypearray_num(&t0->tuples);
   for (i=0; i<numtypes; i++) {
      t = datatypearray_get(&t0->tuples, i);
      PQLASSERT(t->rep == DT_TUPLE);
      arity = datatypearray_num(&t->tuple.members);
      PQLASSERT(arity >= 1);
      PQLASSERT(datatypearray_get(&t->tuple.members, 0) == t0);

      if (arity != newarity) {
	 continue;
      }
      for (j=0; j<arity; j++) {
	 if (datatypearray_get(members, j) !=
	     datatypearray_get(&t->tuple.members, j)) {
	    break;
	 }
	 if (headings != NULL &&
	     columnarray_get(headings, j) !=
	     columnarray_get(&t->tuple.headings, j)) {
	    break;
	 }
      }
      if (j == arity) {
	 /* found it */
	 return t;
      }
   }

   /* create it */
   strcpy(buf, "(");
   for (j=0; j<newarity; j++) {
      if (j>0) {
	 snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), ", ");
      }
      snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), "%s : %s",
	       datatypearray_get(members, j)->name,
	       headings ? columnarray_get(headings, j)->name : "(NULL)");
   }
   snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), ")");

   t = datatype_create(pql, buf, DT_TUPLE);
   datatypearray_init(&t->tuple.members);
   datatypearray_setsize(pql, &t->tuple.members, newarity);
   columnarray_init(&t->tuple.headings);
   columnarray_setsize(pql, &t->tuple.headings, newarity);
   for (j=0; j<newarity; j++) {
      datatypearray_set(&t->tuple.members, j, datatypearray_get(members, j));
      if (headings != NULL) {
	 col = columnarray_get(headings, j);
	 column_typeincref(col);
      }
      else {
	 col = NULL;
      }
      columnarray_set(&t->tuple.headings, j, col);
   }

   /* store it */
   datatypearray_add(t0->dtm->pql, &t0->tuples, t, NULL);

   return t;
}

/*
 * Given t0 = (t00, t01, ..., t0n), t1 = (t10, t11, ..., t1m)
 * create (t00, ..., t0n, t10, ..., t1m).
 */
struct datatype *datatype_tuple_concat(struct pqlcontext *pql,
				       struct datatype *t0,
				       struct datatype *t1) {
   unsigned i, j, numtypes, arity, t0arity, t1arity;
   struct datatype *t, *t00;
   struct column *col;
   char buf[1024];

   /* must all belong together */
   PQLASSERT(t0->dtm == pql->dtm);
   PQLASSERT(t1->dtm == pql->dtm);

   PQLASSERT(datatype_istuple(t0));
   PQLASSERT(datatype_istuple(t1));

   if (t0 == datatype_unit(pql)) {
      return t1;
   }
   if (t1 == datatype_unit(pql)) {
      return t0;
   }

   PQLASSERT(t0->rep == DT_TUPLE);
   PQLASSERT(t1->rep == DT_TUPLE);

   /* size of result */
   t0arity = datatypearray_num(&t0->tuple.members);
   t1arity = datatypearray_num(&t1->tuple.members);

   PQLASSERT(t0arity >= 1);
   PQLASSERT(t1arity >= 1);

   /* check if the type already exists */
   t00 = datatypearray_get(&t0->tuple.members,  0);
   numtypes = datatypearray_num(&t00->tuples);
   for (i=0; i<numtypes; i++) {
      t = datatypearray_get(&t00->tuples, i);
      PQLASSERT(t->rep == DT_TUPLE);
      arity = datatypearray_num(&t->tuple.members);
      PQLASSERT(arity >= 1);
      PQLASSERT(datatypearray_get(&t->tuple.members, 0) == t00);

      if (arity != t0arity + t1arity) {
	 continue;
      }
      if (columnarray_get(&t0->tuple.headings, 0) !=
	  columnarray_get(&t->tuple.headings, 0)) {
	 continue;
      }
      for (j=1; j<t0arity; j++) {
	 if (datatypearray_get(&t0->tuple.members, j) !=
	     datatypearray_get(&t->tuple.members, j)) {
	    break;
	 }
	 if (columnarray_get(&t0->tuple.headings, j) !=
	     columnarray_get(&t->tuple.headings, j)) {
	    break;
	 }
      }
      if (j != t0arity) {
	 continue;
      }
      for (j=0; j<t1arity; j++) {
	 if (datatypearray_get(&t1->tuple.members, j) !=
	     datatypearray_get(&t->tuple.members, t0arity + j)) {
	    break;
	 }
	 if (columnarray_get(&t1->tuple.headings, j) !=
	     columnarray_get(&t->tuple.headings, t0arity + j)) {
	    break;
	 }
      }
      if (j != t1arity) {
	 continue;
      }
      /* found it */
      return t;
   }

   /* not found; make it */
   strcpy(buf, "(");
   for (j=0; j<t0arity; j++) {
      if (j>0) {
	 snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), ", ");
      }
      snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), "%s : %s",
	       datatypearray_get(&t0->tuple.members, j)->name,
	       column_getnamex(columnarray_get(&t0->tuple.headings, j)));
   }
   for (j=0; j<t1arity; j++) {
      snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), ", %s : %s",
	       datatypearray_get(&t1->tuple.members, j)->name,
	       column_getnamex(columnarray_get(&t1->tuple.headings, j)));
   }
   snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), ")");

   t = datatype_create(pql, buf, DT_TUPLE);
   datatypearray_init(&t->tuple.members);
   datatypearray_setsize(pql, &t->tuple.members, t0arity + t1arity);
   columnarray_init(&t->tuple.headings);
   columnarray_setsize(pql, &t->tuple.headings, t0arity + t1arity);
   for (j=0; j<t0arity; j++) {
      datatypearray_set(&t->tuple.members, j,
			datatypearray_get(&t0->tuple.members, j));
      col = columnarray_get(&t0->tuple.headings, j);
      column_typeincref(col);
      columnarray_set(&t->tuple.headings, j, col);
   }
   for (j=0; j<t1arity; j++) {
      datatypearray_set(&t->tuple.members, t0arity + j,
			datatypearray_get(&t1->tuple.members, j));
      col = columnarray_get(&t1->tuple.headings, j);
      column_typeincref(col);
      columnarray_set(&t->tuple.headings, t0arity + j, col);
   }

   /* store it */
   datatypearray_add(t00->dtm->pql, &t00->tuples, t, NULL);

   return t;
}

/*
 * Given t0 = (t00, t01, ..., t0n) and t1,
 * create (t00, ..., t0n, t1).
 */
struct datatype *datatype_tuple_append(struct pqlcontext *pql,
				       struct datatype *t0,
				       struct datatype *t1,
				       struct column *c1) {
   unsigned i, j, numtypes, arity, t0arity;
   struct datatype *t, *t00;
   struct column *col;
   char buf[1024];
   size_t len;

   PQLASSERT(t0->dtm == pql->dtm);
   PQLASSERT(t1->dtm == pql->dtm);

   /* treat t1 as a non-tuple and cons it onto the end as column c1 */

   /*
    * Tuples containing tuples are not allowed. t0 must be a tuple,
    * possibly unit, and t1 must not be.
    */

   PQLASSERT(t0->rep == DT_TUPLE || t0 == datatype_unit(pql));
   PQLASSERT(t1->rep != DT_TUPLE && t1 != datatype_unit(pql));

   if (t0 == datatype_unit(pql)) {
      return datatype_tuple_monople(pql, t1, c1);
   }

   t0arity = datatypearray_num(&t0->tuple.members);

   /* check if the type already exists */
   t00 = datatypearray_get(&t0->tuple.members, 0);
   numtypes = datatypearray_num(&t00->tuples);
   for (i=0; i<numtypes; i++) {
      t = datatypearray_get(&t00->tuples, i);
      PQLASSERT(t->rep == DT_TUPLE);
      arity = datatypearray_num(&t->tuple.members);
      PQLASSERT(arity >= 1);
      PQLASSERT(datatypearray_get(&t->tuple.members, 0) == t00);

      if (arity != t0arity+1) {
	 continue;
      }
      if (columnarray_get(&t0->tuple.headings, 0) !=
	  columnarray_get(&t->tuple.headings, 0)) {
	 continue;
      }
      for (j=1; j<t0arity; j++) {
	 if (datatypearray_get(&t0->tuple.members, j) !=
	     datatypearray_get(&t->tuple.members, j)) {
	    break;
	 }
	 if (columnarray_get(&t0->tuple.headings, j) !=
	     columnarray_get(&t->tuple.headings, j)) {
	    break;
	 }
      }
      if (j != t0arity) {
	 continue;
      }
      if (datatypearray_get(&t->tuple.members, t0arity) != t1) {
	 continue;
      }
      if (columnarray_get(&t->tuple.headings, t0arity) != c1) {
	 continue;
      }
      /* found it */
      return t;
   }

   /* not found, make one */

   /* extend the existing parens in t0's name */
   len = strlen(t0->name);
   PQLASSERT(len > 2);
   PQLASSERT(t0->name[0] == '(');
   PQLASSERT(t0->name[len-1] == ')');
   snprintf(buf, sizeof(buf), "(%.*s, %s : %s)", (int)(len-2), t0->name+1,
	    t1->name, column_getnamex(c1));

   t = datatype_create(pql, buf, DT_TUPLE);
   datatypearray_init(&t->tuple.members);
   datatypearray_setsize(pql, &t->tuple.members, t0arity+1);
   columnarray_init(&t->tuple.headings);
   columnarray_setsize(pql, &t->tuple.headings, t0arity+1);
   for (j=0; j<t0arity; j++) {
      datatypearray_set(&t->tuple.members, j,
			datatypearray_get(&t0->tuple.members, j));
      col = columnarray_get(&t0->tuple.headings, j);
      column_typeincref(col);
      columnarray_set(&t->tuple.headings, j, col);
   }
   datatypearray_set(&t->tuple.members, t0arity, t1);
   column_typeincref(c1);
   columnarray_set(&t->tuple.headings, t0arity, c1);

   /* store it */
   datatypearray_add(t00->dtm->pql, &t00->tuples, t, NULL);

   return t;
}

struct datatype *datatype_tuple_strip(struct pqlcontext *pql,
				      struct datatype *t0, unsigned ix) {
   unsigned i, j, numtypes, arity, t0arity;
   struct datatype *t, *t00;
   struct column *col;
   char buf[1024];

   PQLASSERT(t0->dtm == pql->dtm);

   t0arity = datatype_arity(t0);
   PQLASSERT(ix < t0arity);

   if (t0arity == 1) {
      return datatype_unit(pql);
   }

   PQLASSERT(t0->rep == DT_TUPLE);

   /* check if the type already exists */
   t00 = datatypearray_get(&t0->tuple.members, ix == 0 ? 1 : 0);
   numtypes = datatypearray_num(&t00->tuples);
   for (i=0; i<numtypes; i++) {
      t = datatypearray_get(&t00->tuples, i);
      PQLASSERT(t->rep == DT_TUPLE);
      arity = datatypearray_num(&t->tuple.members);
      PQLASSERT(arity >= 1);
      PQLASSERT(datatypearray_get(&t->tuple.members, 0) == t00);

      if (arity != t0arity-1) {
	 continue;
      }
      for (j=0; j<ix; j++) {
	 if (datatypearray_get(&t0->tuple.members, j) !=
	     datatypearray_get(&t->tuple.members, j)) {
	    break;
	 }
	 if (columnarray_get(&t0->tuple.headings, j) !=
	     columnarray_get(&t->tuple.headings, j)) {
	    break;
	 }
      }
      if (j != ix) {
	 continue;
      }
      for (j=ix+1; j<t0arity; j++) {
	 if (datatypearray_get(&t0->tuple.members, j) !=
	     datatypearray_get(&t->tuple.members, j-1)) {
	    break;
	 }
	 if (columnarray_get(&t0->tuple.headings, j) !=
	     columnarray_get(&t->tuple.headings, j-1)) {
	    break;
	 }
      }
      if (j != t0arity) {
	 continue;
      }
      /* found it */
      return t;
   }

   /* not found, make it */
   strcpy(buf, "(");
   for (j=0; j<ix; j++) {
      if (j>0) {
	 snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), ", ");
      }
      snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), "%s : %s",
	       datatypearray_get(&t0->tuple.members, j)->name,
	       column_getnamex(columnarray_get(&t0->tuple.headings, j)));
   }
   for (j=ix+1; j<t0arity; j++) {
      if (j>1) {
	 snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), ", ");
      }
      snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), "%s : %s",
	       datatypearray_get(&t0->tuple.members, j)->name,
	       column_getnamex(columnarray_get(&t0->tuple.headings, j)));
   }
   snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), ")");

   t = datatype_create(pql, buf, DT_TUPLE);
   datatypearray_init(&t->tuple.members);
   datatypearray_setsize(pql, &t->tuple.members, t0arity-1);
   columnarray_init(&t->tuple.headings);
   columnarray_setsize(pql, &t->tuple.headings, t0arity-1);
   for (j=0; j<ix; j++) {
      datatypearray_set(&t->tuple.members, j,
			datatypearray_get(&t0->tuple.members, j));
      col = columnarray_get(&t0->tuple.headings, j);
      column_typeincref(col);
      columnarray_set(&t->tuple.headings, j, col);
   }
   for (j=ix+1; j<t0arity; j++) {
      datatypearray_set(&t->tuple.members, j-1,
			datatypearray_get(&t0->tuple.members, j));
      col = columnarray_get(&t0->tuple.headings, j);
      column_typeincref(col);
      columnarray_set(&t->tuple.headings, j-1, col);
   }

   /* store it */
   datatypearray_add(t00->dtm->pql, &t00->tuples, t, NULL);

   return t;
}

struct datatype *datatype_tupleset_strip(struct pqlcontext *pql,
					 struct datatype *t,
					 unsigned ix) {
   struct datatype *t2;
   bool isset = false, issequence = false;

   if (datatype_isset(t)) {
      isset = true;
      t = datatype_set_member(t);
   }
   else if (datatype_issequence(t)) {
      issequence = true;
      t = datatype_sequence_member(t);
   }

   t2 = datatype_tuple_strip(pql, t, ix);

   if (issequence) {
      t2 = datatype_sequence(pql, t2);
   }
   else if (isset) {
      t2 = datatype_set(pql, t2);
   }

   return t2;
}

////////////////////////////////////////////////////////////

bool datatype_eq_nonames(const struct datatype *t0,
			 const struct datatype *t1) {
   if (t0 == t1) {
      return true;
   }

   /* t0 and t1 can now only be the same if they are or contain tuples */

   if (t0->rep != t1->rep) {
      return false;
   }
   switch (t0->rep) {
    case DT_BASE:
     return false;
    case DT_SET:
     return datatype_eq_nonames(t0->set.member, t1->set.member);
    case DT_SEQUENCE:
     return datatype_eq_nonames(t0->sequence.member, t1->sequence.member);
    case DT_TUPLE:
     {
	unsigned i, num;

	num = datatypearray_num(&t0->tuple.members);
	if (num != datatypearray_num(&t1->tuple.members)) {
	   return false;
	}
	for (i=0; i<num; i++) {
	   if (!datatype_eq_nonames(datatypearray_get(&t0->tuple.members, i),
				    datatypearray_get(&t1->tuple.members, i))){
	      return false;
	   }
	}
	return true;
     }
     break;
   }
   PQLASSERT(0);
   return false;
}

bool datatype_eq_withnames(const struct datatype *t0,
			   const struct datatype *t1) {
   return t0 == t1;
}

/*
 * Check structure (set organization, tuple arity) and column names
 * only, considering all scalar types equal.
 *
 * This is used for checking matching column structures in a few
 * places.  It should probably go away in favor of a better column
 * name strategy in generalize/specialize. I think the best approach
 * is to make those fail if the column names don't match precisely;
 * however, this requires tracking column names in values, which we're
 * not ready for yet. XXX: future.
 */
bool datatype_eq_namesonly(const struct datatype *t0,
			   const struct datatype *t1) {
   if (t0->rep != t1->rep) {

      /* XXX this should NOT be a special case */
      if (datatype_isabsbottom(t0) &&
	  datatype_issequence(t1) &&
	  datatype_ispathelement(datatype_sequence_member(t1))) {
	 return true;
      }
      if (datatype_isabsbottom(t1) &&
	  datatype_issequence(t0) &&
	  datatype_ispathelement(datatype_sequence_member(t0))) {
	 return true;
      }

      return false;
   }

   switch (t0->rep) {
    case DT_BASE:
     return true;
    case DT_SET:
     return datatype_eq_namesonly(t0->set.member, t1->set.member);
    case DT_SEQUENCE:
     return datatype_eq_namesonly(t0->sequence.member, t1->sequence.member);
    case DT_TUPLE:
     {
	unsigned i, num;
	struct datatype *t0sub, *t1sub;
	struct column *c0sub, *c1sub;

	num = datatypearray_num(&t0->tuple.members);
	if (num != datatypearray_num(&t1->tuple.members)) {
	   return false;
	}
	for (i=0; i<num; i++) {
	   t0sub = datatypearray_get(&t0->tuple.members, i);
	   t1sub = datatypearray_get(&t1->tuple.members, i);
	   if (!datatype_eq_namesonly(t0sub, t1sub)) {
	      return false;
	   }
	   c0sub = columnarray_get(&t0->tuple.headings, i);
	   c1sub = columnarray_get(&t1->tuple.headings, i);
	   if (c0sub != c1sub) {
	      return false;
	   }
	}
	return true;
     }
     break;
   }
   PQLASSERT(0);
   return false;
}

//bool datatype_eq(const struct datatype *t0, const struct datatype *t1) {
//   return datatype_eq_nonames(t0, t1);
//}

////////////////////////////////////////////////////////////

bool datatype_isbool(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_bool);
}

bool datatype_isint(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_int);
}

bool datatype_isdouble(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_double);
}

bool datatype_isstring(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_string);
}

bool datatype_isstruct(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_struct);
}

bool datatype_ispathelement(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_pathelement);
}

bool datatype_isdistinguisher(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_distinguisher);
}

/*
 * A type is a tuple if (1) it's unit, or (2) its representation is a
 * tuple.
 */
bool datatype_istuple(const struct datatype *t) {
   if (t == &t->dtm->type_unit) {
      return true;
   }
   return t->rep == DT_TUPLE;
}

bool datatype_isset(const struct datatype *t) {
   return t->rep == DT_SET;
}

bool datatype_issequence(const struct datatype *t) {
   return t->rep == DT_SEQUENCE;
}

bool datatype_isabsdbedge(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_absdbedge);
}

bool datatype_isabsnumber(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_absnumber);
}

bool datatype_isabsatom(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_absatom);
}

bool datatype_isabsdbobj(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_absdbobj);
}

bool datatype_isabstop(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_abstop);
}

bool datatype_isabsbottom(const struct datatype *t) {
   return datatype_eq_withnames(t, &t->dtm->type_absbottom);
}

// XXX these should be removed

bool datatype_isdbedge(const struct datatype *t) {
   return datatype_isabsdbedge(t);
}

bool datatype_isdbobj(const struct datatype *t) {
   return datatype_isabsdbobj(t);
}

////////////////////////////////////////////////////////////

/*
 * Database edges may be strings or integers.
 */
bool datatype_isanydbedge(const struct datatype *t) {
   return datatype_isabsbottom(t) || datatype_isabsdbedge(t) || datatype_isstring(t) || datatype_isint(t);
}

/*
 * Numbers are integers or floats.
 */
bool datatype_isanynumber(const struct datatype *t) {
   return datatype_isabsbottom(t) || datatype_isabsnumber(t) || datatype_isint(t) || datatype_isdouble(t);
}

/*
 * Atoms are numbers, strings, or booleans.
 */
bool datatype_isanyatom(const struct datatype *t) {
   return datatype_isabsbottom(t) ||
      datatype_isabsatom(t) ||
      datatype_isanynumber(t) ||
      datatype_isstring(t) || datatype_isbool(t);
}

/*
 * Database objects are atoms or structs.
 */
bool datatype_isanydbobj(const struct datatype *t) {
   return datatype_isabsbottom(t) ||
      datatype_isabsdbobj(t) ||
      datatype_isanyatom(t) || datatype_isstruct(t);
}

////////////////////////////////////////////////////////////

/*
 * Find the least common supertype of T1 and T2. Returns NULL
 * if no match exists.
 */
struct datatype *datatype_match_generalize(struct pqlcontext *pql,
					   struct datatype *t1,
					   struct datatype *t2) {
   struct datatype *result;

   if (datatype_eq_withnames(t1, t2)) {
      return t1;
   }

   if (datatype_isabsbottom(t1)) {
      return t2;
   }
   if (datatype_isabsbottom(t2)) {
      return t1;
   }

   /* Compound types */

   if (datatype_isset(t1) && datatype_isset(t2)) {
      result = datatype_match_generalize(pql,
					 datatype_set_member(t1),
					 datatype_set_member(t2));
      if (result == NULL) {
	 return NULL;
      }
      return datatype_set(pql, result);
   }
   if (datatype_issequence(t1) && datatype_issequence(t2)) {
      result = datatype_match_generalize(pql,
					 datatype_sequence_member(t1),
					 datatype_sequence_member(t2));
      if (result == NULL) {
	 return NULL;
      }
      return datatype_sequence(pql, result);
   }

   if (datatype_istuple(t1) && datatype_istuple(t2)) {
      struct datatypearray members;
      struct columnarray cols;
      struct datatype *st1, *st2, *stm, *ret;
      struct column *sc1, *sc2;
      unsigned i, arity;

      arity = datatype_arity(t1);
      if (arity != datatype_arity(t2)) {
	 return NULL;
      }

      datatypearray_init(&members);
      datatypearray_setsize(t1->dtm->pql, &members, arity);
      columnarray_init(&cols);
      columnarray_setsize(t1->dtm->pql, &cols, arity);

      for (i=0; i<arity; i++) {
	 st1 = datatype_getnth(t1, i);
	 st2 = datatype_getnth(t2, i);
	 stm = datatype_match_generalize(pql, st1, st2);
	 if (stm == NULL) {
	    datatypearray_setsize(t1->dtm->pql, &members, 0);
	    datatypearray_cleanup(t1->dtm->pql, &members);
	    columnarray_setsize(t1->dtm->pql, &cols, 0);
	    columnarray_cleanup(t1->dtm->pql, &cols);
	    return NULL;
	 }
	 datatypearray_set(&members, i, stm);

	 /* if headings not the same, generalize to NULL */
	 sc1 = datatype_tuple_getcolumn(t1, i);
	 sc2 = datatype_tuple_getcolumn(t2, i);
	 if (sc1 != sc2) {
	    columnarray_set(&cols, i, NULL);
	 }
	 else {
	    columnarray_set(&cols, i, sc1);
	 }
      }

      ret = datatype_tuple_specific(pql, &members, &cols);
      
      datatypearray_setsize(t1->dtm->pql, &members, 0);
      datatypearray_cleanup(t1->dtm->pql, &members);
      columnarray_setsize(t1->dtm->pql, &cols, 0);
      columnarray_cleanup(t1->dtm->pql, &cols);

      return ret;
   }

   if (datatype_istuple(t1) || datatype_istuple(t2)) {
      /* since they aren't both tuples, fail top */
      return NULL;
   }

   /* Base types and abstract supertypes */

   /* Don't generate dbedge unless it's already present */
   if ((datatype_isabsdbedge(t1) && datatype_isanydbedge(t2)) ||
       (datatype_isanydbedge(t1) && datatype_isabsdbedge(t2))) {
      return datatype_absdbedge(pql);
   }

   /* number */
   if (datatype_isanynumber(t1) && datatype_isanynumber(t2)) {
      return datatype_absnumber(pql);
   }

   /* atom */
   if (datatype_isanyatom(t1) && datatype_isanyatom(t2)) {
      return datatype_absatom(pql);
   }

   /* dbobj */
   if (datatype_isanydbobj(t1) && datatype_isanydbobj(t2)) {
      return datatype_absdbobj(pql);
   }

   /* no luck */
   return NULL;
}

/*
 * Find the greatest common subtype of T1 and T2. If none, returns
 * NULL.
 */
struct datatype *datatype_match_specialize(struct pqlcontext *pql,
					   struct datatype *t1,
					   struct datatype *t2) {
   struct datatype *result;

   if (datatype_eq_withnames(t1, t2)) {
      return t1;
   }

   if (datatype_isabsbottom(t1) || datatype_isabsbottom(t2)) {
      return datatype_absbottom(pql);
   }

   if (datatype_isabstop(t1)) {
      return t2;
   }
   if (datatype_isabstop(t2)) {
      return t1;
   }

   /* Compound types */

   if (datatype_isset(t1) && datatype_isset(t2)) {
      result = datatype_match_specialize(pql,
					 datatype_set_member(t1),
					 datatype_set_member(t2));
      if (result == NULL) {
	 return NULL;
      }
      return datatype_set(pql, result);
   }
   if (datatype_issequence(t1) && datatype_issequence(t2)) {
      result = datatype_match_specialize(pql,
					 datatype_sequence_member(t1),
					 datatype_sequence_member(t2));
      if (result == NULL) {
	 return NULL;
      }
      return datatype_sequence(pql, result);
   }

   if (datatype_istuple(t1) && datatype_istuple(t2)) {
      struct datatypearray members;
      struct columnarray cols;
      struct datatype *st1, *st2, *stm, *ret;
      struct column *sc1, *sc2;
      unsigned i, arity;

      arity = datatype_arity(t1);
      if (arity != datatype_arity(t2)) {
	 return NULL;
      }

      datatypearray_init(&members);
      datatypearray_setsize(t1->dtm->pql, &members, arity);
      columnarray_init(&cols);
      columnarray_setsize(t1->dtm->pql, &cols, arity);

      for (i=0; i<arity; i++) {
	 st1 = datatype_getnth(t1, i);
	 st2 = datatype_getnth(t2, i);
	 stm = datatype_match_specialize(pql, st1, st2);
	 if (stm == NULL) {
	    datatypearray_setsize(t1->dtm->pql, &members, 0);
	    datatypearray_cleanup(t1->dtm->pql, &members);
	    columnarray_setsize(t1->dtm->pql, &cols, 0);
	    columnarray_cleanup(t1->dtm->pql, &cols);
	    return NULL;
	 }
	 datatypearray_set(&members, i, stm);

	 /* if one heading is null, pick the other */
	 sc1 = datatype_tuple_getcolumn(t1, i);
	 sc2 = datatype_tuple_getcolumn(t2, i);
	 if (sc1 == NULL && sc2 == NULL) {
	    columnarray_set(&cols, i, NULL);
	 }
	 else if (sc1 == NULL) {
	    columnarray_set(&cols, i, sc2);
	 }
	 else if (sc2 == NULL) {
	    columnarray_set(&cols, i, sc1);
	 }
	 else {
	    // both are defined and they aren't the same; fail
	    // XXX is this really what we want?
	    datatypearray_setsize(t1->dtm->pql, &members, 0);
	    datatypearray_cleanup(t1->dtm->pql, &members);
	    columnarray_setsize(t1->dtm->pql, &cols, 0);
	    columnarray_cleanup(t1->dtm->pql, &cols);
	    return NULL;
	 }
      }

      ret = datatype_tuple_specific(pql, &members, &cols);
      
      datatypearray_setsize(t1->dtm->pql, &members, 0);
      datatypearray_cleanup(t1->dtm->pql, &members);
      columnarray_setsize(t1->dtm->pql, &cols, 0);
      columnarray_cleanup(t1->dtm->pql, &cols);

      return ret;
   }

   if (datatype_istuple(t1) || datatype_istuple(t2)) {
      /* since they aren't both tuples, fail */
      return NULL;
   }

   /* Base types and abstract supertypes */

   /* Don't involve dbedge unless it's already present */
   if (datatype_isabsdbedge(t1) && datatype_isanydbedge(t2)) {
      return t2;
   }
   if (datatype_isanydbedge(t1) && datatype_isabsdbedge(t2)) {
      return t1;
   }

   /* number */
   if (datatype_isabsnumber(t1) && datatype_isanynumber(t2)) {
      return t2;
   }
   if (datatype_isanynumber(t1) && datatype_isabsnumber(t2)) {
      return t1;
   }

   /* atom */
   if (datatype_isabsatom(t1) && datatype_isanyatom(t2)) {
      return t2;
   }
   if (datatype_isanyatom(t1) && datatype_isabsatom(t2)) {
      return t1;
   }

   /* dbobj */
   if (datatype_isabsdbobj(t1) && datatype_isanydbobj(t2)) {
      return t2;
   }
   if (datatype_isanydbobj(t1) && datatype_isabsdbobj(t2)) {
      return t1;
   }

   /* no luck */
   return NULL;
}


/*
 * Like generalize but returns top (the most general type) instead
 * of failing.
 */
struct datatype *datatype_match_generalize_any(struct pqlcontext *pql,
					       struct datatype *t1,
					       struct datatype *t2) {
   struct datatype *result;

   result = datatype_match_generalize(pql, t1, t2);
   if (result == NULL) {
      return datatype_abstop(pql);
   }
   return result;
}

/*
 * Like specialize but returns top (the most specific type) instead of
 * failing.
 */
struct datatype *datatype_match_specialize_any(struct pqlcontext *pql,
					       struct datatype *t1,
					       struct datatype *t2) {
   struct datatype *result;

   result = datatype_match_specialize(pql, t1, t2);
   if (result == NULL) {
      return datatype_absbottom(pql);
   }
   return result;
}

////////////////////////////////////////////////////////////

const char *datatype_getname(const struct datatype *t) {
   return t->name;
}

unsigned datatype_arity(const struct datatype *t) {
   if (datatype_eq_withnames(t, &t->dtm->type_unit)) {
      return 0;
   }
   if (t->rep != DT_TUPLE) {
      return 1;
   }

   PQLASSERT(datatypearray_num(&t->tuple.members) ==
	     columnarray_num(&t->tuple.headings));
   return datatypearray_num(&t->tuple.members);
}

struct datatype *datatype_getnth(/*const*/ struct datatype *t, unsigned n) {
   unsigned arity;

   /* unit has no members; therefore cannot get members from it */
   PQLASSERT(t != &t->dtm->type_unit);

   arity = datatype_arity(t);
   PQLASSERT(n < arity);

   /* allow getting index 0 from a non-tuple */
   // XXX this should be removed
   if (arity == 1 && t->rep != DT_TUPLE) {
      //PQLASSERT(n == 0); -- redundant
      return t;
   }

   PQLASSERT(t->rep == DT_TUPLE);

   return datatypearray_get(&t->tuple.members, n);
}

unsigned datatype_nonset_arity(const struct datatype *t) {
   if (datatype_isset(t)) {
      t = datatype_set_member(t);
   }
   else if (datatype_issequence(t)) {
      t = datatype_sequence_member(t);
   }
   return datatype_arity(t);
}

struct column *datatype_tuple_getcolumn(struct datatype *t, unsigned ix) {
   PQLASSERT(t->rep == DT_TUPLE);
   return columnarray_get(&t->tuple.headings, ix);
}

struct datatype *datatype_set_member(const struct datatype *t) {
   PQLASSERT(t->rep == DT_SET);
   return t->set.member;
}

struct datatype *datatype_sequence_member(const struct datatype *t) {
   PQLASSERT(t->rep == DT_SEQUENCE);
   return t->sequence.member;
}

////////////////////////////////////////////////////////////

/*
 * This would be much tidier to write with closures...
 */

#define SET_WRAP_DECL \
   bool isset=false, isseq=false

#define SET_WRAP_START(t) \
   (datatype_isset(t) ? isset=true, datatype_set_member(t) : \
    datatype_issequence(t) ? isseq=true, datatype_sequence_member(t) : (t))

#define SET_WRAP_AGAIN_STRICT(t) \
   (datatype_isset(t) && isset ? datatype_set_member(t) : \
    datatype_issequence(t) && isseq ? datatype_sequence_member(t) : \
    isset || isseq || datatype_isset(t) || datatype_issequence(t) ? NULL : (t))

#define SET_WRAP_AGAIN_UNSTRICT(t) \
   (datatype_isset(t) && isset ? datatype_set_member(t) : \
    datatype_issequence(t) && isseq ? datatype_sequence_member(t) : \
    isset || isseq ? NULL : (t))


#define SET_WRAP_END(rt) \
   (isset ? datatype_set((rt)->dtm->pql, rt) : \
    isseq ? datatype_sequence((rt)->dtm->pql, rt) : (rt))

struct datatype *datatype_projectscalar(struct datatype *t, unsigned ix) {
   struct datatype *rt;
   SET_WRAP_DECL;

   rt = datatype_getnth(SET_WRAP_START(t), ix);
   return SET_WRAP_END(rt);
}

struct datatype *datatype_projecttuple(struct datatype *t,
				       const struct indexarray *ixes) {
   struct datatypearray arr;
   struct columnarray cols;
   unsigned i, num;
   struct datatype *rt;
   SET_WRAP_DECL;

   t = SET_WRAP_START(t);

   num = indexarray_num(ixes);

   if (num == 0) {
      return datatype_unit(t->dtm->pql);
   }

   datatypearray_init(&arr);
   datatypearray_setsize(t->dtm->pql, &arr, num);
   columnarray_init(&cols);
   columnarray_setsize(t->dtm->pql, &cols, num);

   for (i=0; i<num; i++) {
      datatypearray_set(&arr, i,
			datatypearray_get(&t->tuple.members,
					  indexarray_get(ixes, i)));
      columnarray_set(&cols, i,
		      columnarray_get(&t->tuple.headings,
				      indexarray_get(ixes, i)));
   }

   rt = datatype_tuple_specific(t->dtm->pql, &arr, &cols);

   datatypearray_setsize(t->dtm->pql, &arr, 0);
   datatypearray_cleanup(t->dtm->pql, &arr);
   columnarray_setsize(t->dtm->pql, &cols, 0);
   columnarray_cleanup(t->dtm->pql, &cols);

   rt = SET_WRAP_END(rt);

   return rt;
}

struct datatype *datatype_strip_one(struct datatype *t, unsigned ix) {
   return datatype_tupleset_strip(t->dtm->pql, t, ix);
}

struct datatype *datatype_strip(/*const*/ struct datatype *t,
				const struct indexarray *ixes) {
   struct datatype *rt;
   unsigned arity, num, i;
   SET_WRAP_DECL;

   t = SET_WRAP_START(t);

   arity = datatype_arity(t);
   num = indexarray_num(ixes);
   PQLASSERT(num <= arity);

   if (num == arity) {
      rt = datatype_unit(t->dtm->pql);
   }
   else {
      for (i=0; i<num-1; i++) {
	 PQLASSERT(indexarray_get(ixes, i) < indexarray_get(ixes, i+1));
      }

      rt = t;
      for (i=num; i-- > 0; ) {
	 rt = datatype_tuple_strip(rt->dtm->pql, rt, indexarray_get(ixes, i));
      }
   }

   rt = SET_WRAP_END(rt);
   return rt;
}

/*
 * Project out COL regardless of where it is in the nesting tree of type T.
 */
struct datatype *datatype_subprojectscalar(struct datatype *t,
					   struct column *col) {
   if (datatype_isset(t)) {
      t = datatype_set_member(t);
      t = datatype_subprojectscalar(t, col);
      if (t == NULL) {
	 return NULL;
      }
      return datatype_set(t->dtm->pql, t);
   }
   if (datatype_issequence(t)) {
      t = datatype_sequence_member(t);
      t = datatype_subprojectscalar(t, col);
      if (t == NULL) {
	 return NULL;
      }
      return datatype_sequence(t->dtm->pql, t);
   }
   if (datatype_istuple(t)) {
      struct datatype *st;
      unsigned i, arity;

      arity = datatype_arity(t);
      for (i=0; i<arity; i++) {
	 st = datatype_getnth(t, i);
	 if (datatype_tuple_getcolumn(t, i) == col) {
	    return st;
	 }
	 st = datatype_subprojectscalar(st, col);
	 if (st != NULL) {
	    return st;
	 }
      }
      return NULL;
   }
   return NULL;
}

/*
 * Project out COLS regardless of where they are in the nesting tree
 * of type T.
 */

#if 0
/*
 * This is all well and good but it produces columns in the order
 * found in T, not the order specified in COLS.
 */

struct datatype *datatype_subprojecttuple(struct datatype *t,
					  const struct columnarray *cols) {
   if (datatype_isset(t)) {
      t = datatype_set_member(t);
      t = datatype_subprojecttuple(t, cols);
      if (t == NULL) {
	 return NULL;
      }
      return datatype_set(t->dtm->pql, t);
   }
   if (datatype_issequence(t)) {
      t = datatype_sequence_member(t);
      t = datatype_subprojecttuple(t, cols);
      if (t == NULL) {
	 return NULL;
      }
      return datatype_sequence(t->dtm->pql, t);
   }
   if (datatype_istuple(t)) {
      struct datatype *resulttype, *foundtype, *subtype;
      struct column *subcol;
      unsigned i, arity;

      arity = datatype_arity(t);
      resulttype = NULL;
      for (i=0; i<arity; i++) {
	 subtype = datatype_getnth(t, i);
	 subcol = datatype_tuple_getcolumn(t, i);
	 foundtype = NULL;
	 if (columnarray_contains(cols, subcol)) {
	    foundtype = subtype;
	 }
	 else {
	    subtype = datatype_subprojecttuple(subtype, cols);
	    if (subtype != NULL) {
	       foundtype = subtype;
	    }
	 }
	 if (foundtype != NULL) {
	    if (resulttype == NULL) {
	       resulttype = datatype_unit(t->dtm->pql);
	    }
	    resulttype = datatype_tuple_append(t->dtm->pql,
					       resulttype, foundtype, subcol);
	 }
      }
      return resulttype;
   }
   return NULL;
}

#else

/*
 * Recursive helper: collect columns starting at cols[pos]
 */
static struct datatype *datatype_subprojecttuple_rec(struct datatype *t,
					const struct columnarray *cols,
					unsigned *pos) {
   
   if (datatype_isset(t)) {
      t = datatype_set_member(t);
      t = datatype_subprojecttuple_rec(t, cols, pos);
      if (t == NULL) {
	 return NULL;
      }
      return datatype_set(t->dtm->pql, t);
   }
   if (datatype_issequence(t)) {
      t = datatype_sequence_member(t);
      t = datatype_subprojecttuple_rec(t, cols, pos);
      if (t == NULL) {
	 return NULL;
      }
      return datatype_sequence(t->dtm->pql, t);
   }
   if (datatype_istuple(t)) {
      struct datatype *ttype, *foundtype, *ret;
      struct column *tcol, *thiscol;
      unsigned i, numcols, arity;
      bool more;

      numcols = columnarray_num(cols);
      arity = datatype_arity(t);

      ret = NULL;

      more = true;
      while (more && *pos < numcols) {
	 more = false;

	 thiscol = columnarray_get(cols, *pos);

	 for (i=0; i<arity; i++) {
	    ttype = datatype_getnth(t, i);
	    tcol = datatype_tuple_getcolumn(t, i);

	    /*
	     * If you have something like
	     *
	     *    (int as A, set((int as B, int as C)) as D)
	     *
	     * and you try to project out e.g. both B and D, the logic
	     * here will fail to find B and the typechecker will
	     * reject the results. Such projections are disallowed
	     * because if they weren't the result type would have to
	     * be this:
	     *
	     *    (set(int as B) as D, set((int as B, int as C)) as D)
	     *
	     * and duplicate column names are not allowed.
	     */

	    foundtype = NULL;
	    if (tcol == thiscol) {
	       foundtype = ttype;
	       (*pos)++;
	    }
	    else {
	       foundtype = datatype_subprojecttuple_rec(ttype, cols, pos);
	    }
	    if (foundtype != NULL) {
	       more = true;
	       if (ret == NULL) {
		  ret = datatype_tuple_monople(t->dtm->pql, foundtype, tcol);
	       }
	       else {
		  ret = datatype_adjoin(ret, foundtype, tcol);
	       }
	    }
	 }
      }

      return ret;
   }
   return NULL;
}

struct datatype *datatype_subprojecttuple(struct datatype *t,
					  const struct columnarray *cols) {
   unsigned pos = 0;

   return datatype_subprojecttuple_rec(t, cols, &pos);
}

#endif

/*
 * Strip out COLS regardless of where they are in the nesting tree of
 * type T.
 */
struct datatype *datatype_substrip(struct datatype *t,
				   const struct columnarray *cols) {
   if (datatype_isset(t)) {
      t = datatype_set_member(t);
      t = datatype_substrip(t, cols);
      if (t == NULL) {
	 return NULL;
      }
      return datatype_set(t->dtm->pql, t);
   }
   if (datatype_issequence(t)) {
      t = datatype_sequence_member(t);
      t = datatype_substrip(t, cols);
      if (t == NULL) {
	 return NULL;
      }
      return datatype_sequence(t->dtm->pql, t);
   }
   if (datatype_istuple(t)) {
      struct datatype *resulttype, *subtype;
      struct column *subcol;
      unsigned i, arity;

      arity = datatype_arity(t);
      resulttype = datatype_unit(t->dtm->pql);
      for (i=0; i<arity; i++) {
	 subtype = datatype_getnth(t, i);
	 subcol = datatype_tuple_getcolumn(t, i);
	 if (!columnarray_contains(cols, subcol)) {
	    subtype = datatype_substrip(subtype, cols);
	    if (subtype != NULL) {
	       resulttype = datatype_tuple_append(t->dtm->pql,
						  resulttype, subtype, subcol);
	    }
	 }
      }
      return resulttype;
   }
   return t;
}

struct datatype *datatype_rename(struct datatype *t,
				 struct column *oldcol, struct column *newcol){
   struct datatype *rt, *st;
   struct column *col;
   struct datatypearray types;
   struct columnarray cols;
   unsigned arity, i;
   bool found = false;
   SET_WRAP_DECL;

   t = SET_WRAP_START(t);

   if (!datatype_istuple(t)) {
      return NULL;
   }

   arity = datatype_arity(t);
   if (arity < 1) {
      /* requested column name cannot possibly exist */
      return NULL;
   }
   PQLASSERT(t->rep == DT_TUPLE);

   datatypearray_init(&types);
   columnarray_init(&cols);
   datatypearray_setsize(t->dtm->pql, &types, arity);
   columnarray_setsize(t->dtm->pql, &cols, arity);
   for (i=0; i<arity; i++) {
      st = datatypearray_get(&t->tuple.members, i);
      col = columnarray_get(&t->tuple.headings, i);

      datatypearray_set(&types, i, st);
      if (col == oldcol) {
	 /* assertion failure implies column name duplicated */
	 PQLASSERT(!found);
	 found = true;
	 columnarray_set(&cols, i, newcol);
      }
      else {
	 columnarray_set(&cols, i, col);
      }
   }
   if (!found) {
      for (i=0; i<arity; i++) {
	 st = datatypearray_get(&types, i);
	 st = datatype_rename(st, oldcol, newcol);
	 if (st != NULL) {
	    /* assertion failure implies column name duplicated */
	    PQLASSERT(!found);
	    found = true;
	    datatypearray_set(&types, i, st);
	 }
      }
   }

   if (found) {
      rt = datatype_tuple_specific(t->dtm->pql, &types, &cols);
   }
   else {
      rt = NULL;
   }

   datatypearray_setsize(t->dtm->pql, &types, 0);
   datatypearray_cleanup(t->dtm->pql, &types);
   columnarray_setsize(t->dtm->pql, &cols, 0);
   columnarray_cleanup(t->dtm->pql, &cols);

   if (rt == NULL) {
      return NULL;
   }

   rt = SET_WRAP_END(rt);
   return rt;
}

struct datatype *datatype_join(struct datatype *t1,
			       struct datatype *t2) {
   struct datatype *rt;
   SET_WRAP_DECL;

   t1 = SET_WRAP_START(t1);
   t2 = SET_WRAP_AGAIN_STRICT(t2);
   if (t2 == NULL) {
      return NULL;
   }

   rt = datatype_tuple_concat(t1->dtm->pql, t1, t2);

   rt = SET_WRAP_END(rt);
   return rt;
}

struct datatype *datatype_nest(struct datatype *t,
			       const struct indexarray *ixes,
			       struct column *newcol) {
   struct datatype *rt, *t1, *t2;
   SET_WRAP_DECL;

   t = SET_WRAP_START(t);

   t1 = datatype_strip(t, ixes);
   t2 = datatype_projecttuple(t, ixes);

   rt = datatype_adjoin(t1, datatype_set(t->dtm->pql, t2), newcol);

   rt = SET_WRAP_END(rt);
   return rt;
}

struct datatype *datatype_unnest(struct datatype *t, unsigned ix) {
   struct datatype *rt, *t1, *t2;
   struct column *c2;
   SET_WRAP_DECL;

   t = SET_WRAP_START(t);

   PQLASSERT(datatype_istuple(t));

   t1 = datatype_strip_one(t, ix);
   t2 = datatype_projectscalar(t, ix);
   c2 = datatype_tuple_getcolumn(t, ix);

   if (!datatype_isset(t2)) {
      return NULL;
   }

   t2 = datatype_set_member(t2);

   if (datatype_istuple(t2)) {
      rt = datatype_tuple_concat(t->dtm->pql, t1, t2);
   }
   else {
      rt = datatype_tuple_append(t->dtm->pql, t1, t2, c2);
   }

   rt = SET_WRAP_END(rt);
   return rt;
}

struct datatype *datatype_distinguish(struct datatype *t, struct column *col) {
   return datatype_adjoin(t, datatype_distinguisher(t->dtm->pql), col);
}

struct datatype *datatype_adjoin(struct datatype *t1, struct datatype *t2,
		      struct column *c2) {
   struct datatype *rt;
   SET_WRAP_DECL;

   t1 = SET_WRAP_START(t1);
   t2 = SET_WRAP_AGAIN_UNSTRICT(t2);
   if (t2 == NULL) {
      return NULL;
   }

   rt = datatype_tuple_append(t1->dtm->pql, t1, t2, c2);

   rt = SET_WRAP_END(rt);
   return rt;
}

////////////////////////////////////////////////////////////

#if 0 /* debugging code */

static unsigned datatype_countcolumn(struct datatypemanager *dtm,
				     struct datatype *t,
				     struct column *col) {
   struct datatype *dt;
   unsigned i, num;
   unsigned count = 0;

   if (t->rep == DT_TUPLE) {
      num = columnarray_num(&t->tuple.headings);
      for (i=0; i<num; i++) {
	 if (columnarray_get(&t->tuple.headings, i) == col) {
	    count++;
	 }
      }
   }

   if (t->set_of != NULL) {
      PQLASSERT(t->set_of->rep == DT_SET);
      PQLASSERT(t->set_of->set.member == t);
      
      count += datatype_countcolumn(dtm, t->set_of, col);
   }
   if (t->sequence_of != NULL) {
      PQLASSERT(t->sequence_of->rep == DT_SEQUENCE);
      PQLASSERT(t->sequence_of->sequence.member == t);

      count += datatype_countcolumn(dtm, t->sequence_of, col);
   }
   num = datatypearray_num(&t->tuples);
   for (i=0; i<num; i++) {
      dt = datatypearray_get(&t->tuples, i);
      PQLASSERT(dt->rep == DT_TUPLE);
      PQLASSERT(datatypearray_get(&dt->tuple.members, 0) == t);

      count += datatype_countcolumn(dtm, dt, col);
   }

   return count;
}

static void datatypemanager_checkcolumn(struct datatypemanager *dtm,
					struct column *col) {
   unsigned count = 0;

   count += datatype_countcolumn(dtm, &dtm->type_unit, col);
   count += datatype_countcolumn(dtm, &dtm->type_bool, col);
   count += datatype_countcolumn(dtm, &dtm->type_int, col);
   count += datatype_countcolumn(dtm, &dtm->type_double, col);
   count += datatype_countcolumn(dtm, &dtm->type_string, col);
   count += datatype_countcolumn(dtm, &dtm->type_struct, col);
   count += datatype_countcolumn(dtm, &dtm->type_pathelement, col);
   count += datatype_countcolumn(dtm, &dtm->type_distinguisher, col);

   count += datatype_countcolumn(dtm, &dtm->type_absdbedge, col);
   count += datatype_countcolumn(dtm, &dtm->type_absnumber, col);
   count += datatype_countcolumn(dtm, &dtm->type_absatom, col);
   count += datatype_countcolumn(dtm, &dtm->type_absdbobj, col);
   count += datatype_countcolumn(dtm, &dtm->type_abstop, col);
   count += datatype_countcolumn(dtm, &dtm->type_absbottom, col);
   
   PQLASSERT(count == col->typerefcount);
}

#endif /* 0 - debugging code */
