/*
 * Copyright 2009, 2010
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef DATATYPE_H
#define DATATYPE_H

#include <stdbool.h>
#include "array.h"
#include "intarray.h"

struct pqlcontext;


/*
 * The datatypemanager owns all the storage for all datatypes; it is
 * stashed in the pqlcontext.
 *
 * All datatypes are immutable once created, each is unique, and all
 * belong to the datatypemanager; client code need not destroy them.
 *
 * While two references to the same type will always yield the same
 * datatype object, and thus datatypes can be compared by pointer
 * equality, client code should always use datatype_eq. This makes it
 * possible to hunt down callers and check for needed adjustments when
 * the fine shadings of "same type" change.
 */

struct datatypemanager; /* Opaque. */
struct datatype; /* Opaque. */

#ifndef DATATYPE_INLINE
#define DATATYPE_INLINE INLINE
#endif
DECLARRAY(column);
DECLARRAY(datatype);
DEFARRAY(column, DATATYPE_INLINE);
DEFARRAY(datatype, DATATYPE_INLINE);


/* Manager. */
struct datatypemanager *datatypemanager_create(struct pqlcontext *);
void datatypemanager_destroy(struct datatypemanager *);

/*
 * Column headers for tuple types. The NAME may be null, in which case
 * a name is constructed from the internal ID number.
 */
struct column *column_create(struct pqlcontext *, const char *name);
const char *column_getname(struct column *col);
void column_incref(struct column *);
void column_decref(struct pqlcontext *, struct column *);

/*
 * Constructors.
 */

/* Concrete atom types. */
struct datatype *datatype_bool(struct pqlcontext *);
struct datatype *datatype_int(struct pqlcontext *);
struct datatype *datatype_double(struct pqlcontext *);
struct datatype *datatype_string(struct pqlcontext *);

/* Concrete compound types. */
struct datatype *datatype_struct(struct pqlcontext *);
struct datatype *datatype_pathelement(struct pqlcontext *);

/* Distinguisher. */
struct datatype *datatype_distinguisher(struct pqlcontext *);

/* Tuple type. */
struct datatype *datatype_unit(struct pqlcontext *);
struct datatype *datatype_tuple_monople(struct pqlcontext *,
					struct datatype *t0,
					struct column *n0);
struct datatype *datatype_tuple_pair(struct pqlcontext *,
				     struct datatype *t0,
				     struct datatype *t1,
				     struct column *n0,
				     struct column *n1);
struct datatype *datatype_tuple_triple(struct pqlcontext *,
				       struct datatype *t0,
				       struct datatype *t1,
				       struct datatype *t2,
				       struct column *n0,
				       struct column *n1,
				       struct column *n2);
struct datatype *datatype_tuple_specific(struct pqlcontext *,
					 const struct datatypearray *,
					 const struct columnarray *);

/* Collection types. */
struct datatype *datatype_set(struct pqlcontext *, struct datatype *t);
struct datatype *datatype_sequence(struct pqlcontext *, struct datatype *t);


/* Abstract types. */

// XXX rename uses of these to insert "abs"
struct datatype *datatype_dbobj(struct pqlcontext *);
struct datatype *datatype_dbedge(struct pqlcontext *);

struct datatype *datatype_absdbedge(struct pqlcontext *);
struct datatype *datatype_absnumber(struct pqlcontext *);
struct datatype *datatype_absatom(struct pqlcontext *);
struct datatype *datatype_absdbobj(struct pqlcontext *);
struct datatype *datatype_abstop(struct pqlcontext *);
struct datatype *datatype_absbottom(struct pqlcontext *);

/*
 * Tests for exact types.
 */

bool datatype_eq_nonames(const struct datatype *, const struct datatype *);
bool datatype_eq_withnames(const struct datatype *, const struct datatype *);
bool datatype_eq_namesonly(const struct datatype *, const struct datatype *);

bool datatype_isbool(const struct datatype *);
bool datatype_isint(const struct datatype *);
bool datatype_isdouble(const struct datatype *);
bool datatype_isstring(const struct datatype *);
bool datatype_isstruct(const struct datatype *);
bool datatype_ispathelement(const struct datatype *);
bool datatype_isdistinguisher(const struct datatype *);
bool datatype_istuple(const struct datatype *);
bool datatype_isset(const struct datatype *);
bool datatype_issequence(const struct datatype *);

// XXX rename uses of these to insert "abs"
bool datatype_isdbobj(const struct datatype *);
bool datatype_isdbedge(const struct datatype *);

bool datatype_isabsdbedge(const struct datatype *);
bool datatype_isabsnumber(const struct datatype *);
bool datatype_isabsatom(const struct datatype *);
bool datatype_isabsdbobj(const struct datatype *);
bool datatype_isabstop(const struct datatype *);
bool datatype_isabsbottom(const struct datatype *);

/*
 * Tests for type classes.
 */

bool datatype_isanydbedge(const struct datatype *);
bool datatype_isanynumber(const struct datatype *);
bool datatype_isanyatom(const struct datatype *);
bool datatype_isanydbobj(const struct datatype *);
//bool datatype_isanyany(const struct datatype *); -- pointless, use "true" :-)

/*
 * Matching.
 *
 * The first forms return NULL on things that shouldn't be matched at
 * all, like tuples of differing arity. The _any forms return top or
 * bottom respectively in those cases.
 */

struct datatype *datatype_match_generalize(struct pqlcontext *pql,
					   struct datatype *t1,
					   struct datatype *t2);
struct datatype *datatype_match_specialize(struct pqlcontext *pql,
					   struct datatype *t1,
					   struct datatype *t2);

struct datatype *datatype_match_generalize_any(struct pqlcontext *pql,
					       struct datatype *t1,
					       struct datatype *t2);
struct datatype *datatype_match_specialize_any(struct pqlcontext *pql,
					       struct datatype *t1,
					       struct datatype *t2);

/*
 * Inspection functions.
 *
 * Note that all types are treated as tuples -- those that aren't
 * really tuples function as tuples of arity 1. This makes
 * datatype_arity and datatype_getnth work on everything and removes a
 * pile of special cases.
 */

/* polymorphic */
const char *datatype_getname(const struct datatype *);
unsigned datatype_arity(const struct datatype *);
struct datatype *datatype_getnth(struct datatype *, unsigned ix);
unsigned datatype_nonset_arity(const struct datatype *);

/* tuple */
struct column *datatype_tuple_getcolumn(struct datatype *, unsigned ix);

/* set/sequence */
struct datatype *datatype_set_member(const struct datatype *);
struct datatype *datatype_sequence_member(const struct datatype *);

/*
 * Transformations.
 */
struct datatype *datatype_tuple_append(struct pqlcontext *,
				       struct datatype *t0,
				       struct datatype *t1,
				       struct column *c1);
struct datatype *datatype_tuple_concat(struct pqlcontext *,
				       struct datatype *t0,
				       struct datatype *t1);

struct datatype *datatype_tuple_strip(struct pqlcontext *pql,
				      struct datatype *, unsigned ix);
struct datatype *datatype_tupleset_strip(struct pqlcontext *pql,
					 struct datatype *, unsigned ix);


struct datatype *datatype_projectscalar(struct datatype *, unsigned ix);
struct datatype *datatype_projecttuple(struct datatype *,
				       const struct indexarray *ixes);
struct datatype *datatype_strip_one(struct datatype *, unsigned ix);
struct datatype *datatype_strip(struct datatype *,
				const struct indexarray *ixes);
struct datatype *datatype_subprojectscalar(struct datatype *t,
					   struct column *col);
struct datatype *datatype_subprojecttuple(struct datatype *t,
					  const struct columnarray *cols);
struct datatype *datatype_substrip(struct datatype *t,
				   const struct columnarray *notcols);
struct datatype *datatype_rename(struct datatype *,
				 struct column *oldcol, struct column *newcol);
struct datatype *datatype_join(struct datatype *, struct datatype *);
struct datatype *datatype_nest(struct datatype *,
			       const struct indexarray *ixes,
			       struct column *newcol);
struct datatype *datatype_unnest(struct datatype *, unsigned ix);
struct datatype *datatype_distinguish(struct datatype *, struct column *);
struct datatype *datatype_adjoin(struct datatype *, struct datatype *,
				 struct column *);


#endif /* DATATYPE_H */
