/*
 * Copyright 2010
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Build def-use chains for variables and then def-use graphs for columns,
 * and use them to do assorted optimizations that require inspecting the
 * usage points and definition points together.
 *
 * Optimizations we do in here:
 *
 * Additional optimizations we probably should do in here:
 *    - remove unused columns
 *    - convert adjoined columns used in predicates to let-bindings
 *      within those predicates
 *    - if a var is defined as a projection of something, and used
 *      only by projecting more specific things from it, remove it.
 */

#include <stdlib.h>

#include "array.h"
#include "datatype.h"
#include "tcalc.h"
#include "passes.h"

/*
 * Declare some array types.
 */
struct trace;
struct dvar;
struct dcolref;
struct dcol;
DECLARRAY(trace);
DECLARRAY(dvar);
DECLARRAY(dcolref);
DECLARRAY(dcol);
DEFARRAY(trace, );
DEFARRAY(dvar, );
DEFARRAY(dcolref, );
DEFARRAY(dcol, );

/*
 * This records the path from the root expression to an expression of
 * interest. exprs[0] is the root.
 */
struct trace {
   struct tcexprarray exprs;
};

/*
 * This is the def-use information for a single variable. Each reference
 * is a trace; there is one definition and an ordered list of uses.
 */
struct dvar {
   struct tcvar *var;
   struct trace *def;
   struct tracearray uses;
};

/*
 * This holds one reference to a column. There are five ways a column
 * can be referenced, and it can be referenced by either an expression
 * or a variable. The FROM and TO arrays link each usage to the next
 * and previous usages, which are not necessarily unique: a column in
 * a variable can have many TO entries if the variable is used
 * repeatedly, and certain operators like UNION generate multiple FROM
 * entries.
 */
enum dcolrefways {
   CR_DEF,		/* creates the column */
   CR_PASS,		/* passes column through (appears in type only) */
   CR_USE,		/* explicitly uses the column */
   CR_END,		/* column ends here but it's needed */
   CR_STRIP,		/* column ends here and it's not needed */
};
struct dcolref {
   bool isvar;
   union {
      struct tcexpr *expr;
      struct tcvar *var;
   };
   enum dcolrefways how;
   struct dcolrefarray from;
   struct dcolrefarray to;
};

/*
 * The complete information for a column: the column itself, and the
 * complete graph of all references. DEFS and USES list the references
 * of those types.
 *
 * XXX: at the moment it's obvious that DEFS will be wanted, but it's
 * not so clear if USES is useful or if we'll want ENDS, etc.
 */
struct dcol {
   struct column *col;
   struct dcolrefarray all;

   struct dcolrefarray defs;
   struct dcolrefarray uses;
};

/*
 * The complete context for this module.
 */
struct defuse {
   struct pqlcontext *pql;
   struct trace *curtrace;
   struct dvararray vars;
   struct dcolarray cols;
};

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
//                                                        //
//                     A N A L Y S I S                    //
//                                                        //
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// random utility functions

/*
 * Return true if T contains a column COL somewhere.
 */
static bool datatype_has_column(struct datatype *t, struct column *col) {
   return datatype_subprojectscalar(t, col) != NULL;
}

////////////////////////////////////////////////////////////
// trace methods

/*
 * Create a trace object.
 */
static struct trace *trace_create(struct defuse *ctx) {
   struct trace *ret;

   ret = domalloc(ctx->pql, sizeof(*ret));
   tcexprarray_init(&ret->exprs);
   return ret;
}

/*
 * Clone an existing trace object.
 */
static struct trace *trace_clone(struct defuse *ctx, struct trace *ot) {
   struct trace *nt;
   unsigned i, num;

   nt = trace_create(ctx);
   num = tcexprarray_num(&ot->exprs);
   tcexprarray_setsize(ctx->pql, &nt->exprs, num);
   for (i=0; i<num; i++) {
      tcexprarray_set(&nt->exprs, i, tcexprarray_get(&ot->exprs, i));
   }
   return nt;
}

/*
 * Destroy a trace object.
 */
static void trace_destroy(struct defuse *ctx, struct trace *t) {
   tcexprarray_setsize(ctx->pql, &t->exprs, 0);
   tcexprarray_cleanup(ctx->pql, &t->exprs);
   dofree(ctx->pql, t, sizeof(*t));
}

/*
 * Push an expression onto the end of a trace.
 */
static void trace_push(struct defuse *ctx, struct trace *t, struct tcexpr *te){
   tcexprarray_add(ctx->pql, &t->exprs, te, NULL);
}

/*
 * Pop an expression off the end of a trace.
 */
static void trace_pop(struct defuse *ctx, struct trace *t) {
   unsigned num;

   num = tcexprarray_num(&t->exprs);
   PQLASSERT(num > 0);
   tcexprarray_setsize(ctx->pql, &t->exprs, num - 1);
}

/*
 * Get the "main" expression of the trace (the bottom)
 */
static struct tcexpr *trace_getexpr(struct trace *t) {
   unsigned num;

   num = tcexprarray_num(&t->exprs);
   PQLASSERT(num > 0);
   return tcexprarray_get(&t->exprs, num - 1);
}

/*
 * Get a parent expression from the trace, HEIGHT steps from the bottom.
 * trace_getparent(t, 0) is the same as trace_getexpr(t).
 */
static struct tcexpr *trace_getparent(struct trace *t, unsigned height) {
   unsigned num, slot;

   num = tcexprarray_num(&t->exprs);
   PQLASSERT(num > 0);
   if (height >= num) {
      return NULL;
   }
   slot = num - height - 1;
   return tcexprarray_get(&t->exprs, slot);
}

/*
 * Insert an additional parent expression from the trace, *under* the
 * entry HEIGHT steps from the bottom. That is, if there are 3
 * elements [0, 1, 2], inserting at height 0 produces [0, 1, 2, N],
 * and inserting at height 1 produces [0, 1, N, 2]. Use the same height
 * as with trace_getparent() to insert *under* the expression retrieved
 * by the latter.
 */
static void trace_insertparent(struct defuse *ctx,
			       struct trace *t, unsigned height,
			       struct tcexpr *ins) {
   unsigned num, slot;

   num = tcexprarray_num(&t->exprs);
   PQLASSERT(num > 0);
   PQLASSERT(height <= num);

   slot = num - height;

   if (slot == num) {
      tcexprarray_add(ctx->pql, &t->exprs, ins, NULL);
   }
   else {
      tcexprarray_insert(ctx->pql, &t->exprs, slot);
      tcexprarray_set(&t->exprs, slot, ins);
   }
}

/*
 * Clear out a whole array of traces.
 */
static void tracearray_destroyall(struct defuse *ctx, struct tracearray *arr) {
   unsigned i, num;

   num = tracearray_num(arr);
   for (i=0; i<num; i++) {
      trace_destroy(ctx, tracearray_get(arr, i));
   }
   tracearray_setsize(ctx->pql, arr, 0);
}

////////////////////////////////////////////////////////////
// dvar methods

/*
 * Create a dvar.
 */
static struct dvar *dvar_create(struct defuse *ctx, struct tcvar *var) {
   struct dvar *d;
   
   d = domalloc(ctx->pql, sizeof(*d));
   d->var = var;
   d->def = NULL;
   tracearray_init(&d->uses);

   return d;
}

/*
 * Destroy a dvar.
 */
static void dvar_destroy(struct defuse *ctx, struct dvar *d) {
   tracearray_destroyall(ctx, &d->uses);
   tracearray_cleanup(ctx->pql, &d->uses);
   trace_destroy(ctx, d->def);
   dofree(ctx->pql, d, sizeof(*d));
}

////////////////////////////////////////////////////////////
// dcolref methods

/*
 * Create a dcolref that says the expression TE references something
 * in the manner HOW.
 */
static struct dcolref *dcolref_create(struct defuse *ctx, struct tcexpr *te,
				      enum dcolrefways how) {
   struct dcolref *dcr;

   dcr = domalloc(ctx->pql, sizeof(*dcr));
   dcr->isvar = false;
   dcr->expr = te;
   dcr->how = how;
   dcolrefarray_init(&dcr->from);
   dcolrefarray_init(&dcr->to);

   return dcr;
}

/*
 * Create a dcolref that says the variable VAR references something in
 * the manner HOW.
 */
static struct dcolref *dcolref_createvar(struct defuse *ctx, struct tcvar *var,
				      enum dcolrefways how) {
   struct dcolref *dcr;

   dcr = domalloc(ctx->pql, sizeof(*dcr));
   dcr->isvar = true;
   dcr->var = var;
   dcr->how = how;
   dcolrefarray_init(&dcr->from);
   dcolrefarray_init(&dcr->to);

   return dcr;
}

/*
 * Destroy a dcolref.
 */
static void dcolref_destroy(struct defuse *ctx, struct dcolref *dcr) {
   dcolrefarray_setsize(ctx->pql, &dcr->from, 0);
   dcolrefarray_setsize(ctx->pql, &dcr->to, 0);
   dcolrefarray_cleanup(ctx->pql, &dcr->from);
   dcolrefarray_cleanup(ctx->pql, &dcr->to);
   dofree(ctx->pql, dcr, sizeof(*dcr));
}

/*
 * Link two dcolrefs up.
 */
static void dcolref_link_prevthis(struct defuse *ctx,
				  struct dcolref *prev,
				  struct dcolref *this) {
   dcolrefarray_add(ctx->pql, &prev->to, this, NULL);
   dcolrefarray_add(ctx->pql, &this->from, prev, NULL);
}

/*
 * Clean out a whole array of dcolrefs.
 */
static void dcolrefarray_destroyall(struct defuse *ctx,
				    struct dcolrefarray *arr) {
   unsigned i, num;

   num = dcolrefarray_num(arr);
   for (i=0; i<num; i++) {
      dcolref_destroy(ctx, dcolrefarray_get(arr, i));
   }
   dcolrefarray_setsize(ctx->pql, arr, 0);
}

////////////////////////////////////////////////////////////
// dcol methods

/*
 * Create a dcol.
 */
static struct dcol *dcol_create(struct defuse *ctx, struct column *col) {
   struct dcol *dc;

   dc = domalloc(ctx->pql, sizeof(*dc));
   dc->col = col;
   dcolrefarray_init(&dc->defs);
   dcolrefarray_init(&dc->uses);
   dcolrefarray_init(&dc->all);

   return dc;
}

/*
 * Destroy a dcol.
 */
static void dcol_destroy(struct defuse *ctx, struct dcol *dc) {
   dcolrefarray_setsize(ctx->pql, &dc->defs, 0);
   dcolrefarray_setsize(ctx->pql, &dc->uses, 0);
   dcolrefarray_destroyall(ctx, &dc->all);

   dcolrefarray_cleanup(ctx->pql, &dc->defs);
   dcolrefarray_cleanup(ctx->pql, &dc->uses);
   dcolrefarray_cleanup(ctx->pql, &dc->all);

   dofree(ctx->pql, dc, sizeof(*dc));
}

/*
 * Add a newly created dcolref to a dcol.
 */
static void dcol_add_dcolref(struct defuse *ctx, struct dcol *dc,
			     struct dcolref *dcr) {
   dcolrefarray_add(ctx->pql, &dc->all, dcr, NULL);
   switch (dcr->how) {
    case CR_DEF: dcolrefarray_add(ctx->pql, &dc->defs, dcr, NULL); break;
    case CR_USE: dcolrefarray_add(ctx->pql, &dc->uses, dcr, NULL); break;
    case CR_PASS: break;
    case CR_END: /* XXX? */ break;
    case CR_STRIP: /* XXX? */ break;
   }
}

/*
 * Find the dcolref assocated with the expression TE in DC.
 */
static struct dcolref *dcol_find_expr(struct defuse *ctx, struct dcol *dc,
				      struct tcexpr *te) {
   struct dcolref *dcr;
   unsigned i, num;

   (void)ctx;

   num = dcolrefarray_num(&dc->all);
   for (i=0; i<num; i++) {
      dcr = dcolrefarray_get(&dc->all, i);
      if (!dcr->isvar && dcr->expr == te) {
	 return dcr;
      }
   }
   return NULL;
}

/*
 * Find the dcolref assocated with the variable VAR in DC.
 */
static struct dcolref *dcol_find_var(struct defuse *ctx, struct dcol *dc,
				     struct tcvar *var) {
   struct dcolref *dcr;
   unsigned i, num;

   (void)ctx;

   num = dcolrefarray_num(&dc->all);
   for (i=0; i<num; i++) {
      dcr = dcolrefarray_get(&dc->all, i);
      if (dcr->isvar && dcr->var == var) {
	 return dcr;
      }
   }
   return NULL;
}

////////////////////////////////////////////////////////////
// defuse methods

/*
 * Initialize the module context.
 */
static void defuse_init(struct defuse *ctx, struct pqlcontext *pql) {
   ctx->pql = pql;
   ctx->curtrace = NULL;
   dvararray_init(&ctx->vars);
   dcolarray_init(&ctx->cols);

   ctx->curtrace = trace_create(ctx);
}

/*
 * Clean up the module context.
 */
static void defuse_cleanup(struct defuse *ctx) {
   unsigned i, num;

   num = dcolarray_num(&ctx->cols);
   for (i=0; i<num; i++) {
      dcol_destroy(ctx, dcolarray_get(&ctx->cols, i));
   }
   dcolarray_setsize(ctx->pql, &ctx->cols, 0);

   num = dvararray_num(&ctx->vars);
   for (i=0; i<num; i++) {
      dvar_destroy(ctx, dvararray_get(&ctx->vars, i));
   }
   dvararray_setsize(ctx->pql, &ctx->vars, 0);

   trace_destroy(ctx, ctx->curtrace);
   ctx->curtrace = NULL;

   dcolarray_cleanup(ctx->pql, &ctx->cols);
   dvararray_cleanup(ctx->pql, &ctx->vars);
}

/*
 * Find the dvar for a given variable.
 */
static struct dvar *defuse_find_dvar(struct defuse *ctx, struct tcvar *var) {
   struct dvar *dv;
   unsigned i, num;

   num = dvararray_num(&ctx->vars);
   for (i=0; i<num; i++) {
      dv = dvararray_get(&ctx->vars, i);
      if (dv->var == var) {
	 return dv;
      }
   }
   PQLASSERT(0);
   return NULL;
}

/*
 * Add a new dvar for a variable.
 */
static void defuse_add_dvar(struct defuse *ctx, struct tcvar *var) {
   struct dvar *dv;

   dv = dvar_create(ctx, var);
   dv->def = trace_clone(ctx, ctx->curtrace);
   dvararray_add(ctx->pql, &ctx->vars, dv, NULL);
}

/*
 * Add a usage record for a variable.
 */
static void defuse_use_dvar(struct defuse *ctx, struct tcvar *var) {
   struct dvar *dv;
   struct trace *t;

   dv = defuse_find_dvar(ctx, var);
   t = trace_clone(ctx, ctx->curtrace);
   tracearray_add(ctx->pql, &dv->uses, t, NULL);
}

/*
 * Get the dcol for a column. If it doesn't exist yet, create it.
 */
static struct dcol *defuse_get_dcol(struct defuse *ctx, struct column *col) {
   struct dcol *dc;
   unsigned i, num;

   num = dcolarray_num(&ctx->cols);
   for (i=0; i<num; i++) {
      dc = dcolarray_get(&ctx->cols, i);
      if (dc->col == col) {
	 return dc;
      }
   }
   dc = dcol_create(ctx, col);
   dcolarray_add(ctx->pql, &ctx->cols, dc, NULL);
   return dc;
}

////////////////////////////////////////////////////////////
// column list tools

/*
 * Remove duplicates from DCOLS.
 */
static void colsuniq(struct defuse *ctx, struct dcolarray *dcols) {
   struct dcol *ci, *cj;
   unsigned i, j, num;

   (void)ctx;

   /* XXX should do this smarter */
   num = dcolarray_num(dcols);
   for (i=num; i-- > 1; ) {
      ci = dcolarray_get(dcols, i);
      for (j=i-1; j-- > 0; ) {
	 cj = dcolarray_get(dcols, j);
	 if (ci == cj) {
	    dcolarray_remove(dcols, j);
	    num--;
	    break;
	 }
      }
   }
}

/*
 * Remove the columns in TARGETS from the column array ARR.
 */
static void removecols(struct defuse *ctx, struct dcolarray *arr,
		       const struct dcolarray *targets) {
   struct dcol *cj;
   unsigned i, num;
   unsigned j, numtargets;
   unsigned *indexes;

   /* this should be smarter too */

   num = dcolarray_num(arr);
   numtargets = dcolarray_num(targets);
   indexes = domalloc(ctx->pql, numtargets * sizeof(*indexes));

   for (j=0; j<numtargets; j++) {
      cj = dcolarray_get(targets, j);
      indexes[j] = num;
      for (i=0; i<num; i++) {
	 if (dcolarray_get(arr, i) == cj) {
	    indexes[j] = i;
	 }
      }
      /* all columns had better be found */
      PQLASSERT(indexes[j] < num);
   }

   qsort(indexes, numtargets, sizeof(*indexes), unsigned_sortfunc);

   for (j=numtargets; j-- > 0; ) {
      dcolarray_remove(arr, indexes[j]);
   }

   dofree(ctx->pql, indexes, numtargets * sizeof(*indexes));
}

////////////////////////////////////////////////////////////
// column list collection

/*
 * Collect one column into DCOLS
 */
static void collectcols_one(struct defuse *ctx, struct dcolarray *dcols,
			    struct column *col) {
   struct dcol *dc;

   dc = defuse_get_dcol(ctx, col);
   dcolarray_add(ctx->pql, dcols, dc, NULL);
}

/*
 * Collect an array of columns into DCOLS
 */
static void collectcols_array(struct defuse *ctx, struct dcolarray *dcols,
			      struct columnarray *cols) {
   unsigned i, num;

   num = columnarray_num(cols);
   for (i=0; i<num; i++) {
      collectcols_one(ctx, dcols, columnarray_get(cols, i));
   }
}

/*
 * Collect columns from a type into DCOLS
 */
static void collectcols_type(struct defuse *ctx, struct dcolarray *dcols,
			     struct datatype *t) {
   if (datatype_isset(t)) {
      collectcols_type(ctx, dcols, datatype_set_member(t));
   }
   else if (datatype_issequence(t)) {
      collectcols_type(ctx, dcols, datatype_sequence_member(t));
   }
   else if (datatype_istuple(t)) {
      unsigned i, num;

      num = datatype_arity(t);
      for (i=0; i<num; i++) {
	 collectcols_type(ctx, dcols, datatype_getnth(t, i));
	 collectcols_one(ctx, dcols, datatype_tuple_getcolumn(t, i));
      }
   }
   else {
      /* primitive type; no columns */
   }
}

/*
 * Collect columns from a type and assert that the resulting list is
 * the same as DCOLS; that is, the columns are already there in the
 * same order.
 */
static void collectcols_type_already(struct defuse *ctx,
				     struct dcolarray *dcols,
				     struct datatype *t) {
   struct dcolarray dcols2;
   unsigned i, num;

   dcolarray_init(&dcols2);
   collectcols_type(ctx, &dcols2, t);

   num = dcolarray_num(&dcols2);
   PQLASSERT(num == dcolarray_num(dcols));
   for (i=0; i<num; i++) {
      PQLASSERT(dcolarray_get(&dcols2, i) == dcolarray_get(dcols, i));
   }

   dcolarray_setsize(ctx->pql, &dcols2, 0);
   dcolarray_cleanup(ctx->pql, &dcols2);
}

////////////////////////////////////////////////////////////
// Single column usage
// (this would all be a lot easier with some type polymorphism available)

/*
 * Add a reference to column DC from the expression TE in the manner
 * HOW.
 */
static void refercol(struct defuse *ctx, struct dcol *dc,
		     struct tcexpr *te, enum dcolrefways how) {
   dcol_add_dcolref(ctx, dc, dcolref_create(ctx, te, how));
}

/*
 * Add a reference to column DC from the variable VAR in the manner
 * HOW.
 */
static void refercol_var(struct defuse *ctx, struct dcol *dc,
			 struct tcvar *var, enum dcolrefways how) {
   dcol_add_dcolref(ctx, dc, dcolref_createvar(ctx, var, how));
}

/*
 * Link the references to the column DC made by expressions PREV and
 * THIS.
 */
static void linkcol(struct defuse *ctx, struct dcol *dc,
		    struct tcexpr *prev, struct tcexpr *this) {
   struct dcolref *prevref, *thisref;

   prevref = dcol_find_expr(ctx, dc, prev);
   PQLASSERT(prevref != NULL);

   thisref = dcol_find_expr(ctx, dc, this);
   PQLASSERT(thisref != NULL);

   dcolref_link_prevthis(ctx, prevref, thisref);
}

/*
 * Likewise, but don't hork if there's no reference from PREV.
 */
static void linkcol_missingok(struct defuse *ctx, struct dcol *dc,
			      struct tcexpr *prev, struct tcexpr *this) {
   struct dcolref *prevref, *thisref;

   prevref = dcol_find_expr(ctx, dc, prev);
   if (prevref == NULL) {
      return;
   }

   thisref = dcol_find_expr(ctx, dc, this);
   PQLASSERT(thisref != NULL);

   dcolref_link_prevthis(ctx, prevref, thisref);
}

/*
 * Likewise, but from a variable back to an expression.
 */
static void linkcol_exprvar(struct defuse *ctx, struct dcol *dc,
			    struct tcexpr *prev, struct tcvar *this) {
   struct dcolref *prevref, *thisref;

   prevref = dcol_find_expr(ctx, dc, prev);
   PQLASSERT(prevref != NULL);

   thisref = dcol_find_var(ctx, dc, this);
   PQLASSERT(thisref != NULL);

   dcolref_link_prevthis(ctx, prevref, thisref);
}

/*
 * Likewise, but from an expression back to a variable.
 */
static void linkcol_varexpr(struct defuse *ctx, struct dcol *dc,
			    struct tcvar *prev, struct tcexpr *this) {
   struct dcolref *prevref, *thisref;

   prevref = dcol_find_var(ctx, dc, prev);
   PQLASSERT(prevref != NULL);

   thisref = dcol_find_expr(ctx, dc, this);
   PQLASSERT(thisref != NULL);

   dcolref_link_prevthis(ctx, prevref, thisref);
}

////////////////////////////////////////////////////////////
// Column list usage, basic operations.
// (this would all be a lot easier with some type polymorphism available)

/*
 * Add a reference to each column of DCOLS from the expression TE in
 * the manner HOW.
 */
static void refercols(struct defuse *ctx, struct dcolarray *dcols,
		      struct tcexpr *te, enum dcolrefways how) {
   struct dcol *dc;
   unsigned i, num;

   num = dcolarray_num(dcols);
   for (i=0; i<num; i++) {
      dc = dcolarray_get(dcols, i);
      refercol(ctx, dc, te, how);
   }
}

/*
 * Same, but from a variable.
 */
static void refercols_var(struct defuse *ctx, struct dcolarray *dcols,
			  struct tcvar *var, enum dcolrefways how) {
   struct dcol *dc;
   unsigned i, num;

   num = dcolarray_num(dcols);
   for (i=0; i<num; i++) {
      dc = dcolarray_get(dcols, i);
      refercol_var(ctx, dc, var, how);
   }
}

/*
 * For each column in DCOLS, link the column references for THIS back
 * to the one for PREV.
 */
static void linkcols(struct defuse *ctx, struct dcolarray *dcols,
		     struct tcexpr *prev, struct tcexpr *this) {
   struct dcol *dc;
   unsigned i, num;

   num = dcolarray_num(dcols);
   for (i=0; i<num; i++) {
      dc = dcolarray_get(dcols, i);
      linkcol(ctx, dc, prev, this);
   }
}

/*
 * Likewise, but don't hork if PREV doesn't actually have a reference.
 */
static void linkcols_missingok(struct defuse *ctx, struct dcolarray *dcols,
			       struct tcexpr *prev, struct tcexpr *this) {
   struct dcol *dc;
   unsigned i, num;

   num = dcolarray_num(dcols);
   for (i=0; i<num; i++) {
      dc = dcolarray_get(dcols, i);
      linkcol_missingok(ctx, dc, prev, this);
   }
}

/*
 * Likewise, but from var to expression.
 */
static void linkcols_exprvar(struct defuse *ctx, struct dcolarray *dcols,
			     struct tcexpr *prev, struct tcvar *this) {
   struct dcol *dc;
   unsigned i, num;

   num = dcolarray_num(dcols);
   for (i=0; i<num; i++) {
      dc = dcolarray_get(dcols, i);
      linkcol_exprvar(ctx, dc, prev, this);
   }
}

/*
 * Likewise, but from expression to var.
 */
static void linkcols_varexpr(struct defuse *ctx, struct dcolarray *dcols,
			     struct tcvar *prev, struct tcexpr *this) {
   struct dcol *dc;
   unsigned i, num;

   num = dcolarray_num(dcols);
   for (i=0; i<num; i++) {
      dc = dcolarray_get(dcols, i);
      linkcol_varexpr(ctx, dc, prev, this);
   }
}

/*
 * Clear out a column list because we've finished handling it.
 */
static void donecols(struct defuse *ctx, struct dcolarray *dcols) {
   dcolarray_setsize(ctx->pql, dcols, 0);
}

////////////////////////////////////////////////////////////
// Column list usage, compound operations.
// (this would all be a lot easier with some type polymorphism available)

/*
 * Declare that the expression THIS defines some columns DCOLS.
 */
static void defcols(struct defuse *ctx, struct dcolarray *dcols,
		    struct tcexpr *this) {
   refercols(ctx, dcols, this, CR_END);
   donecols(ctx, dcols);
}

/*
 * Declare that the expression THIS passes through some columns DCOLS
 * from PREV.
 */
static void passcols(struct defuse *ctx, struct dcolarray *dcols,
		     struct tcexpr *prev, struct tcexpr *this) {
   refercols(ctx, dcols, this, CR_PASS);
   linkcols(ctx, dcols, prev, this);
   donecols(ctx, dcols);
}

/*
 * Likewise, but from an expression to a variable.
 */
static void passcolsvar(struct defuse *ctx, struct dcolarray *dcols,
			struct tcexpr *prev, struct tcvar *this) {
   
   refercols_var(ctx, dcols, this, CR_PASS);
   linkcols_exprvar(ctx, dcols, prev, this);
   donecols(ctx, dcols);
}

/*
 * Likewise, but from a variable to an expression.
 */
static void passvarcols(struct defuse *ctx, struct dcolarray *dcols,
			struct tcvar *prev, struct tcexpr *this) {
   refercols(ctx, dcols, this, CR_PASS);
   linkcols_varexpr(ctx, dcols, prev, this);
   donecols(ctx, dcols);
}

/*
 * Declare that the expression THIS uses some columns DCOLS from PREV
 * and also passes them on.
 */
static void usecols(struct defuse *ctx, struct dcolarray *dcols,
		    struct tcexpr *prev, struct tcexpr *this) {
   refercols(ctx, dcols, this, CR_USE);
   linkcols(ctx, dcols, prev, this);
   donecols(ctx, dcols);
}

/*
 * Declare that the expression THIS uses some columns DCOLS from PREV,
 * and does not pass them on.
 */
static void endcols(struct defuse *ctx, struct dcolarray *dcols,
		    struct tcexpr *prev, struct tcexpr *this) {
   refercols(ctx, dcols, this, CR_END);
   linkcols(ctx, dcols, prev, this);
   donecols(ctx, dcols);
}

////////////////////////////////////////////////////////////
// Compound-compound column usage.
// (this would all be a lot easier with some type polymorphism available)


/*
 * Like passcols, but link back to two ancestor expressions.
 * (e.g. for UNION.)
 */
static void doublepasscols(struct defuse *ctx, struct dcolarray *dcols,
			   struct tcexpr *prev1, struct tcexpr *prev2,
			   struct tcexpr *this) {
   refercols(ctx, dcols, this, CR_PASS);
   linkcols(ctx, dcols, prev1, this);
   linkcols(ctx, dcols, prev2, this);
   donecols(ctx, dcols);
}

/*
 * Like endcols, but link back to two ancestor expressions.
 * (e.g. for SUBSETOF)
 */
static void doubleendcols(struct defuse *ctx, struct dcolarray *dcols,
			  struct tcexpr *prev1, struct tcexpr *prev2,
			  struct tcexpr *this) {
   refercols(ctx, dcols, this, CR_END);
   linkcols(ctx, dcols, prev1, this);
   linkcols(ctx, dcols, prev2, this);
   donecols(ctx, dcols);
}

/*
 * Like passcols, but link back to a whole pile of ancestor
 * expressions, some of which may not actually need the links.
 * For TCE_FUNC invocations.
 */
static void multipasscols(struct defuse *ctx, struct dcolarray *dcols,
			  struct tcexprarray *prevs, struct tcexpr *this) {
   struct tcexpr *prev;
   unsigned i, num;

   refercols(ctx, dcols, this, CR_PASS);
   num = tcexprarray_num(prevs);
   for (i=0; i<num; i++) {
      prev = tcexprarray_get(prevs, i);
      linkcols_missingok(ctx, dcols, prev, this);
   }
   donecols(ctx, dcols);
}

/*
 * Like endcols, but link back to a whole pile of ancestor
 * expressions, some of which may not actually need the links.
 * For TCE_FUNC invocations.
 */
static void multiendcols(struct defuse *ctx, struct dcolarray *dcols,
			 struct tcexprarray *prevs, struct tcexpr *this) {
   struct tcexpr *prev;
   unsigned i, num;

   refercols(ctx, dcols, this, CR_END);
   num = tcexprarray_num(prevs);
   for (i=0; i<num; i++) {
      prev = tcexprarray_get(prevs, i);
      linkcols_missingok(ctx, dcols, prev, this);
   }
   donecols(ctx, dcols);
}

////////////////////////////////////////////////////////////
// Compound and compound-compound column usage of *some* columns.
// (this would all be a lot easier with some type polymorphism available)


/*
 * Like defcols but define a single column COL, to be taken out of
 * DCOLS.
 */
static void defonecol(struct defuse *ctx, struct dcolarray *dcols,
		      struct column *col,
		      struct tcexpr *this) {
   struct dcolarray dcols2;
   
   dcolarray_init(&dcols2);
   collectcols_one(ctx, &dcols2, col);
   removecols(ctx, dcols, &dcols2);
   defcols(ctx, &dcols2, this);
   dcolarray_cleanup(ctx->pql, &dcols2);
}

/*
 * Like usecols but use a single column COL, to be taken out of DCOLS.
 */
static void useonecol(struct defuse *ctx, struct dcolarray *dcols,
		      struct column *col,
		      struct tcexpr *prev, struct tcexpr *this) {
   struct dcolarray dcols2;
   
   dcolarray_init(&dcols2);
   collectcols_one(ctx, &dcols2, col);
   removecols(ctx, dcols, &dcols2);
   usecols(ctx, &dcols2, prev, this);
   dcolarray_cleanup(ctx->pql, &dcols2);
}

/*
 * Like usecols but use the columns in COLS and take them out of
 * DCOLS.
 */
static void usesomecols(struct defuse *ctx, struct dcolarray *dcols,
			struct columnarray *cols,
			struct tcexpr *prev, struct tcexpr *this) {
   struct dcolarray dcols2;
   
   dcolarray_init(&dcols2);
   collectcols_array(ctx, &dcols2, cols);
   removecols(ctx, dcols, &dcols2);
   usecols(ctx, &dcols2, prev, this);
   dcolarray_cleanup(ctx->pql, &dcols2);
}

/*
 * Like passcols but use the columns found in the type T and take them
 * out of DCOLS.
 */
static void passtypecols(struct defuse *ctx, struct dcolarray *dcols,
			 struct datatype *t,
			 struct tcexpr *prev, struct tcexpr *this) {
   struct dcolarray dcols2;
   
   dcolarray_init(&dcols2);
   collectcols_type(ctx, &dcols2, t);
   removecols(ctx, dcols, &dcols2);
   passcols(ctx, &dcols2, prev, this);
   dcolarray_cleanup(ctx->pql, &dcols2);
}

/*
 * Like doublepasscols but use the columns found in the type T and
 * take them out of DCOLS.
 */
static void doublepasstypecols(struct defuse *ctx, struct dcolarray *dcols,
			       struct datatype *t,
			       struct tcexpr *prev1, struct tcexpr *prev2,
			       struct tcexpr *this) {
   struct dcolarray dcols2;
   
   dcolarray_init(&dcols2);
   collectcols_type(ctx, &dcols2, t);
   removecols(ctx, dcols, &dcols2);
   doublepasscols(ctx, &dcols2, prev1, prev2, this);
   dcolarray_cleanup(ctx->pql, &dcols2);
}

/*
 * Like multipasscols but use the columns found in the type T and take
 * them out of DCOLS.
 */
static void multipasstypecols(struct defuse *ctx, struct dcolarray *dcols,
			      struct datatype *t,
			      struct tcexprarray *prevs, struct tcexpr *this) {
   struct dcolarray dcols2;
   
   dcolarray_init(&dcols2);
   collectcols_type(ctx, &dcols2, t);
   removecols(ctx, dcols, &dcols2);
   multipasscols(ctx, &dcols2, prevs, this);
   dcolarray_cleanup(ctx->pql, &dcols2);
}

////////////////////////////////////////////////////////////
// recursive build

/*
 * Build the chains and graphs for expression TE.
 *
 * In order to be able to manipulate vars usefully we keep track of
 * where each one entry appears; this is a "trace" and is the chain
 * of expressions back up to the tree root. This function maintains
 * the "current trace" in the context by pushing and popping the
 * passed expression.
 *
 * The structures for columns are a lot more complicated than that;
 * because columns appear in the types, everywhere a column goes we
 * have to keep track of it anyway, so we don't need traces. I think.
 *
 * Each distinct column gets its own def-use graph, but each
 * expression node must appear only once in each such graph.
 * (Otherwise it's too hard to make the linking work.) This means that
 * for each node we need to collect up the full list of columns it
 * touches and the possibly different ways each one is touched. The
 * primitives for doing this are (lengthily) above. The basic idea is
 * that involved columns are collected into an array COLS, and then
 * removed again by calling passcols() or whatever. COLS must be empty
 * at the bottom of the function or dcolarray_cleanup will assert.
 */
static void tcexpr_dbuild(struct defuse *ctx, struct tcexpr *te) {
   struct dcolarray cols;

   dcolarray_init(&cols);
   trace_push(ctx, ctx->curtrace, te);

   switch (te->type) {

    case TCE_FILTER:
     tcexpr_dbuild(ctx, te->filter.sub);
     collectcols_type(ctx, &cols, te->datatype);
     passcols(ctx, &cols, te->filter.sub, te);

     defuse_add_dvar(ctx, te->filter.predvar);
     collectcols_type(ctx, &cols, te->datatype);
     passcolsvar(ctx, &cols, te, te->filter.predvar);
     tcexpr_dbuild(ctx, te->filter.predexpr);
     break;

    case TCE_PROJECTTUPLE:
     tcexpr_dbuild(ctx, te->projecttuple.sub);
     collectcols_array(ctx, &cols, te->projecttuple.cols);
     usecols(ctx, &cols, te->projecttuple.sub, te);
     break;

    case TCE_PROJECTSCALAR:
     tcexpr_dbuild(ctx, te->projectscalar.sub);
     collectcols_one(ctx, &cols, te->projectscalar.col);
     usecols(ctx, &cols, te->projectscalar.sub, te);
     break;

    case TCE_RENAME:
     /* should not appear here */
     PQLASSERT(0);
     break;

    case TCE_JOIN:
     tcexpr_dbuild(ctx, te->join.left);
     tcexpr_dbuild(ctx, te->join.right);
     collectcols_type(ctx, &cols, te->join.left->datatype);
     passcols(ctx, &cols, te->join.left, te);
     collectcols_type(ctx, &cols, te->join.right->datatype);
     passcols(ctx, &cols, te->join.right, te);
     if (te->join.predexpr != NULL && te->join.predvar != NULL) {
	defuse_add_dvar(ctx, te->join.predvar);
	collectcols_type(ctx, &cols, te->datatype);
	passcolsvar(ctx, &cols, te, te->join.predvar);
	tcexpr_dbuild(ctx, te->join.predexpr);
     }
     break;

    case TCE_ORDER:
     tcexpr_dbuild(ctx, te->order.sub);
     collectcols_type(ctx, &cols, te->datatype);
     usesomecols(ctx, &cols, te->order.cols, te->order.sub, te);
     passcols(ctx, &cols, te->order.sub, te);
     break;

    case TCE_UNIQ:
     tcexpr_dbuild(ctx, te->uniq.sub);
     collectcols_type(ctx, &cols, te->datatype);
     usesomecols(ctx, &cols, te->uniq.cols, te->uniq.sub, te);
     passcols(ctx, &cols, te->uniq.sub, te);
     break;

    case TCE_NEST:
     tcexpr_dbuild(ctx, te->nest.sub);
     collectcols_type(ctx, &cols, te->datatype);
     defonecol(ctx, &cols, te->nest.newcol, te);
     usesomecols(ctx, &cols, te->nest.cols, te->nest.sub, te);
     passcols(ctx, &cols, te->nest.sub, te);
     break;

    case TCE_UNNEST:
     tcexpr_dbuild(ctx, te->unnest.sub);
     collectcols_type(ctx, &cols, te->datatype);
     if (datatype_has_column(te->datatype, te->unnest.col)) {
	/* unnesting a scalar column; the column name is preserved */
	useonecol(ctx, &cols, te->unnest.col, te->unnest.sub, te);
	passcols(ctx, &cols, te->unnest.sub, te);
     }
     else {
	/* the unnest column goes away */
	passcols(ctx, &cols, te->unnest.sub, te);
	collectcols_one(ctx, &cols, te->unnest.col);
	endcols(ctx, &cols, te->unnest.sub, te);
     }
     break;

    case TCE_DISTINGUISH:
     tcexpr_dbuild(ctx, te->distinguish.sub);
     collectcols_type(ctx, &cols, te->datatype);
     defonecol(ctx, &cols, te->distinguish.newcol, te);
     passcols(ctx, &cols, te->distinguish.sub, te);
     break;

    case TCE_ADJOIN:
     tcexpr_dbuild(ctx, te->adjoin.left);

     defuse_add_dvar(ctx, te->adjoin.var);
     collectcols_type(ctx, &cols, te->adjoin.left->datatype);
     passcolsvar(ctx, &cols, te->adjoin.left, te->adjoin.var);

     tcexpr_dbuild(ctx, te->adjoin.expr);
     collectcols_type(ctx, &cols, te->adjoin.expr->datatype);
     passcols(ctx, &cols, te->adjoin.expr, te);

     collectcols_type(ctx, &cols, te->datatype);
     defonecol(ctx, &cols, te->adjoin.newcol, te);
     passcols(ctx, &cols, te->adjoin.left, te);
     break;

    case TCE_STEP:
     /* these don't exist yet */
     PQLASSERT(0);
     break;

    case TCE_REPEAT:
     tcexpr_dbuild(ctx, te->repeat.sub);

     collectcols_type(ctx, &cols, te->repeat.sub->datatype);
     useonecol(ctx, &cols, te->repeat.subendcolumn, te->repeat.sub, te);
     passcols(ctx, &cols, te->repeat.sub, te);

     defuse_add_dvar(ctx, te->repeat.loopvar);

     collectcols_one(ctx, &cols, te->repeat.bodystartcolumn);
     defcols(ctx, &cols, te);

     collectcols_one(ctx, &cols, te->repeat.bodystartcolumn);
     passcolsvar(ctx, &cols, te, te->repeat.loopvar);

     tcexpr_dbuild(ctx, te->repeat.body);

     if (te->repeat.bodybodycolumn != NULL) {
	collectcols_one(ctx, &cols, te->repeat.bodybodycolumn);
     }
     if (te->repeat.bodypathcolumn != NULL) {
	collectcols_one(ctx, &cols, te->repeat.bodypathcolumn);
     }
     collectcols_one(ctx, &cols, te->repeat.bodyendcolumn);
     usecols(ctx, &cols, te->repeat.body, te);

     collectcols_one(ctx, &cols, te->repeat.repeatpathcolumn);
     if (te->repeat.repeatendcolumn != NULL) {
	collectcols_one(ctx, &cols, te->repeat.repeatendcolumn);
     }
     defcols(ctx, &cols, te);

     break;

    case TCE_SCAN:
     collectcols_one(ctx, &cols, te->scan.leftobjcolumn);
     collectcols_one(ctx, &cols, te->scan.edgecolumn);
     collectcols_one(ctx, &cols, te->scan.rightobjcolumn);
     defcols(ctx, &cols, te);
     break;

    case TCE_BOP:
     tcexpr_dbuild(ctx, te->bop.left);
     tcexpr_dbuild(ctx, te->bop.right);
     collectcols_type(ctx, &cols, te->bop.left->datatype);
     collectcols_type_already(ctx, &cols, te->bop.right->datatype);
     doublepasstypecols(ctx, &cols, te->datatype,
			te->bop.left, te->bop.right, te);
     doubleendcols(ctx, &cols, te->bop.left, te->bop.right, te);
     break;

    case TCE_UOP:
     tcexpr_dbuild(ctx, te->uop.sub);
     collectcols_type(ctx, &cols, te->uop.sub->datatype);
     passtypecols(ctx, &cols, te->datatype, te->uop.sub, te);
     endcols(ctx, &cols, te->uop.sub, te);
     break;

    case TCE_FUNC:
     {
	struct tcexpr *tcsub;
	unsigned i, num;

	num = tcexprarray_num(&te->func.args);
	for (i=0; i<num; i++) {
	   tcsub = tcexprarray_get(&te->func.args, i);
	   tcexpr_dbuild(ctx, tcsub);
	}

	for (i=0; i<num; i++) {
	   tcsub = tcexprarray_get(&te->func.args, i);
	   collectcols_type(ctx, &cols, tcsub->datatype);
	}
	colsuniq(ctx, &cols);

	multipasstypecols(ctx, &cols, te->datatype, &te->func.args, te);
	multiendcols(ctx, &cols, &te->func.args, te);
     }
     break;

    case TCE_MAP:
     tcexpr_dbuild(ctx, te->map.set);
     defuse_add_dvar(ctx, te->map.var);
     collectcols_type(ctx, &cols, te->map.set->datatype);
     passcolsvar(ctx, &cols, te->map.set, te->map.var);

     tcexpr_dbuild(ctx, te->map.result);
     collectcols_type(ctx, &cols, te->datatype);
     passcols(ctx, &cols, te->map.result, te);
     break;

    case TCE_LET:
     tcexpr_dbuild(ctx, te->let.value);
     defuse_add_dvar(ctx, te->let.var);
     collectcols_type(ctx, &cols, te->let.value->datatype);
     passcolsvar(ctx, &cols, te->let.value, te->let.var);

     tcexpr_dbuild(ctx, te->let.body);
     collectcols_type(ctx, &cols, te->datatype);
     passcols(ctx, &cols, te->let.body, te);
     break;

    case TCE_READVAR:
     defuse_use_dvar(ctx, te->readvar);
     collectcols_type(ctx, &cols, te->readvar->datatype);
     passvarcols(ctx, &cols, te->readvar, te);
     break;

    case TCE_READGLOBAL:
     break;

    case TCE_CREATEPATHELEMENT:
     tcexpr_dbuild(ctx, te->createpathelement);
     collectcols_type(ctx, &cols, te->createpathelement->datatype);
     passcols(ctx, &cols, te->createpathelement, te);
     break;

    case TCE_SPLATTER:
     tcexpr_dbuild(ctx, te->splatter.value);
     tcexpr_dbuild(ctx, te->splatter.name);
     /* XXX */
     collectcols_type(ctx, &cols, te->splatter.value->datatype);
     passcols(ctx, &cols, te->splatter.value, te);
     break;

    case TCE_TUPLE:
     {
	struct tcexpr *tcsub;
	unsigned i, num;

	num = tcexprarray_num(&te->tuple.exprs);
	for (i=0; i<num; i++) {
	   tcsub = tcexprarray_get(&te->tuple.exprs, i);
	   tcexpr_dbuild(ctx, tcsub);
	}

	for (i=0; i<num; i++) {
	   tcsub = tcexprarray_get(&te->tuple.exprs, i);
	   collectcols_type(ctx, &cols, tcsub->datatype);
	   passcols(ctx, &cols, tcsub, te);
	}

	collectcols_array(ctx, &cols, te->tuple.columns);
	defcols(ctx, &cols, te);
     }
     break;

    case TCE_VALUE:
     break;
   }

   trace_pop(ctx, ctx->curtrace);
   dcolarray_cleanup(ctx->pql, &cols);
}

static void tcstmt_dbuild(struct defuse *ctx, struct tcstmt *ts) {
   switch (ts->type) {
    case TCS_EXPR:
     tcexpr_dbuild(ctx, ts->expr);
     break;
    case TCS_BEGIN:
    case TCS_ABORT:
    case TCS_COMMIT:
     break;
   }
}

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
//                                                        //
//                   T R A N S F O R M S                  //
//                                                        //
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// combine projections across variable bindings

/*
 * That is, given let X = project C1..Cn (E) and usages of the form
 * project Ck (X), get rid of the projection in the definition of X.
 *
 * We can also do this with map.
 */

static void dprojcombine_one(struct defuse *ctx, struct dvar *dv) {
   struct trace *deftrace, *usetrace;
   struct tcexpr *def, *use;
   struct tcexpr *defsub, *defsubsub, *ins, *usesub;
   unsigned numuses, i;

   deftrace = dv->def;
   def = trace_getexpr(deftrace);
   if (def->type != TCE_LET && def->type != TCE_MAP) {
      return;
   }

   if (def->type == TCE_LET) {
      defsub = def->let.value;
   }
   else {
      defsub = def->map.set;
   }

   if (defsub->type != TCE_PROJECTTUPLE) {
      return;
   }

   numuses = tracearray_num(&dv->uses);
   for (i=0; i<numuses; i++) {
      usetrace = tracearray_get(&dv->uses, i);
      use = trace_getparent(usetrace, 1);
      if (use == NULL ||
	  (use->type != TCE_PROJECTSCALAR &&
	   use->type != TCE_PROJECTTUPLE)) {
	 return;
      }
   }

   /*
    * Every use is a projection. Push the definition's projection to
    * the uses; baseopt will combine them later.
    */
   defsubsub = defsub->projecttuple.sub;
   defsub->projecttuple.sub = NULL;

   if (def->type == TCE_LET) {
      PQLASSERT(def->let.var == dv->var);
      def->let.value = defsubsub;
      def->let.var->datatype = defsubsub->datatype;
   }
   else {
      PQLASSERT(def->map.var == dv->var);
      def->map.set = defsubsub;
      def->map.var->datatype = datatype_set_member(defsubsub->datatype);

      /* we're moving this inside the map, so it becomes scalar */
      defsub->datatype = datatype_set_member(defsub->datatype);
   }

   for (i=0; i<numuses; i++) {
      usetrace = tracearray_get(&dv->uses, i);
      use = trace_getparent(usetrace, 1);
      PQLASSERT(use != NULL);

      ins = tcexpr_clone(ctx->pql, defsub);
      trace_insertparent(ctx, usetrace, 1, ins);

      if (use->type == TCE_PROJECTSCALAR) {
	 usesub = use->projectscalar.sub;
	 use->projectscalar.sub = ins;
      }
      else /*if (use->type == TCE_PROJECTTUPLE)*/ {
	 usesub = use->projecttuple.sub;
	 use->projecttuple.sub = ins;
      }

      ins->projecttuple.sub = usesub;
      usesub->datatype = dv->var->datatype;
   }

   tcexpr_destroy(ctx->pql, defsub);
}

static void dprojcombine(struct defuse *ctx) {
   unsigned i, numvars;

   numvars = dvararray_num(&ctx->vars);
   for (i=0; i<numvars; i++) {
      dprojcombine_one(ctx, dvararray_get(&ctx->vars, i));
   }
}

////////////////////////////////////////////////////////////
// entry point

struct tcstmt *defuse(struct pqlcontext *pql, struct tcstmt *ts) {
   struct defuse ctx;

   defuse_init(&ctx, pql);
   tcstmt_dbuild(&ctx, ts);

   dprojcombine(&ctx);

   defuse_cleanup(&ctx);

   return ts;
}
