/*
 * Copyright (c) 2010 David A. Holland.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Author nor the names of any contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef INTARRAY_H
#define INTARRAY_H

#include <string.h>

#include "inlinedefs.h"
#include "memdefs.h"

#define INTARRAYS_CHECKED

#ifdef INTARRAYS_CHECKED
#include "utils.h"
#define intarrayassert PQLASSERT
#else
#define intarrayassert(x) ((void)(x))
#endif

////////////////////////////////////////////////////////////
// type and ops

#define DECLINTARRAY(T, N) \
   struct N##array {							\
      T *v;								\
      unsigned num, max;						\
   };									\
									\
   struct N##array *N##array_create(struct pqlcontext *);		\
   void N##array_destroy(struct pqlcontext *, struct N##array *);	\
   void N##array_init(struct N##array *);				\
   void N##array_cleanup(struct pqlcontext *, struct N##array *);	\
   unsigned N##array_num(const struct N##array *);			\
   T N##array_get(const struct N##array *, unsigned index);		\
   void N##array_set(struct N##array *, unsigned index, T);		\
   int N##array_setsize(struct pqlcontext *, struct N##array *, unsigned num);\
   int N##array_add(struct pqlcontext *, struct N##array *, T,		\
			unsigned *index_ret);				\
   int N##array_insert(struct pqlcontext *, struct N##array *,	 	\
			unsigned index);				\
   void N##array_remove(struct N##array *, unsigned index)

////////////////////////////////////////////////////////////
// inlining for operations

#define DEFINTARRAY(T, N, INLINE) \
   INLINE void N##array_init(struct N##array *a) {			\
      a->num = a->max = 0;						\
      a->v = NULL;							\
   }									\
									\
   INLINE void N##array_cleanup(struct pqlcontext *pql, struct N##array *a) { \
      intarrayassert(a->num == 0);					\
      dofree(pql, a->v, a->max * sizeof(a->v[0]));			\
   /*#ifdef ARRAYS_CHECKED*/						\
      a->v = NULL;							\
   /*#endif*/								\
   }									\
									\
   INLINE struct N##array *N##array_create(struct pqlcontext *pql) {	\
      struct N##array *a = domalloc(pql, sizeof(*a));			\
      if (a == NULL) {							\
         return NULL;							\
      }									\
      N##array_init(a);							\
      return a;								\
   }									\
									\
   INLINE void N##array_destroy(struct pqlcontext *pql, struct N##array *a) {\
      N##array_cleanup(pql, a);						\
      dofree(pql, a, sizeof(*a));					\
   }									\
									\
   INLINE unsigned N##array_num(const struct N##array *a) {		\
      return a->num;							\
   }									\
									\
   INLINE T N##array_get(const struct N##array *a, unsigned index) {	\
      intarrayassert(index < a->num);					\
      return a->v[index];						\
   }									\
									\
   INLINE void N##array_set(struct N##array *a, unsigned index, T val) { \
      intarrayassert(index < a->num);					\
      a->v[index] = val;						\
   }									\
									\
   INLINE int N##array_setsize(struct pqlcontext *pql, struct N##array *a, \
				unsigned num) { 			\
      void *np = a->v;							\
      if (intarraysupport_setsize(pql, &np, sizeof(T), &a->num, &a->max,num)){\
         return -1;							\
      }									\
      a->v = np;							\
      return 0;								\
   }									\
									\
   INLINE int N##array_add(struct pqlcontext *pql, struct N##array *a, 	\
				T val, unsigned *index_ret) { 		\
      unsigned index = a->num;						\
      if (N##array_setsize(pql, a, index+1)) {				\
         return -1;							\
      }									\
      a->v[index] = val;						\
      if (index_ret != NULL) {						\
	 *index_ret = index;						\
      }									\
      return 0;								\
   }									\
									\
   INLINE int N##array_insert(struct pqlcontext *pql, struct N##array *a, \
				unsigned index) { 			\
      intarrayassert(index < a->num);					\
      unsigned movers = a->num - index;					\
      if (N##array_setsize(pql, a, a->num + 1)) {			\
	 return -1;							\
      }									\
      memmove(a->v + index+1, a->v + index, movers*sizeof(*a->v));	\
      return 0;								\
   }									\
									\
   INLINE void N##array_remove(struct N##array *a, unsigned index) {	\
      intarrayassert(index < a->num);					\
      unsigned movers = a->num - (index + 1);				\
      memmove(a->v + index, a->v + index+1, movers*sizeof(*a->v));	\
      a->num--;								\
   }


////////////////////////////////////////////////////////////
// support functions

int intarraysupport_setsize(struct pqlcontext *pql, void **ptr, size_t sz,
			    unsigned *num, unsigned *max, unsigned newnum);


////////////////////////////////////////////////////////////
// some intarray types

/*
 * Usage:
 *
 * DECLINTARRAY(type, name) declares "struct namearray", which is
 * an array of the integer (or at least base) type "type", plus the
 * operations on it.
 *
 * DEFINTARRAY is the same as DECLINTARRAY except that it defines the
 * operations, and takes an extra argument INLINE. For C99 this
 * should be INLINE in header files and empty in the master source
 * file, the same as elsewhere.
 *
 * Example usage:
 * 
 * DECLINTARRAY(unsigned long, ulong);
 *
 * #ifndef INTARRAYINLINE
 * #define INTARRAYINLINE INLINE
 * #endif
 *
 * DEFINTARRAY(unsigned long, ulong, MYINLINE)
 *
 * intarray.c does "#define INTARRAYINLINE" before including intarray.h.
 */

#ifndef INTARRAYINLINE
#define INTARRAYINLINE INLINE
#endif

DECLINTARRAY(int, int);
DEFINTARRAY(int, int, INTARRAYINLINE);

/* indexarray is an array of array indexes (unsigned) */
DECLINTARRAY(unsigned, index);
DEFINTARRAY(unsigned, index, INTARRAYINLINE);


#endif /* INTINTARRAYBASE_H */
