/*
 * Copyright 2010
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Join order planner.
 *
 * This differs slightly from an ordinary planner in that we don't
 * bother to consider "scan" as an access method; since we require the
 * backend to provide step lookup and all queries are rooted, if no
 * indexes match we'll always start from the root(s) and step
 * outwards. It is possible that there might be cases where a full
 * scan is better, e.g. a very selective predicate out at the end of
 * an otherwise very unselective path. If we find these cases in any
 * significant quantity we'll revisit the design.
 *
 * The basic method is to first build a graph representing all the
 * data sources in the query (scans and readglobals) as nodes and all
 * the relationships among them (distilled from filter predicates) as
 * edges. Then we choose a covering path through this graph: first we
 * look for possible *roots* (indexes and readglobals) and then look
 * for *routes*. so to speak.
 *
 * Note that it's not really a graph, formally, because a predicate
 * like "X + Y + Z > 3" becomes an "edge" that connects three nodes,
 * not two.
 */

#include <stdlib.h> /* for qsort() */

#include "array.h"
#include "datatype.h"
#include "tcalc.h"
#include "tcutil.h"
#include "passes.h"

/*
 * some array types
 */
struct jtail;
struct jpred;
struct jplan;
struct jwork;
struct jexpr;
DECLARRAY(jtail);
DECLARRAY(jpred);
DECLARRAY(jplan);
DECLARRAY(jwork);
DECLARRAY(jexpr);
DEFARRAY(jtail, );
DEFARRAY(jpred, );
DEFARRAY(jplan, );
DEFARRAY(jwork, );
DEFARRAY(jexpr, );

/*
 * Type for cost estimation.
 */
struct cost {
   unsigned long long height;	/* estimated number of tuples */
   unsigned width;		/* tuple arity */
   unsigned long long work;	/* number of work units to generate */
};

/*
 * Structure for a leaf of an expression: a sub-jexpr, plus the
 * pointer to the corresponding slot in the tcexpr tree.
 */
struct jtail {
   bool isarray;
   union {
      struct tcexpr **slot;
      struct {
	 struct tcexprarray *arr;
	 unsigned ix;
      } array;
   };
   struct jexpr *sub;
};

/*
 * Predicate relating two or more jexprs in a join.
 */
struct jpred {
   struct tcvar *var;		/* variable for predicate */
   struct tcexpr *expr;		/* predicate expression */
   struct indexarray sources;	/* the sources we relate */
};

/*
 * Kinds of join we know.
 * XXX this should get larger.
 */
enum joinmethods {
   J_BASIC,			/* TCE_JOIN */
   J_STEP,			/* TCE_STEP */
   J_REVSTEP,			/* TCE_STEP with reverse flag */
};

/*
 * Plan component. Each non-leaf plan component specifies two things
 * to join, how to do it, and which columns need to be produced in the
 * result.
 */
struct jplan {
   bool isleaf;
   union {
      struct {
	 unsigned sourceix;			/* to je->join.sources[] */
      } leaf;
      struct {
	 struct jplan *left;
	 struct jplan *right;
	 enum joinmethods method;
	 struct indexarray predicates;		/* to je->join.predicates[] */
      } join;
   };
   struct cost cost;
   struct columnarray producecols;
};

/*
 * Potential plan component. Not clear if this should really be
 * different from jplan... but for now it is.
 */
enum jworktypes {
   JW_LEAF,
   JW_STEP,
   JW_JOIN,
};
struct jwork {
   struct cost cost;
   enum jworktypes type;
   union {
      struct {
	 unsigned sourceix;
      } leaf;
      struct {
	 unsigned leftplanix;
	 unsigned rightsourceix;
	 unsigned predix;
      } step;
      struct {
	 unsigned leftplanix;
	 unsigned rightplanix;
	 unsigned predix;
      } join;
   };
};

/*
 * Expression representation that hides away a bunch of messy details.
 * Each jexpr is a bundle of tcexprs.
 */
enum jexprtypes {
   JE_SCAN,
   JE_READGLOBAL,
   JE_JOIN,
   JE_REPEAT,
   JE_OTHER,
};
struct jexpr {
   enum jexprtypes type;

   /* fields common to all types */
   struct tcexpr *head;		/* top of whole tcexpr bundle */
   struct columnarray cols;	/* the columns produced at the head */

   union {
      struct {
	 struct tcexpr *scan;		/* the source expression */
	 struct tcexpr *filter;		/* the filter, if any */
      } scan;
      struct {
	 struct tcexpr *readglobal;	/* the scan expression */
	 struct tcexpr *filter;		/* the filter, if any */
      } readglobal;
      struct {
	 struct tcexpr **toptail;	/* bottom of head tree w/o joins */
	 struct jexprarray sources;	/* exprs to join */
	 struct jpredarray predicates;	/* predicates relating them */
	 struct jplan *plan;		/* how we'll do it */
      } join;
      struct {
	 struct tcexpr *repeat;		/* the repeat expression */
	 struct tcexpr *filter;		/* the filter, if any */
	 struct jexpr *sub;		/* LHS */
	 struct jexpr *body;		/* RHS */
      } repeat;
      struct {
	 struct jtailarray tails;	/* leaves of this bundle */
      } other;
   };
};

struct joinorder {
   struct pqlcontext *pql;
};

////////////////////////////////////////////////////////////
// context management

static void joinorder_init(struct joinorder *ctx, struct pqlcontext *pql) {
   ctx->pql = pql;
}

static void joinorder_cleanup(struct joinorder *ctx) {
   (void)ctx;
}

////////////////////////////////////////////////////////////
// cost

static const struct cost free_cost = { 1, 1, 0 };
#define FREE (&free_cost)

static int cost_cmp(const struct cost *a, const struct cost *b) {
   if (a->work < b->work) {
      return -1;
   }
   else if (a->work > b->work) {
      return 1;
   }
   return 0;
}

////////////////////////////////////////////////////////////
// some array tools

#define NONULLS(T) \
   static void T##array_nonulls(struct pqlcontext *pql,			\
				struct T##array *arr) {			\
      struct T *item;							\
      unsigned i, pos, num;						\
									\
      num = T##array_num(arr);						\
      for (i=pos=0; i<num; i++) {					\
	 item = T##array_get(arr, i);					\
	 if (item != NULL) {						\
	    if (pos < i) {						\
	       T##array_set(arr, pos, item);				\
	    }								\
	    pos++;							\
	 }								\
      }									\
      T##array_setsize(pql, arr, pos);					\
   }

NONULLS(jtail);
NONULLS(jplan);
//NONULLS(jexpr);

#define TRANSFER(T) \
   static void T##array_transfer(struct pqlcontext *pql,		\
				 struct T##array *to,			\
				 struct T##array *from) {		\
      unsigned i, num;							\
									\
      num = T##array_num(from);						\
      for (i=0; i<num; i++) {						\
         T##array_add(pql, to, T##array_get(from, i), NULL);		\
      }									\
      T##array_setsize(pql, from, 0);					\
   }

TRANSFER(jtail);
TRANSFER(jpred);
//TRANSFER(jexpr);

#define DESTROYALL(T) \
   static void T##array_destroyall(struct pqlcontext *pql,		\
				   struct T##array *arr) {		\
      struct T *item;							\
      unsigned i, num;							\
									\
      num = T##array_num(arr);						\
      for (i=0; i<num; i++) {						\
	 item = T##array_get(arr, i);					\
         T##_destroy(pql, item);					\
      }									\
      T##array_setsize(pql, arr, 0);					\
   }

/* used below */

static void columnarray_copy_norefs(struct pqlcontext *pql,
				    struct columnarray *to,
				    const struct columnarray *from) {
   unsigned i, num;

   num = columnarray_num(from);
   for (i=0; i<num; i++) {
      columnarray_add(pql, to, columnarray_get(from, i), NULL);
   }
}

static void jplanstack_push(struct pqlcontext *pql,
			    struct jplanarray *stk,
			    struct jplan *jp) {
   jplanarray_add(pql, stk, jp, NULL);
}

static struct jplan *jplanstack_pop(struct pqlcontext *pql,
				    struct jplanarray *stk) {
   struct jplan *jp;
   unsigned num;

   num = jplanarray_num(stk);
   PQLASSERT(num > 0);
   jp = jplanarray_get(stk, num - 1);
   jplanarray_setsize(pql, stk, num - 1);
   return jp;
}

static void indexarray_sortuniq(struct pqlcontext *pql,
				struct indexarray *arr) {
   unsigned i, j, num;
   unsigned prev, here;

   num = indexarray_num(arr);

   if (num == 0) {
      return;
   }

   /* XXX this breaks the abstraction; move to intarray.h? */
   qsort(arr->v, num, sizeof(unsigned), unsigned_sortfunc);

   prev = indexarray_get(arr, 0);
   for (i=j=1; i<num; i++) {
      here = indexarray_get(arr, i);
      if (here != prev) {
	 if (i != j) {
	    indexarray_set(arr, j, here);
	 }
	 j++;
	 prev = here;
      }
   }
   indexarray_setsize(pql, arr, j);
}

static int jwork_sortfunc(const void *av, const void *bv) {
   struct jwork *a = *(void **)av;
   struct jwork *b = *(void **)bv;
   int r;

   r = cost_cmp(&a->cost, &b->cost);
   if (r) {
      return r;
   }

   /*
    * XXX we should probably make this deterministic when the costs
    * are the same in order to avoid random fluctuations in behavior
    * caused by the mood of qsort. That or use a stable sort...
    */

   return 0;
}

static void jworkarray_sort(struct joinorder *ctx, struct jworkarray *arr) {
   unsigned num;

   (void)ctx;
   num = jworkarray_num(arr);
   
   /* XXX this breaks the abstraction; move to intarray.h? */
   qsort(arr->arr.v, num, sizeof(void *), jwork_sortfunc);
}

////////////////////////////////////////////////////////////
// some tcexpr tools

/*
 * Find out which columns the expression TE references.
 * XXX get rid of this bogus external decl and move the code
 * from baseopt to a suitable utility module.
 */
void tcexpr_getusedcolumns(struct joinorder *ctx, struct tcexpr *te,
			   struct columnarray *fill);

////////////////////////////////////////////////////////////
// jtail

static void jexpr_destroy(struct joinorder *ctx, struct jexpr *je);

static struct jtail *mkjtail(struct joinorder *ctx,
			     struct tcexpr **slot, struct jexpr *sub) {
   struct jtail *jt;

   jt = domalloc(ctx->pql, sizeof(*jt));
   jt->isarray = false;
   jt->slot = slot;
   jt->sub = sub;
   return jt;
}

static struct jtail *mkjtail_array(struct joinorder *ctx,
				   struct tcexprarray *arr, unsigned ix,
				   struct jexpr *sub) {
   struct jtail *jt;

   jt = domalloc(ctx->pql, sizeof(*jt));
   jt->isarray = true;
   jt->array.arr = arr;
   jt->array.ix = ix;
   jt->sub = sub;
   return jt;
}

static void jtail_destroy(struct joinorder *ctx, struct jtail *jt) {
   /* slot/arr are unowned references */

   /* owned reference */
   if (jt->sub != NULL) {
      jexpr_destroy(ctx, jt->sub);
   }
   dofree(ctx->pql, jt, sizeof(*jt));
}

/*
 * Assign the tcexpr slot of a jtail.
 */
static void jtail_assign(struct jtail *jt, struct tcexpr *te) {
   if (jt->isarray) {
      PQLASSERT(tcexprarray_get(jt->array.arr, jt->array.ix) == NULL);
      tcexprarray_set(jt->array.arr, jt->array.ix, te);
   }
   else {
      PQLASSERT(*jt->slot == NULL);
      *jt->slot = te;
   }
}

/*
 * Eliminate the tail and reconnect the tcexpr tree by making the
 * tcexpr slot point to jt->sub->head. Destroys jt->sub as well as jt.
 */
static void jtail_reconnect(struct joinorder *ctx, struct jtail *jt) {
   jtail_assign(jt, jt->sub->head);
   jt->sub->head = NULL;
   jtail_destroy(ctx, jt);
}

////////////////////////////////////////////////////////////
// jpred

static struct jpred *mkjpred(struct joinorder *ctx,
			     struct tcvar *var, struct tcexpr *expr) {
   struct jpred *jp;

   jp = domalloc(ctx->pql, sizeof(*jp));
   jp->var = var;
   jp->expr = expr;
   indexarray_init(&jp->sources);

   return jp;
}

static void jpred_destroy(struct joinorder *ctx, struct jpred *jp) {
   if (jp->expr != NULL) {
      tcexpr_destroy(ctx->pql, jp->expr);
   }
   if (jp->var != NULL) {
      tcvar_decref(ctx->pql, jp->var);
   }
   indexarray_setsize(ctx->pql, &jp->sources, 0);
   indexarray_cleanup(ctx->pql, &jp->sources);

   dofree(ctx->pql, jp, sizeof(*jp));
}

////////////////////////////////////////////////////////////
// jplan

static struct jplan *mkjplan_leaf(struct joinorder *ctx,
				  const struct cost *cost,
				  unsigned sourceix) {
   struct jplan *jp;

   jp = domalloc(ctx->pql, sizeof(*jp));
   jp->isleaf = true;
   jp->leaf.sourceix = sourceix;
   jp->cost = *cost;
   columnarray_init(&jp->producecols);
   return jp;
}

static struct jplan *mkjplan_join(struct joinorder *ctx,
				  const struct cost *cost,
				  struct jplan *left, struct jplan *right,
				  enum joinmethods method) {
   struct jplan *jp;

   jp = domalloc(ctx->pql, sizeof(*jp));
   jp->isleaf = false;
   jp->join.left = left;
   jp->join.right = right;
   jp->join.method = method;
   indexarray_init(&jp->join.predicates);
   jp->cost = *cost;
   columnarray_init(&jp->producecols);
   return jp;
}

static void jplan_destroy(struct joinorder *ctx, struct jplan *jp) {
   if (!jp->isleaf) {
      if (jp->join.left) {
	 jplan_destroy(ctx, jp->join.left);
      }
      if (jp->join.right) {
	 jplan_destroy(ctx, jp->join.right);
      }
      indexarray_setsize(ctx->pql, &jp->join.predicates, 0);
      indexarray_cleanup(ctx->pql, &jp->join.predicates);
   }
   /* producecols[] are unowned, so no decrefs */
   columnarray_setsize(ctx->pql, &jp->producecols, 0);
   columnarray_cleanup(ctx->pql, &jp->producecols);
   dofree(ctx->pql, jp, sizeof(*jp));
}

////////////////////////////////////////////////////////////
// jwork

static struct jwork *jwork_create(struct pqlcontext *pql,
				  const struct cost *cost,
				  enum jworktypes type) {
   struct jwork *jw;

   jw = domalloc(pql, sizeof(*jw));
   jw->type = type;
   jw->cost = *cost;
   return jw;
}

static struct jwork *jwork_create_leaf(struct joinorder *ctx,
				       const struct cost *cost,
				       unsigned sourceix) {
   struct jwork *jw;

   jw = jwork_create(ctx->pql, cost, JW_LEAF);
   jw->leaf.sourceix = sourceix;
   return jw;
}

static struct jwork *jwork_create_step(struct joinorder *ctx,
				       const struct cost *cost,
				       unsigned leftplanix,
				       unsigned rightsourceix,
				       unsigned predix) {
   struct jwork *jw;

   jw = jwork_create(ctx->pql, cost, JW_STEP);
   jw->step.leftplanix = leftplanix;
   jw->step.rightsourceix = rightsourceix;
   jw->step.predix = predix;
   return jw;
}

static struct jwork *jwork_create_join(struct joinorder *ctx,
				       const struct cost *cost,
				       unsigned leftplanix,
				       unsigned rightplanix,
				       unsigned predix) {
   struct jwork *jw;

   jw = jwork_create(ctx->pql, cost, JW_JOIN);
   jw->join.leftplanix = leftplanix;
   jw->join.rightplanix = rightplanix;
   jw->join.predix = predix;
   return jw;
}

static void jwork_destroy(struct pqlcontext *pql, struct jwork *jw) {
   switch (jw->type) {
    case JW_LEAF:
     /* nothing */
     break;
    case JW_STEP:
     /* nothing */
     break;
    case JW_JOIN:
     /* nothing */
     break;
   }
   dofree(pql, jw, sizeof(*jw));
}

/* defined above for jworkarray */
DESTROYALL(jwork);

////////////////////////////////////////////////////////////
// jexpr

static void jexpr_fillcols(struct joinorder *ctx, struct jexpr *je,
			   struct datatype *t) {
   struct column *col;
   unsigned arity, i;

   if (datatype_isset(t)) {
      t = datatype_set_member(t);
   }
   if (!datatype_istuple(t)) {
      /* no columns */
      return;
   }

   /* XXX what do we (want to) do with nested columns? */

   arity = datatype_arity(t);
   columnarray_setsize(ctx->pql, &je->cols, arity);
   for (i=0; i<arity; i++) {
      col = datatype_tuple_getcolumn(t, i);
      column_incref(col);
      columnarray_set(&je->cols, i, col);
   }
}

static void jexpr_emptycols(struct joinorder *ctx, struct jexpr *je) {
   unsigned num, i;

   num = columnarray_num(&je->cols);
   for (i=0; i<num; i++) {
      column_decref(ctx->pql, columnarray_get(&je->cols, i));
   }
   columnarray_setsize(ctx->pql, &je->cols, 0);
}

static struct jexpr *mkjexpr(struct joinorder *ctx, enum jexprtypes type,
			     struct tcexpr *head, struct datatype *headtype) {
   struct jexpr *je;

   je = domalloc(ctx->pql, sizeof(*je));
   je->type = type;
   je->head = head;
   columnarray_init(&je->cols);
   jexpr_fillcols(ctx, je, headtype);

   return je;
}

static void jexpr_destroy(struct joinorder *ctx, struct jexpr *je) {
   if (je == NULL) {
      return;
   }

   switch (je->type) {
    case JE_SCAN:
     /* unowned refs */
     je->scan.scan = NULL;
     je->scan.filter = NULL;
     break;

    case JE_READGLOBAL:
     /* unowned refs */
     je->readglobal.readglobal = NULL;
     je->readglobal.filter = NULL;
     break;

    case JE_JOIN:
     {
	unsigned i, num;

	/* owned refs */

	num = jexprarray_num(&je->join.sources);
	for (i=0; i<num; i++) {
	   jexpr_destroy(ctx, jexprarray_get(&je->join.sources, i));
	}
	jexprarray_setsize(ctx->pql, &je->join.sources, 0);
	jexprarray_cleanup(ctx->pql, &je->join.sources);

	num = jpredarray_num(&je->join.predicates);
	for (i=0; i<num; i++) {
	   jpred_destroy(ctx, jpredarray_get(&je->join.predicates, i));
	}
	jpredarray_setsize(ctx->pql, &je->join.predicates, 0);
	jpredarray_cleanup(ctx->pql, &je->join.predicates);

	if (je->join.plan != NULL) {
	   jplan_destroy(ctx, je->join.plan);
	}
     }
     break;

    case JE_REPEAT:
     /* unowned refs */
     je->repeat.repeat = NULL;
     je->repeat.filter = NULL;
     /* owned refs */
     jexpr_destroy(ctx, je->repeat.sub);
     jexpr_destroy(ctx, je->repeat.body);
     break;

    case JE_OTHER:
     {
	struct jtail *jt;
	unsigned i, num;

	num = jtailarray_num(&je->other.tails);
	for (i=0; i<num; i++) {
	   jt = jtailarray_get(&je->other.tails, i);
	   if (jt != NULL) {
	      jtail_destroy(ctx, jt);
	   }
	}
	jtailarray_setsize(ctx->pql, &je->other.tails, 0);
	jtailarray_cleanup(ctx->pql, &je->other.tails);
     }
     break;
   }

   jexpr_emptycols(ctx, je);
   columnarray_cleanup(ctx->pql, &je->cols);
   if (je->head != NULL) {
      tcexpr_destroy(ctx->pql, je->head);
      je->head = NULL;
   }
   dofree(ctx->pql, je, sizeof(*je));
}

static struct jexpr *mkjexpr_scan(struct joinorder *ctx,
				  struct tcexpr *head,
				  struct tcexpr *scan,
				  struct tcexpr *filter) {
   struct jexpr *je;

   je = mkjexpr(ctx, JE_SCAN, head, head->datatype);
   je->scan.scan = scan;
   je->readglobal.filter = filter;

   return je;
}

static struct jexpr *mkjexpr_readglobal(struct joinorder *ctx,
					struct tcexpr *head,
					struct tcexpr *readglobal,
					struct tcexpr *filter) {
   struct jexpr *je;

   je = mkjexpr(ctx, JE_READGLOBAL, head, head->datatype);
   je->readglobal.readglobal = readglobal;
   je->readglobal.filter = filter;

   return je;
}

/*
 * HEAD may be null, so the type is passed explicitly
 */
static struct jexpr *mkjexpr_join(struct joinorder *ctx,
				  struct tcexpr *head,
				  struct datatype *headtype,
				  struct tcexpr **toptail) {
   struct jexpr *je;

   je = mkjexpr(ctx, JE_JOIN, head, headtype);
   je->join.toptail = toptail;
   jexprarray_init(&je->join.sources);
   jpredarray_init(&je->join.predicates);
   je->join.plan = NULL;

   return je;
}

static struct jexpr *mkjexpr_repeat(struct joinorder *ctx,
				    struct tcexpr *head,
				    struct tcexpr *repeat,
				    struct tcexpr *filter,
				    struct jexpr *sub, struct jexpr *body) {
   struct jexpr *je;

   je = mkjexpr(ctx, JE_REPEAT, head, head->datatype);
   je->repeat.repeat = repeat;
   je->repeat.filter = filter;
   je->repeat.sub = sub;
   je->repeat.body = body;

   return je;
}

static struct jexpr *mkjexpr_other(struct joinorder *ctx,
				   struct tcexpr *head) {
   struct jexpr *je;

   je = mkjexpr(ctx, JE_OTHER, head, head->datatype);
   jtailarray_init(&je->other.tails);

   return je;
}

////////////////////////////////////////////////////////////
// jbuild

/*
 * Given a tcexpr tree, build a jexpr tree around it.
 *
 * This uses the following patterns:
 *
 *    projecttuple? filter? scan	=>	JE_SCAN
 *    filter? readglobal		=>	JE_READGLOBAL
 *    projecttuple? join		=>	JE_JOIN
 *    projecttuple? filter? repeat	=>	JE_REPEAT
 *    *					=>	JE_OTHER
 *
 * Adjacent JE_JOINs and JE_OTHERs are combined later.
 */

static struct jexpr *jbuild(struct joinorder *ctx, struct tcexpr *te);

static struct jexpr *jbuild_once(struct joinorder *ctx, struct tcexpr *te) {

   /*
    * scan cases
    */
   if (te->type == TCE_PROJECTTUPLE &&
       te->projecttuple.sub->type == TCE_FILTER &&
       te->projecttuple.sub->filter.sub->type == TCE_SCAN) {
      return mkjexpr_scan(ctx, te, te->projecttuple.sub->filter.sub,
			  te->projecttuple.sub);
   }
   if (te->type == TCE_PROJECTTUPLE &&
       te->projecttuple.sub->type == TCE_SCAN) {
      return mkjexpr_scan(ctx, te, te->projecttuple.sub, NULL);
   }
   if (te->type == TCE_FILTER &&
       te->filter.sub->type == TCE_SCAN) {
      return mkjexpr_scan(ctx, te, te->filter.sub, te);
   }
   if (te->type == TCE_SCAN) {
      return mkjexpr_scan(ctx, te, te, NULL);
   }

   /*
    * readglobal cases
    */
   if (te->type == TCE_FILTER &&
       te->filter.sub->type == TCE_READGLOBAL) {
      return mkjexpr_readglobal(ctx, te, te->filter.sub, te);
   }
   if (te->type == TCE_READGLOBAL) {
      return mkjexpr_readglobal(ctx, te, te, NULL);
   }

   /*
    * join cases
    */
   {
      struct tcexpr *join, **toptail;
      struct datatype *headtype;
      struct jpred *jp;
      struct jexpr *jleft, *jright, *je;

      if (te->type == TCE_PROJECTTUPLE &&
	  te->projecttuple.sub->type == TCE_JOIN
	 /*
	  * Disable this case for now. If the project is not at the
	  * head of a join subtree but in between two joins, jmerge
	  * causes it to be lost, and if the project discards columns
	  * this causes type problems that aren't easily rectified.
	  * XXX.
	  *
	  * The right fix is probably to make sure there are no
	  * projections in the middle of join subtrees, and then to
	  * push projections down later on after planning is complete.
	  */
	  && 0
	  ) {
	 join = te->projecttuple.sub;
	 te->projecttuple.sub = NULL;
	 toptail = &te->projecttuple.sub;
	 headtype = te->datatype;
	 goto foundjoin;
      }
      if (te->type == TCE_JOIN) {
	 join = te;
	 te = NULL;
	 toptail = NULL;
	 headtype = join->datatype;
     foundjoin:
  	 jleft = jbuild(ctx, join->join.left);
	 join->join.left = NULL;

	 jright = jbuild(ctx, join->join.right);
	 join->join.right = NULL;

	 je = mkjexpr_join(ctx, te, headtype, toptail);
	 jexprarray_add(ctx->pql, &je->join.sources, jleft, NULL);
	 jexprarray_add(ctx->pql, &je->join.sources, jright, NULL);

	 jp = mkjpred(ctx, join->join.predvar, join->join.predexpr);
	 join->join.predvar = NULL;
	 join->join.predexpr = NULL;
	 /* jp->sources is filled in later */

	 jpredarray_add(ctx->pql, &je->join.predicates, jp, NULL);

	 tcexpr_destroy(ctx->pql, join);

	 return je;
      }
   }

   /*
    * repeat cases
    */
   {
      struct tcexpr *rep, *filt;
      struct jexpr *jleft, *jright;

      if (te->type == TCE_PROJECTTUPLE &&
	  te->projecttuple.sub->type == TCE_FILTER &&
	  te->projecttuple.sub->filter.sub->type == TCE_REPEAT) {
	 rep = te->projecttuple.sub->filter.sub;
	 filt = te->projecttuple.sub;
	 goto foundrep;
      }
      if (te->type == TCE_PROJECTTUPLE &&
	  te->projecttuple.sub->type == TCE_REPEAT) {
	 rep = te->projecttuple.sub;
	 filt = NULL;
	 goto foundrep;
      }
      if (te->type == TCE_FILTER &&
	  te->filter.sub->type == TCE_REPEAT) {
	 rep = te->filter.sub;
	 filt = te;
	 goto foundrep;
      }
      if (te->type == TCE_REPEAT) {
	 rep = te;
	 filt = NULL;
     foundrep:
	 jleft = jbuild(ctx, rep->repeat.sub);
	 jright = jbuild(ctx, rep->repeat.body);
	 /* these pointers are taken over by the jexprs */
	 rep->repeat.sub = NULL;
	 rep->repeat.body = NULL;
	 return mkjexpr_repeat(ctx, te, rep, filt, jleft, jright);
      }
   }

   return NULL;
}

static void jbuild_othersub(struct joinorder *ctx, struct jexpr *je,
			    struct tcexpr **botptr) {
   struct jtail *jt;

   PQLASSERT(je->type == JE_OTHER);
   jt = mkjtail(ctx, botptr, jbuild(ctx, *botptr));
   jtailarray_add(ctx->pql, &je->other.tails, jt, NULL);
   *botptr = NULL;
}

static void jbuild_othersubarray(struct joinorder *ctx, struct jexpr *je,
				 struct tcexpr *botexpr,
				 struct tcexprarray *botarray, unsigned ix) {
   struct jtail *jt;

   PQLASSERT(je->type == JE_OTHER);
   jt = mkjtail_array(ctx, botarray, ix, jbuild(ctx, botexpr));
   jtailarray_add(ctx->pql, &je->other.tails, jt, NULL);
   tcexprarray_set(botarray, ix, NULL);
}

static struct jexpr *jbuild(struct joinorder *ctx, struct tcexpr *te) {
   struct jexpr *je;
   unsigned i, num;

   je = jbuild_once(ctx, te);
   if (je != NULL) {
      return je;
   }

   je = mkjexpr_other(ctx, te);
   switch (te->type) {
    case TCE_FILTER:
     jbuild_othersub(ctx, je, &te->filter.sub);
     break;
    case TCE_PROJECTTUPLE:
     jbuild_othersub(ctx, je, &te->projecttuple.sub);
     break;
    case TCE_PROJECTSCALAR:
     jbuild_othersub(ctx, je, &te->projectscalar.sub);
     break;
    case TCE_RENAME:
     jbuild_othersub(ctx, je, &te->rename.sub);
     break;
    case TCE_JOIN:
     /* should always be matched elsewhere */
     PQLASSERT(0);
     break;
    case TCE_ORDER:
     jbuild_othersub(ctx, je, &te->order.sub);
     break;
    case TCE_UNIQ:
     jbuild_othersub(ctx, je, &te->uniq.sub);
     break;
    case TCE_NEST:
     jbuild_othersub(ctx, je, &te->nest.sub);
     break;
    case TCE_UNNEST:
     jbuild_othersub(ctx, je, &te->unnest.sub);
     break;
    case TCE_DISTINGUISH:
     jbuild_othersub(ctx, je, &te->distinguish.sub);
     break;
    case TCE_ADJOIN:
     jbuild_othersub(ctx, je, &te->adjoin.left);
     break;
    case TCE_STEP:
     /* not allowed here */
     PQLASSERT(0);
     break;
    case TCE_REPEAT:
    case TCE_SCAN:
     /* should always be matched elsewhere */
     PQLASSERT(0);
     break;
    case TCE_BOP:
     jbuild_othersub(ctx, je, &te->bop.left);
     jbuild_othersub(ctx, je, &te->bop.right);
     break;
    case TCE_UOP:
     jbuild_othersub(ctx, je, &te->uop.sub);
     break;
    case TCE_FUNC:
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	jbuild_othersubarray(ctx, je, tcexprarray_get(&te->func.args, i),
			     &te->func.args, i);
     }
     break;
    case TCE_MAP:
     jbuild_othersub(ctx, je, &te->map.set);
     jbuild_othersub(ctx, je, &te->map.result);
     break;
    case TCE_LET:
     jbuild_othersub(ctx, je, &te->let.value);
     jbuild_othersub(ctx, je, &te->let.body);
     break;
    case TCE_READVAR:
     break;
    case TCE_READGLOBAL:
     break;
    case TCE_CREATEPATHELEMENT:
     jbuild_othersub(ctx, je, &te->createpathelement);
     break;
    case TCE_SPLATTER:
     jbuild_othersub(ctx, je, &te->splatter.value);
     break;
    case TCE_TUPLE:
     num = tcexprarray_num(&te->tuple.exprs);
     for (i=0; i<num; i++) {
	jbuild_othersubarray(ctx, je, tcexprarray_get(&te->tuple.exprs, i),
			     &te->tuple.exprs, i);
     }
     break;
    case TCE_VALUE:
     break;
   }

   return je;
}

////////////////////////////////////////////////////////////
// jmerge

/*
 * Combine adjacent like nodes in the jexpr tree.
 */


static void jmerge_join_sub(struct joinorder *ctx,
			    struct jexprarray *newsources,
			    struct jpredarray *predicates,
			    struct jexpr *je) {
   struct jexpr *jsub;
   unsigned i, num;

   PQLASSERT(je->type == JE_JOIN);

   num = jexprarray_num(&je->join.sources);
   for (i=0; i<num; i++) {
      jsub = jexprarray_get(&je->join.sources, i);
      if (jsub->type == JE_JOIN) {
	 jmerge_join_sub(ctx, newsources, predicates, jsub);

	 jpredarray_transfer(ctx->pql, predicates, &jsub->join.predicates);
	 if (jsub->head != NULL) {
	    /*
	     * This can only be a single projecttuple, which might
	     * reorder columns (we don't care) or drop columns (we
	     * don't care either).  Assert that it's what we expect,
	     * but just throw it away along with the jexpr.
	     *
	     * XXX reordering is fine but dropping columns causes
	     * problems.
	     */
	    PQLASSERT(jsub->head->type == TCE_PROJECTTUPLE);
	    PQLASSERT(jsub->head->projecttuple.sub == NULL);
	    PQLASSERT(jsub->join.toptail == &jsub->head->projecttuple.sub);
	 }
	 jexpr_destroy(ctx, jsub);
      }
      else {
	 jexprarray_add(ctx->pql, newsources, jsub, NULL);
      }
      jexprarray_set(&je->join.sources, i, NULL);
   }
}

static void jmerge_join(struct joinorder *ctx, struct jexpr *je) {
   struct jexprarray newsources;
   struct jpredarray newpredicates;
   unsigned i, num;

   /*
    * In order to simplify various things we want the sources array
    * to come out as an in-order depth-first traversal. So while we
    * could do the merge without all the extra recursive goo, that
    * would either be painful or yield a breadth-first traversal.
    */
   jexprarray_init(&newsources);
   jpredarray_init(&newpredicates);

   jmerge_join_sub(ctx, &newsources, &newpredicates, je);
   jpredarray_transfer(ctx->pql, &newpredicates, &je->join.predicates);

   num = jexprarray_num(&newsources);
   jexprarray_setsize(ctx->pql, &je->join.sources, num);
   for (i=0; i<num; i++) {
      jexprarray_set(&je->join.sources, i, jexprarray_get(&newsources, i));
   }
   jexprarray_setsize(ctx->pql, &newsources, 0);

   num = jpredarray_num(&newpredicates);
   jpredarray_setsize(ctx->pql, &je->join.predicates, num);
   for (i=0; i<num; i++) {
      jpredarray_set(&je->join.predicates, i,
		     jpredarray_get(&newpredicates, i));
   }
   jpredarray_setsize(ctx->pql, &newpredicates, 0);

   jexprarray_cleanup(ctx->pql, &newsources);
   jpredarray_cleanup(ctx->pql, &newpredicates);
}

static void jmerge_other(struct joinorder *ctx, struct jexpr *je) {
   struct jtail *jt;
   unsigned i;
   bool didanything = false;

   PQLASSERT(je->type == JE_OTHER);

   for (i=0; i<jtailarray_num(&je->other.tails); i++) {
      jt = jtailarray_get(&je->other.tails, i);
      if (jt->sub->type == JE_OTHER) {
	 jtailarray_transfer(ctx->pql,
			     &je->other.tails, &jt->sub->other.tails);
	 jtail_reconnect(ctx, jt);
	 jtailarray_set(&je->other.tails, i, NULL);
	 didanything = true;
      }
   }
   if (didanything) {
      jtailarray_nonulls(ctx->pql, &je->other.tails);
   }
}

static void jmerge(struct joinorder *ctx, struct jexpr *je) {
   unsigned i, num;

   switch (je->type) {
    case JE_SCAN:
     break;
    case JE_READGLOBAL:
     break;
    case JE_REPEAT:
     jmerge(ctx, je->repeat.sub);
     jmerge(ctx, je->repeat.body);
     break;
    case JE_JOIN:
     jmerge_join(ctx, je);
     num = jexprarray_num(&je->join.sources);
     for (i=0; i<num; i++) {
	jmerge(ctx, jexprarray_get(&je->join.sources, i));
     }
     break;
    case JE_OTHER:
     jmerge_other(ctx, je);
     num = jtailarray_num(&je->other.tails);
     for (i=0; i<num; i++) {
	jmerge(ctx, jtailarray_get(&je->other.tails, i)->sub);
     }
     break;
   }
}

////////////////////////////////////////////////////////////
// jpredicate

/*
 * Inspect predicates and set up jpred->sources arrays to build the
 * "graph" for each merged join.
 */

/*
 * If the predicate expression in JP can be split into multiple
 * independent pieces, do so. Add the new pieces as new jpreds
 * to FILL.
 */
static void jpred_split(struct joinorder *ctx, struct jpred *jp,
			struct jpredarray *fill) {
   /* XXX for now punt */
   (void)ctx;
   (void)jp;
   (void)fill;
}

/*
 * Find out which columns the predicate expression in JP references.
 */
static void jpred_getcolumns(struct joinorder *ctx, struct jpred *jp,
			     struct columnarray *fill) {
   tcexpr_getusedcolumns(ctx, jp->expr, fill);
}

/*
 * XXX there are several copies of this floating around; unify them
 */
static bool columnarray_contains(const struct columnarray *arr,
				 const struct column *col) {
   unsigned i, num;

   num = columnarray_num(arr);
   for (i=0; i<num; i++) {
      if (columnarray_get(arr, i) == col) {
	 return true;
      }
   }
   return false;
}

/*
 * Find the index of the jexpr that produces column COL.
 * If not found, crash.
 */
static unsigned jexprarray_findcol(struct jexprarray *arr,
				   struct column *col) {
   struct jexpr *je;
   unsigned i, num;

   num = jexprarray_num(arr);
   for (i=0; i<num; i++) {
      je = jexprarray_get(arr, i);
      if (columnarray_contains(&je->cols, col)) {
	 return i;
      }
   }
   PQLASSERT(0);
   return (unsigned)-1;
}

static void jpredicate_join(struct joinorder *ctx, struct jexpr *je) {
   struct column *col;
   struct columnarray cols;
   struct jpred *jp;
   unsigned i, numpreds, j, numcols, k, numsources, ix;
   bool *found;

   PQLASSERT(je->type == JE_JOIN);

   /*
    * We start with one predicate for each TCE_JOIN node we saw.
    * Split these if we can. Note: while the array may grow as
    * we work, there's no need to process the expressions that
    * have been added, so don't refetch the array size.
    */
   numpreds = jpredarray_num(&je->join.predicates);
   for (i=0; i<numpreds; i++) {
      jp = jpredarray_get(&je->join.predicates, i);
      jpred_split(ctx, jp, &je->join.predicates);
   }

   /*
    * Now for each predicate figure out which sources it depends on.
    * Refetch the array size (numpreds) before looping because it may
    * have changed and we now want everything.
    */
   columnarray_init(&cols);
   numsources = jexprarray_num(&je->join.sources);
   found = domalloc(ctx->pql, sizeof(bool)*numsources);

   numpreds = jpredarray_num(&je->join.predicates);
   for (i=0; i<numpreds; i++) {
      jp = jpredarray_get(&je->join.predicates, i);

      if (jp->expr == NULL) {
	 /* cartesian product with no join condition */
	 PQLASSERT(indexarray_num(&jp->sources) == 0);
	 continue;
      }

      jpred_getcolumns(ctx, jp, &cols);

      for (k=0; k<numsources; k++) {
	 found[k] = false;
      }

      numcols = columnarray_num(&cols);
      for (j=0; j<numcols; j++) {
	 col = columnarray_get(&cols, j);
	 ix = jexprarray_findcol(&je->join.sources, col);
	 found[ix] = true;
      }
      columnarray_setsize(ctx->pql, &cols, 0);

      for (k=0; k<numsources; k++) {
	 if (found[k]) {
	    indexarray_add(ctx->pql, &jp->sources, k, NULL);
	 }
      }
   }

   dofree(ctx->pql, found, sizeof(bool)*numsources);
   columnarray_cleanup(ctx->pql, &cols);
}

static void jpredicate(struct joinorder *ctx, struct jexpr *je) {
   unsigned i, num;

   switch (je->type) {
    case JE_SCAN:
    case JE_READGLOBAL:
     break;
    case JE_JOIN:
     num = jexprarray_num(&je->join.sources);
     for (i=0; i<num; i++) {
	jpredicate(ctx, jexprarray_get(&je->join.sources, i));
     }
     jpredicate_join(ctx, je);
     break;
    case JE_REPEAT:
     jpredicate(ctx, je->repeat.sub);
     jpredicate(ctx, je->repeat.body);
     break;
    case JE_OTHER:
     num = jtailarray_num(&je->other.tails);
     for (i=0; i<num; i++) {
	jpredicate(ctx, jtailarray_get(&je->other.tails, i)->sub);
     }
     break;
   }
}

////////////////////////////////////////////////////////////
// junbuild

/*
 * Given a jexpr tree, reconstruct the tcexpr tree and throw away
 * the jexpr bits.
 */

static struct tcexpr *junbuild(struct joinorder *ctx, struct jexpr *je);

/*
 * Return a tuple type holding the columns in RED that are not in
 * BLUE.
 */
static struct datatype *get_missing_columns(struct joinorder *ctx,
					    struct datatype *red,
					    struct datatype *blue) {
   struct datatype *ret, *it, *jt;
   struct column *ic, *jc;
   unsigned redarity, bluearity, i, j;
   bool found;

   PQLASSERT(datatype_istuple(red));
   PQLASSERT(datatype_istuple(blue));
   redarity = datatype_arity(red);
   bluearity = datatype_arity(blue);

   ret = datatype_unit(ctx->pql);

   for (i=0; i<redarity; i++) {
      found = false;
      it = datatype_getnth(red, i);
      ic = datatype_tuple_getcolumn(red, i);
      for (j=0; j<bluearity; j++) {
	 jt = datatype_getnth(blue, j);
	 jc = datatype_tuple_getcolumn(blue, j);
	 if (datatype_eq_withnames(it, jt) && ic == jc) {
	    found = true;
	    break;
	 }
      }
      if (!found) {
	 ret = datatype_tuple_append(ctx->pql, ret, it, ic);
      }
   }

   return ret;
}

static bool needproject(struct datatype *t, struct columnarray *cols) {
   unsigned i, num;

   PQLASSERT(datatype_isset(t));
   t = datatype_set_member(t);
   PQLASSERT(datatype_istuple(t));

   num = datatype_arity(t);
   if (num != columnarray_num(cols)) {
      return true;
   }
   for (i=0; i<num; i++) {
      if (datatype_tuple_getcolumn(t, i) != columnarray_get(cols, i)) {
	 return true;
      }
   }
   return false;
}

/*
 * Convert PREDS, which is an array of indexes into je->predicates[],
 * into a single tcvar/tcexpr predicate.
 */
static void junbuild_dopredicates(struct joinorder *ctx, struct jexpr *je,
				  struct indexarray *preds,
				  struct tcvar **predvar_ret,
				  struct tcexpr **predexpr_ret) {
   struct jpred *jp;
   unsigned ix;

   if (indexarray_num(preds) == 0) {
      *predvar_ret = NULL;
      *predexpr_ret = NULL;
      return;
   }

   // good enough for now (XXX)
   PQLASSERT(indexarray_num(preds) == 1);
   (void)ctx;

   ix = indexarray_get(preds, 0);
   jp = jpredarray_get(&je->join.predicates, ix);

   /* not always true... */
   //PQLASSERT(jp->var != NULL);
   //PQLASSERT(jp->expr != NULL);
   *predvar_ret = jp->var;
   *predexpr_ret = jp->expr;
   jp->var = NULL;
   jp->expr = NULL;
}

static struct tcexpr *junbuild_plan(struct joinorder *ctx,
				    struct jexpr *je,
				    struct jplan *jp) {
   struct tcexpr *ret, *tl, *predexpr;
   struct tcvar *predvar;
   struct datatype *oldtype;
   struct column *col;
   unsigned i, num;

   ret = NULL; /* because gcc 4.1 is lame */

   if (jp->isleaf) {
      ret = junbuild(ctx, jexprarray_get(&je->join.sources,jp->leaf.sourceix));
      jexprarray_set(&je->join.sources, jp->leaf.sourceix, NULL);
      jplan_destroy(ctx, jp);
      return ret;
   }

   junbuild_dopredicates(ctx, je, &jp->join.predicates, &predvar, &predexpr);

   /* XXX what about the type of predvar? */

   tl = junbuild_plan(ctx, je, jp->join.left);
   jp->join.left = NULL;

   switch (jp->join.method) {
    case J_BASIC:
#if 1 /* temporary */
    case J_STEP:
    case J_REVSTEP:
#endif
     {
	struct tcexpr *tr, *tmpl;
	struct datatype *rettype, *oldtype, *newtype, *moretype;

	tr = junbuild_plan(ctx, je, jp->join.right);
	jp->join.right = NULL;
	rettype = datatype_join(tl->datatype, tr->datatype);

	if (predvar != NULL) {
	   oldtype = predvar->datatype;
	   newtype = datatype_set_member(rettype);

	   predvar->datatype = newtype;

	   /*
	    * Wrap uses of predvar with an expression that turns
	    * NEWTYPE into OLDTYPE.
	    */

	   if (datatype_eq_withnames(oldtype, newtype)) {
	      /* nothing */
	   }
	   else if (datatype_arity(oldtype) <= datatype_arity(newtype)) {
	      /* oldtype is smaller; just run a project */
	      tmpl = psplit_reorderwrapper(ctx->pql, NULL, oldtype,
					   datatype_unit(ctx->pql));
	      predexpr = wrapvar(ctx->pql, predexpr, predvar, tmpl);
	      tcexpr_destroy(ctx->pql, tmpl);
	   }
	   else {
	      /* oldtype is larger; adjoin needed columns, then reorder */
	      moretype = get_missing_columns(ctx, oldtype, newtype);
	      tmpl = psplit_adjoinwrapper(ctx->pql, NULL, newtype, moretype);
	      tmpl = psplit_reorderwrapper(ctx->pql, tmpl, oldtype,
					   datatype_unit(ctx->pql));
	      predexpr = wrapvar(ctx->pql, predexpr, predvar, tmpl);
	      tcexpr_destroy(ctx->pql, tmpl);
	   }
	}

	ret = mktcexpr_join(ctx->pql, tl, tr, predvar, predexpr);
	ret->datatype = rettype;
     }
     break;
#if 0 /* broken */
    case J_STEP:
    case J_REVSTEP:
#endif
     {
	struct jexpr *jer;
	struct tcexpr *scan, *filter;
	struct column *tlcol;
	struct pqlvalue *edgename;

	PQLASSERT(jp->join.right->isleaf == true);
	jer = jexprarray_get(&je->join.sources, jp->join.right->leaf.sourceix);

	PQLASSERT(jer->type == JE_SCAN);
	scan = jer->scan.scan;
	filter = jer->scan.filter;

	PQLASSERT(scan->type == TCE_SCAN);

	if (filter != NULL) {
	   struct tcexpr *filterpred;
	   struct tcvar *filtervar;

	   PQLASSERT(filter->type == TCE_FILTER);
	   filtervar = filter->filter.predvar;
	   filterpred = filter->filter.predexpr;
	   if (filterpred->type == TCE_BOP &&
	       filterpred->bop.left->type == TCE_PROJECTSCALAR &&
	       filterpred->bop.left->projectscalar.col == scan->scan.edgecolumn &&
	       filterpred->bop.left->projectscalar.sub->type == TCE_READVAR &&
	       filterpred->bop.left->projectscalar.sub->readvar == filtervar &&
	       filterpred->bop.right->type == TCE_VALUE) {
	      edgename = filterpred->bop.right->value;
	      filterpred->bop.right->value = NULL;
	   }
	   else {
	      struct tcexpr *readvar, *proj, *let, *bop;

	      edgename = NULL;

	      tcvar_incref(predvar);
	      readvar = mktcexpr_readvar(ctx->pql, predvar);
	      readvar->datatype = predvar->datatype;

	      proj = mktcexpr_projecttuple(ctx->pql, readvar);
	      if (scan->scan.leftobjcolumn != NULL) {
		 column_incref(scan->scan.leftobjcolumn);
		 columnarray_add(ctx->pql, proj->projecttuple.cols,
				 scan->scan.leftobjcolumn, NULL);
	      }
	      if (scan->scan.edgecolumn != NULL) {
		 column_incref(scan->scan.edgecolumn);
		 columnarray_add(ctx->pql, proj->projecttuple.cols,
				 scan->scan.edgecolumn, NULL);
	      }
	      if (scan->scan.rightobjcolumn != NULL) {
		 column_incref(scan->scan.rightobjcolumn);
		 columnarray_add(ctx->pql, proj->projecttuple.cols,
				 scan->scan.rightobjcolumn, NULL);
	      }
	      proj->datatype = scan->datatype;

	      let = mktcexpr_let(ctx->pql, filtervar, proj, filterpred);
	      let->datatype = filterpred->datatype;

	      bop = mktcexpr_bop(ctx->pql, let, F_AND, predexpr);
	      bop->datatype = predexpr->datatype;

	      filter->filter.predvar = NULL;
	      filter->filter.predexpr = NULL;
	      predexpr = bop;
	   }
	}

	{
	   unsigned predix;
	   struct jpred *pred;

	   /*
	    * XXX this will not work; it assumes we haven't already done
	    * junbuild_dopredicates, but we have, and the previous block
	    * of code even relies on that. Hopeless. XXX
	    */

	   /* first predicate should be the one that defines the step */
	   predix = indexarray_get(&jp->join.predicates, 0);
	   pred = jpredarray_get(&je->join.predicates, predix);

	   indexarray_remove(&jp->join.predicates, 0);
	   jpredarray_set(&je->join.predicates, predix, NULL);

	   PQLASSERT(pred != NULL);
	   PQLASSERT(pred->var != NULL);
	   PQLASSERT(pred->expr != NULL);

	   PQLASSERT(pred->expr->type == TCE_BOP);
	   PQLASSERT(pred->expr->bop.op == F_EQ);
	   PQLASSERT(pred->expr->bop.left->type == TCE_PROJECTSCALAR);
	   PQLASSERT(pred->expr->bop.right->type == TCE_PROJECTSCALAR);
	   PQLASSERT(pred->expr->bop.left->projectscalar.sub->type == TCE_READVAR);
	   PQLASSERT(pred->expr->bop.right->projectscalar.sub->type == TCE_READVAR);
	   PQLASSERT(pred->expr->bop.left->projectscalar.sub->readvar == pred->var);
	   PQLASSERT(pred->expr->bop.right->projectscalar.sub->readvar == pred->var);

	   if (jp->join.method == J_REVSTEP) {
	      if (pred->expr->bop.left->projectscalar.col ==
		  scan->scan.rightobjcolumn) {
		 tlcol = pred->expr->bop.right->projectscalar.col;
		 pred->expr->bop.right->projectscalar.col = NULL;
	      }
	      else {
		 PQLASSERT(pred->expr->bop.right->projectscalar.col ==
			   scan->scan.rightobjcolumn);
		 tlcol = pred->expr->bop.left->projectscalar.col;
		 pred->expr->bop.left->projectscalar.col = NULL;
	      }
	   }
	   else {
	      if (pred->expr->bop.left->projectscalar.col ==
		  scan->scan.leftobjcolumn) {
		 tlcol = pred->expr->bop.right->projectscalar.col;
		 pred->expr->bop.right->projectscalar.col = NULL;
	      }
	      else {
		 PQLASSERT(pred->expr->bop.right->projectscalar.col ==
			   scan->scan.leftobjcolumn);
		 tlcol = pred->expr->bop.left->projectscalar.col;
		 pred->expr->bop.left->projectscalar.col = NULL;
	      }
	   }

	   jpred_destroy(ctx, pred);
	}

	ret = mktcexpr_step(ctx->pql, tl, tlcol, edgename,
			    jp->join.method == J_REVSTEP /*reversed */,
			    scan->scan.leftobjcolumn,
			    scan->scan.edgecolumn,
			    scan->scan.rightobjcolumn,
			    predvar, predexpr);
	ret->datatype = datatype_join(tl->datatype,
				      jer->head->datatype);

	scan->scan.leftobjcolumn = NULL;
	scan->scan.edgecolumn = NULL;
	scan->scan.rightobjcolumn = NULL;

	if (jp->join.right != NULL) {
	   jplan_destroy(ctx, jp->join.right);
	   jp->join.right = NULL;
	}
     }
     break;
   }

   if (needproject(ret->datatype, &jp->producecols)) {
      oldtype = ret->datatype;
      ret = mktcexpr_projecttuple(ctx->pql, ret);
      num = columnarray_num(&jp->producecols);
      for (i=0; i<num; i++) {
	 col = columnarray_get(&jp->producecols, i);
	 column_incref(col);
	 columnarray_add(ctx->pql, ret->projecttuple.cols, col, NULL);
      }
      ret->datatype = datatype_subprojecttuple(oldtype, &jp->producecols);
   }

   jplan_destroy(ctx, jp);

   return ret;
}

static struct tcexpr *junbuild(struct joinorder *ctx, struct jexpr *je) {
   struct tcexpr *te, *ret;
   struct jtail *jt;
   unsigned i, num;

   switch (je->type) {
    case JE_SCAN:
     /* nothing needed */
     break;
    case JE_READGLOBAL:
     /* nothing needed */
     break;
    case JE_JOIN:
     te = junbuild_plan(ctx, je, je->join.plan);
     je->join.plan = NULL;
     if (je->join.toptail == NULL) {
	PQLASSERT(je->head == NULL);
	je->head = te;
     }
     else {
	*je->join.toptail = te;
     }
     break;
    case JE_REPEAT:
     PQLASSERT(je->repeat.repeat->repeat.sub == NULL);
     PQLASSERT(je->repeat.repeat->repeat.body == NULL);
     je->repeat.repeat->repeat.sub = junbuild(ctx, je->repeat.sub);
     je->repeat.repeat->repeat.body = junbuild(ctx, je->repeat.body);
     je->repeat.sub = NULL;
     je->repeat.body = NULL;
     break;
    case JE_OTHER:
     num = jtailarray_num(&je->other.tails);
     for (i=0; i<num; i++) {
	jt = jtailarray_get(&je->other.tails, i);
	te = junbuild(ctx, jt->sub);
	jt->sub = NULL;
	jtail_assign(jt, te);
	jtail_destroy(ctx, jt);
	jtailarray_set(&je->other.tails, i, NULL);
     }
     break;
   }
   ret = je->head;
   je->head = NULL;
   jexpr_destroy(ctx, je);
   return ret;
}

////////////////////////////////////////////////////////////
// cost estimation

/*
 * XXX all bogus
 */

static void estimate_leaf(struct joinorder *ctx, struct jexpr *je,
			  struct cost *ret) {
   (void)ctx;
   (void)je;
   ret->height = 1;
   ret->width = 1;
   ret->work = 1;
}

static void estimate_join(struct joinorder *ctx,
			  struct jplan *left, struct jplan *right,
			  struct jpred *pred,
			  struct cost *ret) {
   (void)ctx;
   (void)pred;
   ret->height = left->cost.height * right->cost.height;
   ret->width = left->cost.width + right->cost.width;
   ret->work = left->cost.work + right->cost.work + ret->height * ret->width;
}

static void estimate_step(struct joinorder *ctx,
			  struct jplan *left, struct jexpr *right,
			  struct jpred *pred,
			  struct cost *ret) {
   (void)ctx;
   (void)right;
   (void)pred;
   ret->height = left->cost.height * 2 /*right->cost.height*/;
   ret->width = left->cost.width + 1 /*right->cost.width*/;
   ret->work = left->cost.work + 0 /*right->cost.work*/ +
      left->cost.height + ret->height;
}

////////////////////////////////////////////////////////////
// planner

/*
 * This code creates the trivial plan, that is, it creates the plan
 * such that junbuild reconstructs the input query.
 */
static void jplan_join_trivial(struct joinorder *ctx, struct jexpr *je) {
   struct jpred *pred;
   struct jexpr *leftje, *rightje;
   struct jplan *leftplan, *rightplan, *curplan;
   struct jplanarray saveplans;
   bool *used;
   unsigned leftsourceix, rightsourceix, numsourceixes;
   unsigned numpreds, predix;
   unsigned numsources, numnewsources, cursourcecount;
   unsigned k;

   PQLASSERT(je->type == JE_JOIN);
   PQLASSERT(je->join.plan == NULL);

   /*
    * The jexpr node was built in depth-first order, and so far we
    * don't have any predicate slicing code. So the first predicate
    * will always link the first source to the second, and the first
    * source will always be a readvar. After that, we can proceed by
    * inspecting the size of each predicate's sources array: if it's
    * larger by one we join with the next source; if it's exactly two
    * we save what we've got and start fresh; otherwise we combine
    * with the last thing we saved. If we get a completely empty
    * predicate, that means join the last *two* things in a straight
    * cartesian product with no filter condition.
    */

   jplanarray_init(&saveplans);
   numsourceixes = jexprarray_num(&je->join.sources);
   used = domalloc(ctx->pql, sizeof(*used) * numsourceixes);
   for (k=0; k<numsourceixes; k++) {
      used[k] = false;
   }

   pred = jpredarray_get(&je->join.predicates, 0);
   PQLASSERT(pred != NULL);

   if (indexarray_num(&pred->sources) == 0) {
      /* null predicate; do cartesian product of first two sources */
      PQLASSERT(pred->var == NULL);
      PQLASSERT(pred->expr == NULL);
      leftsourceix = 0;
      rightsourceix = 1;
   }
   else {
      PQLASSERT(indexarray_num(&pred->sources) == 2);
      leftsourceix = indexarray_get(&pred->sources, 0);
      rightsourceix = indexarray_get(&pred->sources, 1);
      PQLASSERT(leftsourceix == 0);
      PQLASSERT(rightsourceix == 1);
   }

   leftje = jexprarray_get(&je->join.sources, leftsourceix);
   rightje = jexprarray_get(&je->join.sources, rightsourceix);

   leftplan = mkjplan_leaf(ctx, FREE, 0);
   rightplan = mkjplan_leaf(ctx, FREE, 1);
   columnarray_copy_norefs(ctx->pql, &leftplan->producecols, &leftje->cols);
   columnarray_copy_norefs(ctx->pql, &rightplan->producecols, &rightje->cols);

   curplan = mkjplan_join(ctx, FREE, leftplan, rightplan, J_BASIC);
   indexarray_add(ctx->pql, &curplan->join.predicates, 0, NULL);
   columnarray_copy_norefs(ctx->pql, &curplan->producecols,
			   &leftplan->producecols);
   columnarray_copy_norefs(ctx->pql, &curplan->producecols,
			   &rightplan->producecols);

   used[leftsourceix] = true;
   used[rightsourceix] = true;
   cursourcecount = 2;

   numpreds = jpredarray_num(&je->join.predicates);
   for (predix = 1; predix < numpreds; predix++) {
      pred = jpredarray_get(&je->join.predicates, predix);

      numsources = indexarray_num(&pred->sources);
      numnewsources = 0;
      for (k=0; k<numsources; k++) {
	 if (indexarray_get(&pred->sources, k) >= cursourcecount) {
	    numnewsources++;
	 }
      }
      PQLASSERT(numnewsources <= 2);

      if (numsources == 0 && jplanarray_num(&saveplans) == 0) {
	 leftplan = curplan;

	 numnewsources++;
	 rightsourceix = cursourcecount;

	 rightje = jexprarray_get(&je->join.sources, rightsourceix);
	 rightplan = mkjplan_leaf(ctx, FREE, rightsourceix);
	 columnarray_copy_norefs(ctx->pql, &rightplan->producecols,
				 &rightje->cols);
	 PQLASSERT(!used[rightsourceix]);
	 used[rightsourceix] = true;
      }
      else if (/*numsources == cursourcecount + 1*/ numnewsources == 1) {
	 leftplan = curplan;

	 rightsourceix = indexarray_get(&pred->sources, numsources - 1);
	 PQLASSERT(rightsourceix == cursourcecount);
	 rightje = jexprarray_get(&je->join.sources, rightsourceix);
	 rightplan = mkjplan_leaf(ctx, FREE, rightsourceix);
	 columnarray_copy_norefs(ctx->pql, &rightplan->producecols,
				 &rightje->cols);
	 PQLASSERT(!used[rightsourceix]);
	 used[rightsourceix] = true;
      }
      else if (numnewsources == 2) {
	 PQLASSERT(numsources == 2);
	 jplanstack_push(ctx->pql, &saveplans, curplan);

	 leftsourceix = indexarray_get(&pred->sources, 0);
	 leftje = jexprarray_get(&je->join.sources, leftsourceix);
	 leftplan = mkjplan_leaf(ctx, FREE, leftsourceix);
	 columnarray_copy_norefs(ctx->pql, &leftplan->producecols,
				 &leftje->cols);

	 rightsourceix = indexarray_get(&pred->sources, 1);
	 rightje = jexprarray_get(&je->join.sources, rightsourceix);
	 rightplan = mkjplan_leaf(ctx, FREE, rightsourceix);
	 columnarray_copy_norefs(ctx->pql, &rightplan->producecols,
				 &rightje->cols);
	 PQLASSERT(!used[leftsourceix]);
	 used[leftsourceix] = true;
	 PQLASSERT(!used[rightsourceix]);
	 used[rightsourceix] = true;
      }
      else {
	 rightplan = curplan;
	 leftplan = jplanstack_pop(ctx->pql, &saveplans);
      }

      curplan = mkjplan_join(ctx, FREE, leftplan, rightplan, J_BASIC);
      indexarray_add(ctx->pql, &curplan->join.predicates, predix, NULL);
      columnarray_copy_norefs(ctx->pql, &curplan->producecols,
			      &leftplan->producecols);
      columnarray_copy_norefs(ctx->pql, &curplan->producecols,
			      &rightplan->producecols);

      //cursourcecount = numsources;
      cursourcecount += numnewsources;
   }

   je->join.plan = curplan;

   jplanarray_cleanup(ctx->pql, &saveplans);

   PQLASSERT(cursourcecount == numsourceixes);
   dofree(ctx->pql, used, sizeof(*used) * numsourceixes);
}

/*
 * This code does simple greedy planning, that is, it always picks the
 * cheapest thing to do next. This may or may not be suitable in the
 * long run but it ought to do adequately for the time being: because
 * we're always joining one table against itself, there are symmetries
 * that don't exist in the general relational join case. Or something
 * like that.
 */

#define NOWHERE ((unsigned)-1)

static void jplan_join_greedy(struct joinorder *ctx, struct jexpr *je) {

   /* Currently available subplans to work with. */
   struct jplanarray subplans;

   /* subplans[sourcemap[ix]] is where to find je->sources[ix] */
   struct indexarray sourcemap;

   /* subplans[predmap[ix]] is where to find je->predicates[ix] */
   struct indexarray predmap;

   /* List of possible work we can do */
   struct jworkarray work;

   /* Tables used while inspecting predicates */
   struct indexarray psources, psubplans;

   /* various counters */
   unsigned i, numsources, sourcesleft;
   unsigned j, numpredicates;
   unsigned k, num;

   /* scratch variables */
   struct jexpr *source;
   struct jpred *pred;
   struct jplan *newplan;
   struct jwork *jw;
   unsigned sourceix, subplanix;
   unsigned ns, np;
   unsigned lplanix, rplanix;
   struct jplan *lplan, *rplan;
   unsigned rsourceix;
   struct jexpr *rsource;
   struct cost cost;


   /*
    * Initialize locals.
    */
   jplanarray_init(&subplans);
   indexarray_init(&sourcemap);
   indexarray_init(&predmap);
   jworkarray_init(&work);
   indexarray_init(&psources);
   indexarray_init(&psubplans);

   numsources = jexprarray_num(&je->join.sources);
   sourcesleft = numsources;
   numpredicates = jpredarray_num(&je->join.predicates);

   /*
    * Initialize sourcemap to show none of the sources have been
    * integrated yet. Likewise for predmap.
    */
   indexarray_setsize(ctx->pql, &sourcemap, numsources);
   for (i=0; i<numsources; i++) {
      indexarray_set(&sourcemap, i, NOWHERE);
   }
   indexarray_setsize(ctx->pql, &predmap, numpredicates);
   for (i=0; i<numpredicates; i++) {
      indexarray_set(&predmap, i, NOWHERE);
   }

   /*
    * Continue looking for work to do until we're down to a single
    * plan component that covers everything.
    */
   while (sourcesleft > 0 || jplanarray_num(&subplans) > 1) {
      /*
       * There are four kinds of things we can do:
       *    (1) start a new subplan from one of the remaining sources;
       *    (2) combine an in-progress subplan with one of the
       *        remaining sources by applying one of the predicates in
       *        je->join.predicates[];
       *    (3) combine two subplans using one of the predicates;
       *    (4) take an open cartesian product of two subplans for which
       *        no predicate exists.
       */

      /* (1) Find all the new subplans we can start. */
      for (i=0; i<numsources; i++) {
	 if (indexarray_get(&sourcemap, i) != NOWHERE) {
	    continue;
	 }
	 source = jexprarray_get(&je->join.sources, i);
	 switch (source->type) {
	  case JE_SCAN:
	   /* we don't do scans, so don't consider starting here */
	   break;
	  case JE_JOIN:
	   /* not allowed */
	   PQLASSERT(0);
	   break;
	  case JE_READGLOBAL:
	  case JE_REPEAT:
	  case JE_OTHER:
	   estimate_leaf(ctx, source, &cost);
	   jw = jwork_create_leaf(ctx, &cost, i);
	   jworkarray_add(ctx->pql, &work, jw, NULL);
	   break;
	 }
      }

      /* (2/3) check all the predicates */
      for (j=0; j<numpredicates; j++) {
	 pred = jpredarray_get(&je->join.predicates, j);
	 if (indexarray_get(&predmap, j) != NOWHERE) {
	    /* already used */
	    continue;
	 }

	 /* figure out what this predicate connects */
	 num = indexarray_num(&pred->sources);
	 for (k=0; k<num; k++) {
	    sourceix = indexarray_get(&pred->sources, k);
	    subplanix = indexarray_get(&sourcemap, sourceix);
	    if (subplanix == NOWHERE) {
	       indexarray_add(ctx->pql, &psources, sourceix, NULL);
	    }
	    else {
	       indexarray_add(ctx->pql, &psubplans, subplanix, NULL);
	    }
	 }
	 indexarray_sortuniq(ctx->pql, &psources);
	 indexarray_sortuniq(ctx->pql, &psubplans);
	 ns = indexarray_num(&psources);
	 np = indexarray_num(&psubplans);

	 if (ns == 0 && np == 2) {
	    /* (3) two subplans */
	    lplanix = indexarray_get(&psubplans, 0);
	    rplanix = indexarray_get(&psubplans, 1);
	    lplan = jplanarray_get(&subplans, lplanix);
	    rplan = jplanarray_get(&subplans, rplanix);
	    /* XXX does left/right order matter? */
	    estimate_join(ctx, lplan, rplan, pred, &cost);
	    jw = jwork_create_join(ctx, &cost, lplanix, rplanix, j);
	    jworkarray_add(ctx->pql, &work, jw, NULL);
	 }
	 else if (ns == 1 && np == 1) {
	    /* (2) subplan and previously unused source */
	    lplanix = indexarray_get(&psubplans, 0);
	    lplan = jplanarray_get(&subplans, lplanix);
	    rsourceix = indexarray_get(&psources, 0);
	    rsource = jexprarray_get(&je->join.sources, rsourceix);
	    switch (rsource->type) {
	     case JE_SCAN:
	      estimate_step(ctx, lplan, rsource, pred, &cost);
	      jw = jwork_create_step(ctx, &cost, lplanix, rsourceix, j);
	      jworkarray_add(ctx->pql, &work, jw, NULL);
	      break;
	     case JE_READGLOBAL:
	      /*
	       * If we get here, it means that the subplan for reading
	       * the global was more expensive than getting to the
	       * global some other way. This probably means we should
	       * try to use a reverse lookup index for the global if
	       * that's available. XXX (future)
	       */
	      break;
	     case JE_JOIN:
	      /* not allowed */
	      PQLASSERT(0);
	      break;
	     case JE_REPEAT:
	     case JE_OTHER:
	      /*
	       * No leaf subplan for this source yet? Wait until one's
	       * been established.
	       */
	      break;
	    }
	 }
	 indexarray_setsize(ctx->pql, &psources, 0);
	 indexarray_setsize(ctx->pql, &psubplans, 0);
      }

      /*
       * (4) cartesian products
       *
       * As this is inherently expensive, don't do it unless we're
       * otherwise stuck. Assert that we do have at least two subplans
       * to operate on.
       *
       * Because paths are rooted we should never get into a situation
       * where there are sources left that the logic above can't pull
       * in: we can always start at the path root and work forwards.
       * Therefore, before we reach the point of needing to do
       * unfiltered cartesian products, every source should have been
       * added to a subplan; if the outer loop hasn't stopped we must
       * have more than one subplan in the table.
       *
       * Which is nice, but it's not true. We might have a predicate
       * that relates three things, which can be e.g. two existing
       * subplans and a remaining source. Because we don't do our own
       * predicate splitting, this can prevent roping in the remaining
       * source without first taking the cartesian product of the two
       * existing subplans.
       *
       * XXX it's not clear if, given predicate splitting, there are
       * still other cases where there'll be untouched sources
       * remaining.
       */
      if (jworkarray_num(&work) == 0) {
	 //PQLASSERT(sourcesleft == 0); // wrong
	 num = jplanarray_num(&subplans);
	 PQLASSERT(num > 1);
	 for (lplanix=0; lplanix<num; lplanix++) {
	    lplan = jplanarray_get(&subplans, lplanix);
	    for (rplanix=lplanix+1; rplanix<num; rplanix++) {
	       rplan = jplanarray_get(&subplans, rplanix);
	       /* XXX does left/right order matter? */
	       estimate_join(ctx, lplan, rplan, NULL, &cost);
	       jw = jwork_create_join(ctx, &cost, lplanix, rplanix, NOWHERE);
	       jworkarray_add(ctx->pql, &work, jw, NULL);
	    }
	 }
      }

      /*
       * Now we have a list of things we can do. Sort it by cost and
       * take the winner.
       */
      jworkarray_sort(ctx, &work);
      jw = jworkarray_get(&work, 0);
      switch (jw->type) {

       case JW_LEAF:
	/* make a leaf plan using the indicated source */
	source = jexprarray_get(&je->join.sources, jw->leaf.sourceix);
	newplan = mkjplan_leaf(ctx, &jw->cost, jw->leaf.sourceix);
	columnarray_copy_norefs(ctx->pql, &newplan->producecols,
				&source->cols);
	/* add it to the current subplans */
	jplanarray_add(ctx->pql, &subplans, newplan, &subplanix);
	/* update sourcemap */
	indexarray_set(&sourcemap, jw->leaf.sourceix, subplanix);
	sourcesleft--;
	break;

       case JW_STEP:
	/* get the left-hand subplan */
	lplan = jplanarray_get(&subplans, jw->step.leftplanix);
	PQLASSERT(lplan != NULL);

	/* make the right-hand subplan from the indicated source */
	rsource = jexprarray_get(&je->join.sources, jw->step.rightsourceix);
	rplan = mkjplan_leaf(ctx, FREE, jw->step.rightsourceix);
	columnarray_copy_norefs(ctx->pql, &rplan->producecols, &rsource->cols);

	/* make the new join subplan */
	newplan = mkjplan_join(ctx, &jw->cost, lplan, rplan, J_STEP);
	columnarray_copy_norefs(ctx->pql, &newplan->producecols,
				&lplan->producecols);
	columnarray_copy_norefs(ctx->pql, &newplan->producecols,
				&rplan->producecols);
	if (jw->step.predix != NOWHERE) {
	   indexarray_add(ctx->pql, &newplan->join.predicates,
			  jw->step.predix, NULL);
	}

	/* replace the left-hand subplan in the current subplans */
	jplanarray_set(&subplans, jw->step.leftplanix, newplan);

	/* update sourcemap and predmap */
	indexarray_set(&sourcemap, jw->step.rightsourceix,jw->step.leftplanix);
	sourcesleft--;
	if (jw->step.predix != NOWHERE) {
	   indexarray_set(&predmap, jw->step.predix, jw->step.leftplanix);
	}

	break;

       case JW_JOIN:
	/* get the left-hand subplan */
	lplan = jplanarray_get(&subplans, jw->join.leftplanix);

	/* get the right-hand subplan */
	rplan = jplanarray_get(&subplans, jw->join.rightplanix);

	/* make the new join subplan */
	newplan = mkjplan_join(ctx, &jw->cost, lplan, rplan, J_BASIC);
	columnarray_copy_norefs(ctx->pql, &newplan->producecols,
				&lplan->producecols);
	columnarray_copy_norefs(ctx->pql, &newplan->producecols,
				&rplan->producecols);
	if (jw->join.predix != NOWHERE) {
	   indexarray_add(ctx->pql, &newplan->join.predicates,
			  jw->join.predix, NULL);
	}

	/* replace the left-hand subplan in the current subplans */
	jplanarray_set(&subplans, jw->join.leftplanix, newplan);

	/* clear the right-hand subplan in the current subplans */
	jplanarray_set(&subplans, jw->join.rightplanix, NULL);
	jplanarray_nonulls(ctx->pql, &subplans);

	/* update sourcemap and predmap */
	for (i=0; i<numsources; i++) {
	   if (indexarray_get(&sourcemap, i) == jw->join.rightplanix) {
	      indexarray_set(&sourcemap, i, jw->join.leftplanix);
	   }
	}
	for (j=0; j<numpredicates; j++) {
	   if (indexarray_get(&predmap, j) == jw->join.rightplanix) {
	      indexarray_set(&predmap, j, jw->join.leftplanix);
	   }
	}
	if (jw->join.predix != NOWHERE) {
	   indexarray_set(&predmap, jw->join.predix, jw->join.leftplanix);
	}

	break;
      }

      /*
       * Ideally we would destroy only the work element we did, update
       * the rest, and add only the new ones that are now possible
       * after what we just did. But for now at least that's way too
       * complicated, so just nuke it all and start over.
       */
      jworkarray_destroyall(ctx->pql, &work);
   }

   /*
    * Install result
    */
   je->join.plan = jplanarray_get(&subplans, 0);

   /*
    * Clean up locals
    */
   jplanarray_setsize(ctx->pql, &subplans, 0);
   jplanarray_cleanup(ctx->pql, &subplans);
   indexarray_setsize(ctx->pql, &sourcemap, 0);
   indexarray_cleanup(ctx->pql, &sourcemap);
   indexarray_setsize(ctx->pql, &predmap, 0);
   indexarray_cleanup(ctx->pql, &predmap);
   jworkarray_cleanup(ctx->pql, &work);
   indexarray_cleanup(ctx->pql, &psources);
   indexarray_cleanup(ctx->pql, &psubplans);
}

/*
 * Switch. For now at least I want both pieces of planner code compiled,
 * so we'll pretend to allow choosing at runtime.
 */
static void jplan_join(struct joinorder *ctx, struct jexpr *je) {
   PQLASSERT(je->type == JE_JOIN);
   PQLASSERT(je->join.plan == NULL);

   if (1) {
      jplan_join_greedy(ctx, je);
   }
   else {
      jplan_join_trivial(ctx, je);
   }
}

static void jplan(struct joinorder *ctx, struct jexpr *je) {
   unsigned i, num;

   switch (je->type) {
    case JE_SCAN:
     /* nothing to do */
     break;
    case JE_READGLOBAL:
     /* nothing to do */
     break;
    case JE_JOIN:
     num = jexprarray_num(&je->join.sources);
     for (i=0; i<num; i++) {
	jplan(ctx, jexprarray_get(&je->join.sources, i));
     }
     jplan_join(ctx, je);
     break;
    case JE_REPEAT:
     jplan(ctx, je->repeat.sub);
     jplan(ctx, je->repeat.body);
     break;
    case JE_OTHER:
     num = jtailarray_num(&je->other.tails);
     for (i=0; i<num; i++) {
	jplan(ctx, jtailarray_get(&je->other.tails, i)->sub);
     }
     break;
   }
}

////////////////////////////////////////////////////////////
// overall

static struct tcexpr *tcexpr_joinorder(struct joinorder *ctx,
				       struct tcexpr *te) {
   struct jexpr *je;

   je = jbuild(ctx, te);
   jmerge(ctx, je);
   jpredicate(ctx, je);

   jplan(ctx, je);

   te = junbuild(ctx, je);
   return te;
}

////////////////////////////////////////////////////////////
// statements

static struct tcstmt *tcstmt_joinorder(struct joinorder *ctx,
				       struct tcstmt *ts) {
   switch (ts->type) {
    case TCS_EXPR:
     ts->expr = tcexpr_joinorder(ctx, ts->expr);
     break;
    case TCS_BEGIN:
    case TCS_ABORT:
    case TCS_COMMIT:
     break;
   }
   return ts;
}

////////////////////////////////////////////////////////////
// entry point

struct tcstmt *joinorder(struct pqlcontext *pql, struct tcstmt *ts) {
   struct joinorder ctx;

   joinorder_init(&ctx, pql);
   ts = tcstmt_joinorder(&ctx, ts);
   joinorder_cleanup(&ctx);

   return ts;
}
