/*
 * Copyright 2009, 2010
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Apply and remove all rename nodes.
 *
 * Because all our column names are distinguished (even if the user
 * recklessly gives some of them the same string names), there's no
 * danger of any nonsense about variable capture or conflating things
 * that are supposed to be different.
 *
 * Also, because all the names arise structurally (as opposed to being
 * read from the database on the fly) we *can* do this and have no
 * renames left afterwards.
 */

#include "datatype.h"
#include "tcalc.h"
#include "passes.h"

struct norenames {
   struct pqlcontext *pql;
   struct tcexprarray renamelist;
};

////////////////////////////////////////////////////////////
// context management

static void norenames_init(struct norenames *ctx, struct pqlcontext *pql) {
   ctx->pql = pql;
   tcexprarray_init(&ctx->renamelist);
}

static void norenames_cleanup(struct norenames *ctx) {
   PQLASSERT(tcexprarray_num(&ctx->renamelist) == 0);
   tcexprarray_cleanup(ctx->pql, &ctx->renamelist);
}

////////////////////////////////////////////////////////////
// the works

static struct tcexpr *tcexpr_norenames(struct norenames *ctx,
				       struct tcexpr *te);

/*
 * Traverse a datatype and rename columns inside it. Note: while we
 * could use datatype_rename for the common case, here we have to do a
 * complete traversal rather than just adjusting the top level.
 */
static struct datatype *datatype_norenames_one(struct norenames *ctx,
					       struct datatype *t,
					       struct column *insidecol,
					       struct column *outsidecol) {
   if (datatype_isset(t)) {
      t = datatype_norenames_one(ctx, datatype_set_member(t),
				 insidecol, outsidecol);
      return datatype_set(ctx->pql, t);
   }
   if (datatype_issequence(t)) {
      t = datatype_norenames_one(ctx, datatype_sequence_member(t),
				 insidecol, outsidecol);
      return datatype_sequence(ctx->pql, t);
   }
   if (datatype_istuple(t)) {
      /* bleh, what an ugly mess */
      struct datatypearray types;
      struct columnarray cols;
      struct datatype *subtype;
      struct column *subcol;
      unsigned i, num;

      datatypearray_init(&types);
      columnarray_init(&cols);

      num = datatype_arity(t);
      datatypearray_setsize(ctx->pql, &types, num);
      columnarray_setsize(ctx->pql, &cols, num);
      for (i=0; i<num; i++) {
	 subtype = datatype_getnth(t, i);
	 subcol = datatype_tuple_getcolumn(t, i);

	 subtype = datatype_norenames_one(ctx, subtype, insidecol, outsidecol);
	 if (subcol == insidecol) {
	    // no - this is an unowned reference
	    //column_decref(ctx->pql, insidecol);
	    //column_incref(outsidecol);
	    subcol = outsidecol;
	 }

	 datatypearray_set(&types, i, subtype);
	 columnarray_set(&cols, i, subcol);
      }

      t = datatype_tuple_specific(ctx->pql, &types, &cols);

      
      datatypearray_setsize(ctx->pql, &types, 0);
      columnarray_setsize(ctx->pql, &cols, 0);
      datatypearray_cleanup(ctx->pql, &types);
      columnarray_cleanup(ctx->pql, &cols);
      
      return t;
   }
   /* nothing else contains subtypes */
   return t;
}

static struct datatype *datatype_norenames(struct norenames *ctx,
					   struct datatype *t) {
   unsigned i, num;
   struct tcexpr *rn;
   struct datatype *nt;

   num = tcexprarray_num(&ctx->renamelist);
   for (i=num; i-- > 0; ) {
      rn = tcexprarray_get(&ctx->renamelist, i);
      PQLASSERT(rn->type == TCE_RENAME);
      PQLASSERT(rn->rename.insidecol != NULL);
      PQLASSERT(rn->rename.outsidecol != NULL);
      nt = datatype_norenames_one(ctx, t,
				  rn->rename.insidecol, rn->rename.outsidecol);
      if (nt != NULL) {
	 t = nt;
	 /*
	  * Note: it is tempting to break from the loop here, but
	  * doing so would be wrong: there can be and sometimes are
	  * chains of renames applied to the same starting column.
	  */
      }
   }
   return t;
}

static void columnarray_replace(struct norenames *ctx,
				struct columnarray *arr,
				struct column *oldcol,
				struct column *newcol) {
   unsigned i, num;

   num = columnarray_num(arr);
   for (i=0; i<num; i++) {
      if (columnarray_get(arr, i) == oldcol) {
	 column_decref(ctx->pql, oldcol);
	 column_incref(newcol);
	 columnarray_set(arr, i, newcol);
      }
   }
}

static void columnarray_norenames(struct norenames *ctx,
				  struct columnarray *arr) {
   unsigned i, num;
   struct tcexpr *rn;

   if (arr == NULL) {
      return;
   }

   num = tcexprarray_num(&ctx->renamelist);
   for (i=num; i-- > 0; ) {
      rn = tcexprarray_get(&ctx->renamelist, i);
      PQLASSERT(rn->type == TCE_RENAME);
      columnarray_replace(ctx, arr, rn->rename.insidecol,
			  rn->rename.outsidecol);
   }
}

static struct column *column_norenames(struct norenames *ctx,
				       struct column *col) {
   unsigned i, num;
   struct tcexpr *rn;

   if (col == NULL) {
      return NULL;
   }

   num = tcexprarray_num(&ctx->renamelist);
   for (i=num; i-- > 0; ) {
      rn = tcexprarray_get(&ctx->renamelist, i);
      PQLASSERT(rn->type == TCE_RENAME);
      if (col == rn->rename.insidecol) {
	 column_decref(ctx->pql, rn->rename.insidecol);
	 column_incref(rn->rename.outsidecol);
	 col = rn->rename.outsidecol;
      }
   }
   return col;
}

static void tcvar_norenames(struct norenames *ctx, struct tcvar *var) {
   if (var == NULL) {
      return;
   }
   var->datatype = datatype_norenames(ctx, var->datatype);
}

static void tcexprarray_norenames(struct norenames *ctx,
				  struct tcexprarray *arr) {
   struct tcexpr *sub;
   unsigned i, num;
   
   num = tcexprarray_num(arr);
   for (i=0; i<num; i++) {
      sub = tcexprarray_get(arr, i);
      sub = tcexpr_norenames(ctx, sub);
      tcexprarray_set(arr, i, sub);
   }
}

static struct tcexpr *tcexpr_norenames(struct norenames *ctx,
				       struct tcexpr *te) {
   if (te == NULL) {
      return NULL;
   }
   switch (te->type) {
    case TCE_FILTER:
     te->filter.sub = tcexpr_norenames(ctx, te->filter.sub);
     tcvar_norenames(ctx, te->filter.predvar);
     te->filter.predexpr = tcexpr_norenames(ctx, te->filter.predexpr);
     break;
    case TCE_PROJECTTUPLE:
     te->projecttuple.sub = tcexpr_norenames(ctx, te->projecttuple.sub);
     columnarray_norenames(ctx, te->projecttuple.cols);
     break;
    case TCE_PROJECTSCALAR:
     te->projectscalar.sub = tcexpr_norenames(ctx, te->projectscalar.sub);
     te->projectscalar.col = column_norenames(ctx, te->projectscalar.col);
     break;
    case TCE_RENAME:
     {
	struct tcexpr *sub;
	unsigned pos;

	tcexprarray_add(ctx->pql, &ctx->renamelist, te, &pos);
	sub = te->rename.sub;
	te->rename.sub = NULL;
	sub = tcexpr_norenames(ctx, sub);
	PQLASSERT(tcexprarray_num(&ctx->renamelist) == pos+1);
	tcexprarray_setsize(ctx->pql, &ctx->renamelist, pos);

	tcexpr_destroy(ctx->pql, te);
	return sub;
     }
     break;
    case TCE_JOIN:
     te->join.left = tcexpr_norenames(ctx, te->join.left);
     te->join.right = tcexpr_norenames(ctx, te->join.right);
     tcvar_norenames(ctx, te->join.predvar);
     te->join.predexpr = tcexpr_norenames(ctx, te->join.predexpr);
     break;
    case TCE_ORDER:
     te->order.sub = tcexpr_norenames(ctx, te->order.sub);
     columnarray_norenames(ctx, te->order.cols);
     break;
    case TCE_UNIQ:
     te->uniq.sub = tcexpr_norenames(ctx, te->uniq.sub);
     columnarray_norenames(ctx, te->uniq.cols);
     break;
    case TCE_NEST:
     te->nest.sub = tcexpr_norenames(ctx, te->nest.sub);
     columnarray_norenames(ctx, te->nest.cols);
     te->nest.newcol = column_norenames(ctx, te->nest.newcol);
     break;
    case TCE_UNNEST:
     te->unnest.sub = tcexpr_norenames(ctx, te->unnest.sub);
     te->unnest.col = column_norenames(ctx, te->unnest.col);
     break;
    case TCE_DISTINGUISH:
     te->distinguish.sub = tcexpr_norenames(ctx, te->distinguish.sub);
     te->distinguish.newcol = column_norenames(ctx, te->distinguish.newcol);
     break;
    case TCE_ADJOIN:
     te->adjoin.left = tcexpr_norenames(ctx, te->adjoin.left);
     tcvar_norenames(ctx, te->adjoin.var);
     te->adjoin.expr = tcexpr_norenames(ctx, te->adjoin.expr);
     te->adjoin.newcol = column_norenames(ctx, te->adjoin.newcol);
     break;
    case TCE_STEP:
     /* optimization result, should not appear here */
     PQLASSERT(0);
     break;
    case TCE_REPEAT:
     te->repeat.sub = tcexpr_norenames(ctx, te->repeat.sub);
     te->repeat.subendcolumn = column_norenames(ctx, te->repeat.subendcolumn);
     tcvar_norenames(ctx, te->repeat.loopvar);
     te->repeat.bodystartcolumn = column_norenames(ctx,
						  te->repeat.bodystartcolumn);
     te->repeat.body = tcexpr_norenames(ctx, te->repeat.body);
     te->repeat.bodybodycolumn = column_norenames(ctx,
						  te->repeat.bodybodycolumn);
     te->repeat.bodypathcolumn = column_norenames(ctx,
						  te->repeat.bodypathcolumn);
     te->repeat.bodyendcolumn = column_norenames(ctx,
						  te->repeat.bodyendcolumn);
     te->repeat.repeatpathcolumn = column_norenames(ctx,
						  te->repeat.repeatpathcolumn);
     te->repeat.repeatendcolumn = column_norenames(ctx,
						  te->repeat.repeatendcolumn);
     break;
    case TCE_SCAN:
     te->scan.leftobjcolumn = column_norenames(ctx, te->scan.leftobjcolumn);
     te->scan.edgecolumn = column_norenames(ctx, te->scan.edgecolumn);
     te->scan.rightobjcolumn = column_norenames(ctx, te->scan.rightobjcolumn);
     break;
    case TCE_BOP:
     te->bop.left = tcexpr_norenames(ctx, te->bop.left);
     te->bop.right = tcexpr_norenames(ctx, te->bop.right);
     break;
    case TCE_UOP:
     te->uop.sub = tcexpr_norenames(ctx, te->uop.sub);
     break;
    case TCE_FUNC:
     tcexprarray_norenames(ctx, &te->func.args);
     break;
    case TCE_MAP:
     tcvar_norenames(ctx, te->map.var);
     te->map.set = tcexpr_norenames(ctx, te->map.set);
     te->map.result = tcexpr_norenames(ctx, te->map.result);
     break;
    case TCE_LET:
     tcvar_norenames(ctx, te->let.var);
     te->let.value = tcexpr_norenames(ctx, te->let.value);
     te->let.body = tcexpr_norenames(ctx, te->let.body);
     break;
    case TCE_READVAR:
     tcvar_norenames(ctx, te->readvar);
     break;
    case TCE_READGLOBAL:
     break;
    case TCE_CREATEPATHELEMENT:
     te->createpathelement = tcexpr_norenames(ctx, te->createpathelement);
     break;
    case TCE_SPLATTER:
     te->splatter.value = tcexpr_norenames(ctx, te->splatter.value);
     te->splatter.name = tcexpr_norenames(ctx, te->splatter.name);
     break;
    case TCE_TUPLE:
     tcexprarray_norenames(ctx, &te->tuple.exprs);
     columnarray_norenames(ctx, te->tuple.columns);
     break;
    case TCE_VALUE:
     break;
   }
   te->datatype = datatype_norenames(ctx, te->datatype);
   return te;
}

static struct tcstmt *tcstmt_norenames(struct norenames *ctx,
				       struct tcstmt *ts) {
   switch (ts->type) {
    case TCS_EXPR:
     ts->expr = tcexpr_norenames(ctx, ts->expr);
     break;
    case TCS_BEGIN:
    case TCS_ABORT:
    case TCS_COMMIT:
     break;
   }
   return ts;
}

////////////////////////////////////////////////////////////
// entry points

struct tcexpr *norenames_expr(struct pqlcontext *pql, struct tcexpr *te) {
   struct norenames ctx;

   norenames_init(&ctx, pql);
   te = tcexpr_norenames(&ctx, te);
   norenames_cleanup(&ctx);
   return te;
}

struct tcstmt *norenames(struct pqlcontext *pql, struct tcstmt *ts) {
   struct norenames ctx;

   norenames_init(&ctx, pql);
   ts = tcstmt_norenames(&ctx, ts);
   norenames_cleanup(&ctx);
   return ts;
}
