/*
 * Copyright 2009
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "utils.h"
#include "pql.h"
#include "pttree.h"  // for ptmanager
#include "tcalc.h"  // for tcexpr_destroy
#include "passes.h"
#include "pqlcontext.h"

struct pqlquery {
   struct pqlcontext *pql;
   struct tcstmt *ts;
};

struct pqlquery *pqlquery_compile(struct pqlcontext *pql,
				  const char *textbuf, size_t len) {
   struct pqlquery *q;
   struct ptstmt *ps;
   struct tcstmt *ts;

   /*
    * Parse
    */

   /* Should not be a parse in progress... */
   PQLASSERT(pql->pcb == NULL);

   ps = parse(pql, textbuf, len);

   /* ...and the parser is supposed to clean up after itself. */
   PQLASSERT(pql->pcb == NULL);

   if (ps == NULL) {
      return NULL;
   }
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_PARSER, ptdump(pql, ps));
   }

   /*
    * Resolve variable references
    */

   ps = resolvevars(pql, ps);
   if (ps == NULL) {
      ptmanager_destroyall(pql->ptm);
      return NULL;
   }
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_RESOLVE, ptdump(pql, ps));
   }

   /*
    * Normalize everything, especially paths.
    */
   ps = normalize(pql, ps);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_NORMALIZE, ptdump(pql, ps));
   }

   /*
    * Unify common path segments.
    */
   ps = unify(pql, ps);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_UNIFY, ptdump(pql, ps));
   }

   /*
    * Move all paths to the from-clause. (Note: not for the
    * set-canonicalization world, if that ever reappears.)
    */
   ps = movepaths(pql, ps);
   if (ps == NULL) {
      ptmanager_destroyall(pql->ptm);
      return NULL;
   }
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_MOVEPATHS, ptdump(pql, ps));
   }

   /*
    * Arrange to deal with variables whose bindings are skippable.
    */
   ps = bindnil(pql, ps);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_BINDNIL, ptdump(pql, ps));
   }

   /*
    * Turn quantifier expressions into map expressions.
    */
   ps = dequantify(pql, ps);

   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_DEQUANTIFY, ptdump(pql, ps));
   }

   /*
    * Convert to tuple calculus
    */

   ts = tuplify(pql, ps);
   //ptexpr_destroy(ps); -- handled by region allocator
   ptmanager_destroyall(pql->ptm);
   ps = NULL;

   if (ts == NULL) {
      /* not supposed to happen */
      PQLASSERT(0);
      return NULL;
   }

   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_TUPLIFY, tcdump(pql, ts, false));
   }

   /*
    * Infer types
    */
   typeinf(pql, ts);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_TYPEINF, tcdump(pql, ts, true));
   }

   /*
    * Check the resulting types
    */
   if (typecheck(pql, ts) < 0) {
      tcstmt_destroy(pql, ts);
      return NULL;
   }

   /*
    * First, scratch out all the rename operations.
    */
   ts = norenames(pql, ts);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_NORENAMES, tcdump(pql, ts, false));
   }
   if (typecheck(pql, ts) < 0) {
      complain(pql, 0, 0, "Typecheck failed after norenames");
      tcstmt_destroy(pql, ts);
      return NULL;
   }

   /*
    * Do a quick pass that removes unnecessary let-bindings.
    */
   ts = unlet(pql, ts);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_UNLET, tcdump(pql, ts, false));
   }
   if (typecheck(pql, ts) < 0) {
      complain(pql, 0, 0, "Typecheck failed after unlet");
      tcstmt_destroy(pql, ts);
      return NULL;
   }

   /*
    * Do basic optimizations.
    */
   ts = baseopt(pql, ts);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_BASEOPT, tcdump(pql, ts, false));
   }
   if (typecheck(pql, ts) < 0) {
      complain(pql, 0, 0, "Typecheck failed after baseopt");
      tcstmt_destroy(pql, ts);
      return NULL;
   }

   /*
    * Build def-use chains and do things those enable.
    */
   ts = defuse(pql, ts);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_DEFUSE, tcdump(pql, ts, false));
   }
   if (typecheck(pql, ts) < 0) {
      complain(pql, 0, 0, "Typecheck failed after defuse");
      tcstmt_destroy(pql, ts);
      return NULL;
   }

   /*
    * Figure out a good evaluation order.
    */
   ts = joinorder(pql, ts);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_JOINORDER, tcdump(pql, ts, false));
   }
   if (typecheck(pql, ts) < 0) {
      complain(pql, 0, 0, "Typecheck failed after joinorder");
      tcstmt_destroy(pql, ts);
      return NULL;
   }

   /*
    * Run another baseopt after joinorder, because it sometimes leaves
    * crap behind.
    */
   ts = baseopt(pql, ts);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_JOINORDERBASEOPT, tcdump(pql, ts, false));
   }
   if (typecheck(pql, ts) < 0) {
      complain(pql, 0, 0, "Typecheck failed after post-joinorder baseopt");
      tcstmt_destroy(pql, ts);
      return NULL;
   }

   /*
    * Find joins that can be handled as graph step operations.
    */
   ts = stepjoins(pql, ts);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_STEPJOINS, tcdump(pql, ts, false));
   }
   if (typecheck(pql, ts) < 0) {
      complain(pql, 0, 0, "Typecheck failed after stepjoins");
      tcstmt_destroy(pql, ts);
      return NULL;
   }

   /*
    * Do another unlet and then baseopt pass.
    */
   ts = unlet(pql, ts);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_FINALUNLET, tcdump(pql, ts, false));
   }
   if (typecheck(pql, ts) < 0) {
      complain(pql, 0, 0, "Typecheck failed after final unlet");
      tcstmt_destroy(pql, ts);
      return NULL;
   }

   ts = baseopt(pql, ts);
   if (pql->dodumps) {
      pqlcontext_adddump(pql, DUMP_FINALBASEOPT, tcdump(pql, ts, false));
   }
   if (typecheck(pql, ts) < 0) {
      complain(pql, 0, 0, "Typecheck failed after final baseopt");
      tcstmt_destroy(pql, ts);
      return NULL;
   }

   q = domalloc(pql, sizeof(*q));
   q->pql = pql;
   q->ts = ts;
   return q;
}

void pqlquery_destroy(struct pqlquery *q) {
   tcstmt_destroy(q->pql, q->ts);
   q->ts = NULL;
   dofree(q->pql, q, sizeof(*q));
}

struct pqlvalue *pqlquery_run(struct pqlcontext *pql,
			      const struct pqlquery *q) {
   PQLASSERT(pql == q->pql);

   return eval(pql, q->ts);
}
