/*
 * Copyright 2009
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Pick joins that can be done as graph steps.
 */

#include "datatype.h"
#include "pqlvalue.h"
#include "tcalc.h"
#include "passes.h"

struct stepjoins {
   struct pqlcontext *pql;
};

////////////////////////////////////////////////////////////
// context management

static void stepjoins_init(struct stepjoins *ctx, struct pqlcontext *pql) {
   ctx->pql = pql;
}

static void stepjoins_cleanup(struct stepjoins *ctx) {
   (void)ctx;
}

////////////////////////////////////////////////////////////
// tools

/*
 * Given:
 *    lambda V1: P1
 *    lambda V2: P2
 * produce
 *    lambda V3: P3
 * where
 *    P3 === P1 and P2
 *
 * Pretend NULL expressions are "true".
 *
 * The inputs are consumed; replacements are produced.
 *
 * XXX this should be shared with the similar code in baseopt.c.
 */
static void combine_predicates(struct stepjoins *ctx,
			       struct tcvar *v1,
			       struct tcexpr *p1,
			       struct tcvar *v2,
			       struct tcexpr *p2,
			       struct tcvar **vret,
			       struct tcexpr **pret) {
   struct tcexpr *bop, *readvar, *proj, *let;
   struct datatype *vt1, *vt2, *subtype1, *subtype2;
   struct column *subcol1, *subcol2;
   unsigned arity1, arity2, prefixlen, i;

   if (p1 == NULL) {
      PQLASSERT(v1 == NULL);
      *vret = v2;
      *pret = p2;
      return;
   }
   if (p2 == NULL) {
      PQLASSERT(v2 == NULL);
      *vret = v1;
      *pret = p1;
      return;
   }

   PQLASSERT(datatype_isbool(p1->datatype));
   PQLASSERT(datatype_isbool(p2->datatype));

   if (p1->type == TCE_VALUE &&
       pqlvalue_isbool(p1->value) &&
       pqlvalue_bool_get(p1->value) == true) {
      tcexpr_destroy(ctx->pql, p1);
      if (v1 != NULL) {
	 tcvar_decref(ctx->pql, v1);
      }
      *vret = v2;
      *pret = p2;
      return;
   }
#if 0 /* does not work w/o more help - p1 has the wrong type */
   if (p2->type == TCE_VALUE &&
       pqlvalue_isbool(p2->value) &&
       pqlvalue_bool_get(p2->value) == true) {
      tcexpr_destroy(ctx->pql, p1);
      tcvar_decref(ctx->pql, v1);
      *vret = v1;
      *pret = p1;
      return;
   }
#endif

   if (v1 == NULL) {
      *vret = v2;
      *pret = mktcexpr_bop(ctx->pql, p1, F_AND, p2);
      return;
   }
   if (v2 == NULL) {
      *vret = v1;
      *pret = mktcexpr_bop(ctx->pql, p1, F_AND, p2);
      return;
   }

   vt1 = v1->datatype;
   vt2 = v2->datatype;

   /*
    * We expect that vt1 is the suffix of vt2; that is, p2 is
    * something x p1.
    *
    * I'm not entirely sure if this order makes sense, but it matches
    * the existing call site and I don't want to change it around
    * without a reason.
    */
   PQLASSERT(datatype_istuple(vt1));
   PQLASSERT(datatype_istuple(vt2));

   arity1 = datatype_arity(vt1);
   arity2 = datatype_arity(vt2);
   PQLASSERT(arity1 <= arity2);
   prefixlen = arity2 - arity1;
   for (i=prefixlen; i<arity2; i++) {
      subtype1 = datatype_getnth(vt1, i-prefixlen);
      subcol1 = datatype_tuple_getcolumn(vt1, i-prefixlen);
      subtype2 = datatype_getnth(vt2, i);
      subcol2 = datatype_tuple_getcolumn(vt2, i);
      PQLASSERT(datatype_eq_withnames(subtype1, subtype2));
      PQLASSERT(subcol1 == subcol2);
   }

   /*
    * Because p2 is a wider tuple than p1, use p2 as the resulting
    * lambda expression (leaving its var alone) and let-bind p1's
    * var.
    *
    * Blah, I really want baseopt's wrapvar() here. That would be
    * far superior to having to find this project later and push
    * it through the let-binding.
    *
    * XXX fix this when wrapvar() gets moved out of baseopt.
    */
   bop = mktcexpr_bop(ctx->pql, p1, F_AND, p2);
   bop->datatype = datatype_bool(ctx->pql);

   tcvar_incref(v2);
   readvar = mktcexpr_readvar(ctx->pql, v2);
   readvar->datatype = v2->datatype;

   proj = mktcexpr_projecttuple(ctx->pql, readvar);
   columnarray_setsize(ctx->pql, proj->projecttuple.cols, arity1);
   for (i=0; i<arity1; i++) {
      subcol1 = datatype_tuple_getcolumn(vt1, i);
      column_incref(subcol1);
      columnarray_set(proj->projecttuple.cols, i, subcol1);
   }
   proj->datatype = vt1;

   let = mktcexpr_let(ctx->pql, v1, proj, bop);
   let->datatype = bop->datatype;

   *vret = v2;
   *pret = let;
}

/*
 * Check if COL is at the top level of tuple type T.
 */
static bool tuple_contains_toplevel(struct datatype *t, struct column *col) {
   unsigned i, num;

   PQLASSERT(datatype_istuple(t));
   num = datatype_arity(t);
   for (i=0; i<num; i++) {
      if (datatype_tuple_getcolumn(t, i) == col) {
	 return true;
      }
   }
   return false;
}

////////////////////////////////////////////////////////////
// the real work

/*
 * Inspect the predicate on a join to see if (part of) it corresponds
 * to a step operation from one of the top-level columns in TYPE. If
 * so, update the predicate to remove that condition, set *BACK_RET
 * based on whether the step is forward or backward, set *COL_RET to
 * the column in question, and return true. If not, return false.
 */
static bool check_join_predicate(struct stepjoins *ctx,
				 struct tcexpr **predicate_inout,
				 struct datatype *type,
				 struct tcvar *lvar,
				 struct column *leftcol,
				 struct column *rightcol,
				 bool *back_ret,
				 struct column **col_ret) {
   struct tcexpr *predicate;
   struct column *c1, *c2;

   predicate = *predicate_inout;

   /*
    * We expect the predicate to be of the form T1 && T2 && ..., since
    * that's what's generated upstream. If one of the subexpressions
    * compares a column in COLS to either LEFTCOL or RIGHTCOL, we win.
    * Prune the subexpression in question.
    */

   if (predicate->type == TCE_BOP && predicate->bop.op == F_AND) {
      if (check_join_predicate(ctx, &predicate->bop.left, type, 
			       lvar, leftcol, rightcol, back_ret, col_ret)) {
	 return true;
      }
      if (check_join_predicate(ctx, &predicate->bop.right, type,
			       lvar, leftcol, rightcol, back_ret, col_ret)) {
	 return true;
      }
   }

   if (predicate->type == TCE_BOP && predicate->bop.op == F_EQ) {
      if (predicate->bop.left->type == TCE_PROJECTSCALAR &&
	  predicate->bop.left->projectscalar.sub->type == TCE_READVAR &&
	  predicate->bop.left->projectscalar.sub->readvar == lvar &&
	  predicate->bop.right->type == TCE_PROJECTSCALAR &&
	  predicate->bop.right->projectscalar.sub->type == TCE_READVAR &&
	  predicate->bop.right->projectscalar.sub->readvar == lvar) {
	 c1 = predicate->bop.left->projectscalar.col;
	 c2 = predicate->bop.right->projectscalar.col;

	 if (tuple_contains_toplevel(type, c1) && c2 == leftcol) {
	    *back_ret = false;
	    *col_ret = c1;
	    goto matched;
	 }
	 else if (tuple_contains_toplevel(type, c1) && c2 == rightcol) {
	    *back_ret = true;
	    *col_ret = c1;
	    goto matched;
	 }
	 else if (tuple_contains_toplevel(type, c2) && c1 == leftcol) {
	    *back_ret = false;
	    *col_ret = c2;
	    goto matched;
	 }
	 else if (tuple_contains_toplevel(type, c2) && c1 == rightcol) {
	    *back_ret = true;
	    *col_ret = c2;
	    goto matched;
	 }
      }
   }

   return false;

 matched:
   tcexpr_destroy(ctx->pql, predicate);
   *predicate_inout = mktcexpr_value(ctx->pql, pqlvalue_bool(ctx->pql, true));
   (*predicate_inout)->datatype = datatype_bool(ctx->pql);
   return true;
}

/*
 * Inspect a predicate on a join to see if (part of) it compares
 * EDGECOL to a constant. If so, update the predicate to remove that
 * condition, place the constant in *VALUE_RET, and return true. If
 * not, return false.
 */
static bool check_edge_predicate(struct stepjoins *ctx,
				 struct tcexpr **predicate_inout,
				 struct tcvar *lvar,
				 struct column *edgecol,
				 struct pqlvalue **value_ret) {
   struct tcexpr *predicate;

   predicate = *predicate_inout;

   /*
    * We expect the predicate to be of the form T1 && T2 && ..., since
    * that's what's generated upstream.
    */

   if (predicate->type == TCE_BOP && predicate->bop.op == F_AND) {
      if (check_edge_predicate(ctx, &predicate->bop.left,
			       lvar, edgecol, value_ret)) {
	 return true;
      }
      if (check_edge_predicate(ctx, &predicate->bop.right,
			       lvar, edgecol, value_ret)) {
	 return true;
      }
   }

   if (predicate->type == TCE_BOP && predicate->bop.op == F_EQ) {

      if (predicate->bop.left->type == TCE_PROJECTSCALAR &&
	  predicate->bop.left->projectscalar.sub->type == TCE_READVAR &&
	  predicate->bop.left->projectscalar.sub->readvar == lvar &&
	  predicate->bop.left->projectscalar.col == edgecol &&
	  predicate->bop.right->type == TCE_VALUE) {

	 *value_ret = predicate->bop.right->value;
	 predicate->bop.right->value = NULL;
	 goto matched;
      }

      if (predicate->bop.right->type == TCE_PROJECTSCALAR &&
	  predicate->bop.right->projectscalar.sub->type == TCE_READVAR &&
	  predicate->bop.right->projectscalar.sub->readvar == lvar &&
	  predicate->bop.right->projectscalar.col == edgecol &&
	  predicate->bop.left->type == TCE_VALUE) {

	 *value_ret = predicate->bop.left->value;
	 predicate->bop.left->value = NULL;
	 goto matched;
      }

   }

   return false;

 matched:
   tcexpr_destroy(ctx->pql, predicate);
   *predicate_inout = mktcexpr_value(ctx->pql, pqlvalue_bool(ctx->pql, true));
   (*predicate_inout)->datatype = datatype_bool(ctx->pql);
   return true;
}

/*
 * Inspect and perhaps convert a join expression.
 */
static struct tcexpr *try_convert_stepjoin(struct stepjoins *ctx,
					   struct tcexpr *join) {
   struct tcexpr *scan, *scanfilter, *other, *step;
   struct column *othercol;
   struct datatype *othertype;
   struct pqlvalue *edgeval;
   bool onleft, reversed, ok;

   PQLASSERT(join->type == TCE_JOIN);

   /*
    * We have the form of a step join if:
    *
    *   - one side is a scan;
    *   - the predicate compares a column of the other side to one
    *     object of the scan;
    *   - optionally, the predicate compares the edge of the scan to
    *     a constant.
    */

   if (join->join.left->type == TCE_SCAN) {
      onleft = true;
      scan = join->join.left;
      scanfilter = NULL;
      other = join->join.right;
   }
   else if (join->join.left->type == TCE_FILTER &&
	    join->join.left->filter.sub->type == TCE_SCAN) {
      onleft = true;
      scan = join->join.left->filter.sub;
      scanfilter = join->join.left;
      other = join->join.right;
   }
   else if (join->join.right->type == TCE_SCAN) {
      onleft = false;
      scan = join->join.right;
      scanfilter = NULL;
      other = join->join.left;
   }
   else if (join->join.right->type == TCE_FILTER &&
	    join->join.right->filter.sub->type == TCE_SCAN) {
      onleft = false;
      scan = join->join.right->filter.sub;
      scanfilter = join->join.right;
      other = join->join.left;
   }
   else {
      /* no luck */
      return join;
   }

   /*
    * Check the predicate on the join.
    */

   if (join->join.predexpr == NULL) {
      /* nope. */
      return join;
   }

   othertype = other->datatype;
   if (datatype_isset(othertype)) {
      othertype = datatype_set_member(othertype);
   }
   else if (datatype_issequence(othertype)) {
      othertype = datatype_set_member(othertype);
   }
   PQLASSERT(datatype_istuple(othertype));

   ok = check_join_predicate(ctx, &join->join.predexpr,
			     othertype,
			     join->join.predvar,
			     scan->scan.leftobjcolumn,
			     scan->scan.rightobjcolumn,
			     &reversed, &othercol);


   if (!ok) {
      /* no good */
      return join;
   }

   /*
    * Look for a condition on the edge name.
    */

   //PQLASSERT(join->join.predicate->type == TCE_LAMBDA); -- duplicate
   ok = check_edge_predicate(ctx, &join->join.predexpr,
			     join->join.predvar,
			     scan->scan.edgecolumn, &edgeval);
   if (!ok && scanfilter != NULL) {
      ok = check_edge_predicate(ctx,
				&scanfilter->filter.predexpr,
				scanfilter->filter.predvar,
				scan->scan.edgecolumn, &edgeval);
   }
   if (!ok) {
      edgeval = NULL;
   }

   if (scanfilter != NULL) {
      combine_predicates(ctx,
			 scanfilter->filter.predvar,
			 scanfilter->filter.predexpr,
			 join->join.predvar, join->join.predexpr,
			 &join->join.predvar, &join->join.predexpr);
      scanfilter->filter.predvar = NULL;
      scanfilter->filter.predexpr = NULL;
   }

   column_incref(othercol);
   step = mktcexpr_step(ctx->pql,
			other,
			othercol,
			edgeval,
			reversed,
			scan->scan.leftobjcolumn,
			scan->scan.edgecolumn,
			scan->scan.rightobjcolumn,
			join->join.predvar,
			join->join.predexpr);

   scan->scan.leftobjcolumn = NULL;
   scan->scan.edgecolumn = NULL;
   scan->scan.rightobjcolumn = NULL;
   join->join.predvar = NULL;
   join->join.predexpr = NULL;
   if (onleft) {
      PQLASSERT(join->join.right == other);
      join->join.right = NULL;
   }
   else {
      PQLASSERT(join->join.left == other);
      join->join.left = NULL;
   }

   step->datatype = join->datatype;

   /* destroy JOIN and SCAN and any SCANFILTER */
   tcexpr_destroy(ctx->pql, join);

   if (step->step.predexpr->type == TCE_VALUE &&
       pqlvalue_isbool(step->step.predexpr->value) &&
       pqlvalue_bool_get(step->step.predexpr->value) == true) {
      if (step->step.predvar != NULL) {
	 tcvar_decref(ctx->pql, step->step.predvar);
	 step->step.predvar = NULL;
      }
      tcexpr_destroy(ctx->pql, step->step.predexpr);
      step->step.predexpr = NULL;
   }

   return step;
}

////////////////////////////////////////////////////////////
// recursive traversal

static struct tcexpr *tcexpr_stepjoins(struct stepjoins *ctx,
				       struct tcexpr *te) {
   unsigned i, num;
   struct tcexpr *subexpr;

   switch (te->type) {
    case TCE_FILTER:
     te->filter.sub = tcexpr_stepjoins(ctx, te->filter.sub);
     te->filter.predexpr = tcexpr_stepjoins(ctx, te->filter.predexpr);
     break;
    case TCE_PROJECTTUPLE:
     te->projecttuple.sub = tcexpr_stepjoins(ctx, te->projecttuple.sub);
     break;
    case TCE_PROJECTSCALAR:
     te->projectscalar.sub = tcexpr_stepjoins(ctx, te->projectscalar.sub);
     break;
    case TCE_RENAME:
     te->rename.sub = tcexpr_stepjoins(ctx, te->rename.sub);
     break;
    case TCE_JOIN:
     te->join.left = tcexpr_stepjoins(ctx, te->join.left);
     te->join.right = tcexpr_stepjoins(ctx, te->join.right);
     if (te->join.predexpr != NULL) {
	te->join.predexpr = tcexpr_stepjoins(ctx, te->join.predexpr);
     }
     te = try_convert_stepjoin(ctx, te);
     break;
    case TCE_ORDER:
     te->order.sub = tcexpr_stepjoins(ctx, te->order.sub);
     break;
    case TCE_UNIQ:
     te->uniq.sub = tcexpr_stepjoins(ctx, te->uniq.sub);
     break;
    case TCE_NEST:
     te->nest.sub = tcexpr_stepjoins(ctx, te->nest.sub);
     break;
    case TCE_UNNEST:
     te->unnest.sub = tcexpr_stepjoins(ctx, te->unnest.sub);
     break;
    case TCE_DISTINGUISH:
     te->distinguish.sub = tcexpr_stepjoins(ctx, te->distinguish.sub);
     break;
    case TCE_ADJOIN:
     te->adjoin.left = tcexpr_stepjoins(ctx, te->adjoin.left);
     te->adjoin.expr = tcexpr_stepjoins(ctx, te->adjoin.expr);
     break;
    case TCE_STEP:
     /* shouldn't see one of these since we create them */
     te->step.sub = tcexpr_stepjoins(ctx, te->step.sub);
     if (te->step.predexpr != NULL) {
	te->step.predexpr = tcexpr_stepjoins(ctx, te->step.predexpr);
     }
     break;
    case TCE_REPEAT:
     te->repeat.sub = tcexpr_stepjoins(ctx, te->repeat.sub);
     te->repeat.body = tcexpr_stepjoins(ctx, te->repeat.body);
     break;
    case TCE_SCAN:
     /* nothing */
     break;
    case TCE_BOP:
     te->bop.left = tcexpr_stepjoins(ctx, te->bop.left);
     te->bop.right = tcexpr_stepjoins(ctx, te->bop.right);
     break;
    case TCE_UOP:
     te->uop.sub = tcexpr_stepjoins(ctx, te->uop.sub);
     break;
    case TCE_FUNC:
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	subexpr = tcexprarray_get(&te->func.args, i);
	subexpr = tcexpr_stepjoins(ctx, subexpr);
	tcexprarray_set(&te->func.args, i, subexpr);
     }
     break;
    case TCE_MAP:
     te->map.set = tcexpr_stepjoins(ctx, te->map.set);
     te->map.result = tcexpr_stepjoins(ctx, te->map.result);
     break;
    case TCE_LET:
     te->let.value = tcexpr_stepjoins(ctx, te->let.value);
     te->let.body = tcexpr_stepjoins(ctx, te->let.body);
     break;
    case TCE_READVAR:
     break;
    case TCE_READGLOBAL:
     break;
    case TCE_CREATEPATHELEMENT:
     te->createpathelement = tcexpr_stepjoins(ctx, te->createpathelement);
     break;
    case TCE_SPLATTER:
     te->splatter.value = tcexpr_stepjoins(ctx, te->splatter.value);
     te->splatter.name = tcexpr_stepjoins(ctx, te->splatter.name);
     break;
    case TCE_TUPLE:
     num = tcexprarray_num(&te->tuple.exprs);
     for (i=0; i<num; i++) {
	subexpr = tcexprarray_get(&te->tuple.exprs, i);
	subexpr = tcexpr_stepjoins(ctx, subexpr);
	tcexprarray_set(&te->tuple.exprs, i, subexpr);
     }
     break;
    case TCE_VALUE:
     break;
   }
   return te;
}

static struct tcstmt *tcstmt_stepjoins(struct stepjoins *ctx,
				       struct tcstmt *ts) {
   switch (ts->type) {
    case TCS_EXPR:
     ts->expr = tcexpr_stepjoins(ctx, ts->expr);
     break;
    case TCS_BEGIN:
    case TCS_ABORT:
    case TCS_COMMIT:
     break;
   }
   return ts;
}

////////////////////////////////////////////////////////////
// entry point

struct tcstmt *stepjoins(struct pqlcontext *pql, struct tcstmt *ts) {
   struct stepjoins ctx;

   stepjoins_init(&ctx, pql);
   ts = tcstmt_stepjoins(&ctx, ts);
   stepjoins_cleanup(&ctx);

   return ts;
}

