/*
 * Copyright 2009
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h> // for snprintf
#include "utils.h"
#include "datatype.h"
#include "layout.h"
#include "pqlvalue.h"
#include "pqlcontext.h"

#define TCALC_INLINE
#include "tcalc.h"

////////////////////////////////////////////////////////////
// constructors

struct tcglobal *mktcglobal(struct pqlcontext *pql, const char *name) {
   struct tcglobal *tcg;

   tcg = domalloc(pql, sizeof(*tcg));
   tcg->name = dostrdup(pql, name);
   tcg->refcount = 1;

   return tcg;
}

void tcglobal_incref(struct tcglobal *tcg) {
   PQLASSERT(tcg != NULL);
   tcg->refcount++;
}

//////////////////////////////

struct tcvar *mktcvar_fresh(struct pqlcontext *pql) {
   struct tcvar *var;
   
   var = domalloc(pql, sizeof(*var));
   var->id = pql->nextvarid++;
   var->refcount = 1;
   var->datatype = NULL;

   return var;
}

void tcvar_incref(struct tcvar *var) {
   PQLASSERT(var != NULL);
   var->refcount++;
}

//////////////////////////////

static struct tcexpr *mktcexpr(struct pqlcontext *pql, enum tcexprtypes type) {
   struct tcexpr *te;

   te = domalloc(pql, sizeof(*te));
   te->datatype = NULL;
   te->type = type;

   return te;
}

//////////

struct tcexpr *mktcexpr_filter(struct pqlcontext *pql,
			       struct tcexpr *sub,
			       struct tcvar *predvar,
			       struct tcexpr *predexpr) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_FILTER);
   te->filter.sub = sub;
   te->filter.predvar = predvar;
   te->filter.predexpr = predexpr;

   return te;
}

struct tcexpr *mktcexpr_projecttuple(struct pqlcontext *pql,
				     struct tcexpr *sub) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_PROJECTTUPLE);
   te->projecttuple.sub = sub;
   te->projecttuple.cols = columnarray_create(pql);
   te->projecttuple.complement_cols = false;

   return te;
}

struct tcexpr *mktcexpr_projectscalar(struct pqlcontext *pql,
				      struct tcexpr *sub,
				      struct column *col) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_PROJECTSCALAR);
   te->projectscalar.sub = sub;
   te->projectscalar.col = col;

   return te;
}

struct tcexpr *mktcexpr_rename(struct pqlcontext *pql,
			       struct tcexpr *sub, struct column *insidecol,
			       struct column *outsidecol) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_RENAME);
   te->rename.sub = sub;
   te->rename.insidecol = insidecol;
   te->rename.outsidecol = outsidecol;

   return te;
}

struct tcexpr *mktcexpr_join(struct pqlcontext *pql,
			     struct tcexpr *left, struct tcexpr *right,
			     struct tcvar *predvar,
			     struct tcexpr *predexpr) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_JOIN);
   te->join.left = left;
   te->join.right = right;
   te->join.predvar = predvar;
   te->join.predexpr = predexpr;

   return te;
}

struct tcexpr *mktcexpr_order(struct pqlcontext *pql, struct tcexpr *sub) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_ORDER);
   te->order.sub = sub;
   te->order.cols = columnarray_create(pql);

   return te;
}

struct tcexpr *mktcexpr_uniq(struct pqlcontext *pql, struct tcexpr *sub) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_UNIQ);
   te->uniq.sub = sub;
   te->uniq.cols = columnarray_create(pql);

   return te;
}

//////////

struct tcexpr *mktcexpr_nest_none(struct pqlcontext *pql,
				  struct tcexpr *sub,
				  struct column *newcol) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_NEST);
   te->nest.sub = sub;
   te->nest.cols = columnarray_create(pql);
   te->nest.newcol = newcol;
   te->nest.complement_cols = false;

   return te;
}

struct tcexpr *mktcexpr_nest_one(struct pqlcontext *pql,
				 struct tcexpr *sub, struct column *col,
				 struct column *newcol) {
   struct tcexpr *te;

   te = mktcexpr_nest_none(pql, sub, newcol);
   columnarray_add(pql, te->nest.cols, col, NULL);

   return te;
}

struct tcexpr *mktcexpr_nest_set(struct pqlcontext *pql,
				 struct tcexpr *sub,
				 struct columnarray *cols,
				 struct column *newcol) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_NEST);
   te->nest.sub = sub;
   te->nest.cols = cols;
   te->nest.newcol = newcol;
   te->nest.complement_cols = false;

   return te;
}

struct tcexpr *mktcexpr_unnest(struct pqlcontext *pql,
			       struct tcexpr *sub, struct column *col) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_UNNEST);
   te->unnest.sub = sub;
   te->unnest.col = col;

   return te;
}

//////////

struct tcexpr *mktcexpr_distinguish(struct pqlcontext *pql,
			       struct tcexpr *sub, struct column *newcol) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_DISTINGUISH);
   te->distinguish.sub = sub;
   te->distinguish.newcol = newcol;

   return te;
}

struct tcexpr *mktcexpr_adjoin(struct pqlcontext *pql,
			       struct tcexpr *left,
			       struct tcvar *var,
			       struct tcexpr *expr,
			       struct column *newcol) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_ADJOIN);
   te->adjoin.left = left;
   te->adjoin.var = var;
   te->adjoin.expr = expr;
   te->adjoin.newcol = newcol;

   return te;
}

//////////

struct tcexpr *mktcexpr_step(struct pqlcontext *pql,
			     struct tcexpr *sub,
			     struct column *subcolumn,
			     struct pqlvalue *edgename,
			     bool reversed,
			     struct column *leftobjcolumn,
			     struct column *edgecolumn,
			     struct column *rightobjcolumn,
			     struct tcvar *predvar,
			     struct tcexpr *predexpr) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_STEP);
   te->step.sub = sub;
   te->step.subcolumn = subcolumn;
   te->step.edgename = edgename;
   te->step.reversed = reversed;
   te->step.leftobjcolumn = leftobjcolumn;
   te->step.edgecolumn = edgecolumn;
   te->step.rightobjcolumn = rightobjcolumn;
   te->step.predvar = predvar;
   te->step.predexpr = predexpr;

   return te;
}

struct tcexpr *mktcexpr_repeat(struct pqlcontext *pql,
			       struct tcexpr *sub,
			       struct column *subendcolumn,
			       struct tcvar *loopvar,
			       struct column *bodystartcolumn,
			       struct tcexpr *body,
			       struct column *bodybodycolumn,
			       struct column *bodypathcolumn,
			       struct column *bodyendcolumn,
			       struct column *repeatpathcolumn,
			       struct column *repeatendcolumn) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_REPEAT);
   te->repeat.sub = sub;
   te->repeat.subendcolumn = subendcolumn;
   te->repeat.loopvar = loopvar;
   te->repeat.bodystartcolumn = bodystartcolumn;
   te->repeat.body = body;
   te->repeat.bodybodycolumn = bodybodycolumn;
   te->repeat.bodypathcolumn = bodypathcolumn;
   te->repeat.bodyendcolumn = bodyendcolumn;
   te->repeat.repeatpathcolumn = repeatpathcolumn;
   te->repeat.repeatendcolumn = repeatendcolumn;

   return te;
}

//////////

struct tcexpr *mktcexpr_scan(struct pqlcontext *pql,
			     struct column *leftobjcolumn,
			     struct column *edgecolumn,
			     struct column *rightobjcolumn) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_SCAN);
   te->scan.leftobjcolumn = leftobjcolumn;
   te->scan.edgecolumn = edgecolumn;
   te->scan.rightobjcolumn = rightobjcolumn;

   return te;
}

//////////

struct tcexpr *mktcexpr_bop(struct pqlcontext *pql, struct tcexpr *left,
			    enum functions op, struct tcexpr *right) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_BOP);
   te->bop.left = left;
   te->bop.op = op;
   te->bop.right = right;

   return te;
}

struct tcexpr *mktcexpr_uop(struct pqlcontext *pql,
			    enum functions op, struct tcexpr *sub) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_UOP);
   te->uop.op = op;
   te->uop.sub = sub;

   return te;
}

struct tcexpr *mktcexpr_func(struct pqlcontext *pql, enum functions op) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_FUNC);
   te->func.op = op;
   tcexprarray_init(&te->func.args);

   return te;
}

//////////

struct tcexpr *mktcexpr_map(struct pqlcontext *pql, struct tcvar *var,
			    struct tcexpr *set, struct tcexpr *result) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_MAP);
   te->map.var = var;
   te->map.set = set;
   te->map.result = result;

   return te;
}

struct tcexpr *mktcexpr_let(struct pqlcontext *pql, struct tcvar *var,
			    struct tcexpr *value, struct tcexpr *body) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_LET);
   te->let.var = var;
   te->let.value = value;
   te->let.body = body;

   return te;
}

//////////

struct tcexpr *mktcexpr_readvar(struct pqlcontext *pql, struct tcvar *var) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_READVAR);
   te->readvar = var;

   return te;
}

struct tcexpr *mktcexpr_readglobal(struct pqlcontext *pql,
				   struct tcglobal *var) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_READGLOBAL);
   te->readglobal = var;

   return te;
}

//////////

struct tcexpr *mktcexpr_createpathelement(struct pqlcontext *pql,
					  struct tcexpr *sub) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_CREATEPATHELEMENT);
   te->createpathelement = sub;

   return te;
}

struct tcexpr *mktcexpr_splatter(struct pqlcontext *pql,
				 struct tcexpr *value, struct tcexpr *name) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_SPLATTER);
   te->splatter.value = value;
   te->splatter.name = name;

   return te;
}

struct tcexpr *mktcexpr_tuple(struct pqlcontext *pql, unsigned arity) {
   struct tcexpr *te;
   unsigned i;

   te = mktcexpr(pql, TCE_TUPLE);
   tcexprarray_init(&te->tuple.exprs);
   te->tuple.columns = columnarray_create(pql);
   tcexprarray_setsize(pql, &te->tuple.exprs, arity);
   columnarray_setsize(pql, te->tuple.columns, arity);
   for (i=0; i<arity; i++) {
      tcexprarray_set(&te->tuple.exprs, i, NULL);
      columnarray_set(te->tuple.columns, i, NULL);
   }

   return te;
}

struct tcexpr *mktcexpr_monople(struct pqlcontext *pql, struct tcexpr *sub,
				struct column *col) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_TUPLE);
   tcexprarray_init(&te->tuple.exprs);
   te->tuple.columns = columnarray_create(pql);

   tcexprarray_setsize(pql, &te->tuple.exprs, 1);
   tcexprarray_set(&te->tuple.exprs, 0, sub);

   columnarray_setsize(pql, te->tuple.columns, 1);
   columnarray_set(te->tuple.columns, 0, col);

   return te;
}

//////////

struct tcexpr *mktcexpr_value(struct pqlcontext *pql, struct pqlvalue *val) {
   struct tcexpr *te;

   te = mktcexpr(pql, TCE_VALUE);
   te->value = val;

   return te;
}

//////////////////////////////

static struct tcstmt *mktcstmt(struct pqlcontext *pql, enum tcstmttypes type) {
   struct tcstmt *ts;

   ts = domalloc(pql, sizeof(*ts));
   ts->type = type;

   return ts;
}

//////////

struct tcstmt *mktcstmt_expr(struct pqlcontext *pql, struct tcexpr *expr) {
   struct tcstmt *ts;

   ts = mktcstmt(pql, TCS_EXPR);
   ts->expr = expr;

   return ts;
}

struct tcstmt *mktcstmt_begin(struct pqlcontext *pql) {
   struct tcstmt *ts;

   ts = mktcstmt(pql, TCS_BEGIN);

   return ts;
}

struct tcstmt *mktcstmt_abort(struct pqlcontext *pql) {
   struct tcstmt *ts;

   ts = mktcstmt(pql, TCS_ABORT);

   return ts;
}

struct tcstmt *mktcstmt_commit(struct pqlcontext *pql) {
   struct tcstmt *ts;

   ts = mktcstmt(pql, TCS_COMMIT);

   return ts;
}

//////////////////////////////

/*
 * Cloning safely requires making fresh variables anywhere variables
 * are bound. Otherwise, things like adjustments to the datatype end
 * up shared and really strange things happen.
 */

struct clone {
   struct pqlcontext *pql;
   struct tcvararray oldvars;
   struct tcvararray newvars;
};

static void columnarray_copy(struct clone *ctx,
			     struct columnarray *dest,
			     const struct columnarray *src) {
   struct column *col;
   unsigned i, num;

   PQLASSERT(dest != NULL);
   num = columnarray_num(src);
   columnarray_setsize(ctx->pql, dest, num);
   for (i=0; i<num; i++) {
      col = columnarray_get(src, i);
      column_incref(col);
      columnarray_set(dest, i, col);
   }
}

static struct tcvar *tcvar_clone_bind(struct clone *ctx,
				      struct tcvar *oldvar) {
   struct tcvar *newvar;

   if (oldvar != NULL) {
      newvar = mktcvar_fresh(ctx->pql);
      newvar->datatype = oldvar->datatype;
   }
   else {
      newvar = NULL;
   }
   tcvararray_add(ctx->pql, &ctx->oldvars, oldvar, NULL);
   tcvararray_add(ctx->pql, &ctx->newvars, newvar, NULL);
   return newvar;
}

static void tcvar_clone_unbind(struct clone *ctx) {
   unsigned num;

   num = tcvararray_num(&ctx->oldvars);
   PQLASSERT(num == tcvararray_num(&ctx->newvars));
   PQLASSERT(num > 0);

   tcvararray_setsize(ctx->pql, &ctx->oldvars, num-1);
   tcvararray_setsize(ctx->pql, &ctx->newvars, num-1);
}

static struct tcvar *tcvar_clone_lookup(struct clone *ctx,
					struct tcvar *oldvar) {
   struct tcvar *ret;
   unsigned i, num;

   PQLASSERT(oldvar != NULL);

   num = tcvararray_num(&ctx->oldvars);
   PQLASSERT(num == tcvararray_num(&ctx->newvars));
   for (i=0; i<num; i++) {
      if (tcvararray_get(&ctx->oldvars, i) == oldvar) {
	 ret = tcvararray_get(&ctx->newvars, i);
	 tcvar_incref(ret);
	 return ret;
      }
   }
   return NULL;
}

static struct tcexpr *tcexpr_clone2(struct clone *ctx,
				    const struct tcexpr *te) {
   struct tcexpr *ret;
   struct tcvar *boundvar, *foundvar;
   unsigned i, num;

   ret = NULL; // gcc 4.1

   if (te == NULL) {
      return NULL;
   }

   switch (te->type) {
    case TCE_FILTER:
     boundvar = tcvar_clone_bind(ctx, te->filter.predvar);
     ret = mktcexpr_filter(ctx->pql,
			   tcexpr_clone2(ctx, te->filter.sub),
			   boundvar,
			   tcexpr_clone2(ctx, te->filter.predexpr));
     tcvar_clone_unbind(ctx);
     break;

    case TCE_PROJECTTUPLE:
     ret = mktcexpr_projecttuple(ctx->pql,
				 tcexpr_clone2(ctx, te->projecttuple.sub));
     columnarray_copy(ctx, ret->projecttuple.cols, te->projecttuple.cols);
     ret->projecttuple.complement_cols = te->projecttuple.complement_cols;
     break;

    case TCE_PROJECTSCALAR:
     column_incref(te->projectscalar.col);
     ret = mktcexpr_projectscalar(ctx->pql,
				  tcexpr_clone2(ctx, te->projectscalar.sub),
				  te->projectscalar.col);
     break;

    case TCE_RENAME:
     column_incref(te->rename.insidecol);
     column_incref(te->rename.outsidecol);
     ret = mktcexpr_rename(ctx->pql,
			   tcexpr_clone2(ctx, te->rename.sub),
			   te->rename.insidecol,
			   te->rename.outsidecol);
     break;

    case TCE_JOIN:
     boundvar = tcvar_clone_bind(ctx, te->join.predvar);
     ret = mktcexpr_join(ctx->pql,
			 tcexpr_clone2(ctx, te->join.left),
			 tcexpr_clone2(ctx, te->join.right),
			 boundvar,
			 tcexpr_clone2(ctx, te->join.predexpr));
     tcvar_clone_unbind(ctx);
     break;

    case TCE_ORDER:
     ret = mktcexpr_order(ctx->pql, tcexpr_clone2(ctx, te->order.sub));
     columnarray_copy(ctx, ret->order.cols, te->order.cols);
     break;

    case TCE_UNIQ:
     ret = mktcexpr_uniq(ctx->pql, tcexpr_clone2(ctx, te->uniq.sub));
     columnarray_copy(ctx, ret->uniq.cols, te->uniq.cols);
     break;

    case TCE_NEST:
     column_incref(te->nest.newcol);
     ret = mktcexpr_nest_none(ctx->pql,
			      tcexpr_clone2(ctx, te->nest.sub),
			      te->nest.newcol);
     columnarray_copy(ctx, ret->nest.cols, te->nest.cols);
     break;

    case TCE_UNNEST:
     column_incref(te->unnest.col);
     ret = mktcexpr_unnest(ctx->pql,
			   tcexpr_clone2(ctx, te->unnest.sub),
			   te->unnest.col);
     break;

    case TCE_DISTINGUISH:
     column_incref(te->distinguish.newcol);
     ret = mktcexpr_distinguish(ctx->pql,
				tcexpr_clone2(ctx, te->distinguish.sub),
				te->distinguish.newcol);
     break;

    case TCE_ADJOIN:
     column_incref(te->adjoin.newcol);
     boundvar = tcvar_clone_bind(ctx, te->adjoin.var);
     ret = mktcexpr_adjoin(ctx->pql,
			   tcexpr_clone2(ctx, te->adjoin.left),
			   boundvar,
			   tcexpr_clone2(ctx, te->adjoin.expr),
			   te->adjoin.newcol);
     tcvar_clone_unbind(ctx);
     break;

    case TCE_STEP:
     column_incref(te->step.subcolumn);
     column_incref(te->step.leftobjcolumn);
     column_incref(te->step.edgecolumn);
     column_incref(te->step.rightobjcolumn);
     boundvar = tcvar_clone_bind(ctx, te->step.predvar);
     ret = mktcexpr_step(ctx->pql,
			 tcexpr_clone2(ctx, te->step.sub),
			 te->step.subcolumn,
			 pqlvalue_clone(ctx->pql, te->step.edgename),
			 te->step.reversed,
			 te->step.leftobjcolumn,
			 te->step.edgecolumn,
			 te->step.rightobjcolumn,
			 boundvar,
			 tcexpr_clone2(ctx, te->step.predexpr));
     tcvar_clone_unbind(ctx);
     break;

    case TCE_REPEAT:
     boundvar = tcvar_clone_bind(ctx, te->repeat.loopvar);
     column_incref(te->repeat.subendcolumn);
     column_incref(te->repeat.bodystartcolumn);
     column_incref(te->repeat.bodybodycolumn);
     column_incref(te->repeat.bodypathcolumn);
     column_incref(te->repeat.bodyendcolumn);
     column_incref(te->repeat.repeatpathcolumn);
     column_incref(te->repeat.repeatendcolumn);
     ret = mktcexpr_repeat(ctx->pql,
			   tcexpr_clone2(ctx, te->repeat.sub),
			   te->repeat.subendcolumn,
			   boundvar,
			   te->repeat.bodystartcolumn,
			   tcexpr_clone2(ctx, te->repeat.body),
			   te->repeat.bodybodycolumn,
			   te->repeat.bodypathcolumn,
			   te->repeat.bodyendcolumn,
			   te->repeat.repeatpathcolumn,
			   te->repeat.repeatendcolumn);
     tcvar_clone_unbind(ctx);
     break;

    case TCE_SCAN:
     column_incref(te->scan.leftobjcolumn);
     column_incref(te->scan.edgecolumn);
     column_incref(te->scan.rightobjcolumn);
     ret = mktcexpr_scan(ctx->pql,
			 te->scan.leftobjcolumn,
			 te->scan.edgecolumn,
			 te->scan.rightobjcolumn);
     break;

    case TCE_BOP:
     ret = mktcexpr_bop(ctx->pql,
			tcexpr_clone2(ctx, te->bop.left),
			te->bop.op,
			tcexpr_clone2(ctx, te->bop.right));
     break;

    case TCE_UOP:
     ret = mktcexpr_uop(ctx->pql,
			te->uop.op,
			tcexpr_clone2(ctx, te->uop.sub));
     break;

    case TCE_FUNC:
     ret = mktcexpr_func(ctx->pql, te->func.op);
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	tcexprarray_add(ctx->pql, &ret->func.args,
			tcexpr_clone2(ctx, tcexprarray_get(&te->func.args, i)),
			NULL);
     }
     break;

    case TCE_MAP:
     boundvar = tcvar_clone_bind(ctx, te->map.var);
     ret = mktcexpr_map(ctx->pql,
			boundvar,
			tcexpr_clone2(ctx, te->map.set),
			tcexpr_clone2(ctx, te->map.result));
     tcvar_clone_unbind(ctx);
     break;

    case TCE_LET:
     boundvar = tcvar_clone_bind(ctx, te->let.var);
     ret = mktcexpr_let(ctx->pql,
			boundvar,
			tcexpr_clone2(ctx, te->let.value),
			tcexpr_clone2(ctx, te->let.body));
     tcvar_clone_unbind(ctx);
     break;

    case TCE_READVAR:
     /*
      * If the variable is from inside the cloned subtree, look up its
      * clone. If it's from outside, add a reference to the existing
      * var instead.
      */
     foundvar = tcvar_clone_lookup(ctx, te->readvar);
     if (foundvar != NULL) {
	ret = mktcexpr_readvar(ctx->pql, foundvar);
     }
     else {
	tcvar_incref(te->readvar);
	ret = mktcexpr_readvar(ctx->pql, te->readvar);
     }
     break;

    case TCE_READGLOBAL:
     tcglobal_incref(te->readglobal);
     ret = mktcexpr_readglobal(ctx->pql, te->readglobal);
     break;

    case TCE_CREATEPATHELEMENT:
     ret = mktcexpr_createpathelement(ctx->pql,
				      tcexpr_clone2(ctx,
						    te->createpathelement));
     break;

    case TCE_SPLATTER:
     ret = mktcexpr_splatter(ctx->pql,
			     tcexpr_clone2(ctx, te->splatter.value),
			     tcexpr_clone2(ctx, te->splatter.name));
     break;

    case TCE_TUPLE:
     num = tcexprarray_num(&te->tuple.exprs);
     ret = mktcexpr_tuple(ctx->pql, num);
     for (i=0; i<num; i++) {
	tcexprarray_set(&ret->tuple.exprs, i,
			tcexprarray_get(&te->tuple.exprs, i));
     }
     columnarray_copy(ctx, ret->tuple.columns, te->tuple.columns);
     break;

    case TCE_VALUE:
     ret = mktcexpr_value(ctx->pql, pqlvalue_clone(ctx->pql, te->value));
     break;
   }
   ret->datatype = te->datatype;

   return ret;
}

struct tcexpr *tcexpr_clone(struct pqlcontext *pql, const struct tcexpr *te) {
   struct clone ctx;
   struct tcexpr *ret;

   ctx.pql = pql;
   tcvararray_init(&ctx.oldvars);
   tcvararray_init(&ctx.newvars);

   ret = tcexpr_clone2(&ctx, te);

   tcvararray_cleanup(pql, &ctx.oldvars);
   tcvararray_cleanup(pql, &ctx.newvars);

   return ret;
} 

////////////////////////////////////////////////////////////
// destructors

static void columnarray_emptyout(struct pqlcontext *pql,
				 struct columnarray *cols) {
   struct column *col;
   unsigned i, num;

   num = columnarray_num(cols);
   for (i=0; i<num; i++) {
      col = columnarray_get(cols, i);
      column_decref(pql, col);
   }
   columnarray_setsize(pql, cols, 0);
}

void tcglobal_decref(struct pqlcontext *pql, struct tcglobal *tcg) {
   PQLASSERT(tcg != NULL);
   PQLASSERT(tcg->refcount > 0);

   tcg->refcount--;
   if (tcg->refcount == 0) {
      dostrfree(pql, tcg->name);
      dofree(pql, tcg, sizeof(*tcg));
   }
}

void tcvar_decref(struct pqlcontext *pql, struct tcvar *var) {
   PQLASSERT(var != NULL);
   PQLASSERT(var->refcount > 0);

   var->refcount--;
   if (var->refcount == 0) {
      dofree(pql, var, sizeof(*var));
   }
}

static void tcexprarray_destroyall(struct pqlcontext *pql,
				   struct tcexprarray *arr) {
   unsigned i, num;

   num = tcexprarray_num(arr);
   for (i=0; i<num; i++) {
      tcexpr_destroy(pql, tcexprarray_get(arr, i));
   }
   tcexprarray_setsize(pql, arr, 0);
}

void tcexpr_destroy(struct pqlcontext *pql, struct tcexpr *tc) {
   if (tc == NULL) {
      return;
   }
   switch (tc->type) {
    case TCE_FILTER:
     tcexpr_destroy(pql, tc->filter.sub);
     if (tc->filter.predvar != NULL) {
	tcvar_decref(pql, tc->filter.predvar);
     }
     tcexpr_destroy(pql, tc->filter.predexpr);
     break;
    case TCE_PROJECTTUPLE:
     tcexpr_destroy(pql, tc->projecttuple.sub);
     columnarray_emptyout(pql, tc->projecttuple.cols);
     columnarray_destroy(pql, tc->projecttuple.cols);
     break;
    case TCE_PROJECTSCALAR:
     tcexpr_destroy(pql, tc->projectscalar.sub);
     column_decref(pql, tc->projectscalar.col);
     break;
    case TCE_RENAME:
     tcexpr_destroy(pql, tc->rename.sub);
     if (tc->rename.insidecol != NULL) {
	column_decref(pql, tc->rename.insidecol);
     }
     if (tc->rename.outsidecol != NULL) {
	column_decref(pql, tc->rename.outsidecol);
     }
     break;
    case TCE_JOIN:
     tcexpr_destroy(pql, tc->join.left);
     tcexpr_destroy(pql, tc->join.right);
     if (tc->join.predvar != NULL) {
	tcvar_decref(pql, tc->join.predvar);
     }
     tcexpr_destroy(pql, tc->join.predexpr);
     break;
    case TCE_ORDER:
     tcexpr_destroy(pql, tc->order.sub);
     columnarray_emptyout(pql, tc->order.cols);
     columnarray_destroy(pql, tc->order.cols);
     break;
    case TCE_UNIQ:
     tcexpr_destroy(pql, tc->uniq.sub);
     columnarray_emptyout(pql, tc->uniq.cols);
     columnarray_destroy(pql, tc->uniq.cols);
     break;
    case TCE_NEST:
     tcexpr_destroy(pql, tc->nest.sub);
     columnarray_emptyout(pql, tc->nest.cols);
     columnarray_destroy(pql, tc->nest.cols);
     if (tc->nest.newcol != NULL) {
	column_decref(pql, tc->nest.newcol);
     }
     break;
    case TCE_UNNEST:
     tcexpr_destroy(pql, tc->unnest.sub);
     if (tc->unnest.col != NULL) {
	column_decref(pql, tc->unnest.col);
     }
     break;
    case TCE_DISTINGUISH:
     tcexpr_destroy(pql, tc->distinguish.sub);
     if (tc->distinguish.newcol != NULL) {
	column_decref(pql, tc->distinguish.newcol);
     }
     break;
    case TCE_ADJOIN:
     tcexpr_destroy(pql, tc->adjoin.left);
     if (tc->adjoin.var != NULL) {
	tcvar_decref(pql, tc->adjoin.var);
     }
     tcexpr_destroy(pql, tc->adjoin.expr);
     if (tc->adjoin.newcol != NULL) {
	column_decref(pql, tc->adjoin.newcol);
     }
     break;
    case TCE_STEP:
     tcexpr_destroy(pql, tc->step.sub);
     if (tc->step.subcolumn != NULL) {
	column_decref(pql, tc->step.subcolumn);
     }
     if (tc->step.edgename != NULL) {
	pqlvalue_destroy(tc->step.edgename);
     }
     if (tc->step.leftobjcolumn != NULL) {
	column_decref(pql, tc->step.leftobjcolumn);
     }
     if (tc->step.edgecolumn != NULL) {
	column_decref(pql, tc->step.edgecolumn);
     }
     if (tc->step.rightobjcolumn != NULL) {
	column_decref(pql, tc->step.rightobjcolumn);
     }
     if (tc->step.predvar != NULL) {
	tcvar_decref(pql, tc->step.predvar);
     }
     tcexpr_destroy(pql, tc->step.predexpr);
     break;
    case TCE_REPEAT:
     tcexpr_destroy(pql, tc->repeat.sub);
     if (tc->repeat.subendcolumn != NULL) {
	column_decref(pql, tc->repeat.subendcolumn);
     }
     if (tc->repeat.loopvar != NULL) {
	tcvar_decref(pql, tc->repeat.loopvar);
     }
     if (tc->repeat.bodystartcolumn != NULL) {
	column_decref(pql, tc->repeat.bodystartcolumn);
     }
     tcexpr_destroy(pql, tc->repeat.body);
     if (tc->repeat.bodybodycolumn != NULL) {
	column_decref(pql, tc->repeat.bodybodycolumn);
     }
     if (tc->repeat.bodypathcolumn != NULL) {
	column_decref(pql, tc->repeat.bodypathcolumn);
     }
     if (tc->repeat.bodyendcolumn != NULL) {
	column_decref(pql, tc->repeat.bodyendcolumn);
     }
     if (tc->repeat.repeatpathcolumn != NULL) {
	column_decref(pql, tc->repeat.repeatpathcolumn);
     }
     if (tc->repeat.repeatendcolumn != NULL) {
	column_decref(pql, tc->repeat.repeatendcolumn);
     }
     break;
    case TCE_SCAN:
     if (tc->scan.leftobjcolumn != NULL) {
	column_decref(pql, tc->scan.leftobjcolumn);
     }
     if (tc->scan.edgecolumn != NULL) {
	column_decref(pql, tc->scan.edgecolumn);
     }
     if (tc->scan.rightobjcolumn != NULL) {
	column_decref(pql, tc->scan.rightobjcolumn);
     }
     break;
    case TCE_BOP:
     tcexpr_destroy(pql, tc->bop.left);
     tcexpr_destroy(pql, tc->bop.right);
     break;
    case TCE_UOP:
     tcexpr_destroy(pql, tc->uop.sub);
     break;
    case TCE_FUNC:
     tcexprarray_destroyall(pql, &tc->func.args);
     tcexprarray_cleanup(pql, &tc->func.args);
     break;
    case TCE_MAP:
     if (tc->map.var != NULL) {
	tcvar_decref(pql, tc->map.var);
     }
     tcexpr_destroy(pql, tc->map.set);
     tcexpr_destroy(pql, tc->map.result);
     break;
    case TCE_LET:
     if (tc->let.var != NULL) {
	tcvar_decref(pql, tc->let.var);
     }
     tcexpr_destroy(pql, tc->let.value);
     tcexpr_destroy(pql, tc->let.body);
     break;
    case TCE_READVAR:
     if (tc->readvar != NULL) {
	tcvar_decref(pql, tc->readvar);
     }
     break;
    case TCE_READGLOBAL:
     tcglobal_decref(pql, tc->readglobal);
     break;
    case TCE_CREATEPATHELEMENT:
     tcexpr_destroy(pql, tc->createpathelement);
     break;
    case TCE_SPLATTER:
     tcexpr_destroy(pql, tc->splatter.value);
     tcexpr_destroy(pql, tc->splatter.name);
     break;
    case TCE_TUPLE:
     tcexprarray_destroyall(pql, &tc->tuple.exprs);
     tcexprarray_cleanup(pql, &tc->tuple.exprs);
     columnarray_emptyout(pql, tc->tuple.columns);
     columnarray_destroy(pql, tc->tuple.columns);
     break;
    case TCE_VALUE:
     if (tc->value != NULL) {
	pqlvalue_destroy(tc->value);
     }
     break;
   }
   dofree(pql, tc, sizeof(*tc));
}

void tcstmt_destroy(struct pqlcontext *pql, struct tcstmt *ts) {
   switch (ts->type) {
    case TCS_EXPR:
     tcexpr_destroy(pql, ts->expr);
     break;
    case TCS_BEGIN:
    case TCS_ABORT:
    case TCS_COMMIT:
     /* nothing */
     break;
   }
   dofree(pql, ts, sizeof(*ts));
}

////////////////////////////////////////////////////////////
// dump

struct tclayout {
   struct pqlcontext *pql;
   bool showtypes;
};

static struct layout *columnarray_layout(struct pqlcontext *pql,
					 struct columnarray *cols,
					 bool complement) {
   unsigned i, num;
   struct layout *ret = NULL, *sublayout;

   num = columnarray_num(cols);
   for (i=0; i<num; i++) {
      sublayout = mklayout_text(pql, column_getname(columnarray_get(cols, i)));
      if (ret != NULL) {
	 ret = mklayout_triple(pql, ret, mklayout_text(pql, ","),
			       sublayout);
      }
      else {
	 ret = sublayout;
      }
   }
   if (ret == NULL) {
      // XXX should do something better than this
      ret = mklayout_text(pql, "--");
   }
   if (complement) {
      ret = mklayout_triple(pql, mklayout_text(pql, "~("), ret,
			    mklayout_text(pql, ")"));
   }
   return ret;
}

static struct layout *tccol_layout(struct tclayout *ctx, struct column *col) {
   return mklayout_text(ctx->pql, column_getname(col));
}

static struct layout *tctype_layout(struct tclayout *ctx, struct layout *l,
				    struct datatype *t) {
   if (ctx->showtypes) {
      l = mklayout_triple(ctx->pql, l,
			  mklayout_text(ctx->pql, "::"),
			  mklayout_text(ctx->pql, datatype_getname(t)));
   }
   return l;
}

static struct layout *tcvar_layout(struct tclayout *ctx, struct tcvar *var) {
   char buf[32];
   struct layout *ret;

   if (var == NULL) {
      return mklayout_text(ctx->pql, ".NIL");
   }
   snprintf(buf, sizeof(buf), ".K%u", var->id);
   ret = mklayout_text(ctx->pql, buf);
   return tctype_layout(ctx, ret, var->datatype);
}

static struct layout *tcexpr_layout(struct tclayout *ctx, struct tcexpr *te);

static struct layout *fakelambda_layout(struct tclayout *ctx,
					struct tcvar *var,
					struct tcexpr *body) {
   struct layout *ret;

   if (var != NULL) {
      ret = mklayout_triple(ctx->pql, 
			    mklayout_text(ctx->pql, "lambda"),
			    tcvar_layout(ctx, var),
			    mklayout_text(ctx->pql, ":"));
      ret = mklayout_indent(ctx->pql,
			    ret,
			    tcexpr_layout(ctx, body),
			    NULL);
   }
   else {
      ret = tcexpr_layout(ctx, body);
   }

   return ret;
}


static struct layout *tcexpr_layout(struct tclayout *ctx, struct tcexpr *te) {
   struct layout *ret;

   ret = NULL; // for gcc 4.1

   switch (te->type) {

    case TCE_FILTER:
     ret = mklayout_triple(ctx->pql,
			   tcexpr_layout(ctx, te->filter.sub),
			   mklayout_text(ctx->pql, "where"),
			   fakelambda_layout(ctx, te->filter.predvar,
					     te->filter.predexpr));
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_PROJECTTUPLE:
     ret = mklayout_triple(ctx->pql,
			   mklayout_text(ctx->pql, "project"),
			   columnarray_layout(ctx->pql, te->projecttuple.cols,
					     te->projecttuple.complement_cols),
			   mklayout_wrap(ctx->pql, "(",
					 tcexpr_layout(ctx,
						       te->projecttuple.sub),
					 ")"));
     break;

    case TCE_PROJECTSCALAR:
     // XXX this should be visually distinguished (but not yet)
     ret = mklayout_triple(ctx->pql,
			   mklayout_text(ctx->pql, "project"),
			   tccol_layout(ctx, te->projectscalar.col),
			   mklayout_wrap(ctx->pql, "(",
					 tcexpr_layout(ctx,
						       te->projectscalar.sub),
					 ")"));
     break;

    case TCE_RENAME:
     ret = mklayout_quad(ctx->pql,
			 tcexpr_layout(ctx, te->rename.sub),
			 tccol_layout(ctx, te->rename.insidecol),
			 mklayout_text(ctx->pql, "=>"),
			 tccol_layout(ctx, te->rename.outsidecol));
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_JOIN:
     ret = mklayout_triple(ctx->pql,
			   tcexpr_layout(ctx, te->join.left),
			   mklayout_text(ctx->pql, "x"),
			   tcexpr_layout(ctx, te->join.right));
     if (te->join.predvar != NULL || te->join.predexpr != NULL) {
	ret = mklayout_triple(ctx->pql, ret,
			      mklayout_text(ctx->pql, "where"),
			      fakelambda_layout(ctx, te->join.predvar,
						te->join.predexpr));
     }
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_ORDER:
     ret = mklayout_triple(ctx->pql,
			   tcexpr_layout(ctx, te->order.sub),
			   mklayout_text(ctx->pql, "order-by"),
			   columnarray_layout(ctx->pql, te->order.cols,
					      false));
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_UNIQ:
     ret = mklayout_triple(ctx->pql,
			   tcexpr_layout(ctx, te->uniq.sub),
			   mklayout_text(ctx->pql, "uniq-on"),
			   columnarray_layout(ctx->pql, te->uniq.cols,
					      false));
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_NEST:
     ret = mklayout_triple(ctx->pql,
			   tcexpr_layout(ctx, te->nest.sub),
			   mklayout_text(ctx->pql, "nest"),
			   columnarray_layout(ctx->pql, te->nest.cols,
					      te->nest.complement_cols));
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_UNNEST:
     ret = mklayout_triple(ctx->pql,
			   tcexpr_layout(ctx, te->unnest.sub),
			   mklayout_text(ctx->pql, "unnest"),
			   tccol_layout(ctx, te->unnest.col));
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_DISTINGUISH:
     ret = mklayout_triple(ctx->pql,
			   tcexpr_layout(ctx, te->distinguish.sub),
			   mklayout_text(ctx->pql, "|+| DISTINGUISH as"),
			   tccol_layout(ctx, te->distinguish.newcol));
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_ADJOIN:
     ret = mklayout_quint(ctx->pql,
			  tcexpr_layout(ctx, te->adjoin.left),
			  mklayout_text(ctx->pql, "|+|"),
			  fakelambda_layout(ctx, te->adjoin.var,
					    te->adjoin.expr),
			  mklayout_text(ctx->pql, "as"),
			  tccol_layout(ctx, te->adjoin.newcol));
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_STEP:
     ret = mklayout_quint(ctx->pql,
			  tccol_layout(ctx, te->step.leftobjcolumn),
			  mklayout_text(ctx->pql, ","),
			  tccol_layout(ctx, te->step.edgecolumn),
			  mklayout_text(ctx->pql, ","),
			  tccol_layout(ctx, te->step.rightobjcolumn));
     ret = mklayout_pair(ctx->pql,
			 mklayout_text(ctx->pql, "as"),
			 ret);
     ret = mklayout_quint(ctx->pql,
			  mklayout_text(ctx->pql, "step"),
			  tccol_layout(ctx, te->step.subcolumn),
			  mklayout_text(ctx->pql, "."),
			  (te->step.edgename ?
			   pqlvalue_layout(ctx->pql, te->step.edgename) :
			   mklayout_text(ctx->pql, "%")),
			  ret);
     ret = mklayout_leftalign_triple(ctx->pql,
				     tcexpr_layout(ctx, te->step.sub),
				     mklayout_text(ctx->pql, "x"),
				     ret);

     if (te->step.predvar != NULL || te->step.predexpr != NULL) {
	ret = mklayout_triple(ctx->pql, ret,
			      mklayout_text(ctx->pql, "where"),
			      fakelambda_layout(ctx, te->step.predvar,
						te->step.predexpr));
     }
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_REPEAT:
     {
	struct layout *locals, *inputs, *path;
	struct layout *head, *body, *tail;

	inputs = mklayout_triple(ctx->pql,
				 tcexpr_layout(ctx, te->repeat.sub),
				 mklayout_text(ctx->pql, "."),
				 tccol_layout(ctx,
						 te->repeat.subendcolumn));

	locals = mklayout_triple(ctx->pql,
				 tcvar_layout(ctx, te->repeat.loopvar),
				 mklayout_text(ctx->pql, "."),
				 tccol_layout(ctx,
						 te->repeat.bodystartcolumn));

	head = mklayout_triple(ctx->pql, locals,
				  mklayout_text(ctx->pql, "<-"),
				  inputs);
	head = mklayout_triple(ctx->pql,
			       mklayout_text(ctx->pql, "repeat"),
			       head,
			       mklayout_text(ctx->pql, "{"));

	body = tcexpr_layout(ctx, te->repeat.body);

	tail = mklayout_triple(ctx->pql,
			       tccol_layout(ctx,
					       te->repeat.bodyendcolumn),
			       mklayout_text(ctx->pql, "->"),
			       tccol_layout(ctx,
					       te->repeat.repeatendcolumn));

	if (te->repeat.bodypathcolumn != NULL) {
	   path = mklayout_triple(ctx->pql, 
				  tccol_layout(ctx,
						  te->repeat.bodypathcolumn),
				  mklayout_text(ctx->pql, "->"),
				  tccol_layout(ctx,
						 te->repeat.repeatpathcolumn));

	   tail = mklayout_triple(ctx->pql, path,
				  mklayout_text(ctx->pql, ","),
				  tail);
	}

	tail = mklayout_pair(ctx->pql,
			     mklayout_text(ctx->pql, "}"), tail);
			      
	ret = mklayout_indent(ctx->pql, head, body, tail);
     }
     break;

    case TCE_SCAN:
     ret = mklayout_quint(ctx->pql,
			  tccol_layout(ctx, te->scan.leftobjcolumn),
			  mklayout_text(ctx->pql, ","),
			  tccol_layout(ctx, te->scan.edgecolumn),
			  mklayout_text(ctx->pql, ","),
			  tccol_layout(ctx, te->scan.rightobjcolumn));
     ret = mklayout_pair(ctx->pql,
			 mklayout_text(ctx->pql, "scan as"),
			 ret);
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_BOP:
     ret = mklayout_triple(ctx->pql,
			   tcexpr_layout(ctx, te->bop.left),
			   mklayout_text(ctx->pql, function_getname(te->bop.op)),
			   tcexpr_layout(ctx, te->bop.right));
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_UOP:
     ret = mklayout_pair(ctx->pql,
			 mklayout_text(ctx->pql, function_getname(te->uop.op)),
			 tcexpr_layout(ctx, te->uop.sub));
     ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     break;

    case TCE_FUNC:
     {
	struct layout *name, *sub;
	unsigned i, num;

	name = mklayout_text(ctx->pql, function_getname(te->func.op));

	num = tcexprarray_num(&te->func.args);
	ret = mklayout_leftalign_empty(ctx->pql);
	layoutarray_setsize(ctx->pql, &ret->leftalign, num);
	for (i=0; i<num; i++) {
	   sub = tcexpr_layout(ctx, tcexprarray_get(&te->func.args, i));
	   if (i < num-1) {
	      sub = mklayout_pair(ctx->pql, sub,
				  mklayout_text(ctx->pql, ","));
	   }
	   layoutarray_set(&ret->leftalign, i, sub);
	}
	ret = mklayout_wrap(ctx->pql, "(", ret, ")");
	ret = mklayout_pair(ctx->pql, name, ret);
     }
     break;

    case TCE_MAP:
     ret = mklayout_quint(ctx->pql,
			  mklayout_text(ctx->pql, "map"),
			  tcvar_layout(ctx, te->map.var),
			  mklayout_text(ctx->pql, "in"),
			  tcexpr_layout(ctx, te->map.set),
			  mklayout_text(ctx->pql, ":"));
     ret = mklayout_indent(ctx->pql,
			   ret,
			   tcexpr_layout(ctx, te->map.result),
			   NULL);
     break;

    case TCE_LET:
     ret = mklayout_quint(ctx->pql,
			  mklayout_text(ctx->pql, "let"),
			  tcvar_layout(ctx, te->let.var),
			  mklayout_text(ctx->pql, "="),
			  tcexpr_layout(ctx, te->let.value),
			  mklayout_text(ctx->pql, ":"));
     ret = mklayout_indent(ctx->pql,
			   ret,
			   tcexpr_layout(ctx, te->let.body),
			   NULL);
     break;

    case TCE_READVAR:
     ret = tcvar_layout(ctx, te->readvar);
     break;

    case TCE_READGLOBAL:
     ret = mklayout_text(ctx->pql, te->readglobal->name);
     break;

    case TCE_CREATEPATHELEMENT:
     ret = mklayout_wrap(ctx->pql, "PATHELEMENT(",
			 tcexpr_layout(ctx, te->createpathelement), ")");
     break;

    case TCE_SPLATTER:
     ret = mklayout_triple(ctx->pql,
			   tcexpr_layout(ctx, te->splatter.value),
			   mklayout_text(ctx->pql, ","),
			   tcexpr_layout(ctx, te->splatter.name));
     ret = mklayout_wrap(ctx->pql, "SPLATTER(", ret, ")");
     break;

    case TCE_TUPLE:
     {
	struct layout *sub;
	unsigned i, num;

	num = tcexprarray_num(&te->tuple.exprs);
	PQLASSERT(num > 0);
	ret = mklayout_leftalign_empty(ctx->pql);
	layoutarray_setsize(ctx->pql, &ret->leftalign, num);
	for (i=0; i<num; i++) {
	   sub = tcexpr_layout(ctx, tcexprarray_get(&te->tuple.exprs, i));
	   if (i < num-1) {
	      sub = mklayout_pair(ctx->pql, sub,
				  mklayout_text(ctx->pql, ","));
	   }
	   layoutarray_set(&ret->leftalign, i, sub);
	}
	ret = mklayout_wrap(ctx->pql, "(", ret, ")");
     }
     break;

    case TCE_VALUE:
     ret = pqlvalue_layout(ctx->pql, te->value);
     break;

   }

   ret = tctype_layout(ctx, ret, te->datatype);
   return ret;
}

static struct layout *tcstmt_layout(struct tclayout *ctx, struct tcstmt *ts) {
   struct layout *ret;

   switch (ts->type) {
    case TCS_EXPR:
     ret = tcexpr_layout(ctx, ts->expr);
     break;
    case TCS_BEGIN:
     ret = mklayout_text(ctx->pql, "begin");
     break;
    case TCS_ABORT:
     ret = mklayout_text(ctx->pql, "abort");
     break;
    case TCS_COMMIT:
     ret = mklayout_text(ctx->pql, "commit");
     break;
   }

   return ret;
}

char *tcdump_expr(struct pqlcontext *pql, struct tcexpr *te, bool showtypes) {
   struct tclayout ctx;
   struct layout *l;
   char *ret;

   ctx.pql = pql;
   ctx.showtypes = showtypes;
   l = tcexpr_layout(&ctx, te);

   l = layout_format(pql, l, pql->dumpwidth);
   ret = layout_tostring(pql, l);
   layout_destroy(pql, l);

   return ret;
}

char *tcdump(struct pqlcontext *pql, struct tcstmt *ts, bool showtypes) {
   struct tclayout ctx;
   struct layout *l;
   char *ret;

   ctx.pql = pql;
   ctx.showtypes = showtypes;
   l = tcstmt_layout(&ctx, ts);

   l = layout_format(pql, l, pql->dumpwidth);
   ret = layout_tostring(pql, l);
   layout_destroy(pql, l);

   return ret;
}

////////////////////////////////////////////////////////////

#if 0 /* debugging code */

struct column *bogon;

static unsigned column_colrefs(struct column *cn, struct column *col) {
   return cn == col ? 1 : 0;
}

static unsigned columnarray_colrefs(struct columnarray *cs,
				    struct column *col) {
   unsigned i, num, ret;

   ret = 0;
   num = columnarray_num(cs);
   for (i=0; i<num; i++) {
      if (columnarray_get(cs, i) == col) {
	 ret++;
      }
   }
   return ret;
}

static unsigned tcvar_colrefs(struct tcvar *v, struct column *col) {
   (void)v;
   (void)col;
   return 0;
}

static unsigned tcexpr_colrefs(struct tcexpr *te, struct column *col);

static unsigned tcexprarray_colrefs(struct tcexprarray *arr,
				     struct column *col) {
   unsigned i, num, ret;

   ret = 0;
   num = tcexprarray_num(arr);
   for (i=0; i<num; i++) {
      ret += tcexpr_colrefs(tcexprarray_get(arr, i), col);
   }
   return ret;
}

static unsigned tcexpr_colrefs(struct tcexpr *te, struct column *col) {
   unsigned ret;

   if (te == NULL) {
      return 0;
   }

   switch (te->type) {
    case TCE_FILTER:
     ret = tcexpr_colrefs(te->filter.sub, col)
	+ tcvar_colrefs(te->filter.predvar, col)
	+ tcexpr_colrefs(te->filter.predexpr, col);
     break;
    case TCE_PROJECTTUPLE:
     ret = tcexpr_colrefs(te->projecttuple.sub, col)
	+ columnarray_colrefs(te->projecttuple.cols, col);
     break;
    case TCE_PROJECTSCALAR:
     ret = tcexpr_colrefs(te->projectscalar.sub, col)
	+ column_colrefs(te->projectscalar.col, col);
     break;
    case TCE_RENAME:
     ret = tcexpr_colrefs(te->rename.sub, col)
	+ column_colrefs(te->rename.oldcol, col)
	+ column_colrefs(te->rename.newcol, col);
     break;
    case TCE_JOIN:
     ret = tcexpr_colrefs(te->join.left, col)
	+ tcexpr_colrefs(te->join.right, col)
	+ tcvar_colrefs(te->join.predvar, col)
	+ tcexpr_colrefs(te->join.predexpr, col);
     break;
    case TCE_ORDER:
     ret = tcexpr_colrefs(te->order.sub, col)
	+ columnarray_colrefs(te->order.cols, col);
     break;
    case TCE_UNIQ:
     ret = tcexpr_colrefs(te->uniq.sub, col)
	+ columnarray_colrefs(te->uniq.cols, col);
     break;
    case TCE_NEST:
     ret = tcexpr_colrefs(te->nest.sub, col)
	+ columnarray_colrefs(te->nest.cols, col)
	+ column_colrefs(te->nest.newcol, col);
     break;
    case TCE_UNNEST:
     ret = tcexpr_colrefs(te->unnest.sub, col)
	+ column_colrefs(te->unnest.col, col);
     break;
    case TCE_DISTINGUISH:
     ret = tcexpr_colrefs(te->distinguish.sub, col)
	+ column_colrefs(te->distinguish.newcol, col);
     break;
    case TCE_ADJOIN:
     ret = tcexpr_colrefs(te->adjoin.left, col)
	+ tcexpr_colrefs(te->adjoin.func, col)
	+ column_colrefs(te->adjoin.newcol, col);
     break;
    case TCE_STEP:
     ret = tcexpr_colrefs(te->step.sub, col)
	+ column_colrefs(te->step.subcolumn, col)
	+ column_colrefs(te->step.leftobjcolumn, col)
	+ column_colrefs(te->step.edgecolumn, col)
	+ column_colrefs(te->step.rightobjcolumn, col)
	+ tcvar_colrefs(te->step.predvar, col)
	+ tcexpr_colrefs(te->step.predexpr, col);
     break;
    case TCE_REPEAT:
     ret = tcexpr_colrefs(te->repeat.sub, col)
	+ column_colrefs(te->repeat.subendcolumn, col)
	+ tcvar_colrefs(te->repeat.loopvar, col)
	+ column_colrefs(te->repeat.bodystartcolumn, col)
	+ tcexpr_colrefs(te->repeat.body, col)
	+ column_colrefs(te->repeat.bodybodycolumn, col)
	+ column_colrefs(te->repeat.bodypathcolumn, col)
	+ column_colrefs(te->repeat.bodyendcolumn, col)
	+ column_colrefs(te->repeat.repeatpathcolumn, col)
	+ column_colrefs(te->repeat.repeatendcolumn, col);
     break;
    case TCE_SCAN:
     ret = column_colrefs(te->scan.leftobjcolumn, col)
	+ column_colrefs(te->scan.edgecolumn, col)
	+ column_colrefs(te->scan.rightobjcolumn, col);
     break;
    case TCE_BOP:
     ret = tcexpr_colrefs(te->bop.left, col)
	+ tcexpr_colrefs(te->bop.right, col);
     break;
    case TCE_UOP:
     ret = tcexpr_colrefs(te->uop.sub, col);
     break;
    case TCE_FUNC:
     ret = tcexprarray_colrefs(&te->func.args, col);
     break;
    case TCE_MAP:
     ret = tcvar_colrefs(te->map.var, col)
	+ tcexpr_colrefs(te->map.set, col)
	+ tcexpr_colrefs(te->map.result, col);
     break;
    case TCE_LET:
     ret = tcvar_colrefs(te->let.var, col)
	+ tcexpr_colrefs(te->let.value, col)
	+ tcexpr_colrefs(te->let.body, col);
     break;
    case TCE_LAMBDA:
     ret = tcvar_colrefs(te->lambda.var, col)
	+ tcexpr_colrefs(te->lambda.body, col);
     break;
    case TCE_APPLY:
     ret = tcexpr_colrefs(te->apply.lambda, col)
	+ tcexpr_colrefs(te->apply.arg, col);
     break;
    case TCE_READVAR:
     //ret = tcvar_colrefs(te->readvar, col);
     ret = 0;
     break;
    case TCE_READGLOBAL:
     ret = 0;
     break;
    case TCE_CREATEPATHELEMENT:
     ret = tcexpr_colrefs(te->createpathelement, col);
     break;
    case TCE_SPLATTER:
     ret = tcexpr_colrefs(te->splatter.value, col)
	+ tcexpr_colrefs(te->splatter.name, col);
     break;
    case TCE_TUPLE:
     ret = tcexprarray_colrefs(&te->tuple.exprs, col)
	+ columnarray_colrefs(te->tuple.columns, col);
     break;
    case TCE_VALUE:
     ret = 0;
     break;
   }

   return ret;
}

bool column_checkrefcount(struct tcexpr *te, struct column *col,
			   unsigned extrarefs) {
   unsigned a, b;

   if (col == NULL) {
      return true;
   }

   a = tcexpr_colrefs(te, col) + extrarefs;
   b = column_getrefcount(col);
   if (a != b) {
      printf("!!! found %u count %u\n", a, b);
   }
   return a == b;
}

#endif

#if 0 /* more debugging code */

struct tcvar *bogon;

#define column_varrefs(a, b) (0)
#define columnarray_varrefs(a, b) (0)

static unsigned tcvar_varrefs(struct tcvar *v, struct tcvar *var) {
   return v == var;
}

static unsigned tcexpr_varrefs(struct tcexpr *te, struct tcvar *var);

static unsigned tcexprarray_varrefs(struct tcexprarray *arr,
				    struct tcvar *var) {
   unsigned i, num, ret;

   ret = 0;
   num = tcexprarray_num(arr);
   for (i=0; i<num; i++) {
      ret += tcexpr_varrefs(tcexprarray_get(arr, i), var);
   }
   return ret;
}

static unsigned tcexpr_varrefs(struct tcexpr *te, struct tcvar *var) {
   unsigned ret;

   if (te == NULL) {
      return 0;
   }

   switch (te->type) {
    case TCE_FILTER:
     ret = tcexpr_varrefs(te->filter.sub, var)
	+ tcvar_varrefs(te->filter.predvar, var)
	+ tcexpr_varrefs(te->filter.predexpr, var);
     break;
    case TCE_PROJECTTUPLE:
     ret = tcexpr_varrefs(te->projecttuple.sub, var)
	+ columnarray_varrefs(te->projecttuple.cols, var);
     break;
    case TCE_PROJECTSCALAR:
     ret = tcexpr_varrefs(te->projectscalar.sub, var)
	+ column_varrefs(te->projectscalar.col, var);
     break;
    case TCE_RENAME:
     ret = tcexpr_varrefs(te->rename.sub, var)
	+ column_varrefs(te->rename.oldcol, var)
	+ column_varrefs(te->rename.newcol, var);
     break;
    case TCE_JOIN:
     ret = tcexpr_varrefs(te->join.left, var)
	+ tcexpr_varrefs(te->join.right, var)
	+ tcvar_varrefs(te->join.predvar, var)
	+ tcexpr_varrefs(te->join.predexpr, var);
     break;
    case TCE_ORDER:
     ret = tcexpr_varrefs(te->order.sub, var)
	+ columnarray_varrefs(te->order.cols, var);
     break;
    case TCE_UNIQ:
     ret = tcexpr_varrefs(te->uniq.sub, var)
	+ columnarray_varrefs(te->uniq.cols, var);
     break;
    case TCE_NEST:
     ret = tcexpr_varrefs(te->nest.sub, var)
	+ columnarray_varrefs(te->nest.cols, var)
	+ column_varrefs(te->nest.newcol, var);
     break;
    case TCE_UNNEST:
     ret = tcexpr_varrefs(te->unnest.sub, var)
	+ column_varrefs(te->unnest.col, var);
     break;
    case TCE_DISTINGUISH:
     ret = tcexpr_varrefs(te->distinguish.sub, var)
	+ column_varrefs(te->distinguish.newcol, var);
     break;
    case TCE_ADJOIN:
     ret = tcexpr_varrefs(te->adjoin.left, var)
	+ tcexpr_varrefs(te->adjoin.func, var)
	+ column_varrefs(te->adjoin.newcol, var);
     break;
    case TCE_STEP:
     ret = tcexpr_varrefs(te->step.sub, var)
	+ column_varrefs(te->step.subcolumn, var)
	+ column_varrefs(te->step.leftobjcolumn, var)
	+ column_varrefs(te->step.edgecolumn, var)
	+ column_varrefs(te->step.rightobjcolumn, var)
	+ tcvar_varrefs(te->step.predvar, var)
	+ tcexpr_varrefs(te->step.predexpr, var);
     break;
    case TCE_REPEAT:
     ret = tcexpr_varrefs(te->repeat.sub, var)
	+ column_varrefs(te->repeat.subendcolumn, var)
	+ tcvar_varrefs(te->repeat.loopvar, var)
	+ column_varrefs(te->repeat.bodystartcolumn, var)
	+ tcexpr_varrefs(te->repeat.body, var)
	+ column_varrefs(te->repeat.bodybodycolumn, var)
	+ column_varrefs(te->repeat.bodypathcolumn, var)
	+ column_varrefs(te->repeat.bodyendcolumn, var)
	+ column_varrefs(te->repeat.repeatpathcolumn, var)
	+ column_varrefs(te->repeat.repeatendcolumn, var);
     break;
    case TCE_SCAN:
     ret = column_varrefs(te->scan.leftobjcolumn, var)
	+ column_varrefs(te->scan.edgecolumn, var)
	+ column_varrefs(te->scan.rightobjcolumn, var);
     break;
    case TCE_BOP:
     ret = tcexpr_varrefs(te->bop.left, var)
	+ tcexpr_varrefs(te->bop.right, var);
     break;
    case TCE_UOP:
     ret = tcexpr_varrefs(te->uop.sub, var);
     break;
    case TCE_FUNC:
     ret = tcexprarray_varrefs(&te->func.args, var);
     break;
    case TCE_MAP:
     ret = tcvar_varrefs(te->map.var, var)
	+ tcexpr_varrefs(te->map.set, var)
	+ tcexpr_varrefs(te->map.result, var);
     break;
    case TCE_LET:
     ret = tcvar_varrefs(te->let.var, var)
	+ tcexpr_varrefs(te->let.value, var)
	+ tcexpr_varrefs(te->let.body, var);
     break;
    case TCE_LAMBDA:
     ret = tcvar_varrefs(te->lambda.var, var)
	+ tcexpr_varrefs(te->lambda.body, var);
     break;
    case TCE_APPLY:
     ret = tcexpr_varrefs(te->apply.lambda, var)
	+ tcexpr_varrefs(te->apply.arg, var);
     break;
    case TCE_READVAR:
     ret = tcvar_varrefs(te->readvar, var);
     break;
    case TCE_READGLOBAL:
     ret = 0;
     break;
    case TCE_CREATEPATHELEMENT:
     ret = tcexpr_varrefs(te->createpathelement, var);
     break;
    case TCE_SPLATTER:
     ret = tcexpr_varrefs(te->splatter.value, var)
	+ tcexpr_varrefs(te->splatter.name, var);
     break;
    case TCE_TUPLE:
     ret = tcexprarray_varrefs(&te->tuple.exprs, var)
	+ columnarray_varrefs(te->tuple.columns, var);
     break;
    case TCE_VALUE:
     ret = 0;
     break;
   }

   return ret;
}

bool tcvar_checkrefcount(struct tcexpr *te, struct tcvar *var,
			 unsigned extrarefs) {
   unsigned a, b;

   if (var == NULL) {
      return true;
   }

   a = tcexpr_varrefs(te, var) + extrarefs;
   b = var->refcount;
   if (a != b) {
      printf("!!! found %u count %u\n", a, b);
   }
   return a == b;
}

#endif
