/*
 * Copyright 2009, 2010
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef TCALC_H
#define TCALC_H

#include <stdbool.h>

#include "array.h"
#include "functions.h"

struct pqlcontext;


/*
 * (Nested) tuple calculus.
 */

#ifndef TCALC_INLINE
#define TCALC_INLINE INLINE
#endif
struct tcglobal;
struct tcvar;
struct tcexpr;
DECLARRAY(tcglobal);
DECLARRAY(tcvar);
DECLARRAY(tcexpr);
DEFARRAY(tcglobal, TCALC_INLINE);
DEFARRAY(tcvar, TCALC_INLINE);
DEFARRAY(tcexpr, TCALC_INLINE);


/*
 * Standard relational operations:
 *
 *    FILTER(sub, expr) - take subset of SUB where EXPR is true.
 *                        ("select" in relational calculus terms.)
 *
 *    PROJECT_cols(sub) - extract columns
 *
 *    RENAME(sub, col => col) - change column name
 *
 *    |x|_cols - join on columns COLS
 *
 *    ORDER_cols - sort by values in COLS
 *
 *    UNIQ_cols - discard duplicates going by values in COLS
 *
 *
 * Nested relational operations:
 *
 *    NEST_cols_newcol - combine rows whose values are equal outside
 *                of COLS, and collect the values found in COLS into a
 *                single tuple set as the column NEWCOL.
 *
 *    UNNEST_col - replicate row values outside of COL, generating one
 *                for each member of the set found in COL.
 *
 *
 * Nonstandard or nonconvential operations:
 *
 *
 * DISTINGUISH_col(sub)
 *
 *    Pastes a distinguisher onto SUB with the column name COL.
 *
 *
 * ADJOIN_col ( left, func )
 *
 *    Computes func(left) and pastes it on.
 *    Definition:
 *
 *    let t = DISTINGUISH_d(LEFT)
 *        u = map t' over t: (FUNC(t'), PROJECT_d(t'))
 *    in t |x|_d u
 *
 * STEP_col(sub, edge)
 *
 *    Arises from "SUB.EDGE". Definition:
 *
 *    let t = SUB x ALLEDGES
 *    in filter t by project_COL(t) = project_ALLEDGES.LEFT(t)
 *                   and project_ALLEDGES.NAME(t) = EDGE
 *
 *    (where ALLEDGES with schema (LEFT, NAME, RIGHT) is the table
 *    of all object linkages.)
 *
 * REPEAT_d0_seq ( input, body )
 *
 *    Arises from ".EDGE+". Definition:
 *
 *    \input ->				       :: d0, start
 *    result := emptyset()                     :: d0, seq
 *    s := ADJOIN_seq(input, emptysequence())  :: d0, seq, start
 *    repeat
 *       let b0 = DISTINGUISH_d(s)             :: d0, seq, start, d
 *           b1 = PROJECT_start_d(b0)          :: start, d
 *           b2 = map b1' over b1: body(b1')   :: d, q*, start
 *           b3 = PROJECT_d0_seq_d(b0)         :: d0, seq, d
 *           b4 = b3 |x|_d b2                  :: d0, seq, d, q*, start
 *           b5 = ADJOIN_newseq(b4, \k -> PROJECT_seq(k) ++ PROJECT_q*(k))
 *                                             :: d0, seq, d, q*, newseq, start
 *           b6 = PROJECT_newseq_start(b5)     :: d0, newseq, start
 *           b7 = RENAME(b6, newseq -> seq)    :: d0, seq, start
 *       in s := b7
 *       result := result ++ PROJECT_d0_seq(s)
 *    while nonempty(s)
 *    return result                            :: d0, seq
 *
 *    This is huge. Bleh.
 */


/* Global (database) variable */
struct tcglobal {
   char *name;
   unsigned refcount;
};

/* Ordinary tuple variable */
struct tcvar {
   unsigned id;
   unsigned refcount;
   struct datatype *datatype;
};

/* Expression */
enum tcexprtypes {
   TCE_FILTER,				/* keep tuples where column = true */
   TCE_PROJECTTUPLE,			/* extract column(s) as tuple */
   TCE_PROJECTSCALAR,			/* extract one column as scalar */
   TCE_RENAME,				/* change column name */
   TCE_JOIN,				/* join */
   TCE_ORDER,				/* sort by column(s) */
   TCE_UNIQ,				/* uniq by column(s) */

   TCE_NEST,				/* collect column(s) */
   TCE_UNNEST,				/* expand column(s) */

   TCE_DISTINGUISH,			/* add column containing unique id */
   TCE_ADJOIN,				/* add column containing f(x) */

   TCE_STEP,				/* follow graph edge */
   TCE_REPEAT,				/* repeat until convergence */
   TCE_SCAN,				/* inspect whole objects table */

   TCE_BOP,				/* binary operator */
   TCE_UOP,				/* unary operator */
   TCE_FUNC,				/* user-called function */

   TCE_MAP,				/* map over a tuple set */
   TCE_LET,				/* let-bind a tuple set */
   TCE_READVAR,				/* read a tuple set var */
   TCE_READGLOBAL,			/* read from a global variable */

   TCE_CREATEPATHELEMENT,		/* construct a pathelement */
   TCE_SPLATTER,			/* splatter (computed edge name) */
   TCE_TUPLE,				/* construct a tuple */

   TCE_VALUE,				/* constant */
};
struct tcexpr {
   struct datatype *datatype;

   enum tcexprtypes type;
   union {
      /*
       * Standard relational algebra ops on tuple sets.
       */
      struct {
	 struct tcexpr *sub;
	 struct tcvar *predvar;
	 struct tcexpr *predexpr;
      } filter;
      struct {
	 struct tcexpr *sub;
	 struct columnarray *cols;
	 bool complement_cols;	/* true -> do cols = ~cols in typeinf */
      } projecttuple;
      struct {
	 struct tcexpr *sub;
	 struct column *col;
      } projectscalar;
      struct {
	 // XXX how do we do computed names? what if they're not uniform?
	 struct tcexpr *sub;
	 struct column *insidecol;	/* column generated by SUB */
	 struct column *outsidecol;	/* column we give to parent */
      } rename;
      struct {
	 struct tcexpr *left;
	 struct tcexpr *right;
	 struct tcvar *predvar;
	 struct tcexpr *predexpr;
      } join;
      struct {
	 struct tcexpr *sub;
	 struct columnarray *cols;
      } order;
      struct {
	 struct tcexpr *sub;
	 struct columnarray *cols;
      } uniq;

      /*
       * For nest/unnest the columns listed are the ones that are
       * transformed between sets and members. Thus nest is the
       * complement of the language-level group-by: given (A, B, C),
       * "group by C" means "nest on A, B".
       */
      struct {
	 struct tcexpr *sub;
	 struct columnarray *cols;
	 struct column *newcol;
	 bool complement_cols;	/* true -> do cols = ~cols in typeinf */
      } nest;
      struct {
	 struct tcexpr *sub;
	 struct column *col;
      } unnest;

      /*
       * Variant relational algebra ops on tuple sets, as defined above.
       */
      struct {
	 struct tcexpr *sub;
	 struct column *newcol;
      } distinguish;
      struct {
	 struct tcexpr *left;
	 struct tcvar *var;
	 struct tcexpr *expr;
	 struct column *newcol;
      } adjoin;

      /*
       * Graph-related ops, as defined above.
       */
      struct {
	 struct tcexpr *sub;
	 struct column *subcolumn;
	 struct pqlvalue *edgename;
	 bool reversed;
	 struct column *leftobjcolumn;	/* Column IDs to generate */
	 struct column *edgecolumn;
	 struct column *rightobjcolumn;
	 struct tcvar *predvar;
	 struct tcexpr *predexpr;
      } step;
      struct {
	 struct tcexpr *sub;		/* expr for stepping-off point */
	 struct column *subendcolumn;	/* trailing column from subexpr */
	 struct tcvar *loopvar;		/* var used by body */
	 struct column *bodystartcolumn;/* leading column used by body */
	 struct tcexpr *body;		/* repeated expression */
	 struct column *bodybodycolumn;/* column for repeated vars from body */
	 struct column *bodypathcolumn;/* path column generated by body */
	 struct column *bodyendcolumn;/* trailing column generated by body */
	 struct column *repeatpathcolumn;/* path column for repeat */
	 struct column *repeatendcolumn;/* trailing column for whole repeat*/
      } repeat;

      struct {
	 struct column *leftobjcolumn;	/* Column IDs to generate */
	 struct column *edgecolumn;
	 struct column *rightobjcolumn;
      } scan;

      /*
       * Computational operators.
       */
      struct {
	 struct tcexpr *left;
	 enum functions op;
	 struct tcexpr *right;
      } bop;
      struct {
	 enum functions op;
	 struct tcexpr *sub;
      } uop;
      struct {
	 enum functions op;
	 struct tcexprarray args;
      } func;

      /*
       * Tuple variables.
       */
      struct {
	 struct tcvar *var;
	 struct tcexpr *set;
	 struct tcexpr *result;
      } map;
      struct {
	 struct tcvar *var;
	 struct tcexpr *value;
	 struct tcexpr *body;
      } let;
      struct tcvar *readvar;
      struct tcglobal *readglobal;

      /*
       * Type constructors
       */
      struct tcexpr *createpathelement;
      struct {
	 struct tcexpr *value;
	 struct tcexpr *name;
      } splatter;
      struct {
	 struct tcexprarray exprs;
	 struct columnarray *columns;
      } tuple;

      /*
       * And finally, values.
       */
      struct pqlvalue *value;
   };
};

/* Statement */
enum tcstmttypes {
   TCS_EXPR,
   TCS_BEGIN,
   TCS_ABORT,
   TCS_COMMIT,
};
struct tcstmt {
   enum tcstmttypes type;
   union {
      struct tcexpr *expr;
      /* begin, abort, commit -- nothing */
   };
};

////////////////////////////////////////////////////////////
// constructors

struct tcglobal *mktcglobal(struct pqlcontext *, const char *name);
void tcglobal_incref(struct tcglobal *);

struct tcvar *mktcvar_fresh(struct pqlcontext *);
void tcvar_incref(struct tcvar *);

struct tcexpr *tcexpr_clone(struct pqlcontext *, const struct tcexpr *te);

struct tcexpr *mktcexpr_filter(struct pqlcontext *,
			       struct tcexpr *sub,
			       struct tcvar *predvar,
			       struct tcexpr *predexpr);
struct tcexpr *mktcexpr_projecttuple(struct pqlcontext *,
				     struct tcexpr *sub);
struct tcexpr *mktcexpr_projectscalar(struct pqlcontext *,
				      struct tcexpr *sub, struct column *col);
struct tcexpr *mktcexpr_rename(struct pqlcontext *,
			       struct tcexpr *sub, struct column *insidecol,
			       struct column *outsidecol);

struct tcexpr *mktcexpr_join(struct pqlcontext *,
			     struct tcexpr *left, struct tcexpr *right,
			     struct tcvar *predvar,
			     struct tcexpr *predexpr);
struct tcexpr *mktcexpr_order(struct pqlcontext *, struct tcexpr *sub);
struct tcexpr *mktcexpr_uniq(struct pqlcontext *, struct tcexpr *sub);


struct tcexpr *mktcexpr_nest_none(struct pqlcontext *,
				  struct tcexpr *sub,
				  struct column *newcol);
struct tcexpr *mktcexpr_nest_one(struct pqlcontext *,
				 struct tcexpr *sub, struct column *col,
				 struct column *newcol);
struct tcexpr *mktcexpr_nest_set(struct pqlcontext *,
				 struct tcexpr *sub, struct columnarray *cols,
				 struct column *newcol);
struct tcexpr *mktcexpr_unnest(struct pqlcontext *,
			       struct tcexpr *sub, struct column *col);


struct tcexpr *mktcexpr_distinguish(struct pqlcontext *,
			       struct tcexpr *sub, struct column *newcol);
struct tcexpr *mktcexpr_adjoin(struct pqlcontext *,
			       struct tcexpr *left,
			       struct tcvar *var,
			       struct tcexpr *expr,
			       struct column *newcol);


struct tcexpr *mktcexpr_step(struct pqlcontext *,
			     struct tcexpr *sub,
			     struct column *subcolumn,
			     struct pqlvalue *edgename,
			     bool reversed,
			     struct column *leftobjcolumn,
			     struct column *edgecolumn,
			     struct column *rightobjcolumn,
			     struct tcvar *predvar,
			     struct tcexpr *predexpr);
struct tcexpr *mktcexpr_repeat(struct pqlcontext *,
			       struct tcexpr *sub,
			       struct column *subendcolumn,
			       struct tcvar *loopvar,
			       struct column *bodystartcolumn,
			       struct tcexpr *body,
			       struct column *bodybodycolumn,
			       struct column *bodypathcolumn,
			       struct column *bodyendcolumn,
			       struct column *repeatpathcolumn,
			       struct column *repeatendcolumn);

struct tcexpr *mktcexpr_scan(struct pqlcontext *,
			     struct column *leftobjcolumn,
			     struct column *edgecolumn,
			     struct column *rightobjcolumn);


struct tcexpr *mktcexpr_bop(struct pqlcontext *, struct tcexpr *left,
			    enum functions op, struct tcexpr *right);
struct tcexpr *mktcexpr_uop(struct pqlcontext *,
			    enum functions op, struct tcexpr *sub);
struct tcexpr *mktcexpr_func(struct pqlcontext *pql, enum functions op);

struct tcexpr *mktcexpr_map(struct pqlcontext *, struct tcvar *var,
			    struct tcexpr *set, struct tcexpr *result);
struct tcexpr *mktcexpr_let(struct pqlcontext *, struct tcvar *var,
			    struct tcexpr *value, struct tcexpr *body);

struct tcexpr *mktcexpr_readvar(struct pqlcontext *, struct tcvar *var);
struct tcexpr *mktcexpr_readglobal(struct pqlcontext *, struct tcglobal *var);

struct tcexpr *mktcexpr_createpathelement(struct pqlcontext *,
					  struct tcexpr *sub);
struct tcexpr *mktcexpr_splatter(struct pqlcontext *pql,
				 struct tcexpr *value, struct tcexpr *name);
struct tcexpr *mktcexpr_tuple(struct pqlcontext *pql, unsigned arity);
struct tcexpr *mktcexpr_monople(struct pqlcontext *pql, struct tcexpr *val,
				struct column *col);

struct tcexpr *mktcexpr_value(struct pqlcontext *, struct pqlvalue *val);

struct tcstmt *mktcstmt_expr(struct pqlcontext *pql, struct tcexpr *expr);
struct tcstmt *mktcstmt_begin(struct pqlcontext *pql);
struct tcstmt *mktcstmt_abort(struct pqlcontext *pql);
struct tcstmt *mktcstmt_commit(struct pqlcontext *pql);

////////////////////////////////////////////////////////////
// destructors

void tcglobal_decref(struct pqlcontext *, struct tcglobal *);
void tcvar_decref(struct pqlcontext *, struct tcvar *);

void tcexpr_destroy(struct pqlcontext *pql, struct tcexpr *);
void tcstmt_destroy(struct pqlcontext *pql, struct tcstmt *);

////////////////////////////////////////////////////////////
// dump

char *tcdump_expr(struct pqlcontext *pql, struct tcexpr *te, bool showtypes);
char *tcdump(struct pqlcontext *pql, struct tcstmt *ts, bool showtypes);


#endif /* TCALC_H */
