/*
 * Copyright 2009, 2010
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Check the types (and column names) for tuple calculus.
 */

#include <stdarg.h>
#include <stdio.h>

#include "datatype.h"
#include "pqlvalue.h"
#include "tcalc.h"
#include "passes.h"
#include "pqlcontext.h"

struct tc {
   struct pqlcontext *pql;
   bool failed;
};

////////////////////////////////////////////////////////////
// context management

static void tc_init(struct tc *ctx, struct pqlcontext *pql) {
   ctx->pql = pql;
   ctx->failed = false;
}

static void tc_cleanup(struct tc *ctx) {
   (void)ctx;
}

static void tc_fail(struct tc *ctx) {
   ctx->failed = true;
}

static void tc_say(struct tc *ctx, const char *fmt, ...) PF(2,3);

static void tc_say(struct tc *ctx, const char *fmt, ...) {
   char buf[4096];
   va_list ap;
   
   va_start(ap, fmt);
   vsnprintf(buf, sizeof(buf), fmt, ap);
   va_end(ap);

   // XXX should get line numbers
   // XXX should add/use vcomplain
   complain(ctx->pql, 0, 0, "%s", buf);
}

////////////////////////////////////////////////////////////
// tools

////////////////////////////////////////////////////////////
// common type checks

/*
 * Check the variable from a filter. It should be the same as the
 * member type of the set the filter is filtering.
 */
static void tc_check_filtervar(struct tc *ctx,
			       struct datatype *membertype,
			       struct datatype *vartype) {
   if (vartype != NULL && !datatype_eq_withnames(vartype, membertype)) {
      tc_say(ctx, "Filter variable should have type %s; found %s",
	     datatype_getname(membertype),
	     datatype_getname(vartype));
      tc_fail(ctx);
   }
}

/*
 * Check the predicate expression from a filter. It should yield bool.
 */
static void tc_check_filterexpr(struct tc *ctx,
				struct datatype *bodytype) {
   if (bodytype != NULL &&
       !datatype_eq_withnames(bodytype, datatype_bool(ctx->pql))) {
      tc_say(ctx, "Filter should produce bool; found %s",
	     datatype_getname(bodytype));
      tc_fail(ctx);
   }
}

////////////////////////////////////////////////////////////
// column checks

/*
 * Return true if COL exists somewhere in T.
 */
static bool tc_lookfor_column(struct tc *ctx, struct datatype *t,
			      struct column *col) {
   unsigned i, num;
   struct datatype *subtype;
   struct column *subcol;

   if (datatype_isset(t)) {
      return tc_lookfor_column(ctx, datatype_set_member(t), col);
   }
   if (datatype_issequence(t)) {
      return tc_lookfor_column(ctx, datatype_sequence_member(t), col);
   }
   if (datatype_istuple(t)) {
      num = datatype_arity(t);
      for (i=0; i<num; i++) {
	 subcol = datatype_tuple_getcolumn(t, i);
	 if (subcol == col) {
	    return true;
	 }
	 subtype = datatype_getnth(t, i);
	 if (tc_lookfor_column(ctx, subtype, col)) {
	    return true;
	 }
      }
      return false;
   }
   /*
    * Everything else is a primitive type with no names in it.
    */
   return false;
}

/*
 * Fail if COL doesn't exist somewhere in T.
 */
static void tc_check_column_exists(struct tc *ctx, const char *what,
				   struct datatype *t,
				   struct column *col) {
   if (!tc_lookfor_column(ctx, t, col)) {
      tc_say(ctx, "%s uses nonexistent column %s", what, column_getname(col));
      tc_fail(ctx);
   }
}

/*
 * Fail if any of COLS fail to exist somewhere in T.
 */
static void tc_check_columns_exist(struct tc *ctx, const char *what,
				   struct datatype *t,
				   const struct columnarray *cols) {
   unsigned i, num;

   num = columnarray_num(cols);
   for (i=0; i<num; i++) {
      tc_check_column_exists(ctx, what, t, columnarray_get(cols, i));
   }
}

/*
 * Given "from", of type FROMTYPE with columns FROMCOLS, check that
 * CHOICES are all found in FROMCOLS, and produce the following:
 *
 *   - CHOSENTYPE, the type of CHOICES
 *   - UNCHOSENTYPE, the type of the columns not in CHOICES
 *
 * CHOSENTYPE, and/or UNCHOSENTYPE can be null if that data isn't
 * required.
 *
 * WHAT describes the expression this is part of, for complaints.
 */
static void tc_check_columnchoices(struct tc *ctx,
				   const char *what,
				   struct datatype *fromtype,
				   struct columnarray *choices,
				   struct datatype **chosentype,
				   struct datatype **unchosentype) {
   struct column *col;
   struct datatype *membertype, *t;
   unsigned numchoices, i;
   bool isset = false, isseq = false;

   numchoices = columnarray_num(choices);

   if (datatype_isset(fromtype)) {
      membertype = datatype_set_member(fromtype);
      isset = true;
   }
   else if (datatype_issequence(fromtype)) {
      membertype = datatype_sequence_member(fromtype);
      isseq = true;
   }
   else {
      membertype = fromtype;
   }

   if (!datatype_istuple(membertype)) {
      tc_say(ctx, "%s applied to non-tuple", what);
      tc_fail(ctx);
      /* make it up */
      if (chosentype != NULL) {
	 *chosentype = fromtype;
      }
      if (unchosentype != NULL) {
	 *unchosentype = datatype_unit(ctx->pql);
	 if (isset) {
	    *unchosentype = datatype_set(ctx->pql, *unchosentype);
	 } else if (isseq) {
	    *unchosentype = datatype_sequence(ctx->pql, *unchosentype);
	 }
      }
      return;
   }

   if (chosentype != NULL) {
      struct datatypearray types;
      struct columnarray typecols;

      datatypearray_init(&types);
      datatypearray_setsize(ctx->pql, &types, numchoices);
      columnarray_init(&typecols);
      columnarray_setsize(ctx->pql, &typecols, numchoices);

      for (i=0; i<numchoices; i++) {
	 col = columnarray_get(choices, i);
	 t = datatype_subprojectscalar(membertype, col);
	 if (t == NULL) {
	    tc_say(ctx, "%s tried to use nonexistent column %s", what,
		   column_getname(col));
	    tc_fail(ctx);
	    /* use bottom */
	    t = datatype_absbottom(ctx->pql);
	 }
	 datatypearray_set(&types, i, t);
	 columnarray_set(&typecols, i, col);
      }

      *chosentype = datatype_tuple_specific(ctx->pql, &types, &typecols);

      datatypearray_setsize(ctx->pql, &types, 0);
      datatypearray_cleanup(ctx->pql, &types);
      columnarray_setsize(ctx->pql, &typecols, 0);
      columnarray_cleanup(ctx->pql, &typecols);
   }

   /*
    * Do this separately so we get it in the same order as FROMTYPE,
    * but get CHOSENTYPE in the order of CHOICES, which isn't
    * necessarily the same.
    */
   if (unchosentype != NULL) {
      struct columnarray choicelist;

      columnarray_init(&choicelist);
      for (i=0; i<numchoices; i++) {
	 col = columnarray_get(choices, i);
	 columnarray_add(ctx->pql, &choicelist, col, NULL);
      }

      *unchosentype = datatype_substrip(membertype, &choicelist);

      columnarray_setsize(ctx->pql, &choicelist, 0);
      columnarray_cleanup(ctx->pql, &choicelist);
   }

   if (isset) {
      if (chosentype != NULL) {
	 *chosentype = datatype_set(ctx->pql, *chosentype);
      }
      if (unchosentype != NULL) {
	 *unchosentype = datatype_set(ctx->pql, *unchosentype);
      }
   }
   else if (isseq) {
      if (chosentype != NULL) {
	 *chosentype = datatype_sequence(ctx->pql, *chosentype);
      }
      if (unchosentype != NULL) {
	 *unchosentype = datatype_sequence(ctx->pql, *unchosentype);
      }
   }
}

static void tc_collect_columns(struct tc *ctx, struct datatype *t,
			       struct columnarray *fill) {
   unsigned i, num;
   struct column *col;

   if (datatype_isset(t)) {
      tc_collect_columns(ctx, datatype_set_member(t), fill);
   }
   else if (datatype_issequence(t)) {
      tc_collect_columns(ctx, datatype_sequence_member(t), fill);
   }
   else if (datatype_istuple(t)) {
      num = datatype_arity(t);
      for (i=0; i<num; i++) {
	 col = datatype_tuple_getcolumn(t, i);
	 PQLASSERT(col != NULL);
	 columnarray_add(ctx->pql, fill, col, NULL);
	 tc_collect_columns(ctx, datatype_getnth(t, i), fill);
      }
   }
}

static void tc_check_duplicatecolumns(struct tc *ctx, struct datatype *t) {
   struct columnarray collect;
   struct column *thiscol, *othercol;
   unsigned i, j, num;

   if (datatype_isanydbobj(t)) {
      /* shortcut a common case */
      return;
   }

   columnarray_init(&collect);
   tc_collect_columns(ctx, t, &collect);

   /* this should really be done by sort and linear compare */

   num = columnarray_num(&collect);
   for (i=0; i<num; i++) {
      thiscol = columnarray_get(&collect, i);
      for (j=i+1; j<num; j++) {
	 othercol = columnarray_get(&collect, j);
	 if (thiscol == othercol) {
	    tc_say(ctx, "Column name %s duplicated",
		   column_getname(thiscol));
	    tc_fail(ctx);
	    /* don't bother checking thiscol any more */
	    break;
	 }
      }
   }

   // drop the pointers as we didn't take references to them
   columnarray_setsize(ctx->pql, &collect, 0);
   columnarray_cleanup(ctx->pql, &collect);
}

////////////////////////////////////////////////////////////
// builtin functions

static void tc_check_bool(struct tc *ctx,
			  const char *subname,
			  const char *name,
			  struct datatype *t) {
   if (!datatype_match_specialize(ctx->pql, t, datatype_bool(ctx->pql))) {
      tc_say(ctx, "%s %s is not boolean (found %s)",
	     subname, name, datatype_getname(t));
      tc_fail(ctx);
   }
}

static void tc_check_string(struct tc *ctx,
			    const char *subname,
			    const char *name,
			    struct datatype *t) {
   if (!datatype_match_specialize(ctx->pql, t, datatype_string(ctx->pql))) {
      tc_say(ctx, "%s %s is not a string (found %s)",
	     subname, name, datatype_getname(t));
      tc_fail(ctx);
   }
}

static void tc_check_numeric(struct tc *ctx,
			     const char *subname,
			     const char *name,
			     struct datatype *t) {
   /* XXX this shouldn't need to be a special case */
   if (datatype_isabsdbedge(t)) {
      /* absdbedge can be specialized to integer, so it's ok */
      return;
   }
   if (!datatype_match_specialize(ctx->pql, t, datatype_absnumber(ctx->pql))) {
      tc_say(ctx, "%s %s is not a number (found %s)",
	     subname, name, datatype_getname(t));
      tc_fail(ctx);
   }
}

static void tc_check_pathelement(struct tc *ctx,
				 const char *subname,
				 const char *name,
				 struct datatype *t) {
   if (!datatype_match_specialize(ctx->pql, t,
				  datatype_pathelement(ctx->pql))) {
      tc_say(ctx, "%s %s is not a path element (found %s)",
	     subname, name, datatype_getname(t));
      tc_fail(ctx);
   }
}

static struct datatype *tc_check_set(struct tc *ctx,
				     const char *subname,
				     const char *name,
				     struct datatype *t) {
   if (!datatype_isset(t)) {
      tc_say(ctx, "%s %s is not a set (found %s)",
	     subname, name, datatype_getname(t));
      tc_fail(ctx);
      return t;
   }
   return datatype_set_member(t);
}

static int tc_check_numargs(struct tc *ctx, const char *name,
			    struct datatypearray *argtypes, unsigned numargs) {
   unsigned arity;

   arity = datatypearray_num(argtypes);
   if (arity != numargs) {
      tc_say(ctx, "Wrong number of arguments for %s (found %u, expected %u)",
	     name, arity, numargs);
      tc_fail(ctx);
      return -1;
   }
   return 0;
}

static void tc_check_bop(struct tc *ctx, const char *name,
			 struct datatypearray *argtypes,
			 struct datatype **left_ret,
			 struct datatype **right_ret) {
   if (tc_check_numargs(ctx, name, argtypes, 2) < 0) {
      *left_ret = *right_ret = datatype_absbottom(ctx->pql);
   }
   else {
      *left_ret = datatypearray_get(argtypes, 0);
      *right_ret = datatypearray_get(argtypes, 1);
   }
}

static void tc_check_uop(struct tc *ctx, const char *name,
			 struct datatypearray *argtypes,
			 struct datatype **arg_ret) {
   if (tc_check_numargs(ctx, name, argtypes, 1) < 0) {
      *arg_ret = datatype_absbottom(ctx->pql);
   }
   else {
      *arg_ret = datatypearray_get(argtypes, 0);
   }
}

static void tc_check_noargs(struct tc *ctx, const char *name,
			    struct datatypearray *argtypes) {
   tc_check_numargs(ctx, name, argtypes, 0);
}

static void function_tc(struct tc *ctx, enum functions f,
			struct datatypearray *argtypes,
			struct datatype *rettype) {

   struct datatype *left, *right, *leftmember, *rightmember;
   struct datatype *type, *member;
   struct datatype *result;
   const char *name;

   result = NULL; // gcc 4.1

   name = function_getname(f);

   switch (f) {
    case F_UNION:
    case F_INTERSECT:
    case F_EXCEPT:
    case F_UNIONALL:
    case F_INTERSECTALL:
    case F_EXCEPTALL:
     /* set(T) x set(T) -> set(T) */
     tc_check_bop(ctx, name, argtypes, &left, &right);
     leftmember = tc_check_set(ctx, "Left argument of", name, left);
     rightmember = tc_check_set(ctx, "Right argument of", name, right);

     if (datatype_istuple(leftmember) != datatype_istuple(rightmember)) {
	tc_say(ctx, "Cannot take %s of tuple set and scalar set",
	       function_getname(f));
	tc_fail(ctx);
     }
     else if (datatype_istuple(leftmember) /*&& datatype_istuple(rightmember)*/
	      && datatype_arity(leftmember) != datatype_arity(rightmember)) {
	tc_say(ctx, "Cannot take %s of tuple sets of different arity "
	       "(%u vs. %u)",
	       function_getname(f),
	       datatype_arity(leftmember), datatype_arity(rightmember));
	tc_fail(ctx);
     }

     /*
      * If we take a union of two tuple sets with different column
      * names, we keep the column names from the first. The second
      * might have different column names. Note that the types have
      * to agree regardless.
      *
      * 20091211 dholland: no, we need to make sure the column names
      * are the same. Otherwise we run into trouble silently if one
      * side has the same columns but they get into a different order
      * by accident. We'll insert renames somewhere for handling any
      * user-generated nonuniform names.
      */

     if (!datatype_eq_namesonly(leftmember, rightmember)) {
	tc_say(ctx, "Column names in %s do not match: %s, %s", name,
	       datatype_getname(leftmember), datatype_getname(rightmember));
	tc_fail(ctx);
     }

     result = datatype_match_generalize(ctx->pql, leftmember, rightmember);
     if (result == NULL) {
	tc_say(ctx, "Illegal types for %s: %s, %s",
	       function_getname(f),
	       datatype_getname(left), datatype_getname(right));
	tc_fail(ctx);
	result = datatype_abstop(ctx->pql);
     }

     /* this should maybe be an assertion */
     if (!datatype_eq_namesonly(leftmember, result)) {
	tc_say(ctx, "Column names in %s do not match result: %s, %s", name,
	       datatype_getname(leftmember), datatype_getname(result));
	tc_fail(ctx);
     }

     result = datatype_set(ctx->pql, result);
     break;

    case F_SUBSETOF:
     /* set(T) x set(T) -> bool */
     tc_check_bop(ctx, name, argtypes, &left, &right);
     rightmember = tc_check_set(ctx, "Right argument of", name, right);
     leftmember = tc_check_set(ctx, "Left argument of", name, left);
     result = datatype_match_specialize(ctx->pql, leftmember, rightmember);
     if (result == NULL) {
	tc_say(ctx, "Illegal types for %s: %s, %s", name,
	       datatype_getname(left), datatype_getname(right));
	tc_fail(ctx);
     }
     if (!datatype_eq_namesonly(leftmember, rightmember)) {
	tc_say(ctx, "Column names in %s do not match: %s, %s", name,
	       datatype_getname(leftmember), datatype_getname(rightmember));
	tc_fail(ctx);
     }
     result = datatype_bool(ctx->pql);
     break;

    case F_ELEMENTOF:
     /* T x set(T) -> bool */
     tc_check_bop(ctx, name, argtypes, &left, &right);
     rightmember = tc_check_set(ctx, "Right argument of", name, right);
     result = datatype_match_specialize(ctx->pql, left, rightmember);
     if (result == NULL) {
	tc_say(ctx, "Illegal types for %s: %s, %s", name,
	       datatype_getname(left), datatype_getname(right));
	tc_fail(ctx);
     }
     if (!datatype_eq_namesonly(left, rightmember)) {
	tc_say(ctx, "Column names in %s do not match: %s, %s", name,
	       datatype_getname(left), datatype_getname(rightmember));
	tc_fail(ctx);
     }
     result = datatype_bool(ctx->pql);
     break;

    case F_NONEMPTY:
     /* set(T) -> bool */
     tc_check_uop(ctx, name, argtypes, &type);
     member = tc_check_set(ctx, "Argument of", name, type);
     result = datatype_bool(ctx->pql);
     break;

    case F_MAKESET:
     /* T -> set(T) */
     tc_check_uop(ctx, name, argtypes, &type);
     result = datatype_set(ctx->pql, type);
     break;

    case F_GETELEMENT:
     /* set(T) -> T */
     tc_check_uop(ctx, name, argtypes, &type);
     result = tc_check_set(ctx, "Argument of", name, type);
     break;

    case F_MAYBEMAKESET:
     /* not allowed downstream of typeinf */
     PQLASSERT(0);
     break;

    case F_COUNT:
     /* set(T) -> int */
     tc_check_uop(ctx, name, argtypes, &type);
     member = tc_check_set(ctx, "Argument of", name, type);
     result = datatype_int(ctx->pql);
     break;

    case F_SUM:
    case F_MIN:
    case F_MAX:
     /* set(T) -> T, T numeric */
     tc_check_uop(ctx, name, argtypes, &type);
     member = tc_check_set(ctx, "Argument of", name, type);
     tc_check_numeric(ctx, "Argument of", name, member);
     result = member;
     break;

    case F_AVG:
     /* set(T) -> float, T numeric */
     tc_check_uop(ctx, name, argtypes, &type);
     member = tc_check_set(ctx, "Argument of", name, type);
     tc_check_numeric(ctx, "Argument of", name, member);
     result = datatype_double(ctx->pql);
     break;

    case F_ALLTRUE:
    case F_ANYTRUE:
     /* set(bool) -> bool */
     tc_check_uop(ctx, name, argtypes, &type);
     member = tc_check_set(ctx, "Argument of", name, type);
     tc_check_bool(ctx, "Member type of argument of", name, member);
     result = datatype_bool(ctx->pql);
     break;

    case F_AND:
    case F_OR:
     /* bool x bool -> bool */
     tc_check_bop(ctx, name, argtypes, &left, &right);
     tc_check_bool(ctx, "Left argument of", name, left);
     tc_check_bool(ctx, "Right argument of", name, right);
     result = datatype_bool(ctx->pql);
     break;

    case F_NOT:
     /* bool -> bool */
     tc_check_uop(ctx, name, argtypes, &type);
     tc_check_bool(ctx, "Argument of", name, type);
     result = datatype_bool(ctx->pql);
     break;

    case F_NEW:
     /* set OR sequence OR pathelement OR tuple -> struct  OR  T -> T */
     tc_check_uop(ctx, name, argtypes, &type);
     if (datatype_isset(type) ||
	 datatype_issequence(type) ||
	 datatype_ispathelement(type) ||
	 datatype_istuple(type)) {
	result = datatype_struct(ctx->pql);
     }
     else {
	result = type;
     }
     break;

    case F_STRFTIME:
     /* string x int -> string */
     // XXX there should be a separate time type
     tc_check_bop(ctx, name, argtypes, &left, &right);
     tc_check_string(ctx, "Format argument of", name, left);
     tc_check_numeric(ctx, "Timestamp argument of", name, right);
     result = datatype_string(ctx->pql);
     break;

    case F_NOW:
     /* unit -> int */
     // XXX there should be a separate time/timestamp type
     tc_check_noargs(ctx, name, argtypes);
     result = datatype_int(ctx->pql);
     break;

    case F_EQ:
    case F_NOTEQ:
     /* T x T -> bool */
     tc_check_bop(ctx, name, argtypes, &left, &right);
     result = datatype_match_specialize(ctx->pql, left, right);
     if (result == NULL) {
	tc_say(ctx, "Illegal types for %s: %s, %s", name,
	       datatype_getname(left), datatype_getname(right));
	tc_fail(ctx);
     }
     result = datatype_bool(ctx->pql);
     break;

    case F_LT:
    case F_GT:
    case F_LTEQ:
    case F_GTEQ:
     /* T x T -> bool, T numeric */
     tc_check_bop(ctx, name, argtypes, &left, &right);
     tc_check_numeric(ctx, "Left argument of", name, left);
     tc_check_numeric(ctx, "Right argument of", name, right);
     result = datatype_match_specialize(ctx->pql, left, right);
     if (result == NULL) {
	tc_say(ctx, "Illegal types for %s: %s, %s", name,
	       datatype_getname(left), datatype_getname(right));
	tc_fail(ctx);
     }
     result = datatype_bool(ctx->pql);
     break;

    case F_LIKE:
    case F_GLOB:
    case F_GREP:
    case F_SOUNDEX:
     /* string x string -> bool */
     tc_check_bop(ctx, name, argtypes, &left, &right);
     tc_check_string(ctx, "Left argument of", name, left);
     tc_check_string(ctx, "Right argument of", name, right);
     result = datatype_bool(ctx->pql);
     break;

    case F_TOSTRING:
     /* T -> string */
     tc_check_uop(ctx, name, argtypes, &type);
     result = datatype_string(ctx->pql);
     break;

    case F_CONCAT:
     /*
      * string x string -> string  OR  seq(T) x seq(T) -> seq(T)
      * OR pathelemeent x pathelement -> seq(pathelement)
      *
      * (XXX: should we allow seq(pathelement) ++ pathelement like
      * this, or should we insist that the upstream code wrap single
      * pathelements in sequences if it's going to concat them?
      */
     tc_check_bop(ctx, name, argtypes, &left, &right);
     if (datatype_issequence(left) && datatype_issequence(right)) {
	leftmember = datatype_sequence_member(left);
	rightmember = datatype_sequence_member(right);
	result = datatype_match_generalize(ctx->pql, leftmember, rightmember);
	if (result == NULL) {
	   tc_say(ctx, "Illegal types for %s: %s, %s",
		  function_getname(f),
		  datatype_getname(left), datatype_getname(right));
	   tc_fail(ctx);
	   result = datatype_abstop(ctx->pql);
	}
	if (!datatype_eq_namesonly(leftmember, rightmember)) {
	   tc_say(ctx, "Column names in %s do not match: %s, %s", name,
		  datatype_getname(leftmember), datatype_getname(rightmember));
	   tc_fail(ctx);
	}
	result = datatype_sequence(ctx->pql, result);
     }
     else if (datatype_isstring(left) && datatype_isstring(right)) {
	result = left;
     }
     else if (datatype_ispathelement(left) && 
	      datatype_ispathelement(right)) {
	result = datatype_sequence(ctx->pql, left);
     }
     else if (datatype_ispathelement(left) && 
	      datatype_issequence(right) &&
	      datatype_ispathelement(datatype_sequence_member(right))) {
	result = right;
     }
     else if (datatype_issequence(left) &&
	      datatype_ispathelement(datatype_sequence_member(left)) &&
	      datatype_ispathelement(right)) {
	result = left;
     }
     else {
	tc_say(ctx, "Illegal types for %s: %s, %s", name,
	       datatype_getname(left), datatype_getname(right));
	tc_fail(ctx);
	result = left;
     }
     break;

    /* pathelement -> dbobj */
    case F_LEFT:
    case F_RIGHT:
     tc_check_uop(ctx, name, argtypes, &type);
     tc_check_pathelement(ctx, "Argument of", name, type);
     result = datatype_absdbobj(ctx->pql);
     break;

    /* pathelement -> dbedge */
    case F_EDGE:
     tc_check_uop(ctx, name, argtypes, &type);
     tc_check_pathelement(ctx, "Argument of", name, type);
     result = datatype_absdbedge(ctx->pql);
     break;

    case F_CHOOSE:
     /* T x T -> T */
     tc_check_bop(ctx, name, argtypes, &left, &right);
     result = datatype_match_generalize(ctx->pql, left, right);
     if (result == NULL) {
	tc_say(ctx, "Illegal types for %s: %s, %s", name,
	       datatype_getname(left), datatype_getname(right));
	tc_fail(ctx);
	result = left;
     }
     if (!datatype_eq_namesonly(left, right)) {
	tc_say(ctx, "Column names in %s do not match: %s, %s", name,
	       datatype_getname(left), datatype_getname(right));
	tc_fail(ctx);
     }
     break;

    case F_ADD:
    case F_SUB:
    case F_MUL:
    case F_DIV:
    case F_MOD:
     /* T x T -> T, T numeric */
     tc_check_bop(ctx, name, argtypes, &left, &right);
     tc_check_numeric(ctx, "Left argument of", name, left);
     tc_check_numeric(ctx, "Right argument of", name, right);
     result = datatype_match_generalize_any(ctx->pql, left, right);
#if 0 /* wrong */
     result = datatype_match_specialize(ctx->pql, left, right);
     if (result == NULL) {
#if 0 /* tc_check_numeric already issues a message */
	tc_say(ctx, "Illegal types for %s: %s, %s", name,
	       datatype_getname(left), datatype_getname(right));
	tc_fail(ctx);
#else
	PQLASSERT(ctx->failed);
#endif
	result = datatype_absany(ctx->pql);
     }
#endif
     break;

    case F_NEG:
    case F_ABS:
     /* T -> T, T numeric */
     tc_check_uop(ctx, name, argtypes, &type);
     tc_check_numeric(ctx, "Argument of", name, type);
     result = type;
     break;

    case F_IDENTITY:
     /* T -> T */
     tc_check_uop(ctx, name, argtypes, &type);
     result = type;
     break;
   }

   if (!datatype_eq_withnames(result, rettype)) {
      tc_say(ctx, "Result of %s has wrong type (computed %s, had %s)",
	     name, datatype_getname(result), datatype_getname(rettype));
      tc_fail(ctx);
   }
}

////////////////////////////////////////////////////////////
// recursive traversal

static void tcexpr_tc(struct tc *ctx, struct tcexpr *te) {
   PQLASSERT(te->datatype != NULL);

   tc_check_duplicatecolumns(ctx, te->datatype);

   switch (te->type) {
    case TCE_FILTER:
     tcexpr_tc(ctx, te->filter.sub);
     tcexpr_tc(ctx, te->filter.predexpr);
     {
	struct datatype *membertype;

	if (!datatype_eq_withnames(te->datatype, te->filter.sub->datatype)) {
	   tc_say(ctx, "Filter operation changes type of its operand set");
	   tc_fail(ctx);
	}
	if (!datatype_isset(te->datatype)) {
	   tc_say(ctx, "Filter operation applied to non-set");
	   tc_fail(ctx);
	   membertype = te->datatype;
	}
	else {
	   membertype = datatype_set_member(te->datatype);
	}

	if (te->filter.predvar != NULL) {
	   tc_check_filtervar(ctx, membertype, te->filter.predvar->datatype);
	}
	PQLASSERT(te->filter.predexpr != NULL);
	tc_check_filterexpr(ctx, te->filter.predexpr->datatype);
     }
     break;

    case TCE_PROJECTTUPLE:
     PQLASSERT(te->projecttuple.complement_cols == false);
     tcexpr_tc(ctx, te->projecttuple.sub);
     {
	struct datatype *resulttype;

	tc_check_columns_exist(ctx, "Project expression",
			       te->projecttuple.sub->datatype,
			       te->projecttuple.cols);

	if (datatype_isset(te->datatype)) {
	   PQLASSERT(datatype_istuple(datatype_set_member(te->datatype)));
	}
	else {
	   PQLASSERT(datatype_istuple(te->datatype));
	}

	tc_check_columnchoices(ctx, "Project expression",
			       te->projecttuple.sub->datatype,
			       te->projecttuple.cols,
			       &resulttype,
			       NULL);

	if (!datatype_eq_withnames(resulttype, te->datatype)) {
	   tc_say(ctx, "Project result has wrong type (computed %s, found %s)",
		  datatype_getname(resulttype),
		  datatype_getname(te->datatype));
	   tc_fail(ctx);
	}
     }
     break;

    case TCE_PROJECTSCALAR:
     tcexpr_tc(ctx, te->projectscalar.sub);
     {
	struct datatype *starttype;
	struct datatype *resulttype;
	struct columnarray *cols;
	bool isset = false, isseq = false;

	tc_check_column_exists(ctx, "Project-scalar expression",
			       te->projectscalar.sub->datatype,
			       te->projectscalar.col);

	if (datatype_isset(te->datatype)) {
	   PQLASSERT(!datatype_istuple(datatype_set_member(te->datatype)));
	}
	else {
	   PQLASSERT(!datatype_istuple(te->datatype));
	}

	starttype = te->projectscalar.sub->datatype;
	if (datatype_isset(starttype)) {
	   starttype = datatype_set_member(starttype);
	   isset = true;
	}
	else if (datatype_issequence(starttype)) {
	   starttype = datatype_sequence_member(starttype);
	   isseq = true;
	}

	cols = columnarray_create(ctx->pql);
	columnarray_add(ctx->pql, cols, te->projectscalar.col, NULL);
	tc_check_columnchoices(ctx, "Project expression",
			       starttype,
			       cols,
			       &resulttype,
			       NULL);
	// drop the pointers as we didn't take references to them
	columnarray_setsize(ctx->pql, cols, 0);
	columnarray_destroy(ctx->pql, cols);

	// XXX should this be allowed?
	if (!datatype_isabstop(resulttype)) {
	   PQLASSERT(datatype_istuple(resulttype));
	   PQLASSERT(datatype_arity(resulttype) == 1);
	   resulttype = datatype_getnth(resulttype, 0);
	}

	if (isset) {
	   resulttype = datatype_set(ctx->pql, resulttype);
	}
	else if (isseq) {
	   resulttype = datatype_sequence(ctx->pql, resulttype);
	}

	if (!datatype_eq_withnames(resulttype, te->datatype)) {
	   tc_say(ctx, "Project-scalar result has wrong type "
		       "(computed %s, found %s)",
		  datatype_getname(resulttype),
		  datatype_getname(te->datatype));
	   tc_fail(ctx);
	}
     }
     break;

    case TCE_RENAME:
     tcexpr_tc(ctx, te->rename.sub);
     if (!datatype_eq_nonames(te->datatype, te->rename.sub->datatype)) {
	tc_say(ctx, "Rename operation changed type of expression");
	tc_fail(ctx);
     }
     {
	struct datatype *ct;

	ct = datatype_rename(te->rename.sub->datatype,
			     te->rename.insidecol, te->rename.outsidecol);
	if (ct == NULL) {
	   tc_say(ctx, "Target column %s of rename operation not found",
		  column_getname(te->rename.insidecol));
	   tc_fail(ctx);
	}
	else if (!datatype_eq_withnames(te->datatype, ct)) {
	   tc_say(ctx, "Rename operation produced wrong type: %s -> %s",
		  datatype_getname(te->rename.sub->datatype),
		  datatype_getname(te->datatype));
	   tc_fail(ctx);
	}
     }
     break;

    case TCE_JOIN:
     tcexpr_tc(ctx, te->join.left);
     tcexpr_tc(ctx, te->join.right);
     if (te->join.predexpr != NULL) {
	tcexpr_tc(ctx, te->join.predexpr);
     }
     {
	struct datatype *left, *right, *t;
	bool isset = false;

	left = te->join.left->datatype;
	right = te->join.right->datatype;
	if (datatype_isset(left)) {
	   left = datatype_set_member(left);
	   isset = true;
	}
	if (datatype_isset(right)) {
	   right = datatype_set_member(right);
	   isset = true;
	}
	/* XXX should this be allowed? */
	if (datatype_isabstop(left) || datatype_isabstop(right)) {
	   t = datatype_abstop(ctx->pql);
	}
	else {
	   t = datatype_tuple_concat(ctx->pql, left, right);
	}
	if (isset) {
	   t = datatype_set(ctx->pql, t);
	}
	else {
	   /* XXX should this be allowed? */
	}
	if (!datatype_eq_withnames(t, te->datatype)) {
	   tc_say(ctx, "Result of join has wrong type");
	   tc_fail(ctx);
	}
	if (te->join.predvar != NULL || te->join.predexpr != NULL) {
	   if (!isset) {
	      /* this at least can (must) be prohibited */
	      tc_say(ctx, "Filter operation applied to non-set");
	      tc_fail(ctx);
	   }
	   if (te->join.predvar != NULL) {
	      tc_check_filtervar(ctx, datatype_set_member(te->datatype),
				 te->join.predvar->datatype);
	   }
	   if (te->join.predexpr != NULL) {
	      tc_check_filterexpr(ctx, te->join.predexpr->datatype);
	   }
	}
     }
     break;

    case TCE_ORDER:
     tcexpr_tc(ctx, te->order.sub);
     tc_check_columns_exist(ctx, "Order expression",
			    te->order.sub->datatype, te->order.cols);
     if (!datatype_isset(te->order.sub->datatype)) {
	tc_say(ctx, "Order applied to non-set");
	tc_fail(ctx);
	/* give up */
	break;
     }
     if (!datatype_issequence(te->datatype)) {
	tc_say(ctx, "Order result is not a sequence");
	tc_fail(ctx);
	/* give up */
	break;
     }
     if (!datatype_eq_withnames(datatype_set_member(te->order.sub->datatype),
				datatype_sequence_member(te->datatype))) {
	tc_say(ctx, "Order changes type of operand");
	tc_fail(ctx);
     }
     break;

    case TCE_UNIQ:
     tcexpr_tc(ctx, te->uniq.sub);
     tc_check_columns_exist(ctx, "Uniq expression",
			    te->uniq.sub->datatype, te->uniq.cols);
     if (!datatype_issequence(te->uniq.sub->datatype)) {
	tc_say(ctx, "Uniq applied to non-sequence");
	tc_fail(ctx);
	/* give up */
	break;
     }
     if (!datatype_eq_withnames(te->order.sub->datatype, te->datatype)) {
	tc_say(ctx, "Uniq changes type of operand");
	tc_fail(ctx);
     }
     break;

    case TCE_NEST:
     tcexpr_tc(ctx, te->nest.sub);
     {
	struct datatype *chosentype, *unchosentype, *resulttype;

	tc_check_columns_exist(ctx, "Nest expression",
			       te->nest.sub->datatype, te->nest.cols);

	tc_check_columnchoices(ctx, "Nest expression",
			       te->nest.sub->datatype,
			       te->nest.cols,
			       &chosentype,
			       &unchosentype);

	/* tc_check_columnchoices conses the results back up; undo that */
	if (datatype_isset(te->nest.sub->datatype)) {
	   PQLASSERT(datatype_isset(chosentype));
	   PQLASSERT(datatype_isset(unchosentype));
	   chosentype = datatype_set_member(chosentype);
	   unchosentype = datatype_set_member(unchosentype);
	}
	else if (datatype_issequence(te->nest.sub->datatype)) {
	   PQLASSERT(datatype_issequence(chosentype));
	   PQLASSERT(datatype_issequence(unchosentype));
	   chosentype = datatype_sequence_member(chosentype);
	   unchosentype = datatype_sequence_member(unchosentype);
	}

	resulttype = datatype_tuple_append(ctx->pql,
					   unchosentype,
					   datatype_set(ctx->pql, chosentype),
					   te->nest.newcol);

	if (datatype_isset(te->nest.sub->datatype)) {
	   resulttype = datatype_set(ctx->pql, resulttype);
	}
	else if (datatype_issequence(te->nest.sub->datatype)) {
	   resulttype = datatype_sequence(ctx->pql, resulttype);
	}


	if (!datatype_eq_withnames(te->datatype, resulttype)) {
	   tc_say(ctx, "Nest result type is wrong (computed %s, found %s)",
		  datatype_getname(resulttype),
		  datatype_getname(te->datatype));
	   tc_fail(ctx);
	}
     }
     break;

    case TCE_UNNEST:
     tcexpr_tc(ctx, te->unnest.sub);
     {
	struct columnarray *cols;
	struct datatype *chosentype, *unchosentype;
	struct datatype *unnestedtype, *resulttype;

	tc_check_column_exists(ctx, "Unnest expression",
			       te->unnest.sub->datatype, te->unnest.col);

	cols = columnarray_create(ctx->pql);
	columnarray_add(ctx->pql, cols, te->unnest.col, NULL);

	tc_check_columnchoices(ctx, "Unnest expression",
			       te->unnest.sub->datatype,
			       cols,
			       &chosentype,
			       &unchosentype);

	// drop the pointers as we didn't take references to them
	columnarray_setsize(ctx->pql, cols, 0);
	columnarray_destroy(ctx->pql, cols);

	/* tc_check_columnchoices conses the results back up; undo that */
	if (datatype_isset(te->unnest.sub->datatype)) {
	   PQLASSERT(datatype_isset(chosentype));
	   PQLASSERT(datatype_isset(unchosentype));
	   chosentype = datatype_set_member(chosentype);
	   unchosentype = datatype_set_member(unchosentype);
	}
	else if (datatype_issequence(te->unnest.sub->datatype)) {
	   PQLASSERT(datatype_issequence(chosentype));
	   PQLASSERT(datatype_issequence(unchosentype));
	   chosentype = datatype_sequence_member(chosentype);
	   unchosentype = datatype_sequence_member(unchosentype);
	}

	PQLASSERT(datatype_istuple(chosentype));
	PQLASSERT(datatype_arity(chosentype) == 1);
	chosentype = datatype_getnth(chosentype, 0);

	// XXX this should enforce that te->unnest.col is a toplevel column

	if (datatype_isset(chosentype)) {
	   unnestedtype = datatype_set_member(chosentype);
	}
	else if (datatype_issequence(chosentype)) {
	   unnestedtype = datatype_sequence_member(chosentype);
	}
	else {
	   tc_say(ctx, "Cannot unnest non-set column %s",
		  column_getname(te->unnest.col));
	   tc_fail(ctx);
	   /* give up */
	   break;
	}

	if (datatype_istuple(unnestedtype)) {
	   resulttype = datatype_tuple_concat(ctx->pql,
					      unchosentype, unnestedtype);
	}
	else {
	   /* if unnesting a scalar, preserve its column name */
	   resulttype = datatype_tuple_append(ctx->pql,
					      unchosentype, unnestedtype,
					      te->unnest.col);
	}

	if (datatype_isset(te->unnest.sub->datatype)) {
	   resulttype = datatype_set(ctx->pql, resulttype);
	}
	else if (datatype_issequence(te->unnest.sub->datatype)) {
	   resulttype = datatype_sequence(ctx->pql, resulttype);
	}

	if (!datatype_eq_withnames(resulttype, te->datatype)) {
	   tc_say(ctx, "Unnest result type is wrong (computed %s, found %s)",
		  datatype_getname(resulttype),
		  datatype_getname(te->datatype));
	   tc_fail(ctx);
	}
     }
     break;

    case TCE_DISTINGUISH:
     tcexpr_tc(ctx, te->distinguish.sub);
     {
	struct datatype *subtype, *resulttype;
	bool isset;
	
	subtype = te->distinguish.sub->datatype;
	isset = datatype_isset(subtype);
	if (isset) {
	   subtype = datatype_set_member(subtype);
	}

	resulttype = datatype_tuple_append(ctx->pql,
					   subtype,
					   datatype_distinguisher(ctx->pql),
					   te->distinguish.newcol);
	if (isset) {
	   resulttype = datatype_set(ctx->pql, resulttype);
	}
	if (!datatype_eq_withnames(te->datatype, resulttype)) {
	   tc_say(ctx, "Result of distinguish has wrong type");
	   tc_fail(ctx);
	}
     }
     break;

    case TCE_ADJOIN:
     tcexpr_tc(ctx, te->adjoin.left);
     tcexpr_tc(ctx, te->adjoin.expr);
     {
	struct datatype *membertype, *newcoltype, *resulttype;
	bool isset;

	membertype = te->adjoin.left->datatype;
	isset = datatype_isset(membertype);
	if (isset) {
	   membertype = datatype_set_member(membertype);
	}

	if (te->adjoin.var != NULL) {
	   if (!datatype_eq_withnames(te->adjoin.var->datatype, membertype)) {
	      tc_say(ctx, "Adjoin expects %s; found %s",
		     datatype_getname(te->adjoin.var->datatype),
		     datatype_getname(membertype));
	      tc_fail(ctx);
	   }
	}
	newcoltype = te->adjoin.expr->datatype;

	// XXX should this check be part of datatype_tuple_append?
	if (datatype_isabstop(membertype)) {
	   resulttype = datatype_abstop(ctx->pql);
	}
	else {
	   resulttype = datatype_tuple_append(ctx->pql, membertype,
					      newcoltype, te->adjoin.newcol);
	}
	if (isset) {
	   resulttype = datatype_set(ctx->pql, resulttype);
	}
	if (!datatype_eq_withnames(te->datatype, resulttype)) {
	   tc_say(ctx, "Result of adjoin has wrong type "
		  "(computed %s, found %s)", datatype_getname(resulttype),
		  datatype_getname(te->datatype));
	   tc_fail(ctx);
	}
     }
     break;

    case TCE_STEP:
     tcexpr_tc(ctx, te->step.sub);
     if (te->step.predexpr != NULL) {
	tcexpr_tc(ctx, te->step.predexpr);
     }
     {
	struct datatype *left, *right, *member, *set;

	tc_check_column_exists(ctx, "Step-join expression",
			       te->step.sub->datatype,te->step.subcolumn);

	left = te->step.sub->datatype;
	if (datatype_isset(left)) {
	   left = datatype_set_member(left);
	}
	right = datatype_tuple_triple(ctx->pql,
				      datatype_dbobj(ctx->pql),
				      datatype_dbedge(ctx->pql),
				      datatype_dbobj(ctx->pql),
				      te->step.leftobjcolumn,
				      te->step.edgecolumn,
				      te->step.rightobjcolumn);
	member = datatype_tuple_concat(ctx->pql, left, right);
	set = datatype_set(ctx->pql, member);
	if (!datatype_eq_withnames(set, te->datatype)) {
	   tc_say(ctx, "Result of step has wrong type");
	   tc_fail(ctx);
	}

	if (te->step.predvar != NULL) {
	   tc_check_filtervar(ctx, member, te->step.predvar->datatype);
	}
	if (te->step.predexpr != NULL) {
	   tc_check_filterexpr(ctx, te->step.predexpr->datatype);
	}
     }
     break;

    case TCE_REPEAT:
     {
	struct datatype *chosentype, *unchosentype, *starttype, *looptype;
	struct datatype *outputtype, *outputpathtype;
	struct columnarray *cols;

	tc_check_column_exists(ctx, "Repeat expression",
			       te->repeat.sub->datatype,
			       te->repeat.subendcolumn);
	tc_check_column_exists(ctx, "Repeat expression",
			       te->repeat.loopvar->datatype,
			       te->repeat.bodystartcolumn);
	tc_check_column_exists(ctx, "Repeat expression",
			       te->repeat.body->datatype,
			       te->repeat.bodyendcolumn);
	if (te->repeat.bodypathcolumn != NULL) {
	   tc_check_column_exists(ctx, "Repeat expression",
				  te->repeat.body->datatype,
				  te->repeat.bodypathcolumn);
	}

	/* first, check the start expression */
	tcexpr_tc(ctx, te->repeat.sub);
	starttype = te->repeat.sub->datatype;
	if (!datatype_isset(starttype)) {
	   tc_say(ctx, "Repeat start expression is not a set "
		  "(found %s)",
		  datatype_getname(starttype));
	   tc_fail(ctx);
	}
	else {
	   starttype = datatype_set_member(starttype);
	}

	/* check that the start column is valid */
	cols = columnarray_create(ctx->pql);
	columnarray_add(ctx->pql, cols, te->repeat.subendcolumn, NULL);
	tc_check_columnchoices(ctx, "Repeat expression",
			       te->repeat.sub->datatype,
			       cols,
			       &chosentype, NULL);
	looptype = datatype_rename(chosentype, te->repeat.subendcolumn,
				   te->repeat.bodystartcolumn);
	PQLASSERT(looptype != NULL);
	PQLASSERT(datatype_isset(looptype));
	PQLASSERT(datatype_istuple(datatype_set_member(looptype)));
	PQLASSERT(datatype_arity(datatype_set_member(looptype)) == 1);
	if (!datatype_eq_withnames(looptype, te->repeat.loopvar->datatype)) {
	   tc_say(ctx, "Repeat expression iteration var has wrong type "
		  "(computed %s, found %s)",
		  datatype_getname(chosentype),
		  datatype_getname(te->repeat.loopvar->datatype));
	   tc_fail(ctx);
	}

	tcexpr_tc(ctx, te->repeat.body);

	// drop the pointers as we didn't take references to them
	columnarray_setsize(ctx->pql, cols, 0);
	columnarray_destroy(ctx->pql, cols);
	cols = columnarray_create(ctx->pql);

	columnarray_add(ctx->pql, cols, te->repeat.bodystartcolumn, NULL);
	if (te->repeat.bodypathcolumn != NULL) {
	   columnarray_add(ctx->pql, cols, te->repeat.bodypathcolumn, NULL);
	}
	columnarray_add(ctx->pql, cols, te->repeat.bodyendcolumn, NULL);

	tc_check_columnchoices(ctx, "Repeat expression",
			       te->repeat.body->datatype,
			       cols,
			       &chosentype,
			       &unchosentype);

	PQLASSERT(datatype_isset(chosentype));
	PQLASSERT(datatype_isset(unchosentype));

	chosentype = datatype_set_member(chosentype);
	unchosentype = datatype_set_member(unchosentype);

	if (te->repeat.bodypathcolumn != NULL) {
	   PQLASSERT(datatype_arity(chosentype) == 3);
	   outputpathtype = datatype_getnth(chosentype, 1);
	   /* if already a sequence, leave it alone, otherwise make it one */
	   if (!datatype_issequence(outputpathtype)) {
	      outputpathtype = datatype_sequence(ctx->pql, outputpathtype);
	   }
	   outputtype = datatype_getnth(chosentype, 2);
	}
	else {
	   PQLASSERT(datatype_arity(chosentype) == 2);
	   outputtype = datatype_getnth(chosentype, 1);
	   outputpathtype = NULL;
	}

	/* if we'd get sequence(unit), skip it */
	if (datatype_arity(unchosentype) > 0) {
	   PQLASSERT(te->repeat.bodybodycolumn != NULL);
	   unchosentype = datatype_sequence(ctx->pql, unchosentype);
	   unchosentype = datatype_tuple_monople(ctx->pql, unchosentype,
						 te->repeat.bodybodycolumn);
	}
	else {
	   PQLASSERT(te->repeat.bodybodycolumn == NULL);
	}

	if (outputpathtype != NULL) {
	   unchosentype = datatype_tuple_append(ctx->pql, unchosentype,
						outputpathtype,
						te->repeat.repeatpathcolumn);
	}

	unchosentype = datatype_tuple_append(ctx->pql, unchosentype,
					     outputtype,
					     te->repeat.repeatendcolumn);

	unchosentype = datatype_tuple_concat(ctx->pql, starttype,
					     unchosentype);

	if (datatype_isset(te->repeat.body->datatype)) {
	   unchosentype = datatype_set(ctx->pql, unchosentype);
	}

	if (!datatype_eq_withnames(unchosentype, te->datatype)) {
	   tc_say(ctx, "Repeat expression has wrong type "
		  "(computed %s, found %s)",
		  datatype_getname(unchosentype),
		  datatype_getname(te->datatype));
	   tc_fail(ctx);
	}

	// drop the pointers as we didn't take references to them
	columnarray_setsize(ctx->pql, cols, 0);
	columnarray_destroy(ctx->pql, cols);
     }
     break;

    case TCE_SCAN:
     {
	struct datatype *t;

	t = datatype_tuple_triple(ctx->pql,
				  datatype_dbobj(ctx->pql),
				  datatype_dbedge(ctx->pql),
				  datatype_dbobj(ctx->pql),
				  te->scan.leftobjcolumn,
				  te->scan.edgecolumn,
				  te->scan.rightobjcolumn);
	t = datatype_set(ctx->pql, t);
	if (!datatype_eq_withnames(t, te->datatype)) {
	   tc_say(ctx, "Result of scan has wrong type");
	   tc_fail(ctx);
	}
     }
     break;

    case TCE_BOP:
     tcexpr_tc(ctx, te->bop.left);
     tcexpr_tc(ctx, te->bop.right);
     {
	struct datatypearray subtypes;

	datatypearray_init(&subtypes);

	datatypearray_add(ctx->pql, &subtypes, te->bop.left->datatype, NULL);
	datatypearray_add(ctx->pql, &subtypes, te->bop.right->datatype, NULL);

	function_tc(ctx, te->bop.op, &subtypes, te->datatype);

	datatypearray_setsize(ctx->pql, &subtypes, 0);
	datatypearray_cleanup(ctx->pql, &subtypes);
     }
     break;

    case TCE_UOP:
     tcexpr_tc(ctx, te->uop.sub);
     {
	struct datatypearray subtypes;

	datatypearray_init(&subtypes);

	datatypearray_add(ctx->pql, &subtypes, te->uop.sub->datatype, NULL);

	function_tc(ctx, te->uop.op, &subtypes, te->datatype);

	datatypearray_setsize(ctx->pql, &subtypes, 0);
	datatypearray_cleanup(ctx->pql, &subtypes);
     }
     break;


    case TCE_FUNC:
     {
	unsigned i, num;
	struct tcexpr *subexpr;
	struct datatypearray subtypes;

	datatypearray_init(&subtypes);

	num = tcexprarray_num(&te->func.args);
	datatypearray_setsize(ctx->pql, &subtypes, num);

	for (i=0; i<num; i++) {
	   subexpr = tcexprarray_get(&te->func.args, i);
	   tcexpr_tc(ctx, subexpr);
	   datatypearray_set(&subtypes, i, subexpr->datatype);
	}

	function_tc(ctx, te->func.op, &subtypes, te->datatype);

	datatypearray_setsize(ctx->pql, &subtypes, 0);
	datatypearray_cleanup(ctx->pql, &subtypes);
     }
     break;

    case TCE_MAP:
     tcexpr_tc(ctx, te->map.set);
     tcexpr_tc(ctx, te->map.result);
     {
	struct datatype *settype, *membertype;
	struct datatype *vartype, *bodytype;
	struct datatype *resulttype;

	settype = te->map.set->datatype;
	vartype = te->map.var->datatype;
	bodytype = te->map.result->datatype;
	resulttype = datatype_set(ctx->pql, bodytype);

	if (!datatype_isset(settype)) {
	   tc_say(ctx, "Map operator or quantifier applied to non-set type %s",
		  datatype_getname(te->map.set->datatype));
	   tc_fail(ctx);
	   membertype = settype;
	}
	else {
	   membertype = datatype_set_member(settype);
	}

	if (!datatype_eq_withnames(vartype, membertype)) {
	   /* XXX shouldn't open-code .K%u */
	   tc_say(ctx, "Quantifier or map variable .K%u has wrong type %s "
		  "(should be %s)",
		  te->map.var->id,
		  datatype_getname(vartype),
		  datatype_getname(membertype));
	   tc_fail(ctx);
	}
	if (!datatype_eq_withnames(te->datatype, resulttype)) {
	   tc_say(ctx, "Quantifier or map expression produces type %s, but "
		  "body yields type %s",
		  datatype_getname(te->datatype),
		  datatype_getname(resulttype)),
	   tc_fail(ctx);
	}
     }
     break;

    case TCE_LET:
     tcexpr_tc(ctx, te->let.value);
     tcexpr_tc(ctx, te->let.body);
     {
	struct datatype *vartype, *valtype, *bodytype;

	vartype = te->let.var->datatype;
	valtype = te->let.value->datatype;
	bodytype = te->let.body->datatype;

	if (!datatype_eq_withnames(vartype, valtype)) {
	   /* XXX shouldn't open-code .K%u */
	   tc_say(ctx, "Let-bound variable .K%u has wrong type %s "
		  "(should be %s)",
		  te->let.var->id,
		  datatype_getname(vartype),
		  datatype_getname(valtype));
	   tc_fail(ctx);
	}
	if (!datatype_eq_withnames(te->datatype, bodytype)) {
	   tc_say(ctx, "Let expression produces type %s, but body has type %s",
		  datatype_getname(te->datatype), 
		  datatype_getname(bodytype));
	   tc_fail(ctx);
	}
     }
     break;

    case TCE_READVAR:
     if (!datatype_eq_withnames(te->datatype, te->readvar->datatype)) {
	tc_say(ctx, "Read-variable expression has wrong type");
	tc_fail(ctx);
     }
     break;

    case TCE_READGLOBAL:
     if (!datatype_eq_withnames(te->datatype, datatype_set(ctx->pql,
						 datatype_dbobj(ctx->pql)))) {
	tc_say(ctx, "Read-global-variable expression has wrong type");
	tc_fail(ctx);
     }
     break;

    case TCE_CREATEPATHELEMENT:
     tcexpr_tc(ctx, te->createpathelement);
     {
	struct datatype *t0, *t1, *t2;

	if (datatype_istuple(te->createpathelement->datatype)) {
	   t0 = datatype_getnth(te->createpathelement->datatype, 0);
	   t1 = datatype_getnth(te->createpathelement->datatype, 1);
	   t2 = datatype_getnth(te->createpathelement->datatype, 2);
	}
	else {
	   t0 = t1 = t2 = NULL;
	}
	if (!datatype_eq_withnames(t0, datatype_dbobj(ctx->pql)) ||
	    !datatype_eq_withnames(t1, datatype_dbedge(ctx->pql)) ||
	    !datatype_eq_withnames(t2, datatype_dbobj(ctx->pql))) {
	   tc_say(ctx, "Argument to pathelement constructor has wrong type");
	   tc_fail(ctx);
	}
	if (!datatype_eq_withnames(te->datatype, datatype_pathelement(ctx->pql))) {
	   tc_say(ctx, "Result of pathelement constructor has wrong type");
	   tc_fail(ctx);
	}
     }
     break;

    case TCE_SPLATTER:
     tcexpr_tc(ctx, te->splatter.value);
     tcexpr_tc(ctx, te->splatter.name);
     if (!datatype_eq_withnames(te->splatter.value->datatype, te->datatype)) {
	tc_say(ctx, "Result of splatter has wrong type");
	tc_fail(ctx);
     }
     if (datatype_match_specialize(ctx->pql, te->splatter.name->datatype,
				   datatype_absdbedge(ctx->pql)) == NULL) {
	tc_say(ctx, "Edge name in splatter has wrong type");
	tc_fail(ctx);
     }
     break;

    case TCE_TUPLE:
     {
	unsigned i, num;
	struct tcexpr *subexpr;
	struct datatype *subtype;
	struct column *subcol;

	num = tcexprarray_num(&te->tuple.exprs);
	PQLASSERT(num == columnarray_num(te->tuple.columns));
	PQLASSERT(num == datatype_arity(te->datatype));

	for (i=0; i<num; i++) {
	   subexpr = tcexprarray_get(&te->tuple.exprs, i);
	   subtype = datatype_getnth(te->datatype, i);
	   tcexpr_tc(ctx, subexpr);
	   if (!datatype_eq_withnames(subexpr->datatype, subtype)) {
	      tc_say(ctx, "Tuple element %u has wrong type "
		     "(computed %s, have %s)", i,
		     datatype_getname(subexpr->datatype),
		     datatype_getname(subtype));
	      tc_fail(ctx);
	   }
	   subcol = datatype_tuple_getcolumn(te->datatype, i);
	   if (subcol != columnarray_get(te->tuple.columns, i)) {
	      tc_say(ctx, "Tuple element %u has wrong column %s "
		     "(should be %s)", i,
		     column_getname(subcol),
		     column_getname(columnarray_get(te->tuple.columns, i)));
	      tc_fail(ctx);
	   }
	}
     }
     break;

    case TCE_VALUE:
     /*
      * Allow constant nil to be treated as having any type.
      * (Maybe this should be a matching test instead of an equality test?)
      */
     if (!datatype_eq_withnames(te->datatype, pqlvalue_datatype(te->value)) &&
	 !pqlvalue_isnil(te->value)) {
	tc_say(ctx, "Constant expression has wrong type");
	tc_fail(ctx);
     }
     break;
   }
}

static void tcstmt_tc(struct tc *ctx, struct tcstmt *ts) {
   switch (ts->type) {
    case TCS_EXPR:
     tcexpr_tc(ctx, ts->expr);
     break;
    case TCS_BEGIN:
    case TCS_ABORT:
    case TCS_COMMIT:
     break;
   }
}

////////////////////////////////////////////////////////////
// entry point

int typecheck(struct pqlcontext *pql, struct tcstmt *ts) {
   struct tc ctx;
   int ret;

   tc_init(&ctx, pql);
   
   tcstmt_tc(&ctx, ts);

   ret = ctx.failed ? -1 : 0;
   tc_cleanup(&ctx);
   return ret;
}
