/*
 * Copyright 2009
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Figure out types (and column names) for tuple calculus.
 */

#include "datatype.h"
#include "pqlvalue.h"
#include "tcalc.h"
#include "passes.h"

struct typeinf {
   struct pqlcontext *pql;
};

////////////////////////////////////////////////////////////
// context management

static void typeinf_init(struct typeinf *ctx, struct pqlcontext *pql) {
   ctx->pql = pql;
}

static void typeinf_cleanup(struct typeinf *ctx) {
   (void)ctx;
}

////////////////////////////////////////////////////////////
// tools

/*
 * Given an expression TE, which has type and column names data, get
 * the type of column COL. The type of TE must be a tuple or tuple set.
 */
static struct datatype *get_member_type(struct typeinf *ctx, struct tcexpr *te,
					struct column *col) {
   struct datatype *t;

   (void)ctx;

   t = te->datatype;

   /* top must yield top regardless of indexing/arity */
   if (datatype_isabstop(t)) {
      return t;
   }

   PQLASSERT(datatype_istuple(t) ||
	     (datatype_isset(t) && 
	      datatype_istuple(datatype_set_member(t))) ||
	     (datatype_issequence(t) && 
	      datatype_istuple(datatype_sequence_member(t)))
	     );

   if (datatype_isset(t)) {
      t = datatype_set_member(t);
   }
   else if (datatype_issequence(t)) {
      t = datatype_sequence_member(t);
   }

   t = datatype_subprojectscalar(t, col);

   return t;
}

static struct datatype *get_member_types(struct typeinf *ctx,
					 struct tcexpr *te,
					 const struct columnarray *cols) {
   unsigned i, num;
   struct datatypearray types;
   struct columnarray typecols;
   struct datatype *subtype, *ret;
   struct column *col;

   datatypearray_init(&types);
   columnarray_init(&typecols);
   num = columnarray_num(cols);
   for (i=0; i<num; i++) {
      col = columnarray_get(cols, i);
      subtype = get_member_type(ctx, te, col);
      if (subtype != NULL) {
	 datatypearray_add(ctx->pql, &types, subtype, NULL);
	 columnarray_add(ctx->pql, &typecols, col, NULL);
      }
   }
   ret = datatype_tuple_specific(ctx->pql, &types, &typecols);

   datatypearray_setsize(ctx->pql, &types, 0);
   datatypearray_cleanup(ctx->pql, &types);
   columnarray_setsize(ctx->pql, &typecols, 0);
   columnarray_cleanup(ctx->pql, &typecols);

   return ret;
}

static struct datatype *get_member_types_except(struct typeinf *ctx,
						struct tcexpr *te,
						struct columnarray *exclude) {
   struct datatype *t;

   (void)ctx;

   t = te->datatype;

   if (datatype_isset(t)) {
      t = datatype_set_member(t);
   }
   else if (datatype_issequence(t)) {
      t = datatype_sequence_member(t);
   }

   t = datatype_substrip(t, exclude);

   return t;
}

static struct datatype *get_member_types_except_one(struct typeinf *ctx,
						    struct tcexpr *te,
						    struct column *exclude) {
   struct columnarray *cs;
   struct datatype *ret;

   cs = columnarray_create(ctx->pql);
   columnarray_add(ctx->pql, cs, exclude, NULL);

   ret = get_member_types_except(ctx, te, cs);

   // drop the column pointer as we didn't take a reference to it
   columnarray_setsize(ctx->pql, cs, 0);
   columnarray_destroy(ctx->pql, cs);

   return ret;
}

/*
 * Transform ARR into ~ARR, using T for the reference list of columns.
 * Only toplevel columns are handled. Refcounts columns.
 *
 * XXX there's a duplicate copy of this in baseopt.c
 */
static void columnarray_take_complement(struct typeinf *ctx,
					struct columnarray *arr,
					struct datatype *t) {
   struct column *col;
   bool *keep;
   unsigned i, arity, j, num;

   if (datatype_isset(t)) {
      t = datatype_set_member(t);
   }
   else if (datatype_issequence(t)) {
      t = datatype_sequence_member(t);
   }
   PQLASSERT(datatype_istuple(t));

   arity = datatype_arity(t);
   num = columnarray_num(arr);
   PQLASSERT(num <= arity);

   keep = domalloc(ctx->pql, arity * sizeof(*keep));
   for (i=0; i<arity; i++) {
      keep[i] = true;
   }

   for (i=0; i<arity; i++) {
      col = datatype_tuple_getcolumn(t, i);
      for (j=0; j<num; j++) {
	 if (col == columnarray_get(arr, j)) {
	    keep[i] = false;
	 }
      }
   }

   for (j=0; j<num; j++) {
      column_decref(ctx->pql, columnarray_get(arr, j));
   }
   columnarray_setsize(ctx->pql, arr, arity - num);

   for (i=j=0; i<arity; i++) {
      if (keep[i]) {
	 col = datatype_tuple_getcolumn(t, i);
	 column_incref(col);
	 columnarray_set(arr, j++, col);
      }
   }
   PQLASSERT(j == arity - num);

   dofree(ctx->pql, keep, arity * sizeof(*keep));
}

////////////////////////////////////////////////////////////
// function types

#if 0 /* unused */
/*
 * Get the proper number of arguments for a function.
 */
static unsigned get_func_nargs(enum functions f) {
   switch (f) {
    case F_UNION:
    case F_INTERSECT:
    case F_SETDIFFERENCE:
    case F_IN:
     return 2;

    case F_NONEMPTY:
    case F_MAKESET:
    case F_GETELEMENT:
     return 1;

    case F_COUNT:
    case F_SUM:
    case F_AVG:
    case F_MIN:
    case F_MAX:
    case F_ALLTRUE:
    case F_ANYTRUE:
     return 1;

    case F_AND:
    case F_OR:
     return 2;

    case F_NOT:
     return 1;

    case F_NEW:
     /* Multiple args in concrete syntax become one tuple. */
     return 1;

    case F_CTIME:
     return 0;

    case F_EQ:
    case F_NOTEQ:
    case F_LT:
    case F_GT:
    case F_LTEQ:
    case F_GTEQ:
    case F_LIKE:
    case F_GLOB:
    case F_GREP:
    case F_SOUNDEX:
     return 2;

    case F_TOSTRING:
     return 1;

    case F_CONCAT:
     return 1;

    case F_CHOOSE:
     return 2;

    case F_ADD:
    case F_SUB:
    case F_MUL:
    case F_DIV:
    case F_MOD:
     return 2;

    case F_NEG:
    case F_ABS:
     return 1;
   }

   PQLASSERT(0);
   return 0;
}
#endif /* 0: unused */

/*
 * Get the most general required type for one argument of a function.
 *
 * For an out of range argument number this should return whatever
 * seems convenient and not crash or assert. Type errors, including
 * "wrong number of arguments", are issued downstream.
 */
static struct datatype *get_func_argtype(struct typeinf *ctx, enum functions f,
					 unsigned argnum) {

   switch (f) {
    /* set(T1) x set(T2) */
    case F_UNION:
    case F_INTERSECT:
    case F_EXCEPT:
    case F_UNIONALL:
    case F_INTERSECTALL:
    case F_EXCEPTALL:
     return datatype_set(ctx->pql, datatype_abstop(ctx->pql));

    /* T|set(T) x set(T) (these two cases are not distinguished upstream) */
    case F_SUBSETOF:
    case F_ELEMENTOF:
     if (argnum == 0) {
	return datatype_abstop(ctx->pql);
     }
     return datatype_set(ctx->pql, datatype_abstop(ctx->pql));
     break;

    /* set(T) */
    case F_NONEMPTY:
    case F_GETELEMENT:
    case F_COUNT:
     return datatype_set(ctx->pql, datatype_abstop(ctx->pql));

    /* set(number) */
    case F_SUM:
    case F_AVG:
    case F_MIN:
    case F_MAX:
     return datatype_set(ctx->pql, datatype_absnumber(ctx->pql));

    /* set(bool) */
    case F_ALLTRUE:
    case F_ANYTRUE:
     return datatype_set(ctx->pql, datatype_bool(ctx->pql));

    /* T x T */
    case F_EQ:
    case F_NOTEQ:
    case F_CHOOSE:
     return datatype_abstop(ctx->pql);

    /* T */
    case F_MAKESET:
    case F_MAYBEMAKESET:
    case F_NEW:
    case F_TOSTRING:
    case F_IDENTITY:
     return datatype_abstop(ctx->pql);

    /* seq(T) x seq(T) | string x string */
    case F_CONCAT:
     return datatype_abstop(ctx->pql);

    /* string x string */
    case F_LIKE:
    case F_GLOB:
    case F_GREP:
    case F_SOUNDEX:
     return datatype_string(ctx->pql);

    /* string x number */
    case F_STRFTIME:
     if (argnum == 0) {
        return datatype_string(ctx->pql);
     }
     return datatype_absnumber(ctx->pql);

    /* number x number */
    case F_LT:
    case F_GT:
    case F_LTEQ:
    case F_GTEQ:
    case F_ADD:
    case F_SUB:
    case F_MUL:
    case F_DIV:
    case F_MOD:
     return datatype_absnumber(ctx->pql);

    /* number */
    case F_NEG:
    case F_ABS:
     return datatype_absnumber(ctx->pql);

    /* bool x bool */
    case F_AND:
    case F_OR:
     return datatype_bool(ctx->pql);
     break;

    /* bool */
    case F_NOT:
     return datatype_bool(ctx->pql);

    /* pathelement */
    case F_LEFT:
    case F_EDGE:
    case F_RIGHT:
     return datatype_pathelement(ctx->pql);

    /* void */
    case F_NOW:
     return datatype_absbottom(ctx->pql);
   }

   PQLASSERT(0);
   return datatype_absbottom(ctx->pql);
}

static struct datatype *get_func_result(struct typeinf *ctx, enum functions f,
					struct datatypearray *subtypes) {
   struct datatype *left, *right, *sub;

   /* note: this does something reasonable when things are mismatched */
   switch (datatypearray_num(subtypes)) {
    case 2:
     left = datatypearray_get(subtypes, 0);
     right = datatypearray_get(subtypes, 1);
     sub = left;
     break;
    case 1:
     left = right = sub = datatypearray_get(subtypes, 0);
     break;
    default:
     left = right = sub = datatype_absbottom(ctx->pql);
     break;
   }

   switch (f) {
    case F_CONCAT:
     /* XXX should this really be a special case? */
     if (datatype_ispathelement(left)) {
	left = datatype_sequence(ctx->pql, left);
     }
     if (datatype_ispathelement(right)) {
	right = datatype_sequence(ctx->pql, right);
     }
     /* FALLTHROUGH */
    case F_UNION:
    case F_INTERSECT:
    case F_EXCEPT:
    case F_UNIONALL:
    case F_INTERSECTALL:
    case F_EXCEPTALL:
    case F_CHOOSE:
     return datatype_match_generalize_any(ctx->pql, left, right);

    case F_MAKESET:
     return datatype_set(ctx->pql, sub);
    case F_MAYBEMAKESET:
     if (datatype_isset(sub) || datatype_issequence(sub)) {
	return sub;
     }
     return datatype_set(ctx->pql, sub);

    case F_GETELEMENT:
    case F_SUM:
    case F_MIN:
    case F_MAX:
     if (datatype_isset(sub)) {
	return datatype_set_member(sub);
     }
     break;

    case F_COUNT:
    case F_NOW: /* XXX now should return a time value */
     return datatype_int(ctx->pql);

    case F_AVG:
     if (datatype_isset(sub) &&
	 datatype_isanynumber(datatype_set_member(sub))) {
	return datatype_double(ctx->pql);
     }
     break;

    case F_SUBSETOF:
    case F_ELEMENTOF:
    case F_NONEMPTY:
    case F_ALLTRUE:
    case F_ANYTRUE:
    case F_AND:
    case F_OR:
    case F_NOT:
    case F_EQ:
    case F_NOTEQ:
    case F_LT:
    case F_GT:
    case F_LTEQ:
    case F_GTEQ:
    case F_LIKE:
    case F_GLOB:
    case F_GREP:
    case F_SOUNDEX:
     return datatype_bool(ctx->pql);

    case F_NEW:
     if (datatype_isset(sub) ||
	 datatype_issequence(sub) ||
	 datatype_ispathelement(sub) ||
	 datatype_istuple(sub)) {
	return datatype_struct(ctx->pql);
     }
     return sub;

    case F_LEFT:
    case F_RIGHT:
     return datatype_absdbobj(ctx->pql);
     
    case F_EDGE:
     return datatype_absdbedge(ctx->pql);

    case F_STRFTIME:
    case F_TOSTRING:
     return datatype_string(ctx->pql);

    case F_ADD:
    case F_SUB:
    case F_MUL:
    case F_DIV:
    case F_MOD:
     /* provided both are numbers, use the most general number */
     return datatype_match_generalize_any(ctx->pql, left, right);

    case F_NEG:
    case F_ABS:
    case F_IDENTITY:
     return sub;
   }

   return datatype_absbottom(ctx->pql);
}

////////////////////////////////////////////////////////////
// recursive traversal

/*
 * Determine a type for TE, which is used in a context such that
 * ENVIRONMENT is the expected type. The type we choose might be
 * ENVIRONMENT, or a subtype thereof, or something completely
 * different (in which case typecheck will later flag an error...)
 *
 * Also, determine TE's own column name (used if it's amalgamated into
 * a tuple) and, if it's itself a tuple or tuple set, the column names
 * it contains.
 *
 * This module does not issue errors, warnings, or other diagnostics;
 * when the types are wrong it should plug something in that will
 * cause typecheck to issue an error later, and choose that something
 * to hopefully minimize the number of type errors arising from any
 * one mistake.
 */

static void tcexpr_typeinf(struct typeinf *ctx, struct tcexpr *te,
			   struct datatype *environment);


/*
 * Shared code for functions and operators.
 */
static void func_typeinf(struct typeinf *ctx,
			 enum functions op, struct tcexprarray *args,
			 struct datatype **type_ret) {
   struct datatype *subtype, *resulttype;
   struct tcexpr *subexpr;
   struct datatypearray types;
   unsigned i, num;

   datatypearray_init(&types);

   num = tcexprarray_num(args);

   datatypearray_setsize(ctx->pql, &types, num);

   for (i=0; i<num; i++) {
      subtype = get_func_argtype(ctx, op, i);
      subexpr = tcexprarray_get(args, i);
      tcexpr_typeinf(ctx, subexpr, subtype);
      /* use the type actually found */
      datatypearray_set(&types, i, subexpr->datatype);
   }

   /* note that constructing a tuple type out of types[] would be wrong */

   resulttype = get_func_result(ctx, op, &types);
   if (resulttype == NULL) {
      resulttype = datatype_absbottom(ctx->pql);
   }
   *type_ret = resulttype;

   datatypearray_setsize(ctx->pql, &types, 0);
   datatypearray_cleanup(ctx->pql, &types);
}

/*
 * Wrapper for operators (with fixed-size arglists)
 */
static void op_typeinf(struct typeinf *ctx, enum functions op,
		       struct tcexpr *a0, struct tcexpr *a1,
		       struct datatype **type_ret) {
   struct tcexprarray args;

   tcexprarray_init(&args);
   tcexprarray_add(ctx->pql, &args, a0, NULL);
   if (a1 != NULL) {
      tcexprarray_add(ctx->pql, &args, a1, NULL);
   }
   func_typeinf(ctx, op, &args, type_ret);
   tcexprarray_setsize(ctx->pql, &args, 0);
   tcexprarray_cleanup(ctx->pql, &args);
}

/*
 * General case
 */
static void tcexpr_typeinf(struct typeinf *ctx, struct tcexpr *te,
			   struct datatype *environment) {
   struct datatype *subtype, *lefttype, *righttype;
   struct column *subcol;
   unsigned i, num;

   switch (te->type) {
    case TCE_FILTER:
     /*
      * Filter: we are the same type as the subexpression, which
      * should be set(T), and the test expression should be T -> bool.
      */
     tcexpr_typeinf(ctx, te->filter.sub, environment);
     te->datatype = te->filter.sub->datatype;
     if (datatype_isset(te->datatype)) {
	subtype = datatype_set_member(te->datatype);
     }
     else {
	subtype = te->datatype;
     }
     te->filter.predvar->datatype = subtype;
     tcexpr_typeinf(ctx, te->filter.predexpr, datatype_bool(ctx->pql));
     break;

    case TCE_PROJECTTUPLE:
     /*
      * Project: use the type of the subexpression but keep only the
      * selected columns.
      */
     tcexpr_typeinf(ctx, te->projecttuple.sub, environment);

     if (te->projecttuple.complement_cols) {
	te->projecttuple.complement_cols = false;
	columnarray_take_complement(ctx, te->projecttuple.cols,
				    te->projecttuple.sub->datatype);
     }

     te->datatype = get_member_types(ctx, te->projecttuple.sub,
				     te->projecttuple.cols);

     PQLASSERT(datatype_istuple(te->datatype));

     if (datatype_isset(te->projecttuple.sub->datatype)) {
	te->datatype = datatype_set(ctx->pql, te->datatype);
     }
     else if (datatype_issequence(te->projecttuple.sub->datatype)) {
	te->datatype = datatype_sequence(ctx->pql, te->datatype);
     }
     break;

    case TCE_PROJECTSCALAR:
     /*
      * Project: use the type of the subexpression but keep only the
      * one selected column.
      */

     tcexpr_typeinf(ctx, te->projectscalar.sub, environment);
     te->datatype = get_member_type(ctx, te->projectscalar.sub,
				    te->projectscalar.col);
     PQLASSERT(te->datatype != NULL);
     PQLASSERT(!datatype_istuple(te->datatype));

     if (datatype_isset(te->projectscalar.sub->datatype)) {
	te->datatype = datatype_set(ctx->pql, te->datatype);
     }
     else if (datatype_issequence(te->projectscalar.sub->datatype)) {
	te->datatype = datatype_sequence(ctx->pql, te->datatype);
     }
     break;

    case TCE_RENAME:
     /*
      * Rename: use type of subexpression, adjust matching name.
      */
     tcexpr_typeinf(ctx, te->rename.sub, environment);
     te->datatype = datatype_rename(te->rename.sub->datatype,
				    te->rename.insidecol,
				    te->rename.outsidecol);
     if (te->datatype == NULL) {
	/* no such column; fail later, for now use bottom */
	te->datatype = datatype_absbottom(ctx->pql);
     }
     break;

    case TCE_JOIN:
     /* Check the halves of the cartesian product. */
     subtype = datatype_abstop(ctx->pql);
     subtype = datatype_set(ctx->pql, subtype);
     tcexpr_typeinf(ctx, te->join.left, subtype);
     tcexpr_typeinf(ctx, te->join.right, subtype);
     lefttype = te->join.left->datatype;
     righttype = te->join.right->datatype;

     if (datatype_isabstop(lefttype) || datatype_isabstop(righttype)) {
	te->datatype = datatype_abstop(ctx->pql);
     }
     else if (!datatype_isset(lefttype) && !datatype_isset(righttype)) {
	/* adapt if something failed underneath us */
	te->datatype = datatype_tuple_concat(ctx->pql, lefttype, righttype);
     }
     else {
	// XXX should really only have set x set
	if (datatype_isset(lefttype)) {
	   lefttype = datatype_set_member(lefttype);
	}
	if (datatype_isset(righttype)) {
	   righttype = datatype_set_member(righttype);
	}
	te->datatype = datatype_tuple_concat(ctx->pql, lefttype, righttype);
	te->datatype = datatype_set(ctx->pql, te->datatype);
     }
     
     /* Check the condition. */
     subtype = te->datatype;
     if (datatype_isset(subtype)) {
	subtype = datatype_set_member(subtype);
     }
     if (te->join.predvar != NULL) {
	te->join.predvar->datatype = subtype;
     }
     if (te->join.predexpr != NULL) {
	tcexpr_typeinf(ctx, te->join.predexpr, datatype_bool(ctx->pql));
     }
     break;

    case TCE_ORDER:
     /* Order: columns remain, type changes from set to sequence */
     tcexpr_typeinf(ctx, te->order.sub, environment);
     subtype = te->order.sub->datatype;
     if (datatype_isset(subtype)) {
	subtype = datatype_sequence(ctx->pql, datatype_set_member(subtype));
     }
     else {
	/* wrong, oh well */
     }
     te->datatype = subtype;
     break;

    case TCE_UNIQ:
     /* Uniq: type and columns carry over from subexpression. */
     tcexpr_typeinf(ctx, te->uniq.sub, environment);
     te->datatype = te->uniq.sub->datatype;
     break;

    case TCE_NEST:
     /*
      * Nest: keep the columns not specified, strip the columns specified,
      * wrap them in a set, and add that back as a single column.
      */
     tcexpr_typeinf(ctx, te->nest.sub, environment);

     if (te->nest.complement_cols) {
	te->nest.complement_cols = false;
	columnarray_take_complement(ctx, te->nest.cols,te->nest.sub->datatype);
     }

     te->datatype = get_member_types_except(ctx, te->nest.sub, te->nest.cols);

     subtype = get_member_types(ctx, te->nest.sub, te->nest.cols);
     subtype = datatype_set(ctx->pql, subtype);
     te->datatype = datatype_tuple_append(ctx->pql, te->datatype, subtype,
					  te->nest.newcol);
     if (datatype_isset(te->nest.sub->datatype)) {
	te->datatype = datatype_set(ctx->pql, te->datatype);
     }
     break;
     
    case TCE_UNNEST:
     /* Unnest: keep all but specified column, unpack that one */
     tcexpr_typeinf(ctx, te->unnest.sub, environment);

     te->datatype = get_member_types_except_one(ctx, te->unnest.sub,
						te->unnest.col);

     subtype = get_member_type(ctx, te->unnest.sub, te->unnest.col);
     if (subtype == NULL) {
	/* column not there, use bottom */
	subtype = datatype_absbottom(ctx->pql);
     }
     else {
	PQLASSERT(!datatype_istuple(subtype));
     }

     if (datatype_isset(subtype)) {
	subtype = datatype_set_member(subtype);
     }
     else {
	/* Wrong... oh well. */
     }

     if (!datatype_istuple(subtype)) {
	/* if unnesting a scalar, keep its column name */
	te->datatype = datatype_tuple_append(ctx->pql, te->datatype, subtype,
					     te->unnest.col);
     }
     else {
	te->datatype = datatype_tuple_concat(ctx->pql, te->datatype, subtype);
     }

     if (datatype_isset(te->unnest.sub->datatype)) {
	te->datatype = datatype_set(ctx->pql, te->datatype);
     }
     else if (datatype_issequence(te->unnest.sub->datatype)) {
	te->datatype = datatype_sequence(ctx->pql, te->datatype);
     }
     break;

    case TCE_DISTINGUISH:
     /* Distinguish: add distinguisher column */
     tcexpr_typeinf(ctx, te->distinguish.sub, environment);

     subtype = te->distinguish.sub->datatype;
     if (datatype_isset(subtype)) {
	subtype = datatype_set_member(subtype);
     }
     te->datatype = datatype_tuple_append(ctx->pql,
					  subtype,
					  datatype_distinguisher(ctx->pql),
					  te->distinguish.newcol);
     if (datatype_isset(te->distinguish.sub->datatype)) {
	te->datatype = datatype_set(ctx->pql, te->datatype);
     }
     break;

    case TCE_ADJOIN:
     /* Adjoin: add column based on result type of func */
     tcexpr_typeinf(ctx, te->adjoin.left, environment);
     lefttype = te->adjoin.left->datatype;
     if (datatype_isset(lefttype)) {
	lefttype = datatype_set_member(lefttype);
     }
     if (te->adjoin.var != NULL) {
	te->adjoin.var->datatype = lefttype;
     }
     tcexpr_typeinf(ctx, te->adjoin.expr, datatype_abstop(ctx->pql));
     righttype = te->adjoin.expr->datatype;

     /*
      * Adjoin is supposed to be for adjoining scalar values, not
      * pasting tuples together. However, if we get a tuple back from
      * the function, we don't want to create a type like (a, (b, c))
      * because that's not legal; that should either be (a, b, c) or
      * (a, {b,c}).
      */

     /* adjoining to top must yield top */
     if (datatype_isabstop(lefttype)) {
	te->datatype = lefttype;
     }
     else if (datatype_arity(righttype) != 1) {
	// XXX this is gross and should be removed (it should not happen!)
	te->datatype = datatype_tuple_concat(ctx->pql, lefttype, righttype);
	/* this gets ignored */
	(void)te->adjoin.newcol;
     }
     else {
	te->datatype = datatype_tuple_append(ctx->pql, lefttype, righttype,
					     te->adjoin.newcol);
     }
     if (datatype_isset(te->adjoin.left->datatype)) {
	te->datatype = datatype_set(ctx->pql, te->datatype);
     }
     break;

    case TCE_STEP:
     /* optimization result; not allowed here */
     PQLASSERT(0);
     break;

    case TCE_REPEAT:
     /* Repeat. Urgh. */

     /* start expression */
     tcexpr_typeinf(ctx, te->repeat.sub, environment);

     /*
      * Hack. In some cases te->repeat.sub might be a scalar, e.g.
      *    select ... from ... as X where select ... from X(.foo)+ ....
      * and tuplify can't deal with this, at least for now, because it
      * doesn't know the types of columns. The loop variable always
      * needs to hold a set, because even if we start from a scalar
      * it'll be a set on the second iteration. So insert set() in
      * the start expression if needed.
      *
      * XXX: this should not be needed any more.
      */
     if (!datatype_isset(te->repeat.sub->datatype)) {
	te->repeat.sub = mktcexpr_uop(ctx->pql, F_MAKESET, te->repeat.sub);
	te->repeat.sub->datatype = datatype_set(ctx->pql,
					    te->repeat.sub->uop.sub->datatype);
     }

     /* note: repeat.sub and repeat.body must be a tuple or tuple-set type */

     /* loop variable */
     lefttype = get_member_type(ctx, te->repeat.sub, te->repeat.subendcolumn);
     if (lefttype == NULL) {
	/* column not there, use bottom */
	lefttype = datatype_absbottom(ctx->pql);
     }
     lefttype = datatype_tuple_monople(ctx->pql, lefttype,
				       te->repeat.bodystartcolumn);
     PQLASSERT(datatype_isset(te->repeat.sub->datatype));
     lefttype = datatype_set(ctx->pql, lefttype);
     te->repeat.loopvar->datatype = lefttype;

     /* body expression */
     righttype = datatype_set(ctx->pql, datatype_abstop(ctx->pql));
     tcexpr_typeinf(ctx, te->repeat.body, righttype);
     righttype = te->repeat.body->datatype;

     /*
      * Given
      *    sub :: S
      *    body :: set(B, O, P),
      * the result type is
      *    S x set(seq(B), O, P)
      */

     lefttype = te->repeat.sub->datatype;
     if (datatype_isset(lefttype)) {
	lefttype = datatype_set_member(lefttype);
     }

     {
	struct columnarray *cs;

	cs = columnarray_create(ctx->pql);

	columnarray_add(ctx->pql, cs, te->repeat.bodystartcolumn, NULL);
	if (te->repeat.bodypathcolumn != NULL) {
	   columnarray_add(ctx->pql, cs, te->repeat.bodypathcolumn, NULL);
	}
	columnarray_add(ctx->pql, cs, te->repeat.bodyendcolumn, NULL);

	righttype = get_member_types_except(ctx, te->repeat.body, cs);

	if (datatype_arity(righttype) > 0) {
	   righttype = datatype_sequence(ctx->pql, righttype);
	   te->datatype = datatype_tuple_append(ctx->pql, lefttype, righttype,
						te->repeat.bodybodycolumn);
	}
	else {
	   /* loop doesn't bind anything; avoid adding in sequence(unit) */
	   te->datatype = lefttype;
	   column_decref(ctx->pql, te->repeat.bodybodycolumn);
	   te->repeat.bodybodycolumn = NULL;
	}

	columnarray_setsize(ctx->pql, cs, 0);
	columnarray_destroy(ctx->pql, cs);
     }

     {
	if (te->repeat.bodypathcolumn != NULL) {
	   righttype = get_member_type(ctx, te->repeat.body,
				       te->repeat.bodypathcolumn);
	   if (datatype_ispathelement(righttype)) {
	      righttype = datatype_sequence(ctx->pql, righttype);
	   }
	   te->datatype = datatype_tuple_append(ctx->pql, te->datatype,
						righttype,
						te->repeat.repeatpathcolumn);
	}

	righttype = get_member_type(ctx, te->repeat.body,
				    te->repeat.bodyendcolumn);
	te->datatype = datatype_tuple_append(ctx->pql, te->datatype,
					     righttype,
					     te->repeat.repeatendcolumn);
     }

     if (datatype_isset(te->repeat.body->datatype)) {
	te->datatype = datatype_set(ctx->pql, te->datatype);
     }

     break;

    case TCE_SCAN:
     subtype = datatype_tuple_triple(ctx->pql,
				     datatype_dbobj(ctx->pql),
				     datatype_dbedge(ctx->pql),
				     datatype_dbobj(ctx->pql),
				     te->scan.leftobjcolumn,
				     te->scan.edgecolumn,
				     te->scan.rightobjcolumn);
     te->datatype = datatype_set(ctx->pql, subtype);
     break;

    case TCE_BOP:
     op_typeinf(ctx, te->bop.op, te->bop.left, te->bop.right, &te->datatype);
     /*
      * Special-case hack: the IN operator in the concrete syntax
      * comes in here as F_SUBSETOF. However, IN is supposed to mean
      * either F_SUBSETOF or F_ELEMENTOF depending on the types. So if
      * we find types matching F_ELEMENTOF, change te->bop.op.
      */
     if (te->bop.op == F_SUBSETOF && datatype_isset(te->bop.right->datatype)) {
	struct datatype *lefttype;
	struct datatype *rightmember;

	lefttype = te->bop.left->datatype;
	rightmember = datatype_set_member(te->bop.right->datatype);

	/*
	 * XXX: is this the right test? Maybe we should just check the
	 * number of set nestings.
	 */
	if (datatype_match_specialize(ctx->pql, lefttype, rightmember) !=NULL){
	   te->bop.op = F_ELEMENTOF;
	}
     }
     break;

    case TCE_UOP:
     op_typeinf(ctx, te->uop.op, te->uop.sub, NULL, &te->datatype);
     /*
      * Another special-case hack: if we see F_MAYBEMAKESET, convert
      * it to either F_MAKESET or F_NOP, depending on whether it's had
      * any effect.
      */
     if (te->uop.op == F_MAYBEMAKESET) {
	if (datatype_eq_withnames(te->datatype, te->uop.sub->datatype)) {
	   te->uop.op = F_IDENTITY;
	}
	else {
	   te->uop.op = F_MAKESET;
	}
     }
     break;

    case TCE_FUNC:
     func_typeinf(ctx, te->func.op, &te->func.args, &te->datatype);
     break;

    case TCE_MAP:
     tcexpr_typeinf(ctx, te->map.set, environment);
     subtype = te->map.set->datatype;
     if (datatype_isset(subtype)) {
	subtype = datatype_set_member(subtype);
     }
     else {
	/* wrong, oh well */
     }
     te->map.var->datatype = subtype;
     tcexpr_typeinf(ctx, te->map.result, datatype_abstop(ctx->pql));
     te->datatype = datatype_set(ctx->pql, te->map.result->datatype);
     break;

    case TCE_LET:
     tcexpr_typeinf(ctx, te->let.value, environment);
     te->let.var->datatype = te->let.value->datatype;
     tcexpr_typeinf(ctx, te->let.body, datatype_abstop(ctx->pql));
     te->datatype = te->let.body->datatype;
     break;

    case TCE_READVAR:
     /* variable should have been bound already */
     PQLASSERT(te->readvar->datatype != NULL);
     te->datatype = te->readvar->datatype;
     break;

    case TCE_READGLOBAL:
     te->datatype = datatype_set(ctx->pql, datatype_dbobj(ctx->pql));
     break;

    case TCE_CREATEPATHELEMENT:
     subtype = datatype_tuple_triple(ctx->pql,
				     datatype_dbobj(ctx->pql),
				     datatype_dbedge(ctx->pql),
				     datatype_dbobj(ctx->pql),
				     NULL,
				     NULL,
				     NULL);
     tcexpr_typeinf(ctx, te->createpathelement, subtype);
     te->datatype = datatype_pathelement(ctx->pql);
     break;

    case TCE_SPLATTER:
     tcexpr_typeinf(ctx, te->splatter.value, environment);
     tcexpr_typeinf(ctx, te->splatter.name, datatype_absdbedge(ctx->pql));
     // XXX is splatter going to have to be its own data type?
     te->datatype = te->splatter.value->datatype;
     break;

    case TCE_TUPLE:
     {
	struct datatypearray types;
	struct columnarray typecols;
	struct tcexpr *subexpr;

	num = tcexprarray_num(&te->tuple.exprs);
	PQLASSERT(num == columnarray_num(te->tuple.columns));

	datatypearray_init(&types);
	datatypearray_setsize(ctx->pql, &types, num);

	columnarray_init(&typecols);
	columnarray_setsize(ctx->pql, &typecols, num);

	subtype = datatype_abstop(ctx->pql);
	for (i=0; i<num; i++) {
	   subexpr = tcexprarray_get(&te->tuple.exprs, i);
	   tcexpr_typeinf(ctx, subexpr, subtype);
	   datatypearray_set(&types, i, subexpr->datatype);
	   subcol = columnarray_get(te->tuple.columns, i);
	   columnarray_set(&typecols, i, subcol);
	}

	te->datatype = datatype_tuple_specific(ctx->pql, &types, &typecols);

	datatypearray_setsize(ctx->pql, &types, 0);
	datatypearray_cleanup(ctx->pql, &types);
	columnarray_setsize(ctx->pql, &typecols, 0);
	columnarray_cleanup(ctx->pql, &typecols);
     }
     break;

    case TCE_VALUE:
     te->datatype = pqlvalue_datatype(te->value);
     break;
   }
}

static void tcstmt_typeinf(struct typeinf *ctx, struct tcstmt *ts) {
   switch (ts->type) {
    case TCS_EXPR:
     tcexpr_typeinf(ctx, ts->expr, datatype_abstop(ctx->pql));
     break;
    case TCS_BEGIN:
    case TCS_ABORT:
    case TCS_COMMIT:
     break;
   }
}

////////////////////////////////////////////////////////////
// entry point

void typeinf(struct pqlcontext *pql, struct tcstmt *ts) {
   struct typeinf ctx;

   typeinf_init(&ctx, pql);
   tcstmt_typeinf(&ctx, ts);
   typeinf_cleanup(&ctx);
}
