/*
 * Copyright 2010
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Prune let-bindings whose variable is unused or used only once.
 * Because we carry refcounts on the variables, this is easy to do.
 */

#include "array.h"
#include "tcalc.h"
#include "passes.h"

struct unlet {
   struct pqlcontext *pql;
   struct tcvararray vars;
   struct tcexprarray vals;
};

static void unlet_init(struct unlet *ctx, struct pqlcontext *pql) {
   ctx->pql = pql;
   tcvararray_init(&ctx->vars);
   tcexprarray_init(&ctx->vals);
}

static void unlet_cleanup(struct unlet *ctx) {
   tcvararray_cleanup(ctx->pql, &ctx->vars);
   tcexprarray_cleanup(ctx->pql, &ctx->vals);
}

static void unlet_bindvar(struct unlet *ctx, struct tcvar *var,
			  struct tcexpr *val) {
   PQLASSERT(val != NULL);
   tcvararray_add(ctx->pql, &ctx->vars, var, NULL);
   tcexprarray_add(ctx->pql, &ctx->vals, val, NULL);
}

static void unlet_unbindvar(struct unlet *ctx) {
   unsigned num;

   num = tcvararray_num(&ctx->vars);
   PQLASSERT(num == tcexprarray_num(&ctx->vals));
   PQLASSERT(num > 0);
   tcvararray_setsize(ctx->pql, &ctx->vars, num-1);
   tcexprarray_setsize(ctx->pql, &ctx->vals, num-1);
}

static bool unlet_findvar(struct unlet *ctx, struct tcvar *var,
			  unsigned *ix_ret) {
   unsigned num, i;

   num = tcvararray_num(&ctx->vars);
   for (i=0; i<num; i++) {
      if (var == tcvararray_get(&ctx->vars, i)) {
	 *ix_ret = i;
	 return true;
      }
   }
   return false;
}

static struct tcexpr *unlet_getval(struct unlet *ctx, unsigned ix) {
   struct tcexpr *ret;

   ret = tcexprarray_get(&ctx->vals, ix);
   PQLASSERT(ret != NULL);
   tcexprarray_set(&ctx->vals, ix, NULL);
   return ret;
}

static struct tcexpr *unlet_destruct(struct unlet *ctx, struct tcexpr *te) {
   struct tcexpr *ret;

   ret = te->let.body;
   te->let.body = NULL;
   tcexpr_destroy(ctx->pql, te);
   return ret;
}

////////////////////////////////////////////////////////////
// recursive traversal

static struct tcexpr *tcexpr_unlet(struct unlet *ctx, struct tcexpr *te) {
   unsigned i, num;
   struct tcexpr *subexpr;

   switch (te->type) {
    case TCE_FILTER:
     te->filter.sub = tcexpr_unlet(ctx, te->filter.sub);
     if (te->filter.predvar != NULL && te->filter.predvar->refcount == 1) {
	/* unused (the reference is the binding) */
	tcvar_decref(ctx->pql, te->filter.predvar);
	te->filter.predvar = NULL;
     }
     te->filter.predexpr = tcexpr_unlet(ctx, te->filter.predexpr);
     break;
    case TCE_PROJECTTUPLE:
     te->projecttuple.sub = tcexpr_unlet(ctx, te->projecttuple.sub);
     break;
    case TCE_PROJECTSCALAR:
     te->projectscalar.sub = tcexpr_unlet(ctx, te->projectscalar.sub);
     break;
    case TCE_RENAME:
     te->rename.sub = tcexpr_unlet(ctx, te->rename.sub);
     break;
    case TCE_JOIN:
     te->join.left = tcexpr_unlet(ctx, te->join.left);
     te->join.right = tcexpr_unlet(ctx, te->join.right);
     if (te->join.predvar != NULL && te->join.predvar->refcount == 1) {
	/* unused (the reference is the binding) */
	tcvar_decref(ctx->pql, te->join.predvar);
	te->join.predvar = NULL;
     }
     if (te->join.predexpr != NULL) {
	te->join.predexpr = tcexpr_unlet(ctx, te->join.predexpr);
     }
     break;
    case TCE_ORDER:
     te->order.sub = tcexpr_unlet(ctx, te->order.sub);
     break;
    case TCE_UNIQ:
     te->uniq.sub = tcexpr_unlet(ctx, te->uniq.sub);
     break;
    case TCE_NEST:
     te->nest.sub = tcexpr_unlet(ctx, te->nest.sub);
     break;
    case TCE_UNNEST:
     te->unnest.sub = tcexpr_unlet(ctx, te->unnest.sub);
     break;
    case TCE_DISTINGUISH:
     te->distinguish.sub = tcexpr_unlet(ctx, te->distinguish.sub);
     break;
    case TCE_ADJOIN:
     te->adjoin.left = tcexpr_unlet(ctx, te->adjoin.left);
     if (te->adjoin.var != NULL && te->adjoin.var->refcount == 1) {
	/* unused (the reference is the binding) */
	tcvar_decref(ctx->pql, te->adjoin.var);
	te->adjoin.var = NULL;
     }
     te->adjoin.expr = tcexpr_unlet(ctx, te->adjoin.expr);
     break;
    case TCE_STEP:
     te->join.left = tcexpr_unlet(ctx, te->step.sub);
     if (te->step.predvar != NULL && te->step.predvar->refcount == 1) {
	/* unused (the reference is the binding) */
	tcvar_decref(ctx->pql, te->step.predvar);
	te->step.predvar = NULL;
     }
     if (te->step.predexpr != NULL) {
	te->step.predexpr = tcexpr_unlet(ctx, te->step.predexpr);
     }
     break;
    case TCE_REPEAT:
     te->repeat.sub = tcexpr_unlet(ctx, te->repeat.sub);
     te->repeat.body = tcexpr_unlet(ctx, te->repeat.body);
     break;
    case TCE_SCAN:
     /* nothing */
     break;
    case TCE_BOP:
     te->bop.left = tcexpr_unlet(ctx, te->bop.left);
     te->bop.right = tcexpr_unlet(ctx, te->bop.right);
     break;
    case TCE_UOP:
     te->uop.sub = tcexpr_unlet(ctx, te->uop.sub);
     break;
    case TCE_FUNC:
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	subexpr = tcexprarray_get(&te->func.args, i);
	subexpr = tcexpr_unlet(ctx, subexpr);
	tcexprarray_set(&te->func.args, i, subexpr);
     }
     break;
    case TCE_MAP:
     te->map.set = tcexpr_unlet(ctx, te->map.set);
     te->map.result = tcexpr_unlet(ctx, te->map.result);
     break;
    case TCE_LET:
     if (te->let.var->refcount == 1) {
	/* unused (the one reference is the binding) */
	// skip this
	//te->let.value = tcexpr_unlet(ctx, te->let.value);
	te->let.body = tcexpr_unlet(ctx, te->let.body);
	te = unlet_destruct(ctx, te);
     }
     else if (te->let.var->refcount == 2) {
	/* used once */
	te->let.value = tcexpr_unlet(ctx, te->let.value);
	unlet_bindvar(ctx, te->let.var, te->let.value);
	te->let.value = NULL;
	te->let.body = tcexpr_unlet(ctx, te->let.body);
	unlet_unbindvar(ctx);
	te = unlet_destruct(ctx, te);
     }
     else {
	te->let.value = tcexpr_unlet(ctx, te->let.value);
	te->let.body = tcexpr_unlet(ctx, te->let.body);
     }
     break;
    case TCE_READVAR:
     {
	unsigned index;

	if (unlet_findvar(ctx, te->readvar, &index)) {
	   tcexpr_destroy(ctx->pql, te);
	   te = unlet_getval(ctx, index);
	}
     }
     break;
    case TCE_READGLOBAL:
     break;
    case TCE_CREATEPATHELEMENT:
     te->createpathelement = tcexpr_unlet(ctx, te->createpathelement);
     break;
    case TCE_SPLATTER:
     te->splatter.value = tcexpr_unlet(ctx, te->splatter.value);
     te->splatter.name = tcexpr_unlet(ctx, te->splatter.name);
     break;
    case TCE_TUPLE:
     num = tcexprarray_num(&te->tuple.exprs);
     for (i=0; i<num; i++) {
	subexpr = tcexprarray_get(&te->tuple.exprs, i);
	subexpr = tcexpr_unlet(ctx, subexpr);
	tcexprarray_set(&te->tuple.exprs, i, subexpr);
     }
     break;
    case TCE_VALUE:
     break;
   }
   return te;
}

static struct tcstmt *tcstmt_unlet(struct unlet *ctx, struct tcstmt *ts) {
   switch (ts->type) {
    case TCS_EXPR:
     ts->expr = tcexpr_unlet(ctx, ts->expr);
     break;
    case TCS_BEGIN:
    case TCS_ABORT:
    case TCS_COMMIT:
     break;
   }
   return ts;
}

////////////////////////////////////////////////////////////
// entry point

struct tcstmt *unlet(struct pqlcontext *pql, struct tcstmt *ts) {
   struct unlet ctx;

   unlet_init(&ctx, pql);
   ts = tcstmt_unlet(&ctx, ts);
   unlet_cleanup(&ctx);

   return ts;
}
