/*
 * Copyright 2008, 2009, 2010
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

#include "pql.h"
#include "pqlutil.h"

#include "wdb.h"

#include "primarray.h"
#include "backend.h"

/*
 * Database numbers
 */
#define DBNUM_TEMP  0	/* temporaries */
#define DBNUM_MAGIC 1	/* magic objects */
#define DBNUM_MSTR  2	/* multistring values */
#define DBNUM_MAIN  3	/* pnodes */
#define DBNUM_VERS  4	/* pnodeversions */
#define DBNUM_XARGV 5	/* argv values */
#define DBNUM_XENV  6	/* env values */

#define MAGIC_PROVENANCE 0

static tdb *temp_tdb;
static tdb *prov_mstring_tdb;

////////////////////////////////////////////////////////////
// Berkeley DB iteration tools

#if 0 /* not currently used */

/*
 * Visit duplicate records (same key, different values).
 *
 * Records matching KEY are extracted from database DB. User data CLO
 * is passed through to FUNC, which is passed each value in turn.
 * 
 * Visitation stops immediately if FUNC returns nonzero. The number of
 * records visited is returned.
 */
static int dbvisit_dups(DB *db, DBT *key, void *clo,
			int (*func)(void *clo, DBT *val)) {
   DBC *cursor;
   DBT val;
   int result;
   int num = 0;

   bzero(&val, sizeof(val));

   result = db->cursor(db, NULL, &cursor, 0);
   if (result) {
      db->err(db, result, "cursor");
      exit(1); /* XXX */
   }

   result = cursor->c_get(cursor, key, &val, DB_SET);
   if (result && result != DB_NOTFOUND) {
      db->err(db, result, "cursor->c_get(DB_SET)");
      exit(1); /* XXX */
   }

   while (result == 0) {
      num++;
      if (func(clo, &val)) {
	 break;
      }

      result = cursor->c_get(cursor, key, &val, DB_NEXT_DUP);
      if (result && result != DB_NOTFOUND) {
	 db->err(db, result, "cursor->c_get(DB_NEXT_DUP)");
	 exit(1); /* XXX */
      }
   }

   result = cursor->c_close(cursor);
   if (result) {
      db->err(db, result, "cursor->c_close");
      exit(1); /* XXX */
   }

   return num;
}
#endif /* 0 */

/*
 * Visit records via DB_SETRANGE.
 *
 * Records starting at the one matching KEY and continuing forward are
 * extracted from database DB. User data CLO is passed through to
 * FUNC, which is passed each key and value in turn.
 * 
 * Visitation stops immediately if FUNC returns nonzero. The number of
 * records visited is returned.
 */
static int dbvisit_setrange(DB *db, DBT *key, void *clo,
			    int (*func)(void *clo, DBT *key, DBT *val)) {
   DBC *cursor;
   DBT val;
   int result;
   int num = 0;

   bzero(&val, sizeof(val));

   result = db->cursor(db, NULL, &cursor, 0);
   if (result) {
      db->err(db, result, "cursor");
      exit(1); /* XXX */
   }

   result = cursor->c_get(cursor, key, &val, DB_SET_RANGE);
   if (result && result != DB_NOTFOUND) {
      db->err(db, result, "cursor->c_get(DB_SET_RANGE)");
      exit(1); /* XXX */
   }

   while (result == 0) {
      num++;
      if (func(clo, key, &val)) {
	 break;
      }

      result = cursor->c_get(cursor, key, &val, DB_NEXT);
      if (result && result != DB_NOTFOUND) {
	 db->err(db, result, "cursor->c_get(DB_NEXT)");
	 exit(1); /* XXX */
      }
   }

   result = cursor->c_close(cursor);
   if (result) {
      db->err(db, result, "cursor->c_close");
      exit(1); /* XXX */
   }

   return num;
}

/*
 * Visit the entire contents of the database.
 *
 * Records are extracted from database DB. User data CLO is passed
 * through to FUNC, which is passed each key and value in turn.
 * 
 * Visitation stops immediately if FUNC returns nonzero. The number of
 * records visited is returned.
 */
static int dbvisit_everything(DB *db, void *clo,
			      int (*func)(void *clo, DBT *key, DBT *val)) {
   DBC *cursor;
   DBT key, val;
   int result;
   int num = 0;

   bzero(&key, sizeof(key));
   bzero(&val, sizeof(val));

   result = db->cursor(db, NULL, &cursor, 0);
   if (result) {
      db->err(db, result, "cursor");
      exit(1); /* XXX */
   }

   result = cursor->c_get(cursor, &key, &val, DB_FIRST);
   if (result && result != DB_NOTFOUND) {
      db->err(db, result, "cursor->c_get(DB_FIRST)");
      exit(1); /* XXX */
   }

   while (result == 0) {
      num++;
      if (func(clo, &key, &val)) {
	 break;
      }

      result = cursor->c_get(cursor, &key, &val, DB_NEXT);
      if (result && result != DB_NOTFOUND) {
	 db->err(db, result, "cursor->c_get(DB_NEXT)");
	 exit(1); /* XXX */
      }
   }

   result = cursor->c_close(cursor);
   if (result) {
      db->err(db, result, "cursor->c_close");
      exit(1); /* XXX */
   }

   return num;
}

////////////////////////////////////////////////////////////
// Other Berkeley DB retrieval tools

#if 0 /* not used */
/*
 * Look up the token number for a given string WORD.
 */
static int db_get_tnum(const char *word, tnum_t *ret) {
   DBT key, val;
   int result;

   bzero(&key, sizeof(key));
   bzero(&val, sizeof(val));

   key.data = (char *)word;
   key.size = strlen(word);

   result = g_tok2tnumdb->db->get(g_tok2tnumdb->db, NULL, &key, &val, 0);
   if (result && result != DB_NOTFOUND) {
      g_tok2tnumdb->db->err(g_tok2tnumdb->db, result, "tok2tnumdb get");
      exit(1); /* XXX */
   }
   if (result == DB_NOTFOUND) {
      return -1;
   }
   assert(val.size == sizeof(tnum_t));
   *ret = *(tnum_t *)val.data;
   return 0;
}
#endif /* 0 */

/*
 * Get the string expansion for the token number TN. The string is
 * returned as a pqlvalue.
 */
static struct pqlvalue *db_get_token(pqlcontext *pql, tnum_t tn) {
   DBT key, val;
   int result;

   bzero(&key, sizeof(key));
   bzero(&val, sizeof(val));

   key.data = (char *)&tn;
   key.size = sizeof(tn);

   result = g_tnum2tokdb->db->get(g_tnum2tokdb->db, NULL, &key, &val, 0);
   if (result && result != DB_NOTFOUND) {
      g_tok2tnumdb->db->err(g_tok2tnumdb->db, result, "tnum2tokdb get");
      exit(1); /* XXX */
   }
   if (result == DB_NOTFOUND) {
      return NULL;
   }

   return pqlvalue_string_bylen(pql, (const char *) val.data, val.size);
}

////////////////////////////////////////////////////////////
// Value representation tools

/*
 * Instantiate a struct, in either the whole-object graph or the
 * per-version graph.
 */
static struct pqlvalue *makestruct(struct pqlcontext *pql,
				   pnode_t pnum, version_t version,
				   bool do_versions) {
   if (do_versions) {
      return pqlvalue_struct(pql, DBNUM_VERS, pnum, version);
   }
   else {
      return pqlvalue_struct(pql, DBNUM_MAIN, pnum, 0);
   }
}

/*
 * Convert a stored provdb value V (of length VLEN) to a pqlvalue.
 */
static struct pqlvalue *dbval_to_pql(struct pqlcontext *pql,
				     struct provdb_val *v, size_t vlen,
				     bool do_versions) {
   const void *valptr;
   size_t vallen;

   assert(PROVDB_VAL_TOTSIZE(v) == vlen);
   valptr = PROVDB_VAL_VALUE(v);
   vallen = PROVDB_VAL_VALUELEN(v);

   switch (PROVDB_VAL_VALUETYPE(v)) {
    case PROV_TYPE_NIL:
      return pqlvalue_nil(pql);

    case PROV_TYPE_STRING:
      // XXX should not assume this
      assert(!PROVDB_VAL_ISTOKENIZED(v));
      return pqlvalue_string_bylen(pql, (const char *) valptr, vallen);

    case PROV_TYPE_MULTISTRING:
    { 
      // XXX should not assume this
      assert(PROVDB_VAL_ISTOKENIZED(v));

      if (prov_mstring_tdb == NULL) {
	 prov_mstring_tdb = tdb_create();
      }

      /*
       * We ought to index and reuse these, instead of pumping out a
       * fresh object (with a different oid) each time we see one.
       * However, since now we only get here for multistrings that are
       * not ARGV or ENV it's not that big a deal.
       */

      pqloid_t o = tdb_newobject(prov_mstring_tdb);
      pqlvalue *v = pqlvalue_struct(pql, DBNUM_MSTR, o, 0);

      int ntokens = vallen / sizeof(tnum_t);
      for (int i=0; i<ntokens; i++) {
	 pqlvalue *thisedge = pqlvalue_int(pql, i);
	 pqlvalue *thisval = db_get_token(pql, ((const tnum_t *) valptr)[i] );
	 int r = tdb_assign(pql, prov_mstring_tdb, o, thisedge, thisval);
	 // XXX handle errors?
	 assert(r == 0);
	 pqlvalue_destroy(thisedge);
	 pqlvalue_destroy(thisval);
      }
      return v;
    }

    case PROV_TYPE_INT:
     assert(vallen == sizeof(int32_t));
     return pqlvalue_int(pql, *(int32_t *)valptr);

    case PROV_TYPE_REAL:
     assert(vallen == sizeof(double));
     return pqlvalue_float(pql, *(double *)valptr);

    case PROV_TYPE_TIMESTAMP:
     // don't currently have time values in sage (XXX)
     assert(vallen == sizeof(struct prov_timestamp));
     {
	char tmp[128];
	snprintf(tmp, sizeof(tmp), "%lld",
		 (long long) ((struct prov_timestamp *)valptr)->pt_sec);
	return pqlvalue_string(pql, tmp);
     }

    case PROV_TYPE_INODE:
     assert(vallen == sizeof(uint32_t));
     // XXX might overflow (our ints are ordinary signed int)
     return pqlvalue_int(pql, *(uint32_t *)valptr);

    case PROV_TYPE_PNODE:
    {
      assert(vallen == sizeof(pnode_t));
      pnode_t pnum = *(pnode_t *)valptr;
      return makestruct(pql, pnum, 0, do_versions);
    }

    case PROV_TYPE_PNODEVERSION:
    {
      assert(vallen == sizeof(pnode_version));
      pnode_version *pv = (pnode_version *)valptr;
      return makestruct(pql, pv->pnum, pv->version, do_versions);
    }

    case PROV_TYPE_OBJECT:
    case PROV_TYPE_OBJECTVERSION:
     assert(0);
    default:
     break;

   }
   assert(0);
   return NULL;
}

/*
 * Return true if the on-disk attribute name SEENATTR matches the
 * edge name EDGE that came from the PQL engine.
 */
static bool edge_matches(const char *seenattr, const pqlvalue *edge) {
   if (edge == NULL) {
      /* all edges */
      return true;
   }

   if (pqlvalue_isstring(edge)) {
      const char *edgename = pqlvalue_string_get(edge);
      if (!strcmp(seenattr, edgename)) {
	 return true;
      }
   }
   else if (pqlvalue_isint(edge)) {
      char *t;
      errno = 0;
      long x = strtol(seenattr, &t, 0);
      if (errno == 0 && strlen(t) == 0) {
	 if (x == pqlvalue_int_get(edge)) {
	    return true;
	 }
      }
   }
   return false;
}

/*
 * Return true if the dbvalue DBVAL points to the object PNUM.
 */
static bool pnode_matches(provdb_val *dbval, pnode_t pnum, version_t version,
			  bool do_versions) {
   pnode_version *pv;

   switch (PROVDB_VAL_VALUETYPE(dbval)) {

    case PROV_TYPE_PNODE:
     return *(pnode_t *)PROVDB_VAL_VALUE(dbval) == pnum;

    case PROV_TYPE_PNODEVERSION:
     pv = (pnode_version *)PROVDB_VAL_VALUE(dbval);
     if (do_versions) {
	return pv->pnum == pnum && pv->version == version;
     }
     else {
	return pv->pnum == pnum;
     }

    default:
     break;
   }

   return false;
}

////////////////////////////////////////////////////////////
// Search: scan whole database

/*
 * Scan the database of all objects and return a set referring to
 * each one. If the EDGENAME passed in is not null, pair each object
 * with that edge name as needed for implementing followall().
 */

/*
 * Operation context.
 */
struct scan_state {
   struct pqlcontext *pql;		/* PQL context */
   const char *edgename;		/* Edge name to insert, if any */
   pnode_t prev_pnode;			/* Last pnode number seen */
   version_t prev_version;		/* Last version number seen */
   bool do_versions;			/* Are we distinguishing versions? */
   struct pqlvalue *ret;		/* Result set */
};

/*
 * Visitor function: add each distinct object (different pnode number,
 * or pnode number and version) to the result set.
 */
static int scan_sub(void *sv, DBT *key, DBT * /*val*/) {
   struct scan_state *s = (struct scan_state *)sv;
   struct pnode_version *dbkey;
   struct pqlvalue *pqlval;

   assert(key->size == sizeof(struct pnode_version));
   dbkey = (struct pnode_version *)key->data;

   if (dbkey->pnum != s->prev_pnode ||
       (s->do_versions && dbkey->version != s->prev_version)) {
      /*
       * We're now seeing the records for a different object, so add
       * that object to the result set.
       */
      pqlval = makestruct(s->pql, dbkey->pnum, dbkey->version, s->do_versions);
      if (s->edgename != NULL) {
	 pqlval = pqlvalue_pair(s->pql,
				pqlvalue_string(s->pql, s->edgename), pqlval);
      }
      pqlvalue_set_add(s->ret, pqlval);
      s->prev_pnode = dbkey->pnum;
   }
   return 0;
}

/*
 * Driver function; initialize and use visit_everything.
 */
static struct pqlvalue *scan(struct pqlcontext *pql, const char *edgename,
			     bool do_versions) {
   scan_state s;
   s.pql = pql;
   s.edgename = edgename;
   s.prev_pnode = 0;
   s.prev_version = 0;
   s.do_versions = do_versions;
   s.ret = pqlvalue_emptyset(pql);
   dbvisit_everything(g_provdb->db, &s, scan_sub);
   return s.ret;
}

////////////////////////////////////////////////////////////
// Search: forward.

/*
 * Retrieve the values of forward edges (fields) of an object. If the
 * EDGE passed in is not null, retrieve only values for that edge
 * name. Otherwise, return edge name / value pairs for *all* records
 * associated with the object.
 */

/*
 * Operation context.
 */
struct forward_state {
   struct pqlcontext *pql;		/* PQL context */
   pnode_t pnode;			/* object we're examining */
   version_t thisversion;		/* version we're examining */
   version_t highversion;		/* highest version seen */
   const struct pqlvalue *edge;		/* edge name to follow */
   bool do_versions;			/* Are we distinguishing versions? */
   struct pqlvalue *ret;		/* result set */
};

/*
 * Visitor function: add matching values to result set, stop when we
 * reach the next pnode.
 */
static int follow_forward_sub(void *sv, DBT *key, DBT *val) {
   struct forward_state *s = (forward_state *)sv;
   provdb_key *dbkey;
   provdb_val *dbval;
   size_t dbval_len;
   const char *attr;
   struct pqlvalue *pqlval;

   dbkey = (provdb_key *) key->data;
   dbval = (provdb_val *) val->data;
   dbval_len = val->size;

   if (dbkey->pnum != s->pnode) {
      /* We've reached the next pnode; stop. */
      return 1;
   }

   if (s->do_versions && dbkey->version != s->thisversion) {
      /* We've reached the next pnode; stop. */
      return 1;
   }

   if (dbkey->version > s->highversion) {
      s->highversion = dbkey->version;
   }

   attr = PROVDB_VAL_ATTR(dbval);
   if (edge_matches(attr, s->edge)) {
      /*
       * XXX we shouldn't "know" this. On the other hand, multistring
       * values shouldn't exist. The right thing to do is clean
       * *those* up; then this logic can go away.
       */
      if (!strcmp(attr, "ARGV") &&
	  PROVDB_VAL_VALUETYPE(dbval) == PROV_TYPE_MULTISTRING) {
	 pqlval = pqlvalue_struct(s->pql, DBNUM_XARGV, s->pnode,
				  s->do_versions ? 0 : 1);
      }
      else if (!strcmp(attr, "ENV") &&
	       PROVDB_VAL_VALUETYPE(dbval) == PROV_TYPE_MULTISTRING) {
	 pqlval = pqlvalue_struct(s->pql, DBNUM_XENV, s->pnode,
				  s->do_versions ? 0 : 1);
      }
      else {
	 pqlval = dbval_to_pql(s->pql, dbval, dbval_len, s->do_versions);
      }
      if (s->edge == NULL) {
	 /* followall case */
	 pqlval = pqlvalue_pair(s->pql, pqlvalue_string(s->pql, attr), pqlval);
      }
      pqlvalue_set_add(s->ret, pqlval);
   }

   return 0;
}

/*
 * Driver function; initialize and use visit_by_setrange.
 */
static struct pqlvalue *follow_forward(struct pqlcontext *pql,
				       const struct pqlvalue *obj,
				       const struct pqlvalue *edge,
				       bool do_versions) {
   struct forward_state s;
   DBT key;
   struct pnode_version k;
   const char *attr;
   struct pqlvalue *pqlval;

   s.pql = pql;
   s.pnode = pqlvalue_struct_getoid(obj);
   s.thisversion = pqlvalue_struct_getsubid(obj);
   s.highversion = 1;
   s.edge = edge;
   s.do_versions = do_versions;
   s.ret = pqlvalue_emptyset(pql);

   bzero(&key, sizeof(key));
   k.pnum = s.pnode;
   k.version = do_versions ? s.thisversion : 0;
   key.data = &k;
   key.size = sizeof(k);

   dbvisit_setrange(g_provdb->db, &key, &s, follow_forward_sub);

   /*
    * Magic edges.
    */
   attr = "LATEST_VERSION";
   if (!do_versions && edge_matches(attr, edge)) {
      pqlval = pqlvalue_int(pql, s.highversion);
      if (edge == NULL) {
	 /* followall case */
	 pqlval = pqlvalue_pair(pql, pqlvalue_string(pql, attr), pqlval);
      }
      pqlvalue_set_add(s.ret, pqlval);
   }
   attr = "PNODE";
   if (edge_matches(attr, edge)) {
      // XXX this won't necessarily fit, PQL should have either a u64 type
      // of some kind or an exposed oid type.
      pqlval = pqlvalue_int(pql, s.pnode);
      if (edge == NULL) {
	 /* followall case */
	 pqlval = pqlvalue_pair(pql, pqlvalue_string(pql, attr), pqlval);
      }
      pqlvalue_set_add(s.ret, pqlval);
   }
   attr = "VERSION";
   if (do_versions && edge_matches(attr, edge)) {
      pqlval = pqlvalue_int(pql, s.thisversion);
      if (edge == NULL) {
	 /* followall case */
	 pqlval = pqlvalue_pair(pql, pqlvalue_string(pql, attr), pqlval);
      }
      pqlvalue_set_add(s.ret, pqlval);
   }
   attr = "PREV_VERSION";
   if (do_versions && edge_matches(attr, edge) && s.thisversion > 1) {
      pqlval = pqlvalue_struct(pql, DBNUM_VERS, s.pnode, s.thisversion-1);
      if (edge == NULL) {
	 /* followall case */
	 pqlval = pqlvalue_pair(pql, pqlvalue_string(pql, attr), pqlval);
      }
      pqlvalue_set_add(s.ret, pqlval);
   }
   attr = "BASE_VERSION";
   if (do_versions && edge_matches(attr, edge)) {
      pqlval = pqlvalue_struct(pql, DBNUM_VERS, s.pnode, 0);
      if (edge == NULL) {
	 /* followall case */
	 pqlval = pqlvalue_pair(pql, pqlvalue_string(pql, attr), pqlval);
      }
      pqlvalue_set_add(s.ret, pqlval);
   }

   return s.ret;
}

/*
 * Wrapper for followall.
 */
static struct pqlvalue *followall_forward(struct pqlcontext *pql,
					  const struct pqlvalue *obj,
					  bool do_versions) {
   return follow_forward(pql, obj, NULL, do_versions);
}

////////////////////////////////////////////////////////////
// Search: backwards.

/*
 * Retrieve the values of backward edges of (fields pointing to) an
 * object. If the EDGE passed in is not null, retrieve only values for
 * that edge name. Otherwise, return edge name / value pairs for *all*
 * matching records.
 */

/*
 * Operation context for the first pass.
 */
struct backward_state {
   pnode_t pnode;			/* Object we're looking for. */
   version_t version;			/* Version we're looking for. */
   bool do_versions;			/* Are we distinguishing versions? */
   primarray<pnode_t> children;		/* Objects that refer to it. */
   primarray<version_t> childvers;	/* Versions thereof. */
};

/*
 * Visitor function: collect the objects pointing to us and stop when
 * we see data for the next object.
 */
static int follow_backward_sub(void *sv, DBT *key, DBT *val) {
   struct backward_state *s = (backward_state *)sv;
   struct pnode_version *k, *v;

   assert(key->size == sizeof(pnode_version));
   assert(val->size == sizeof(pnode_version));

   k = (pnode_version *)key->data;
   v = (pnode_version *)val->data;

   if (k->pnum != s->pnode) {
      /* reached next object; stop */
      return 1;
   }

   if (s->do_versions && k->version != s->version) {
      /* reached next version; stop */
      return 1;
   }

   s->children.add(v->pnum);
   if (s->do_versions) {
      s->childvers.add(v->version);
   }

   return 0;
}

/*
 * Operation context for the second pass.
 */
struct backward_check_state {
   struct pqlcontext *pql;		/* PQL context */
   pnode_t pnode;			/* Object we're checking. */
   version_t version;			/* Version we're checking. */
   const pqlvalue *edge;		/* Edge we're looking for, or null. */
   pnode_t target;			/* Object we're looking for. */
   version_t targetver;			/* Version we're looking for. */
   bool do_versions;			/* Are we distinguishing versions? */
   struct pqlvalue *ret;		/* Result set. */
};

/*
 * Visitor function: check for records that point from PNODE to TARGET
 * with edge name EDGE. Stop when we've seen all the records for PNODE.
 */
static int follow_backward_check_sub(void *sv, DBT *key, DBT *val) {
   struct backward_check_state *s = (backward_check_state *)sv;
   pnode_version *dbkey;
   provdb_val *dbval;
   const char *attr;
   struct pqlvalue *pqlval;

   assert(key->size == sizeof(pnode_version));

   dbkey = (provdb_key *)key->data;
   dbval = (provdb_val *)val->data;

   if (dbkey->pnum != s->pnode) {
      /* reached next pnode; stop */
      return 1;
   }

   if (s->do_versions && dbkey->version != s->version) {
      /* reached next version; stop */
      return 1;
   }

   attr = PROVDB_VAL_ATTR(dbval);
   if (edge_matches(attr, s->edge) &&
       pnode_matches(dbval, s->target, s->targetver, s->do_versions)) {
      pqlval = makestruct(s->pql, s->pnode, s->version, s->do_versions);
      if (s->edge == NULL) {
	 /* followall case */
	 pqlval = pqlvalue_pair(s->pql, pqlvalue_string(s->pql, attr), pqlval);
      }
      pqlvalue_set_add(s->ret, pqlval);
   }

   return 0;
}

/*
 * Driver function.
 *
 * Because the child index currently has the wrong stuff in it (it
 * doesn't have edge names) we have to first go backwards using the
 * child index, which retrieves all pnodes that point to OBJ, and then
 * search forwards again to see which of them point to OBJ with edge
 * name EDGE. (Or, if EDGE is null, to find out what edge names they
 * point to OBJ under.)
 *
 * Blah.
 */
static struct pqlvalue *follow_backward(struct pqlcontext *pql,
					const struct pqlvalue *obj,
					const struct pqlvalue *edge,
					bool do_versions) {
   DBT key;
   backward_state s;
   pnode_version k;
   struct pqlvalue *ret;

   /*
    * Backwards step: get list of pnodes.
    */

   s.pnode = pqlvalue_struct_getoid(obj);
   s.version = pqlvalue_struct_getsubid(obj);
   s.do_versions = do_versions;

   k.pnum = s.pnode;
   k.version = 0;

   bzero(&key, sizeof(key));
   key.data = &k;
   key.size = sizeof(k);

   dbvisit_setrange(g_childdb->db, &key, &s, follow_backward_sub);

   /*
    * Forwards step: we have the list of pnodes; now check which ones
    * have a suitable edge and/or extract the edge names.
    */

   ret = pqlvalue_emptyset(pql);
   for (int i=0; i<s.children.num(); i++) {
      backward_check_state s2;

      k.pnum = s.children[i];
      k.version = 0;

      bzero(&key, sizeof(key));
      key.data = &k;
      key.size = sizeof(k);

      s2.pql = pql;
      s2.pnode = s.children[i];
      s2.version = do_versions ? s.childvers[i] : 0;
      s2.edge = edge;
      s2.target = s.pnode;
      s2.targetver = s.version;
      s2.do_versions = do_versions;
      s2.ret = ret;

      dbvisit_setrange(g_provdb->db, &key, &s2, follow_backward_check_sub);
   }

   return ret;
}

/*
 * Wrapper for followall case.
 */
static struct pqlvalue *followall_backward(struct pqlcontext *pql,
					   const struct pqlvalue *obj,
					   bool do_versions) {
   return follow_backward(pql, obj, NULL, do_versions);
}

////////////////////////////////////////////////////////////
// Implementation of magic (dynamically generated) objects

/*
 * The "Provenance" object.
 */

static struct pqlvalue *magic_provenance_e_to_r(struct pqlcontext *pql,
						const struct pqlvalue *edge) {
   const char *edgestr;

   if (!pqlvalue_isstring(edge)) {
      return pqlvalue_emptyset(pql);
   }

   edgestr = pqlvalue_string_get(edge);
   if (!strcmp(edgestr, "obj")) {
      return scan(pql, NULL, false);
   }
   // else ... // Provenance.file? Provenance.proc?
   // XXX think about what we want.
   return pqlvalue_emptyset(pql);
}

static struct pqlvalue *magic_provenance_e_to_l(struct pqlcontext *pql,
						const struct pqlvalue *edge) {
   (void)edge;
   return pqlvalue_emptyset(pql);
}

static struct pqlvalue *magic_provenance_to_er(struct pqlcontext *pql) {
   return scan(pql, "obj", false);
}

static struct pqlvalue *magic_provenance_to_le(struct pqlcontext *pql) {
   return pqlvalue_emptyset(pql);
}

/*
 * Dispatchers.
 */

static struct pqlvalue *magic_le_to_r(struct pqlcontext *pql,
				      const struct pqlvalue *obj,
				      const struct pqlvalue *edge) {
   pqloid_t oid;

   oid = pqlvalue_struct_getoid(obj);
   switch (oid) {
    case MAGIC_PROVENANCE:
     return magic_provenance_e_to_r(pql, edge);
    default:
     return pqlvalue_emptyset(pql);
   }
}

static struct pqlvalue *magic_er_to_l(struct pqlcontext *pql,
				      const struct pqlvalue *edge,
				      const struct pqlvalue *obj) {
   pqloid_t oid;

   oid = pqlvalue_struct_getoid(obj);
   switch (oid) {
    case MAGIC_PROVENANCE:
     return magic_provenance_e_to_l(pql, edge);
    default:
     return pqlvalue_emptyset(pql);
   }
}

static struct pqlvalue *magic_l_to_er(struct pqlcontext *pql,
				      const struct pqlvalue *obj) {
   pqloid_t oid;

   oid = pqlvalue_struct_getoid(obj);
   switch (oid) {
    case MAGIC_PROVENANCE:
     return magic_provenance_to_er(pql);
    default:
     return pqlvalue_emptyset(pql);
   }
}

static struct pqlvalue *magic_r_to_le(struct pqlcontext *pql,
				      const struct pqlvalue *obj) {
   pqloid_t oid;

   oid = pqlvalue_struct_getoid(obj);
   switch (oid) {
    case MAGIC_PROVENANCE:
     return magic_provenance_to_le(pql);
    default:
     return pqlvalue_emptyset(pql);
   }
}

////////////////////////////////////////////////////////////
// argv and env arrays

/*
 * Like follow_forward_sub but does different things upon finding the
 * target.
 */
struct multistring_lookup_state {
   struct pqlcontext *pql;
   pnode_t pnode;
   const char *attr;
   const struct pqlvalue *edge;
   struct pqlvalue *ret;
};
static int multistring_lookup_sub(void *sv, DBT *key, DBT *val) {
   struct multistring_lookup_state *s = (multistring_lookup_state *)sv;
   provdb_key *dbkey;
   provdb_val *dbval;
   size_t dbval_len;
   const char *attr;
   tnum_t tok;

   dbkey = (provdb_key *) key->data;
   dbval = (provdb_val *) val->data;
   dbval_len = val->size;

   if (dbkey->pnum != s->pnode) {
      /* We've reached the next pnode; stop. */
      return 1;
   }

   attr = PROVDB_VAL_ATTR(dbval);
   if (!strcmp(attr, s->attr)) {

      assert(PROVDB_VAL_TOTSIZE(dbval) == dbval_len);
      assert(PROVDB_VAL_VALUETYPE(dbval) == PROV_TYPE_MULTISTRING);
      // XXX should not assume this
      assert(PROVDB_VAL_ISTOKENIZED(dbval));

      int ntokens = PROVDB_VAL_VALUELEN(dbval) / sizeof(tnum_t);
      if (s->edge != NULL) {
	 int index = pqlvalue_int_get(s->edge);
	 if (index < 0 || index >= ntokens) {
	    /* nothing */
	 }
	 else {
	    tok = ((const tnum_t *)PROVDB_VAL_VALUE(dbval))[index];
	    pqlvalue_set_add(s->ret, db_get_token(s->pql, tok));
	 }
      }
      else {
	 for (int i=0; i<ntokens; i++) {
	    pqlvalue *thisedge = pqlvalue_int(s->pql, i);
	    tok = ((const tnum_t *)PROVDB_VAL_VALUE(dbval))[i];
	    pqlvalue *thisval = db_get_token(s->pql, tok);
	    pqlvalue_set_add(s->ret, pqlvalue_pair(s->pql, thisedge, thisval));
	 }
      }
      return 1; /* stop immediately */
   }

   return 0;
}

static struct pqlvalue *multistring_l_to_r(struct pqlcontext *pql,
					   pnode_t pnode,
					   const char *attrib,
					   const struct pqlvalue *edge) {
   struct multistring_lookup_state s;
   DBT key;
   struct pnode_version k;

   if (edge != NULL && !pqlvalue_isint(edge)) {
      return pqlvalue_emptyset(pql);
   }

   bzero(&key, sizeof(key));
   k.pnum = pnode;
   k.version = 0;
   key.data = &k;
   key.size = sizeof(k);

   s.pql = pql;
   s.pnode = pnode;
   s.attr = attrib;
   s.edge = edge;
   s.ret = pqlvalue_emptyset(pql);

   dbvisit_setrange(g_provdb->db, &key, &s, multistring_lookup_sub);

   return s.ret;
}

static struct pqlvalue *xargv_le_to_r(struct pqlcontext *pql,
				      const struct pqlvalue *obj,
				      const struct pqlvalue *edge) {
   pnode_t pnode;

   if (!pqlvalue_isint(edge)) {
      return pqlvalue_emptyset(pql);
   }
   pnode = pqlvalue_struct_getoid(obj);
   return multistring_l_to_r(pql, pnode, "ARGV", edge);
}

static struct pqlvalue *xargv_er_to_l(struct pqlcontext *pql,
				      const struct pqlvalue *edge,
				      const struct pqlvalue *obj) {
   pnode_t pnode;
   bool do_versions;

   if (!pqlvalue_isstring(edge) || strcmp(pqlvalue_string_get(edge), "ARGV") != 0) {
      return pqlvalue_emptyset(pql);
   }
   pnode = pqlvalue_struct_getoid(obj);
   do_versions = pqlvalue_struct_getsubid(obj) != 0;
   return makestruct(pql, pnode, 0, do_versions);
}

static struct pqlvalue *xargv_l_to_er(struct pqlcontext *pql,
				      const struct pqlvalue *obj) {
   pnode_t pnode;

   pnode = pqlvalue_struct_getoid(obj);
   return multistring_l_to_r(pql, pnode, "ARGV", NULL);
}

static struct pqlvalue *xargv_r_to_le(struct pqlcontext *pql,
				      const struct pqlvalue *obj) {
   struct pqlvalue *provobj;
   pnode_t pnode;
   bool do_versions;

   pnode = pqlvalue_struct_getoid(obj);
   do_versions = pqlvalue_struct_getsubid(obj) != 0;
   provobj = makestruct(pql, pnode, 0, do_versions);
   return pqlvalue_pair(pql, provobj, pqlvalue_string(pql, "ARGV"));
}

static struct pqlvalue *xenv_le_to_r(struct pqlcontext *pql,
				      const struct pqlvalue *obj,
				      const struct pqlvalue *edge) {
   pnode_t pnode;

   if (!pqlvalue_isint(edge)) {
      return pqlvalue_emptyset(pql);
   }
   pnode = pqlvalue_struct_getoid(obj);
   return multistring_l_to_r(pql, pnode, "ENV", edge);
}

static struct pqlvalue *xenv_er_to_l(struct pqlcontext *pql,
				      const struct pqlvalue *edge,
				      const struct pqlvalue *obj) {
   pnode_t pnode;
   bool do_versions;

   if (!pqlvalue_isstring(edge) || strcmp(pqlvalue_string_get(edge), "ENV") != 0) {
      return pqlvalue_emptyset(pql);
   }
   pnode = pqlvalue_struct_getoid(obj);
   do_versions = pqlvalue_struct_getsubid(obj) != 0;
   return makestruct(pql, pnode, 0, do_versions);
}

static struct pqlvalue *xenv_l_to_er(struct pqlcontext *pql,
				      const struct pqlvalue *obj) {
   pnode_t pnode;

   pnode = pqlvalue_struct_getoid(obj);
   return multistring_l_to_r(pql, pnode, "ENV", NULL);
}

static struct pqlvalue *xenv_r_to_le(struct pqlcontext *pql,
				      const struct pqlvalue *obj) {
   struct pqlvalue *provobj;
   pnode_t pnode;
   bool do_versions;

   pnode = pqlvalue_struct_getoid(obj);
   do_versions = pqlvalue_struct_getsubid(obj) != 0;
   provobj = makestruct(pql, pnode, 0, do_versions);
   return pqlvalue_pair(pql, provobj, pqlvalue_string(pql, "ENV"));
}

////////////////////////////////////////////////////////////
// Entry points for PQL backend operations

/*
 * Initialize the backend.
 */
int backend_init(const char *dbpath) {
   return wdb_startup(dbpath, WDB_O_RDONLY);
}

void backend_cleartemps(void) {
   if (prov_mstring_tdb != NULL) {
      tdb_destroy(prov_mstring_tdb);
      prov_mstring_tdb = NULL;
   }
   if (temp_tdb != NULL) {
      tdb_destroy(temp_tdb);
      temp_tdb = NULL;
   }
}

/*
 * Shut down the back end and tidy up.
 */
int backend_shutdown(void) {
   assert(prov_mstring_tdb == NULL);
   assert(temp_tdb == NULL);
   return wdb_shutdown();
}

/*
 * Read a global variable.
 */
static struct pqlvalue *backend_read_global(struct pqlcontext *pql,
					    const char *name) {
   if (!strcmp(name, "Provenance")) {
      struct pqlvalue *ret, *val;

      val = pqlvalue_struct(pql, DBNUM_MAGIC, MAGIC_PROVENANCE, 0);

      ret = pqlvalue_emptyset(pql);
      pqlvalue_set_add(ret, val);
      return ret;
   }
   else if (!strcmp(name, "VERSIONS")) {
      return scan(pql, NULL, true);
   }
   else {
      return pqlvalue_nil(pql);
   }
}

/*
 * Create a new (temporary) object.
 */
static struct pqlvalue *backend_newobject(struct pqlcontext *pql) {
   pqloid_t oid;

   if (temp_tdb == NULL) {
      temp_tdb = tdb_create();
   }

   oid = tdb_newobject(temp_tdb);
   if (oid == PQLOID_INVALID) {
      return NULL;
   }

   return pqlvalue_struct(pql, DBNUM_TEMP, oid, 0);
}

/*
 * Assign a field of a (temporary) object.
 */
static int backend_assign(struct pqlcontext *pql, struct pqlvalue *obj,
		   const struct pqlvalue *edge, const struct pqlvalue *val) {
   int dbnum;

   assert(pqlvalue_isstruct(obj));

   dbnum = pqlvalue_struct_getdbnum(obj);

   if (dbnum == DBNUM_TEMP) {
      assert(temp_tdb != NULL);
      return tdb_assign(pql, temp_tdb, pqlvalue_struct_getoid(obj), edge, val);
   }
   else {
      errno = EROFS;
      return -1;
   }
}

/*
 * Follow an edge forward from an object.
 */
static struct pqlvalue *backend_le_to_r(struct pqlcontext *pql,
					const struct pqlvalue *obj,
					const struct pqlvalue *edge) {
   int dbnum;

   assert(pqlvalue_isstruct(obj));

   dbnum = pqlvalue_struct_getdbnum(obj);
   switch (dbnum) {
    case DBNUM_TEMP:
     assert(temp_tdb != NULL);
     return tdb_le_to_r(pql, temp_tdb, pqlvalue_struct_getoid(obj), edge);

    case DBNUM_MAGIC:
     return magic_le_to_r(pql, obj, edge);

    case DBNUM_MSTR:
     return tdb_le_to_r(pql, prov_mstring_tdb, pqlvalue_struct_getoid(obj), edge);

    case DBNUM_MAIN:
     return follow_forward(pql, obj, edge, false);

    case DBNUM_VERS:
     return follow_forward(pql, obj, edge, true);

    case DBNUM_XARGV:
     return xargv_le_to_r(pql, obj, edge);

    case DBNUM_XENV:
     return xenv_le_to_r(pql, obj, edge);

    default:
     assert(0);
     return pqlvalue_emptyset(pql);
   }
}

/*
 * Follow an edge backward from an object.
 */
static struct pqlvalue *backend_er_to_l(struct pqlcontext *pql,
					const struct pqlvalue *edge,
					const struct pqlvalue *obj) {
   int dbnum;

   assert(pqlvalue_isstruct(obj));

   dbnum = pqlvalue_struct_getdbnum(obj);
   switch (dbnum) {
    case DBNUM_TEMP:
     assert(temp_tdb != NULL);
     return tdb_er_to_l(pql, temp_tdb, edge, pqlvalue_struct_getoid(obj));

    case DBNUM_MAGIC:
     return magic_er_to_l(pql, edge, obj);

    case DBNUM_MSTR:
     return tdb_er_to_l(pql, prov_mstring_tdb, edge, pqlvalue_struct_getoid(obj));

    case DBNUM_MAIN:
     return follow_backward(pql, obj, edge, false);

    case DBNUM_VERS:
     return follow_backward(pql, obj, edge, true);

    case DBNUM_XARGV:
     return xargv_er_to_l(pql, obj, edge);

    case DBNUM_XENV:
     return xenv_er_to_l(pql, obj, edge);

    default:
     assert(0);
     return pqlvalue_emptyset(pql);
   }
}

/*
 * Follow all edges forward from an object.
 */
static struct pqlvalue *backend_l_to_er(struct pqlcontext *pql,
					const struct pqlvalue *obj) {
   int dbnum;

   assert(pqlvalue_isstruct(obj));

   dbnum = pqlvalue_struct_getdbnum(obj);
   switch (dbnum) {

    case DBNUM_TEMP:
     assert(temp_tdb != NULL);
     // XXX this leaks memory
     return tdb_l_to_er(pql, temp_tdb, pqlvalue_struct_getoid(obj));

    case DBNUM_MAGIC:
     return magic_l_to_er(pql, obj);

    case DBNUM_MSTR:
     return tdb_l_to_er(pql, prov_mstring_tdb, pqlvalue_struct_getoid(obj));

    case DBNUM_MAIN:
     return followall_forward(pql, obj, false);

    case DBNUM_VERS:
     return followall_forward(pql, obj, true);

    case DBNUM_XARGV:
     return xargv_l_to_er(pql, obj);

    case DBNUM_XENV:
     return xenv_l_to_er(pql, obj);

    default:
     assert(0);
     return pqlvalue_emptyset(pql);
   }
}

/*
 * Follow all edges backward from an object.
 */
static struct pqlvalue *backend_r_to_le(struct pqlcontext *pql,
					const struct pqlvalue *obj) {
   int dbnum;

   assert(pqlvalue_isstruct(obj));

   dbnum = pqlvalue_struct_getdbnum(obj);
   switch (dbnum) {

    case DBNUM_TEMP:
     assert(temp_tdb != NULL);
     // XXX this leaks memory
     return tdb_r_to_le(pql, temp_tdb, pqlvalue_struct_getoid(obj));

    case DBNUM_MAGIC:
     return magic_r_to_le(pql, obj);

    case DBNUM_MSTR:
     return tdb_r_to_le(pql, prov_mstring_tdb, pqlvalue_struct_getoid(obj));

    case DBNUM_MAIN:
     return followall_backward(pql, obj, false);

    case DBNUM_VERS:
     return followall_backward(pql, obj, true);

    case DBNUM_XARGV:
     return xargv_r_to_le(pql, obj);

    case DBNUM_XENV:
     return xenv_r_to_le(pql, obj);

    default:
     assert(0);
     return pqlvalue_emptyset(pql);
   }
}

static pqlvalue *backend_read_global(pqlcontext *, const char *name);
static pqlvalue *backend_newobject(pqlcontext *);
static int backend_assign(pqlcontext *,
			  pqlvalue *obj, const pqlvalue *edge, const pqlvalue *val);
//static pqlvalue *backend_follow(pqlcontext *,
//				const pqlvalue *obj, const pqlvalue *edge,
//				bool reversed);
//static pqlvalue *backend_followall(pqlcontext *, const pqlvalue *obj, bool reversed);

const struct pqlbackend_ops myops = {
   backend_read_global,
   backend_newobject,
   backend_assign,
   backend_le_to_r,
   backend_l_to_er,
   backend_er_to_l,
   backend_r_to_le,
};
