 /** Copyright 2006, 2007, 2010
 *  The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <getopt.h>
#include <assert.h>
#include <sys/time.h>
#include <utime.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <time.h>

#include <db.h>
#include "debug.h"

#include "twig.h"
#include "schema.h"
#include "wdb.h"

#include <linux/types.h>
#include <asm/types.h>


static void dump_something(int ch);
static void dump_provdb(void);
static void dump_tnum2tok(void);
static void dump_tok2tnum(void);
static void dump_arg2p(void);
static void dump_env2p(void);
static void dump_p2i(void);
static void dump_i2p(void);
static void dump_name(void);
static void dump_child(void);
static void dump_parent(void);
static void dump_graph(void);
static void dump_guize(void);
static void dump_guizenew(void);
static void dump_test(void);
static void dump_testnew(void);

static void dump_ky2(void);


static void guize(void);
static void prune_guize(void);
static void test(void);
static void compress_test(void);

static void jiance(void);
static int find(char * ,char *);
static int online_detect(char*);

static char * get_string(pnode_t pnum, const char * attr);
static char ** get_multistring(pnode_t pnum, const char * attr);

static void print_attrflags(uint8_t attrflags);
int                         state[1000];
struct pnode_version  bak[1000];
static int j=0;
static int len=0;
static int W=3;
static int depi=0;
static float T=0.3;
static int ii_intrusion=0;
static int flag=0;
int jiance_time=0; 
////////////////////////////////////////////////////////////////////////////////

static void usage(char *name)
{
   fprintf(stderr,
           "Usage: %s [options] [<path>]\n"
           "   -a             dump all tables\n"
           "   -d xyz         dump selected tables in order given\n"
           "   -h             print usage\n"
       "Tables:\n"
           "   T              tnum2tok\n"
           "   N              tok2tnum\n"
           "   a              arg2p\n"
           "   e              env2p\n"
           "   i              i2p\n"
           "   p              p2i\n"
           "   n              name\n"
           "   >              child\n"
           "   <              parent\n"
           "   b              graph\n"
           "   d              provdb\n"
           "   m              rebuild\n"
           "   z                chakanguize\n"
           "   g               shengchengguize\n"
           "   j                jiance\n"
           "If no path is specified \".\" is used.\n"
           "If no tables are specified, all are dumped.\n",
           name );
}

/**
 * The main method for the wdb dump application.
 */
int
main(int argc, char *argv[])
{
   const  char          *path = ".";
   const  char          *dumps = "TNaeipn><dmg";
   int                   ch, i;
   struct timeval start,end;
   float timeopen;

   //
   // Parse options
   //
   while ( -1 != (ch = getopt(argc, argv, "ad:h")) ) {
      switch ( ch ) {
         case 'a':
        dumps = "TNaeipn><dz";
        break;
         case 'd':
        dumps = optarg;
        break;
         case 'h':
         default:
            usage(argv[0]);
            return 1;
      }
   }

   // extra params -> print usage
   if ( optind == argc - 1 ) {
      path = argv[optind];
   }
   else {
      usage(argv[0]);
      return 1;
   }

   //
   // Initialize all 'databases' (database & log)
   //
   gettimeofday(&start,NULL);
   if (wdb_startup(path, WDB_O_RDWR| WDB_O_CREAT) < 0) {
      fprintf(stderr, "wdb_dump: %s: %s\n", path, strerror(errno));
      exit(1);
   }
   gettimeofday(&end,NULL);
   timeopen = 1000000*(end.tv_sec-start.tv_sec)+end.tv_usec-start.tv_usec;
   //printf("======Open DB Time: %f microseconds , %f s ======\n",timeopen,timeopen/1000000);
   //
   // Dump everything
   //
   for (i=0; dumps[i]; i++) {
      dump_something(dumps[i]);
   }

   //
   // Shut everything down (database & log)
   //
   wdb_shutdown();

   exit(0);
}

static void printstring(const char *s, size_t len) {
   size_t i;
   putchar('"');
   for (i=0; i<len; i++) {
      unsigned char c = s[i];
      switch (c) {
       case '"':
       case '\\':
    putchar('\\');
    putchar(c);
    break;
       case '\a': putchar('\\'); putchar('a'); break;
       case '\b': putchar('\\'); putchar('b'); break;
       case '\t': putchar('\\'); putchar('t'); break;
       case '\n': putchar('\\'); putchar('n'); break;
       case '\v': putchar('\\'); putchar('v'); break;
       case '\f': putchar('\\'); putchar('f'); break;
       case '\r': putchar('\\'); putchar('r'); break;
       default:
    if (isascii(c) && (isprint(c) || isspace(c))) {
       putchar(c);
    }
    else {
       printf("\\%03o", c);
    }
    break;
      }
   }
   putchar('"');
}

static void printtoken(tnum_t tok) {
   DBT  dbt_tnum;
   DBT  dbt_token;
   int  ret;

   memset(&dbt_tnum, 0, sizeof(dbt_tnum));
   memset(&dbt_token, 0, sizeof(dbt_token));

   dbt_tnum.data = &tok;
   dbt_tnum.size = sizeof(tok);

   printf(" [%u]", tok);

   ret = g_tnum2tokdb->db->get(g_tnum2tokdb->db, NULL,
                               &dbt_tnum, &dbt_token,
                               0);
   wdb_check_err(ret);

   printstring(dbt_token.data, dbt_token.size);
}

/* For now, we use our own implementation of strnlen, since it's not clear
   what ifdefs we need here to make sure it is not already declared in libc */
static size_t wdb_strnlen(const char *s, size_t maxlen) {
   const char *t = memchr(s, 0, maxlen);
   return t ? (size_t)(t-s) : maxlen;
}

static void dump_something(int ch) {
   switch (ch) {
    case 'T': dump_tnum2tok(); break;
    case 'N': dump_tok2tnum(); break;
    case 'a': dump_arg2p(); break;
    //case 'e': dump_env2p(); break;
    case 'i': dump_i2p(); break;
    case 'c': compress_test(); break;
    case 't': test(); break;
    case 'n': dump_name(); break;
    case '>': dump_child(); break;
    case '<': dump_parent(); break;
    case 'b': dump_graph(); break;
    case 'd': dump_provdb(); break;
    case 'g': guize();break;
    case 'z': dump_guize();break;
    case 'w': dump_guizenew();break;
    case 's': dump_test();break;
    case 'e': dump_testnew();break;
    case 'p': prune_guize();break;
    case 'j': jiance();break;
    case 'k':dump_ky2();break;
    case 'M': modify_graph_construct();break;
    case 'R': read_graph_construct();break;
    case 'E': revert_graph_construct();break;
    case 'A': all_graph_construct();break;
    default:
     fprintf(stderr, "Unknown dump target %c\n", ch);
     exit(1);
   }
}

void dump_provdb(void)
{
   DBC                       *pdbc = NULL;
   DBT                        dbt_key;
   DBT                        dbt_val;
   provdb_key                 key;
   provdb_val                *val;
   struct pnode_version      *ref;
   enum prov_attrname_packed  id = PROV_ATTR_INVALID;
   const char                *attrname = NULL;
   char                      *data;
   size_t                     i, datalen;
   int                        serrno;
   int                        ret;

   printf( "----------------------------------------\n"
           "| Dumping provdb                       |\n"
           "|                                      |\n"
           "----------------------------------------\n" );

   memset(&dbt_key, 0, sizeof(dbt_key));
   memset(&dbt_val, 0, sizeof(dbt_val));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_provdb->db->cursor(g_provdb->db, NULL, &pdbc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = pdbc->c_get(pdbc, &dbt_key, &dbt_val, DB_NEXT)) ) {
      id = PROV_ATTR_INVALID;
      val = malloc(dbt_val.size);
      if ( val == NULL ) {
    serrno = errno;
    fprintf( stderr, "Malloc of provdb record failed %s\n",
         strerror(serrno) );
    errno = serrno;
    // XXX clean up such as closing the cursor
    return;
      }

      memcpy(&key, dbt_key.data, sizeof(key));
      memcpy(val, dbt_val.data, dbt_val.size);

      assert( sizeof(key) == dbt_key.size );
#if DEBUG
      if ( sizeof(*val) >= dbt_val.size ) {
          fprintf( stderr,
                   "sizeof(*val) %lu dbt_val.size %d sizeof(provdb_val) %lu\n"
                   "key: %lu.%u\n"
                   "val: flags %02X valtype %02X code/len %02X/%d vallen %u\n",
                   sizeof(*val), dbt_val.size, sizeof(provdb_val),
                   key.pnum, key.version,
                   val->pdb_flags, val->pdb_valuetype,
                   val->pdb_attrcode, val->pdb_attrlen,
                   val->pdb_valuelen );
      }
#endif /* DEBUG */
      assert( sizeof(*val) <= dbt_val.size );

      attrname = PROVDB_VAL_ATTR(val);
      data = PROVDB_VAL_VALUE(val);
      datalen = PROVDB_VAL_VALUELEN(val);

      print_attrflags(val->pdb_flags);

      printf("%llu.%u %s %u ",
         (unsigned long long)key.pnum, key.version,
             attrname, (unsigned)datalen);

      switch (PROVDB_VAL_VALUETYPE(val)) {

       case PROV_TYPE_NIL:
    printf("---\n");
    break;

       case PROV_TYPE_STRING:
    if (PROVDB_VAL_ISTOKENIZED(val)) {
       tnum_t tok = *(tnum_t *) data;

       assert(datalen == sizeof(tnum_t));
       tok = *(tnum_t *) data;
       printtoken(tok);
       printf("\n");
    }
    else {
       printstring(data, datalen);
       printf("\n");
    }
    break;

       case PROV_TYPE_MULTISTRING:
    if (PROVDB_VAL_ISTOKENIZED(val)) {
       int ntoks, i;
       tnum_t *toks;

       assert(datalen % sizeof(tnum_t) == 0);
       ntoks = datalen / sizeof(tnum_t);
       toks = malloc(datalen);
       memcpy(toks, data, datalen);

       for (i=0; i<ntoks; i++) {
          printtoken(toks[i]);
       }
       free(toks), toks = NULL;
    }
    else {
       size_t x, y;
       for (x = 0; x < datalen; x += y+1) {
          if (x > 0) printf(" ");
          y = wdb_strnlen(data+x, datalen-x);
          printstring(data+x, y);
       }
    }
    printf("\n");
    break;

       case PROV_TYPE_INT:
    assert(datalen == sizeof(int32_t));
    printf( "INT %d\n", *(int32_t *)data );
    break;

       case PROV_TYPE_REAL:
    assert(datalen == sizeof(double));
    printf( "REAL %g\n", *(double *)data );
    break;

       case PROV_TYPE_TIMESTAMP:
    assert(datalen == sizeof(struct prov_timestamp));
    printf( "TIME %u.%09u\n",
        ((struct prov_timestamp *)data)->pt_sec,
        ((struct prov_timestamp *)data)->pt_nsec);
    break;

       case PROV_TYPE_INODE:
    assert(datalen == sizeof(uint32_t));
    printf( "INODE %u\n", *(uint32_t *)data );
    break;

       case PROV_TYPE_PNODE:
    assert(datalen == sizeof(pnode_t));
    printf( "--> %llu\n", (unsigned long long)*(pnode_t *)data);
    break;

       case PROV_TYPE_PNODEVERSION:
    assert(datalen == sizeof(struct pnode_version));
    ref = (struct pnode_version *)data;
    printf( "--> %llu.%u\n",
        (unsigned long long)ref->pnum, ref->version );
    break;

       case PROV_TYPE_OBJECT:
       case PROV_TYPE_OBJECTVERSION:
       default:
    printf("[illegal value type %u]", PROVDB_VAL_VALUETYPE(val));
    for (i=0; i<datalen; i++) {
       printf(" %02x", (unsigned char) data[i]);
    }
    printf("\n");
      }

      free(val), val = NULL;
   }

   // XXX shouldn't this, like, close the cursor?

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

static void print_attrflags(uint8_t attrflags) {
   printf("[%c%c%c] ",
      ((attrflags & PROVDB_ANCESTRY)        ? 'A' : ' '),
      ((attrflags & PROVDB_PACKED)          ? 'P' : ' '),
      ((attrflags & PROVDB_TOKENIZED)       ? 'T' : ' '));
}

void dump_tnum2tok(void)
{
   DBC                  *tnum2tokc = NULL;
   DBT                   dbt_tnum;
   DBT                   dbt_token;
   db_recno_t            tnum;
   char                 *token;
   int                   ret;

   printf( "----------------------------------------\n"
           "| Dumping tnum2token                   |\n"
           "| tnum -> token                        |\n"
           "----------------------------------------\n" );

   memset(&dbt_tnum, 0, sizeof(dbt_tnum));
   memset(&dbt_token, 0, sizeof(dbt_token));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_tnum2tokdb->db->cursor(g_tnum2tokdb->db, NULL, &tnum2tokc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = tnum2tokc->c_get(tnum2tokc, &dbt_tnum, &dbt_token, DB_NEXT)) ) {
      assert( sizeof(tnum) == dbt_tnum.size );

      token = malloc(dbt_token.size + 1);
      memset(token, 0, dbt_token.size + 1);
      memcpy(&tnum, dbt_tnum.data, dbt_tnum.size);
      memcpy(token, dbt_token.data, dbt_token.size);

      printf( "%u %s\n", tnum, token );
      free(token), token = NULL;
   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_tok2tnum(void)
{
   DBC                  *tok2tnumc = NULL;
   DBT                   dbt_tnum;
   DBT                   dbt_token;
   db_recno_t            tnum;
   char                 *token;
   int                   ret;

   printf( "----------------------------------------\n"
           "| Dumping token2tnum                   |\n"
           "| token -> tnum                        |\n"
           "----------------------------------------\n" );

   memset(&dbt_tnum, 0, sizeof(dbt_tnum));
   memset(&dbt_token, 0, sizeof(dbt_token));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_tok2tnumdb->db->cursor(g_tok2tnumdb->db, NULL, &tok2tnumc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = tok2tnumc->c_get(tok2tnumc, &dbt_token, &dbt_tnum, DB_NEXT)) ) {
      assert( sizeof(tnum) == dbt_tnum.size );

      token = malloc(dbt_token.size + 1);
      memset(token, 0, dbt_token.size + 1);
      memcpy(&tnum, dbt_tnum.data, dbt_tnum.size);
      memcpy(token, dbt_token.data, dbt_token.size);

      printf( "%s -> %u\n", token, tnum );
      free(token), token = NULL;
   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_arg2p(void)
{
   DBC                  *arg2pc = NULL;
   DBT                   dbt_tnum;
   DBT                   dbt_pnode;
   db_recno_t            tnum;
   pnode_t               pnum;
   int                   ret;

   printf( "----------------------------------------\n"
           "| Dumping arg tnum -> pnode            |\n"
           "| arg tnum -> pnode                    |\n"
           "----------------------------------------\n" );

   memset(&dbt_tnum, 0, sizeof(dbt_tnum));
   memset(&dbt_pnode, 0, sizeof(dbt_pnode));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_arg2pdb->db->cursor(g_arg2pdb->db, NULL,
                               &arg2pc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = arg2pc->c_get(arg2pc, &dbt_tnum, &dbt_pnode, DB_NEXT)) ) {
      assert( sizeof(tnum) == dbt_tnum.size );
      assert( sizeof(pnum) == dbt_pnode.size );

      memcpy(&tnum, dbt_tnum.data, dbt_tnum.size);
      memcpy(&pnum , dbt_pnode.data, dbt_pnode.size);

      printf( "%u -> %llu\n", (unsigned)tnum, (unsigned long long)pnum );
   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_env2p(void)
{
   DBC                  *env2pc = NULL;
   DBT                   dbt_tnum;
   DBT                   dbt_pnode;
   db_recno_t            tnum;
   pnode_t               pnum;
   int                   ret;

   printf( "----------------------------------------\n"
           "| Dumping env tnum -> pnode            |\n"
           "| env tnum -> pnode                    |\n"
           "----------------------------------------\n" );

   memset(&dbt_tnum, 0, sizeof(dbt_tnum));
   memset(&dbt_pnode, 0, sizeof(dbt_pnode));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_env2pdb->db->cursor(g_env2pdb->db, NULL,
                               &env2pc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = env2pc->c_get(env2pc, &dbt_tnum, &dbt_pnode, DB_NEXT)) ) {
      assert( sizeof(tnum) == dbt_tnum.size );
      assert( sizeof(pnum) == dbt_pnode.size );

      memcpy(&tnum, dbt_tnum.data, dbt_tnum.size);
      memcpy(&pnum , dbt_pnode.data, dbt_pnode.size);

      printf( "%u -> %llu\n", (unsigned)tnum, (unsigned long long)pnum );
   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_name(void)
{
   DBC                  *namec = NULL;
   DBT                   dbt_name;
   DBT                   dbt_pnode;
   char                 *name;
   pnode_t               pnum;
   int                   ret;

   printf( "----------------------------------------\n"
           "| Dumping name                         |\n"
           "| name -> pnode                        |\n"
           "----------------------------------------\n" );

   memset(&dbt_name , 0, sizeof(dbt_name));
   memset(&dbt_pnode, 0, sizeof(dbt_pnode));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_namedb->db->cursor(g_namedb->db, NULL, &namec, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = namec->c_get(namec, &dbt_name, &dbt_pnode, DB_NEXT)) ) {
      assert( sizeof(pnum) == dbt_pnode.size );

      name = malloc(dbt_name.size + 1);
      memcpy(name, dbt_name.data, dbt_name.size);
      name[dbt_name.size] = 0;
      memcpy(&pnum, dbt_pnode.data, dbt_pnode.size);

      printf( "%s -> %llu\n", name, (unsigned long long)pnum );

      free(name), name = NULL;
   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_i2p(void)
{
   DBC                  *i2pc = NULL;
   DBT                   dbt_inode;
   DBT                   dbt_pv;
   uint32_t              inode;
   struct pnode_version  pv;
   int                   ret;

   printf( "----------------------------------------\n"
           "| Dumping i2p                          |\n"
           "| ino -> pnode.version                 |\n"
           "----------------------------------------\n" );

   memset(&dbt_inode, 0, sizeof(dbt_inode));
   memset(&dbt_pv   , 0, sizeof(dbt_pv));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_i2pdb->db->cursor(g_i2pdb->db, NULL, &i2pc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = i2pc->c_get(i2pc, &dbt_inode, &dbt_pv, DB_NEXT)) ) {
      assert( sizeof(inode) == dbt_inode.size );
      assert( sizeof(pv)    == dbt_pv.size );

      memcpy(&inode, dbt_inode.data, dbt_inode.size);
      memcpy(&pv , dbt_pv.data   , dbt_pv.size);

      printf( "%u -> %" PRIu64 ".%"  PRIu32 "\n",
              inode, pv.pnum, pv.version );
   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_p2i(void)
{
   DBC                  *p2ic = NULL;
   DBT                   dbt_inode;
   DBT                   dbt_pnode;
   uint32_t              inode;
   pnode_t               pnum;
   int                   ret;

   printf( "----------------------------------------\n"
           "| Dumping p2i                          |\n"
           "| pnode -> in                          |\n"
           "----------------------------------------\n" );

   memset(&dbt_inode, 0, sizeof(dbt_inode));
   memset(&dbt_pnode, 0, sizeof(dbt_pnode));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_p2idb->db->cursor(g_p2idb->db, NULL, &p2ic, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = p2ic->c_get(p2ic, &dbt_pnode, &dbt_inode, DB_NEXT)) ) {
      assert( sizeof(inode) == dbt_inode.size );
      assert( sizeof(pnum)  == dbt_pnode.size );

      memcpy(&inode, dbt_inode.data, dbt_inode.size);
      memcpy(&pnum , dbt_pnode.data, dbt_pnode.size);

      printf( "%llu -> %u\n", (unsigned long long)pnum, inode );
   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_child(void)
{
   DBC                  *childc = NULL;
   DBT                   dbt_parent;
   DBT                   dbt_child;
   provdb_key            parent;
   provdb_key            child;
   int                   ret;

   printf( "----------------------------------------\n"
           "| Dumping child                        |\n"
           "| parent -> child                      |\n"
           "----------------------------------------\n" );

   memset(&dbt_parent, 0, sizeof(dbt_parent));
   memset(&dbt_child , 0, sizeof(dbt_child));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_childdb->db->cursor(g_childdb->db, NULL, &childc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = childc->c_get(childc, &dbt_parent, &dbt_child, DB_NEXT)) ) {
      assert( sizeof(parent) == dbt_parent.size );
      assert( sizeof(child)  == dbt_child.size );

      memcpy(&parent, dbt_parent.data, dbt_parent.size);
      memcpy(&child , dbt_child.data , dbt_child.size);

      printf( "%llu.%u -> %llu.%u\n",
              (unsigned long long)parent.pnum, parent.version,
              (unsigned long long)child.pnum , child.version );
   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_parent(void)
{
   DBC                  *parentc = NULL;
   DBT                   dbt_child;
   DBT                   dbt_parent;
   provdb_key            child;
   provdb_key            parent;
   int                   ret;

   printf( "----------------------------------------\n"
           "| Dumping parent                        |\n"
           "| child -> parent                      |\n"
           "----------------------------------------\n" );

   memset(&dbt_child, 0, sizeof(dbt_child));
   memset(&dbt_parent , 0, sizeof(dbt_parent));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_parentdb->db->cursor(g_parentdb->db, NULL, &parentc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = parentc->c_get(parentc, &dbt_child, &dbt_parent, DB_NEXT)) ) {
      assert( sizeof(child) == dbt_child.size );
      assert( sizeof(parent)  == dbt_parent.size );

      memcpy(&child , dbt_child.data , dbt_child.size );
      memcpy(&parent, dbt_parent.data, dbt_parent.size);

      printf( "%llu.%u -> %llu.%u\n",
              (unsigned long long)child.pnum, child.version,
              (unsigned long long)parent.pnum , parent.version );
   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}


void dump_graph(void)
{
   DBC                  *graphc = NULL;
   DBT                   dbt_child;
   DBT                   dbt_parent;
   provdb_key            child;
   provdb_key            parent;
   int                   ret;

   printf( "----------------------------------------\n"
           "| Dumping graph                        |\n"
           "| child -> parent                      |\n"
           "----------------------------------------\n" );

   memset(&dbt_child, 0, sizeof(dbt_child));
   memset(&dbt_parent , 0, sizeof(dbt_parent));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_graphdb->db->cursor(g_graphdb->db, NULL, &graphc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = graphc->c_get(graphc, &dbt_child, &dbt_parent, DB_NEXT)) ) {
      assert( sizeof(child) == dbt_child.size );
      assert( sizeof(parent)  == dbt_parent.size );

      memcpy(&child , dbt_child.data , dbt_child.size );
      memcpy(&parent, dbt_parent.data, dbt_parent.size);

      printf( "%llu.%u -> %llu.%u\n",
              (unsigned long long)child.pnum, child.version,
              (unsigned long long)parent.pnum , parent.version );
   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_guize(void)
{
   DBC                  *namec = NULL;
   DBT                   dbt_name;
   DBT                   dbt_pnode;
   char                 *name;
   char                  *pnum;
   int                   ret;
   int                    num;

   printf( "----------------------------------------\n"
           "| Dumping guize                         |\n"
           "| name -> name                        |\n"
           "----------------------------------------\n" );

   memset(&dbt_name , 0, sizeof(dbt_name));
   memset(&dbt_pnode, 0, sizeof(dbt_pnode));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_guizedb->db->cursor(g_guizedb->db, NULL, &namec, 0);
   wdb_check_err(ret);
   num=0;

   // Iterate over the main database
   while ( 0 == (ret = namec->c_get(namec, &dbt_name, &dbt_pnode, DB_NEXT)) ) {

      name = malloc(dbt_name.size + 1);
      memcpy(name, dbt_name.data, dbt_name.size);
      name[dbt_name.size] = 0;
      
      pnum = malloc(dbt_pnode.size + 1);
      memcpy(pnum, dbt_pnode.data, dbt_pnode.size);
      pnum[dbt_pnode.size] = 0;
      printf( "%s -> %s\n", name, pnum );
      num++;

      free(name), name = NULL;
      free(pnum), pnum = NULL;

   }

   printf("num:%d\n",num);
   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_guizenew(void)
{
   DBC                  *namec = NULL;
   DBT                   dbt_name;
   DBT                   dbt_pnode;
   char                 *name;
   char                  *pnum;
   int                   ret;
   int                    num;

   printf( "----------------------------------------\n"
           "| Dumping guize                         |\n"
           "| name -> name                        |\n"
           "----------------------------------------\n" );

   memset(&dbt_name , 0, sizeof(dbt_name));
   memset(&dbt_pnode, 0, sizeof(dbt_pnode));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_guizenewdb->db->cursor(g_guizenewdb->db, NULL, &namec, 0);
   wdb_check_err(ret);
   num=0;

   // Iterate over the main database
   while ( 0 == (ret = namec->c_get(namec, &dbt_name, &dbt_pnode, DB_NEXT)) ) {

      name = malloc(dbt_name.size + 1);
      memcpy(name, dbt_name.data, dbt_name.size);
      name[dbt_name.size] = 0;
      
      pnum = malloc(dbt_pnode.size + 1);
      memcpy(pnum, dbt_pnode.data, dbt_pnode.size);
      pnum[dbt_pnode.size] = 0;
      printf( "%s -> %s\n", name, pnum );
      num++;

      free(name), name = NULL;
      free(pnum), pnum = NULL;

   }

   printf("num:%d\n",num);
   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_test(void)
{
   DBC                  *namec = NULL;
   DBT                   dbt_name;
   DBT                   dbt_pnode;
   char                 *name;
   char                  *pnum;
   int                   ret;
   int                    num;

   printf( "----------------------------------------\n"
           "| Dumping guize                         |\n"
           "| name -> name                        |\n"
           "----------------------------------------\n" );

   memset(&dbt_name , 0, sizeof(dbt_name));
   memset(&dbt_pnode, 0, sizeof(dbt_pnode));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_testdb->db->cursor(g_testdb->db, NULL, &namec, 0);
   wdb_check_err(ret);
   num=0;

   // Iterate over the main database
   while ( 0 == (ret = namec->c_get(namec, &dbt_name, &dbt_pnode, DB_NEXT)) ) {

      name = malloc(dbt_name.size + 1);
      memcpy(name, dbt_name.data, dbt_name.size);
      name[dbt_name.size] = 0;
      
      pnum = malloc(dbt_pnode.size + 1);
      memcpy(pnum, dbt_pnode.data, dbt_pnode.size);
      pnum[dbt_pnode.size] = 0;
      printf( "%s -> %s\n", name, pnum );
      num++;

      free(name), name = NULL;
      free(pnum), pnum = NULL;

   }

   printf("num:%d\n",num);
   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_testnew(void)
{
   DBC                  *namec = NULL;
   DBT                   dbt_name;
   DBT                   dbt_pnode;
   char                 *name;
   char                  *pnum;
   int                   ret;
   int                    num;

   printf( "----------------------------------------\n"
           "| Dumping guize                         |\n"
           "| name -> name                        |\n"
           "----------------------------------------\n" );

   memset(&dbt_name , 0, sizeof(dbt_name));
   memset(&dbt_pnode, 0, sizeof(dbt_pnode));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_testnewdb->db->cursor(g_testnewdb->db, NULL, &namec, 0);
   wdb_check_err(ret);
   num=0;

   // Iterate over the main database
   while ( 0 == (ret = namec->c_get(namec, &dbt_name, &dbt_pnode, DB_NEXT)) ) {

      name = malloc(dbt_name.size + 1);
      memcpy(name, dbt_name.data, dbt_name.size);
      name[dbt_name.size] = 0;
      
      pnum = malloc(dbt_pnode.size + 1);
      memcpy(pnum, dbt_pnode.data, dbt_pnode.size);
      pnum[dbt_pnode.size] = 0;
      printf( "%s -> %s\n", name, pnum );
      num++;

      free(name), name = NULL;
      free(pnum), pnum = NULL;

   }

   printf("num:%d\n",num);
   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void prune_guize(void)
{
   DBC                  *namec = NULL;
   DBT                   dbt_name;
   DBT                   dbt_pnode;
   char                 *name;
   char                  *pnum;
   int                   ret;
   int ret1;
   int                    num;

  
   memset(&dbt_name , 0, sizeof(dbt_name));
   memset(&dbt_pnode, 0, sizeof(dbt_pnode));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_guizedb->db->cursor(g_guizedb->db, NULL, &namec, 0);
   wdb_check_err(ret);
   num=0;

   // Iterate over the main database
   while ( 0 == (ret = namec->c_get(namec, &dbt_name, &dbt_pnode, DB_NEXT)) ) {

      name = malloc(dbt_name.size + 1);
      memcpy(name, dbt_name.data, dbt_name.size);
      name[dbt_name.size] = '\0';
      
      pnum = malloc(dbt_pnode.size + 1);
      memcpy(pnum, dbt_pnode.data, dbt_pnode.size);
      pnum[dbt_pnode.size] = '\0';
  //    printf( "%s -> %s\n", name, pnum );
      if(strstr(name,"mount")||strstr(name,"PIPE")||strstr(name,"tmp")||strstr(name,"/pulse-shm")||strstr(name,"lasagna")||strstr(name,".cache")||strstr(name,"./configure")||strstr(name,"./Configure")||strstr(pnum,"mount")||strstr(pnum,"PIPE")||strstr(pnum,"tmp")||strstr(pnum,"/pulse-shm")||strstr(pnum,"lasagna")||strstr(pnum,".cache")||strstr(pnum,"./configure")||strstr(pnum,"./Configure")||strstr(pnum,"Cache")||strstr(pnum,"cache")||strstr(pnum,"CACHE")||strstr(pnum,"pass")||strstr(pnum,"default")||strstr(name,"Cache")||strstr(name,"cache")||strstr(name,"CACHE")||strstr(name,"pass")||strstr(name,"default"))
     {
        ;
       // namec->c_del(namec,NULL);
        //getchar();
     }
     else if((name[0]=='/')&&(name[1]>='0'&&name[1]<='9')&&(name[2]>='0'&&name[2]<='9')&&(name[3]=='/'))
      ;
     else if((name[0]=='/')&&(name[1]>='0'&&name[1]<='9')&&(name[2]>='0'&&name[2]<='9')&&(name[3]>='0'&&name[3]<='9')&&(name[4]=='/'))
      ;
     else if((name[0]=='/')&&(name[1]>='0'&&name[1]<='9')&&(name[2]>='0'&&name[2]<='9')&&(name[3]>='0'&&name[3]<='9')&&(name[4]>='0'&&name[4]<='9')&&(name[5]=='/'))
      ;
     else if((name[0]=='/')&&(name[1]>='0'&&name[1]<='9')&&(name[2]>='0'&&name[2]<='9')&&(name[3]>='0'&&name[3]<='9')&&(name[4]>='0'&&name[4]<='9')&&(name[5]>='0'&&name[5]<='9')&&(name[6]=='/'))
     ;
     else if((pnum[0]=='/')&&(pnum[1]>='0'&&pnum[1]<='9')&&(pnum[2]>='0'&&pnum[2]<='9')&&(pnum[3]=='/'))
      ;
     else if((pnum[0]=='/')&&(pnum[1]>='0'&&pnum[1]<='9')&&(pnum[2]>='0'&&pnum[2]<='9')&&(pnum[3]>='0'&&pnum[3]<='9')&&(pnum[4]=='/'))
      ;
     else if((pnum[0]=='/')&&(pnum[1]>='0'&&pnum[1]<='9')&&(pnum[2]>='0'&&pnum[2]<='9')&&(pnum[3]>='0'&&pnum[3]<='9')&&(pnum[4]>='0'&&pnum[4]<='9')&&(pnum[5]=='/'))
      ;
     else if((pnum[0]=='/')&&(pnum[1]>='0'&&pnum[1]<='9')&&(pnum[2]>='0'&&pnum[2]<='9')&&(pnum[3]>='0'&&pnum[3]<='9')&&(pnum[4]>='0'&&pnum[4]<='9')&&(pnum[5]>='0'&&pnum[5]<='9')&&(pnum[6]=='/'))
     ;
      else
       ret1 = g_guizenewdb->db->put(g_guizenewdb->db, NULL,
                           &dbt_name, &dbt_pnode,
                           g_guizenewdb->put_flags);
      num++;

      free(name), name = NULL;
      free(pnum), pnum = NULL;

   }

  // printf("num:%d\n",num);
   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void compress_test(void)
{
   DBC                  *namec = NULL;
   DBT                   dbt_name;
   DBT                   dbt_pnode;
   char                 *name;
   char                  *pnum;
   int                   ret;
   int ret1;
   int                    num;

  

   memset(&dbt_name , 0, sizeof(dbt_name));
   memset(&dbt_pnode, 0, sizeof(dbt_pnode));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_testdb->db->cursor(g_testdb->db, NULL, &namec, 0);
   wdb_check_err(ret);
   num=0;
//printf("compress_test\n");
   // Iterate over the main database
   while ( 0 == (ret = namec->c_get(namec, &dbt_name, &dbt_pnode, DB_NEXT)) ) {

      name = malloc(dbt_name.size + 1);
      memcpy(name, dbt_name.data, dbt_name.size);
      name[dbt_name.size] = '\0';
      
      pnum = malloc(dbt_pnode.size + 1);
      memcpy(pnum, dbt_pnode.data, dbt_pnode.size);
      pnum[dbt_pnode.size] = '\0';
      
      if(strstr(name,"mount")||strstr(name,"PIPE")||strstr(name,"tmp")||strstr(name,"/pulse-shm")||strstr(name,"lasagna")||strstr(name,".cache")||strstr(name,"./configure")||strstr(name,"./Configure")||strstr(pnum,"mount")||strstr(pnum,"PIPE")||strstr(pnum,"tmp")||strstr(pnum,"/pulse-shm")||strstr(pnum,"lasagna")||strstr(pnum,".cache")||strstr(pnum,"./configure")||strstr(pnum,"./Configure")||strstr(pnum,"Cache")||strstr(pnum,"cache")||strstr(pnum,"CACHE")||strstr(pnum,"pass")||strstr(pnum,"default")||strstr(name,"Cache")||strstr(name,"cache")||strstr(name,"CACHE")||strstr(name,"pass")||strstr(name,"default"))
     {
        ;
       // namec->c_del(namec,NULL);
        //getchar();
     }
     else if((name[0]=='/')&&(name[1]>='0'&&name[1]<='9')&&(name[2]>='0'&&name[2]<='9')&&(name[3]=='/'))
      ;
     else if((name[0]=='/')&&(name[1]>='0'&&name[1]<='9')&&(name[2]>='0'&&name[2]<='9')&&(name[3]>='0'&&name[3]<='9')&&(name[4]=='/'))
      ;
     else if((name[0]=='/')&&(name[1]>='0'&&name[1]<='9')&&(name[2]>='0'&&name[2]<='9')&&(name[3]>='0'&&name[3]<='9')&&(name[4]>='0'&&name[4]<='9')&&(name[5]=='/'))
      ;
     else if((name[0]=='/')&&(name[1]>='0'&&name[1]<='9')&&(name[2]>='0'&&name[2]<='9')&&(name[3]>='0'&&name[3]<='9')&&(name[4]>='0'&&name[4]<='9')&&(name[5]>='0'&&name[5]<='9')&&(name[6]=='/'))
     ;
     else if((pnum[0]=='/')&&(pnum[1]>='0'&&pnum[1]<='9')&&(pnum[2]>='0'&&pnum[2]<='9')&&(pnum[3]=='/'))
      ;
     else if((pnum[0]=='/')&&(pnum[1]>='0'&&pnum[1]<='9')&&(pnum[2]>='0'&&pnum[2]<='9')&&(pnum[3]>='0'&&pnum[3]<='9')&&(pnum[4]=='/'))
      ;
     else if((pnum[0]=='/')&&(pnum[1]>='0'&&pnum[1]<='9')&&(pnum[2]>='0'&&pnum[2]<='9')&&(pnum[3]>='0'&&pnum[3]<='9')&&(pnum[4]>='0'&&pnum[4]<='9')&&(pnum[5]=='/'))
      ;
     else if((pnum[0]=='/')&&(pnum[1]>='0'&&pnum[1]<='9')&&(pnum[2]>='0'&&pnum[2]<='9')&&(pnum[3]>='0'&&pnum[3]<='9')&&(pnum[4]>='0'&&pnum[4]<='9')&&(pnum[5]>='0'&&pnum[5]<='9')&&(pnum[6]=='/'))
     ;
      else
      {
       ret1 = g_testnewdb->db->put(g_testnewdb->db, NULL,
                           &dbt_name, &dbt_pnode,
                           g_testnewdb->put_flags);
      
    ///  num++;
    //  printf( "%s -> %s\n", name, pnum );
      }free(name), name = NULL;
      free(pnum), pnum = NULL;

   }

   //printf("num:%d\n",num);
   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

void dump_ky2(void)
{
   DBC                  *p2ic = NULL;
   DBT                   dbt_inode;
   DBT                   dbt_pnode;
   pnode_t              inode;
   pnode_t               pnum;
   int                   ret;

   printf( "----------------------------------------\n"
           "| Dumping ky2                          |\n"
           "| c -> p                          |\n"
           "----------------------------------------\n" );

   memset(&dbt_inode, 0, sizeof(dbt_inode));
   memset(&dbt_pnode, 0, sizeof(dbt_pnode));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_ky2db->db->cursor(g_ky2db->db, NULL, &p2ic, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = p2ic->c_get(p2ic, &dbt_pnode, &dbt_inode, DB_NEXT)) ) {
      assert( sizeof(inode) == dbt_inode.size );
      assert( sizeof(pnum)  == dbt_pnode.size );

      memcpy(&inode, dbt_inode.data, dbt_inode.size);
      memcpy(&pnum , dbt_pnode.data, dbt_pnode.size);

      printf( "%llu <- %llu\n", pnum, inode );
   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
}

//ret=1,not find
int find(char * pname ,char * cname)
{
   DBC                       *guizec =NULL;
   DBT                        dbt_guize_key;
   DBT                        dbt_guize_val;
   char                        *c;
   int                           ret,r=1;


   memset(&dbt_guize_key, 0, sizeof(dbt_guize_key));
   memset(&dbt_guize_val, 0, sizeof(dbt_guize_val));
 
   ret = g_guizenewdb->db->cursor(g_guizenewdb->db, NULL, &guizec, 0);

   wdb_check_err(ret);

 
   dbt_guize_key.data = pname;
 
   dbt_guize_key.size = strlen(pname)+1;
 
   ret = guizec->c_get(guizec, &dbt_guize_key, &dbt_guize_val, DB_SET);

   while ( ret != DB_NOTFOUND ) {
    c = malloc(dbt_guize_val.size);
    memcpy(c, dbt_guize_val.data, dbt_guize_val.size);
  //  c[dbt_guize_val.size] = '\0';
   // printf("%s\n%s\n",c,cname);
    if( strcmp(c, cname) ==0 ) {r=0;free(c);c=NULL;break;}
    ret = guizec->c_get(guizec, &dbt_guize_key, &dbt_guize_val, DB_NEXT_DUP);
    free(c),c=NULL;
   }
   return r;

}

int online_detect(char* pname)
{
//printf("Insert: %llu->%llu\n",pname,cname);
   DBC                  *ky1c = NULL;
   DBT                   dbt_name;
   DBT                   dbt_pnode;
   char *               c;

   int                     ret;
   int r=0;
   int aa=0;
   int x=0;
 //  char *  pname=NULL;
   char cname[200];
   char msg[100];
   int ii_detect;
   FILE *file_intrusion;

   if((len+1)>W) return;
   memset(&dbt_name, 0, sizeof(dbt_name));
   memset(&dbt_pnode , 0, sizeof(dbt_pnode));

   // Open a cursor
   ret = g_testnewdb->db->cursor(g_testnewdb->db, NULL, &ky1c, 0);
   wdb_check_err(ret);
 
   dbt_name.data = pname;
   dbt_name.size = strlen(pname)+1;
  
   ret = ky1c->c_get(ky1c, &dbt_name, &dbt_pnode, DB_SET);

   while ( ret != DB_NOTFOUND ) {

    len++;
    //cname = malloc(dbt_pnode.size);
    memcpy(cname, dbt_pnode.data, dbt_pnode.size);
    cname[dbt_pnode.size]='\0';
  

    if(len<=W)
    {
 
     
     //  printf("pname is %s,cname is %s\n",pname,cname);
        if(pname==NULL||cname==NULL) 
         {
            r=0;
         
           }
       else
       {
        r= find(pname,cname);//pipeishifoucunzai
 //       printf("onlinefind:%s->%s\n",pname,cname );
//        jiance_time++;

       }
       
       aa=depi;
       if(r)   
         {
       //     printf("when r!=0,pname is %s,cname is %s.\n",pname,cname);
            
            depi+=1;
           }
       else   depi+=0;


       if (len==W)
       {
         flag+=1;
  
         if((float)depi/(float)(W)>T)
           {
           ii_intrusion+=1;
         //  getchar();
        
              }
            //printf();
          //  if(cname!=NULL)
         // {
         //   free(cname);
          //  cname=NULL;
           // printf("good22\n");
         // }
          return;
        }

    }
   
    else
      return;

    x=len;
    online_detect(cname);
 
    ret = ky1c->c_get(ky1c, &dbt_name, &dbt_pnode, DB_NEXT_DUP);
   
    depi=aa;
    len=x-1;
   }
   
   return 0;
}

void jiance(void)
{
   DBC                  *jchildc = NULL;
   DBT                   dbt_name;
   DBT                   dbt_pnode;
   provdb_key            jchild;
   provdb_key            jparent;
   int                   ret;
   int  r1;
   int  r;
   char msg[100];
   int ii_detect;
   FILE *file_intrusion;
   FILE *file_time;
   char* pname=NULL;
   char* cname=NULL;
   char* name;
   char* pnum;
   char intime[50];
   
   
   struct timeval start_read,end_read;
	float timeuse_read;
   test();
//  printf("testover!\n");
   compress_test();
  

   gettimeofday(&start_read,NULL);   

   memset(&dbt_name, 0, sizeof(dbt_name));
   memset(&dbt_pnode, 0, sizeof(dbt_pnode));

   // Open a cursor
   ret = g_testnewdb->db->cursor(g_testnewdb->db, NULL, &jchildc, 0);//父名->子名
   wdb_check_err(ret);
 

   // Iterate over the main database
   while ( 0 == (ret = jchildc->c_get(jchildc, &dbt_name, &dbt_pnode, DB_NEXT)) ) {
     name = malloc(dbt_name.size);
      memcpy(name, dbt_name.data, dbt_name.size);
    
      
      pnum = malloc(dbt_pnode.size);
      memcpy(pnum, dbt_pnode.data, dbt_pnode.size);
  
   //   printf( "%s -> %s\n", name, pnum );//name父，pnum子

      len++;
  
         if(name==NULL||pnum==NULL) 
         {
            r=0;
         
           }
       else
       {
    //  jiance_time++;
       r= find(name,pnum);//pipeishifoucunzai
  //      printf("jiancefind:%s->%s\n",name,pnum );
      }
   
       if(r)   
         {
    //        printf("when r!=0,jiancepname is %s,cname is %s.\n",name,pnum);
           
            depi+=1;
           }
       else   depi+=0;
    //  printf("good1\n");
        if(name!=NULL)
       {
        free(name);
        name=NULL;
        }
    //   printf("good in jiance\n"); 
      r1= online_detect(pnum);//放子名
         if(pnum!=NULL)
       {
        free(pnum);
        pnum=NULL;
       }
     // getchar();
      len=0;
      depi=0;
 
   }
 

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }

   if(ii_intrusion==0)  
      ii_detect=0;
   else
      ii_detect=1;

   gettimeofday(&end_read,NULL);
   timeuse_read=1000000*(end_read.tv_sec-start_read.tv_sec)+end_read.tv_usec-start_read.tv_usec;

//DB  printf("检测次数：%d\n",jiance_time);
 //  sprintf(msg,"%d\n",ii_detect);
   sprintf(msg,"%d\n%d\n%d\n",ii_detect,ii_intrusion,flag);
   sprintf(intime,"%f\n",timeuse_read);
   file_intrusion=fopen("/home/intrusionpidasdetect.txt","a+");
	file_time=fopen("/home/intrusionpidastime.txt","a+");
   if(file_intrusion==NULL)
      printf("get file error.\n");

   fwrite(msg,sizeof(char),strlen(msg),file_intrusion);
   fclose(file_intrusion);
fwrite(intime,sizeof(char),strlen(intime),file_time);
   fclose(file_time);

}


void test(void)
{
   DBC                  *childc = NULL;
   DBT                   dbt_child;
   DBT                   dbt_parent;
   provdb_key            child;
   provdb_key            parent;
   int                   ret,ret1;
//   int count=0;
   char *  pname;
   char *  cname;

   memset(&dbt_child, 0, sizeof(dbt_child));
   memset(&dbt_parent , 0, sizeof(dbt_parent));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_childdb->db->cursor(g_childdb->db, NULL, &childc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = childc->c_get(childc, &dbt_parent, &dbt_child, DB_NEXT)) ) {
      assert( sizeof(child) == dbt_child.size );
      assert( sizeof(parent)  == dbt_parent.size );

      memcpy(&child , dbt_child.data , dbt_child.size );
      memcpy(&parent, dbt_parent.data, dbt_parent.size);
      
     pname = get_string(parent.pnum,"NAME");
     if(!pname)
        {
            pname = get_string(parent.pnum,"TYPE");
        if(!pname) continue;
     }
     cname = get_string(child.pnum, "NAME");
     if(!cname)
        {
            cname = get_string(child.pnum,"TYPE");
        if(!cname) 
        {free(pname),pname=NULL;
         continue;}
     }

   DBT                        dbt_guize_key;
   DBT                        dbt_guize_val;

   memset(&dbt_guize_key, 0, sizeof(dbt_guize_key));
   memset(&dbt_guize_val, 0, sizeof(dbt_guize_val));

   dbt_guize_key.data = pname;
   dbt_guize_key.size = strlen(pname)+1;

   dbt_guize_val.data = cname;
   dbt_guize_val.size = strlen(cname)+1;

   ret1 = g_testdb->db->put(g_testdb->db, NULL,
                           &dbt_guize_key, &dbt_guize_val,
                           g_testdb->put_flags);
//   count++;
//  printf( "test:%s -> %s\n", pname, cname );
   if ( ret1 == DB_KEYEXIST ) {
#ifdef DEBUG
       fprintf( stderr,
            "name index: key %*s for pnode %llu already recorded\n",
                length, val->pdb_data, key->pnum );
#endif /* DEBUG */
   } else {
       wdb_check_err(ret1);
   }

     free(cname),cname=NULL;
     free(pname),pname=NULL;

   }
 //  printf("count:%d\n",count);
   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }



}

void guize(void)
{
   DBC                  *childc = NULL;
   DBT                   dbt_child;
   DBT                   dbt_parent;
   provdb_key            child;
   provdb_key            parent;
   int                   ret,ret1;
   char *  pname;
   char *  cname;

   memset(&dbt_child, 0, sizeof(dbt_child));
   memset(&dbt_parent , 0, sizeof(dbt_parent));

   //
   // Process the entries
   //

   // Open a cursor
   ret = g_childdb->db->cursor(g_childdb->db, NULL, &childc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = childc->c_get(childc, &dbt_parent, &dbt_child, DB_NEXT)) ) {
      assert( sizeof(child) == dbt_child.size );
      assert( sizeof(parent)  == dbt_parent.size );

      memcpy(&child , dbt_child.data , dbt_child.size );
      memcpy(&parent, dbt_parent.data, dbt_parent.size);

     pname = get_string(parent.pnum,"NAME");
     if(!pname)
        {
            pname = get_string(parent.pnum,"TYPE");
        if(!pname) continue;
     }
     cname = get_string(child.pnum, "NAME");
     if(!cname)
        {
            cname = get_string(child.pnum,"TYPE");
        if(!cname) 
        {free(pname),pname=NULL;
         continue;}
     }

   DBT                        dbt_guize_key;
   DBT                        dbt_guize_val;

   memset(&dbt_guize_key, 0, sizeof(dbt_guize_key));
   memset(&dbt_guize_val, 0, sizeof(dbt_guize_val));

   dbt_guize_key.data = pname;
   dbt_guize_key.size = strlen(pname)+1;

   dbt_guize_val.data = cname;
   dbt_guize_val.size = strlen(cname)+1;

   ret1 = g_guizedb->db->put(g_guizedb->db, NULL,
                           &dbt_guize_key, &dbt_guize_val,
                           g_guizedb->put_flags);

   if ( ret1 == DB_KEYEXIST ) {
#ifdef DEBUG
       fprintf( stderr,
            "name index: key %*s for pnode %llu already recorded\n",
                length, val->pdb_data, key->pnum );
#endif /* DEBUG */
   } else {
       wdb_check_err(ret1);
   }

     free(cname),cname=NULL;
     free(pname),pname=NULL;

   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }



}


void modify_graph_construct(void)
{
   char detect_file[100];
   provdb_key            child;
   provdb_key            parent;
   DBT                   dbt_name;
   DBT                   dbt_pnode;
   DBT                   dbt_child;
   DBT                   dbt_parent;
   unsigned long long nodenum;
   DBC                  *parentc = NULL;
   int ret;
    struct timeval start,end;
     float t;



   memset(detect_file,0,100);

   memset(&dbt_name, 0, sizeof(dbt_name));
   memset(&dbt_pnode , 0, sizeof(dbt_pnode));
   memset(&dbt_child, 0, sizeof(dbt_child));
   memset(&dbt_parent , 0, sizeof(dbt_parent));

   printf("Please input the name of the detect file:\n");
   scanf("%s",detect_file);
   printf("detect_file=%s",detect_file);

   gettimeofday(&start,NULL);

   dbt_name.data = detect_file;
   dbt_name.size = strlen(detect_file);

   ret = g_namedb->db->get(g_namedb->db, NULL, &dbt_name, &dbt_pnode, 0);

   memcpy(&nodenum, dbt_pnode.data, dbt_pnode.size);
  // printf( "%s -> %llu\n", detect_file, (unsigned long long)nodenum);

   // Open a cursor
   ret = g_parentdb->db->cursor(g_parentdb->db, NULL, &parentc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = parentc->c_get(parentc, &dbt_child, &dbt_parent, DB_NEXT)) ) {
    //  printf("xunhuan\n");
      assert( sizeof(child) == dbt_child.size );
      assert( sizeof(parent)  == dbt_parent.size );

      memcpy(&child , dbt_child.data , dbt_child.size );
      memcpy(&parent, dbt_parent.data, dbt_parent.size);
    //  printf("child.pnum=%llu\n",child.pnum);

      if (child.pnum==nodenum)
     {
     // printf("if\n");
      Put_Provenance_Graph(&child,&parent);
      if(parent.pnum != child.pnum){
      FindParent_new(&parent);
      }
     }

   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }
   gettimeofday(&end,NULL);
     t = 1000000*(end.tv_sec-start.tv_sec)+end.tv_usec-start.tv_usec;
printf("======Search Time: %f microseconds , %f s ======\n",t,t/1000000);

}

void revert_graph_construct(void)
{

   provdb_key            child;
   //unsigned long pnum;
   //unsigned int version;

   //printf("Please input the pnum of the network socket:\n");
   //scanf("%lu\n",&pnum);

   //printf("Please input the version of the network socket:\n");
   //scanf("%u\n",&version);
 struct timeval start,end;
     float t;

 gettimeofday(&start,NULL);
   child.pnum = 3;
   child.version = 0;

   FindParent_new(&child);
 gettimeofday(&end,NULL);
     t = 1000000*(end.tv_sec-start.tv_sec)+end.tv_usec-start.tv_usec;
printf("======Search Time: %f microseconds , %f s ======\n",t,t/1000000);

}

void all_graph_construct(void)
{

   provdb_key            child;
   provdb_key            parent;
   DBT                   dbt_child;
   DBT                   dbt_parent;
   DBC                  *parentc = NULL;
   char *pname;
   int ret;
    struct timeval start,end;
     float t;
  gettimeofday(&start,NULL);

   memset(&dbt_child, 0, sizeof(dbt_child));
   memset(&dbt_parent , 0, sizeof(dbt_parent));

   // Open a cursor
   ret = g_parentdb->db->cursor(g_parentdb->db, NULL, &parentc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = parentc->c_get(parentc, &dbt_child, &dbt_parent, DB_NEXT)) ) {
      
     // printf("all_graph_construct FAIL\n");
      assert( sizeof(child) == dbt_child.size );
      assert( sizeof(parent)  == dbt_parent.size );

      memcpy(&child , dbt_child.data , dbt_child.size );
      memcpy(&parent, dbt_parent.data, dbt_parent.size);
   //   printf("%s \n",pname);
      
     if(get_string(child.pnum,"TYPE")!=NULL)
   {
        pname = get_string(child.pnum,"TYPE");

     // printf("child.pnum=%llu\n",child.pnum);
     // printf("pname=%s\n",pname);


      if(strcmp(pname,"SOCKET")==0)
     {
    //  printf("all_graph_construct over\n");
       
      FindParent_new(&parent);
   //   printf("bad\n");
     }
     free(pname),pname=NULL;
     //printf("all_graph_construct FAIL2\n");

   }
}
    gettimeofday(&end,NULL);
     t = 1000000*(end.tv_sec-start.tv_sec)+end.tv_usec-start.tv_usec;
printf("======Search Time: %f microseconds , %f s ======\n",t,t/1000000);
     // printf("over\n");
}

void FindParent(provdb_key *child)
{
   DBC                  *pdbc = NULL;
   DBT                   dbt_child;
   DBT                   dbt_parent;
   provdb_key            parent;
   unsigned long long nodenum;
   int                   ret,ret1;


   memset(&dbt_child, 0, sizeof(dbt_child));
   memset(&dbt_parent , 0, sizeof(dbt_parent));

   dbt_child.data = child;
   dbt_child.size = sizeof(*child);
   // Open a cursor
   ret = g_parentdb->db->cursor(g_parentdb->db, NULL, &pdbc, 0);
   wdb_check_err(ret);

   ret =  pdbc->c_get(pdbc, &dbt_child, &dbt_parent, DB_SET);
  

   while ( ret != DB_NOTFOUND ) {
     assert( sizeof(parent)  == dbt_parent.size );
    memcpy(&parent, dbt_parent.data, dbt_parent.size);

     printf("parent1\n");
      printf("parent.pnum=%llu\n",parent.pnum);
    //  printf("child=%llu\n",dbt_child.data);
      Put_Provenance_Graph(child,&parent);
     printf("parent2\n");
    if(parent.pnum != child->pnum){
      FindParent(&parent);
      }

    ret= pdbc->c_get(pdbc, &dbt_child, &dbt_parent, DB_NEXT_DUP);
  
   }

}

void FindParent_new(provdb_key *child1)
{
   DBC                  *pdbc = NULL;
   DBC                  *parentc = NULL;
   DBT                   dbt_child;
   DBT                   dbt_parent;
   provdb_key            child;
   provdb_key            parent;
   unsigned long long nodenum;
   int                   ret,ret1;


   memset(&dbt_child, 0, sizeof(dbt_child));
   memset(&dbt_parent , 0, sizeof(dbt_parent));

    // Open a cursor
   ret = g_parentdb->db->cursor(g_parentdb->db, NULL, &parentc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = parentc->c_get(parentc, &dbt_child, &dbt_parent, DB_NEXT)) ) {
    //  printf("xunhuan\n");
      assert( sizeof(child) == dbt_child.size );
      assert( sizeof(parent)  == dbt_parent.size );

      memcpy(&child , dbt_child.data , dbt_child.size );
      memcpy(&parent, dbt_parent.data, dbt_parent.size);
     // printf("child.pnum=%llu,parent.pnum=%llu\n",child.pnum,parent.pnum);

      if (child.pnum==child1->pnum)
     {
     // printf("if2\n");
      Put_Provenance_Graph(&child,&parent);
      if(parent.pnum != child.pnum){
      FindParent_new(&parent);
       }
     }

   }

   if ( DB_NOTFOUND != ret ) {
      wdb_check_err(ret);
   }

}



void read_graph_construct(void)
{
   char detect_file[100];
   provdb_key            child;
   provdb_key            parent;
   DBT                   dbt_name;
   DBT                   dbt_pnode;
   DBT                   dbt_child;
   DBT                   dbt_parent;
   unsigned long long nodenum;
   DBC                  *childc = NULL;
   int ret;
   struct timeval start,end;
     float t;

 gettimeofday(&start,NULL);


 //  memset(detect_file,0,100);

 //  memset(&dbt_name, 0, sizeof(dbt_name));
 //  memset(&dbt_pnode , 0, sizeof(dbt_pnode));
   memset(&dbt_child, 0, sizeof(dbt_child));
   memset(&dbt_parent , 0, sizeof(dbt_parent));

/**   printf("Please input the name of the detect file:\n");
   scanf("%s",detect_file);

 /**  dbt_name.data = detect_file;
   dbt_name.size = strlen(detect_file);

   ret = g_namedb->db->get(g_namedb->db, NULL, &dbt_name, &dbt_pnode, 0);

   memcpy(&nodenum, dbt_pnode.data, dbt_pnode.size);
   printf( "%s -> %llu\n", detect_file, (unsigned long long)nodenum);
*/
   // Open a cursor
   ret = g_childdb->db->cursor(g_childdb->db, NULL, &childc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = childc->c_get(childc, &dbt_parent, &dbt_child, DB_NEXT)) ) {
      assert( sizeof(child) == dbt_child.size );
      assert( sizeof(parent)  == dbt_parent.size );

      memcpy(&child , dbt_child.data , dbt_child.size );
      memcpy(&parent, dbt_parent.data, dbt_parent.size);

      if (parent.pnum==203)
     {
      Put_Provenance_Graph(&child,&parent);
      FindChild_new(&child);
     }

   }

gettimeofday(&end,NULL);
     t = 1000000*(end.tv_sec-start.tv_sec)+end.tv_usec-start.tv_usec;
printf("======Search Time: %f microseconds , %f s ======\n",t,t/1000000);

   return 0;

}

void FindChild(provdb_key *parent)
{
   DBC                  *pdbc = NULL;
   DBT                   dbt_child;
   DBT                   dbt_parent;
   provdb_key            child;
   int                   ret,ret1;
   unsigned long long nodenum;

   memset(&dbt_child, 0, sizeof(dbt_child));
   memset(&dbt_parent , 0, sizeof(dbt_parent));

   dbt_parent.data = parent;
   dbt_parent.size = sizeof(*parent);
   // Open a cursor
   ret = g_childdb->db->cursor(g_childdb->db, NULL, &pdbc, 0);
   wdb_check_err(ret);

   ret =  pdbc->c_get(pdbc, &dbt_parent, &dbt_child, DB_SET);
   memcpy(&nodenum, dbt_parent.data, dbt_parent.size);
   printf( "nodenum=%llu\n", (unsigned long long)nodenum);

   while ( ret != DB_NOTFOUND ) {
     assert( sizeof(child)  == dbt_child.size );
    memcpy(&child, dbt_child.data, dbt_child.size);

      Put_Provenance_Graph(&child,parent);
    if(child.pnum != nodenum){
      FindChild(&child);
}

    ret = pdbc->c_get(pdbc, &dbt_parent, &dbt_child, DB_NEXT_DUP);
   }

}


void FindChild_new(provdb_key *parent1)
{
   DBC                  *childc = NULL;
   DBT                   dbt_child;
   DBT                   dbt_parent;
   provdb_key            child;
   provdb_key            parent;
   int                   ret,ret1;
   unsigned long long nodenum;

   memset(&dbt_child, 0, sizeof(dbt_child));
   memset(&dbt_parent , 0, sizeof(dbt_parent));
   
   ret = g_childdb->db->cursor(g_childdb->db, NULL, &childc, 0);
   wdb_check_err(ret);

   // Iterate over the main database
   while ( 0 == (ret = childc->c_get(childc, &dbt_parent, &dbt_child, DB_NEXT)) ) {
      assert( sizeof(child) == dbt_child.size );
      assert( sizeof(parent)  == dbt_parent.size );

      memcpy(&child , dbt_child.data , dbt_child.size );
      memcpy(&parent, dbt_parent.data, dbt_parent.size);
      //printf( "parent.pnum=%llu\n", parent.pnum);

      if (parent.pnum==parent1->pnum)
     {
      Put_Provenance_Graph(&child,&parent);
      if(child.pnum != parent.pnum){
      FindChild_new(&child);
      }
     }

   }

}



void  Put_Provenance_Graph(provdb_key *child,provdb_key *parent)
{
   DBT                        dbt_graph_key;
   DBT                        dbt_graph_val;
   int ret;
   //   printf("Put_Provenance_Graph\n");
   memset(&dbt_graph_key, 0, sizeof(dbt_graph_key));
   memset(&dbt_graph_val, 0, sizeof(dbt_graph_val));
//printf("1\n");
   dbt_graph_key.data = child;
   dbt_graph_key.size = sizeof(*child);

//printf("2\n");
   dbt_graph_val.data = parent;
   dbt_graph_val.size = sizeof(*parent);
//printf("3\n");

   ret = g_graphdb->db->put(g_graphdb->db, NULL,
                           &dbt_graph_key, &dbt_graph_val,
                           g_graphdb->put_flags);
//printf("4\n");

}

static char ** get_multistring(pnode_t pnum, const char * attr)
{
   DBT    dbt_tnum;
   DBT    dbt_token;
   char ** argv;
   char *   element;
   int         ntoks, i;
   tnum_t * toks;

   DBC                       *pdbc;
   DBT                        dbt_key;
   DBT                        dbt_val;
   provdb_key                 key;
   provdb_val                *val=NULL;
   const char                *attrname;
   char                      *data = NULL;
   size_t                     datalen=0;
   int                        serrno;
   int                        ret;

   memset(&dbt_key, 0, sizeof(dbt_key));
   memset(&dbt_val, 0, sizeof(dbt_val));

   ret = g_provdb->db->cursor(g_provdb->db, NULL, &pdbc, 0);
   wdb_check_err(ret);

   key.pnum = pnum;
   key.version = 0;
   dbt_key.data = &key;
   dbt_key.size = sizeof(key);
   ret =  pdbc->c_get(pdbc, &dbt_key, &dbt_val, DB_SET);
   while ( ret != DB_NOTFOUND ) {
    val = malloc(dbt_val.size);
    if ( val == NULL ) {
    serrno = errno;
    fprintf( stderr, "Malloc of provdb record failed %s\n",
         strerror(serrno) );
    errno = serrno;
    return 0;
    }
    memcpy(val, dbt_val.data,dbt_val.size);
    attrname = PROVDB_VAL_ATTR(val);
    data = PROVDB_VAL_VALUE(val);
       datalen = PROVDB_VAL_VALUELEN(val);
    if( strncmp(attr, attrname, 3) ==0 ) break;
    else ret = pdbc->c_get(pdbc, &dbt_key, &dbt_val, DB_NEXT_DUP);
    free(val),val = NULL;
   }

   if ( data == NULL ) return NULL;
   
   ntoks = datalen / sizeof(tnum_t);
   toks = malloc(datalen);
   memcpy(toks, data, datalen);
   free(val),val = NULL;
   argv = (char **)malloc(sizeof(char *)*(ntoks+1));

   for(i=0; i<ntoks;i++) {
    memset(&dbt_tnum, 0, sizeof(dbt_tnum));
    memset(&dbt_token, 0, sizeof(dbt_token));
    dbt_tnum.data = &toks[i];
    dbt_tnum.size = sizeof(toks[i]);
    ret = g_tnum2tokdb->db->get(g_tnum2tokdb->db, NULL,&dbt_tnum, &dbt_token,0);
    wdb_check_err(ret);
    element = (char *)malloc(dbt_token.size+1);
    memcpy(element, dbt_token.data, dbt_token.size);
    element[dbt_token.size] = '\0';
    argv[i] = element;
   }
   free(toks),toks = NULL;
   argv[i] = NULL;
   return argv;
}

static char * get_string(pnode_t pnum, const char * attr)
{
   DBC                       *pdbc = NULL;
   DBT                        dbt_key;
   DBT                        dbt_val;
   provdb_key                 key;
   provdb_val                *val;
   const char                *attrname;
   char                      *data, *attrval = NULL;
   size_t                     datalen;
   int                        serrno;
   int                        ret;
   struct waldo_db * tmp=g_provdb;
   memset(&dbt_key, 0, sizeof(dbt_key));
   memset(&dbt_val, 0, sizeof(dbt_val));

   if(j) tmp=g_provdb;
   ret = tmp->db->cursor(tmp->db, NULL, &pdbc, 0);
   wdb_check_err(ret);

   key.pnum = pnum;
   key.version = 0;
   dbt_key.data = &key;
   dbt_key.size = sizeof(key);
   ret =  pdbc->c_get(pdbc, &dbt_key, &dbt_val, DB_SET);
   while ( ret != DB_NOTFOUND ) {
    val = malloc(dbt_val.size);
    if ( val == NULL ) {
    serrno = errno;
    fprintf( stderr, "Malloc of provdb record failed %s\n",
         strerror(serrno) );
    errno = serrno;
    return 0;
    }
    memcpy(val, dbt_val.data,dbt_val.size);
    attrname = PROVDB_VAL_ATTR(val);
    data = PROVDB_VAL_VALUE(val);
       datalen = PROVDB_VAL_VALUELEN(val);
    if( strncmp(attr, attrname, 4) ==0 ) {
        attrval = (char *)malloc(datalen+1);
        memcpy(attrval, data, datalen);
        attrval[datalen] = '\0';
        free(val),val = NULL;
        break;
    }
    else ret = pdbc->c_get(pdbc, &dbt_key, &dbt_val, DB_NEXT_DUP);
    free(val),val = NULL;
   }
   //if(pnum==3)
      //printf("attrval=%s\n",attrval);
   return attrval;
}
