/*
 * Copyright 2006, 2007, 2010
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <getopt.h>
#include <assert.h>
#include <err.h>

#include <db.h>

#include "schema.h"
#include "wdb.h"

/* possible export formats */
enum formats {
   F_TEXT,
   F_PASSRDF,
};

/* non-null-terminated string that may contain \0's */
struct string {
   char *data;
   size_t len;
};

/* this is a copy of PROV_TYPE_* but as an enum for better compiler checks */
enum valtypes {
   V_NIL,
   V_STRING,
   V_MULTISTRING,
   V_INT,
   V_REAL,
   V_TIMESTAMP,
   V_INODE,
   V_PNODE,
   V_PNODEVERSION,
};
struct triple {
   /* subject position: pnode and version */
   struct pnode_version pv;

   /* predicate position: is-ancestry flag and attribute name */
   bool isancestry;
   char *attrib;

   /* object position: value */
   enum valtypes valtype;
   union {
      /* nil - nothing */
      struct string val_string;
      struct {
	 unsigned num;
	 struct string *v;
      } val_multistring;
      int val_int;
      double val_real;
      struct prov_timestamp val_timestamp;
      uint32_t /*XXX?*/ val_inode;
      pnode_t val_pnode;
      struct pnode_version val_pnodeversion;
   };
};

/* export format */
static enum formats g_format;

/* behavior options - version suppression and edge lowercasing */
static bool g_noversions;
static bool g_lowercase;

////////////////////////////////////////////////////////////
// libc wrappers

/*
 * For now, we use our own implementation of strnlen, since it's not
 * clear what ifdefs we need here to check if it's missing from libc.
 */
static size_t wdb_strnlen(const char *s, size_t maxlen) {
   const char *t = memchr(s, 0, maxlen);
   return t ? (size_t)(t-s) : maxlen;
}

static void *domalloc(size_t len) {
   void *ptr;

   ptr = malloc(len);
   if (ptr == NULL) {
      err(1, "malloc");
   }
   return ptr;
}

static char *dostrdup(const char *s) {
   char *ret;

   ret = strdup(s);
   if (ret == NULL) {
      err(1, "strdup");
   }
   return ret;
}

static void strlower(char *s) {
   size_t i;

   for (i=0; s[i]; i++) {
      s[i] = tolower((unsigned char)s[i]);
   }
}

////////////////////////////////////////////////////////////
// string and triple ops

static void string_init(struct string *s) {
   s->data = NULL;
   s->len = 0; 
}

static void string_cleanup(struct string *s) {
   free(s->data);
}

static void makestring(const char *s, size_t len, struct string *ret) {
   ret->len = len;
   ret->data = domalloc(len);
   memcpy(ret->data, s, len);
}

static void triple_init(struct triple *t) {
   t->pv.pnum = 0;
   t->pv.version = 0;
   t->isancestry = false;
   t->attrib = NULL;
   t->valtype = V_NIL;
}

static void triple_cleanup(struct triple *t) {
   unsigned i;

   free(t->attrib);

   switch (t->valtype) {
    case V_NIL:
     break;
    case V_STRING:
     string_cleanup(&t->val_string);
     break;
    case V_MULTISTRING:
     for (i=0; i<t->val_multistring.num; i++) {
	string_cleanup(&t->val_multistring.v[i]);
     }
     free(t->val_multistring.v);
     break;
    case V_INT:
    case V_REAL:
    case V_TIMESTAMP:
    case V_INODE:
    case V_PNODE:
    case V_PNODEVERSION:
     break;
   }
}

static void triple_set_multistring(struct triple *t, unsigned num) {
   unsigned i;

   t->valtype = V_MULTISTRING;
   t->val_multistring.num = num;
   t->val_multistring.v = domalloc(num * sizeof(*t->val_multistring.v));
   for (i=0; i<num; i++) {
      string_init(&t->val_multistring.v[i]);
   }
}

////////////////////////////////////////////////////////////
// common printing tools

static void printpv(const struct pnode_version *pv) {
   if (g_noversions) {
      printf("%llu", (unsigned long long)pv->pnum);
   }
   else {
      printf("%llu.%u", (unsigned long long)pv->pnum, pv->version);
   }
}

////////////////////////////////////////////////////////////
// text export

static void printstring(const struct string *s) {
   size_t i;

   putchar('"');
   for (i=0; i<s->len; i++) {
      unsigned char c = s->data[i];
      switch (c) {
       case '"':
       case '\\':
	putchar('\\');
	putchar(c);
	break;
       case '\a': putchar('\\'); putchar('a'); break;
       case '\b': putchar('\\'); putchar('b'); break;
       case '\t': putchar('\\'); putchar('t'); break;
       case '\n': putchar('\\'); putchar('n'); break;
       case '\v': putchar('\\'); putchar('v'); break;
       case '\f': putchar('\\'); putchar('f'); break;
       case '\r': putchar('\\'); putchar('r'); break;
       default:
	if (isascii(c) && (isprint(c) || isspace(c))) {
	   putchar(c);
	}
	else {
	   printf("\\%03o", c);
	}
	break;
      }
   }
   putchar('"');
}

static void export_triple_text(const struct triple *t) {
   unsigned i;

   printpv(&t->pv);

   if (t->isancestry) {
      printf(" <%s>", t->attrib);
   }
   else {
      printf(" %s", t->attrib);
   }

   switch (t->valtype) {
    case V_NIL:
     printf(" ---\n");
     break;
    case V_STRING:
     putchar(' ');
     printstring(&t->val_string);
     break;
    case V_MULTISTRING:
     for (i=0; i<t->val_multistring.num; i++) {
	putchar(' ');
	printstring(&t->val_multistring.v[i]);
     }
     break;
    case V_INT:
     printf(" %d\n", t->val_int);
     break;
    case V_REAL:
     printf(" %g\n", t->val_real);
     break;
    case V_TIMESTAMP:
     printf(" %u.%09u\n", t->val_timestamp.pt_sec, t->val_timestamp.pt_nsec);
     break;
    case V_INODE:
     printf(" %lu\n", (unsigned long)t->val_inode);
     break;
    case V_PNODE:
     printf(" %llu\n", (unsigned long long)t->val_pnode);
     break;
    case V_PNODEVERSION:
     putchar(' ');
     printpv(&t->val_pnodeversion);
     break;
   }
   putchar('\n');
}

////////////////////////////////////////////////////////////
// passrdf export

static void rdf_printpv(const struct pnode_version *pv) {
   if (g_noversions || pv->version == 0) {
      printf("thistrace:%llu", (unsigned long long)pv->pnum);
   }
   else {
      printf("thistrace:%llu_%u", (unsigned long long)pv->pnum, pv->version);
   }
}

static void rdf_printstring(const struct string *s) {
   size_t i;

   putchar('"');
   for (i=0; i<s->len; i++) {
      unsigned char c = s->data[i];
      if (c == '"' || c == '&' || c<32 || c>126) {
	 printf("&#%02x;", c);
      }
      else {
	 putchar(c);
      }
   }
   putchar('"');
}

static void export_header_passrdf(void) {
   static const char const passurl[] =
      "http://www.eecs.harvard.edu/syrah/pass";

   printf("@prefix ancestry: <%s/rdf/ancestry/> .\n", passurl);
   printf("@prefix identity: <%s/rdf/identity/> .\n", passurl);
   printf("@prefix passtypes: <%s/rdf/types/> .\n", passurl);
   printf("@prefix index: <%s/rdf/index/> .\n", passurl);

   /* this is probably abusive */
   printf("@prefix thistrace: <%s/rdf/thistrace/> .\n", passurl);
}

static void export_triple_passrdf(const struct triple *t) {
   unsigned i;

   rdf_printpv(&t->pv);

   if (t->isancestry) {
      printf(" ancestry:%s ", t->attrib);
   }
   else {
      printf(" identity:%s ", t->attrib);
   }

   switch (t->valtype) {
    case V_NIL:
     /* XXX */
     printf("thistrace:nil .\n");
     break;
    case V_STRING:
     rdf_printstring(&t->val_string);
     printf(" . \n");
     break;
    case V_MULTISTRING:
     /* XXX this will break if the attribute isn't unique for the pnode */
     /* (but it always will be in normal cases) */
     printf("thistrace:%llu_M%s .\n",
	    (unsigned long long)t->pv.pnum, t->attrib);
     for (i=0; i<t->val_multistring.num; i++) {
	printf("thistrace:%llu_M%s",
	       (unsigned long long)t->pv.pnum, t->attrib);
	printf(" index:%u ", i);
	rdf_printstring(&t->val_multistring.v[i]);
	printf(" . \n");
     }
     break;
    case V_INT:
     printf("\"%d\"^^<xsd:integer> .\n", t->val_int);
     break;
    case V_REAL:
     printf("\"%g\"^^<xsd:decimal> .\n", t->val_real);
     break;
    case V_TIMESTAMP:
     printf("\"%u.%09u\"^^<passtypes:timestamp> .\n",
	    t->val_timestamp.pt_sec, t->val_timestamp.pt_nsec);
     break;
    case V_INODE:
     /* XXX: should this have a pass-specified type? */
     printf("\"%lu\"^^<xsd:integer> .\n", (unsigned long)t->val_inode);
     break;
    case V_PNODE:
     printf("thistrace:%llu .\n", (unsigned long long)t->val_pnode);
     break;
    case V_PNODEVERSION:
     rdf_printpv(&t->val_pnodeversion);
     printf(" . \n");
     break;
   }
}

////////////////////////////////////////////////////////////
// format switch

static void export_header(void) {
   switch (g_format) {
    case F_TEXT: /*export_header_text();*/ break;
    case F_PASSRDF: export_header_passrdf(); break;
   }
}

static void export_triple(const struct triple *t) {
   switch (g_format) {
    case F_TEXT: export_triple_text(t); break;
    case F_PASSRDF: export_triple_passrdf(t); break;
   }
}

////////////////////////////////////////////////////////////
// database scanner

static void checksize(const char *what, size_t expected, size_t found) {
   if (found != expected) {
      errx(1, "wrong size %s (expected %zu, found %zu)",
	   what, expected, found);
   }
}

static void gettoken(tnum_t tok, struct string *ret) {
   DBT tnum_dbt;
   DBT token_dbt;
   int result;

   memset(&tnum_dbt, 0, sizeof(tnum_dbt));
   memset(&token_dbt, 0, sizeof(token_dbt));

   tnum_dbt.data = &tok;
   tnum_dbt.size = sizeof(tok);

   result = g_tnum2tokdb->db->get(g_tnum2tokdb->db, NULL,
				  &tnum_dbt, &token_dbt,
				  0);
   wdb_check_err(result);

   makestring(token_dbt.data, token_dbt.size, ret);
}

static void export(void) {
   DBC *cursor;
   DBT key_dbt, val_dbt;
   provdb_key *key;
   provdb_val *val;
   char *data;
   size_t datalen;
   bool tokenized;
   struct triple t;
   int result;

   /* Initialize things. */
   cursor = NULL;
   memset(&key_dbt, 0, sizeof(key_dbt));
   memset(&val_dbt, 0, sizeof(val_dbt));

   /* Open a cursor for the main provdb. */
   result = g_provdb->db->cursor(g_provdb->db, NULL, &cursor, 0);
   wdb_check_err(result);

   /* Print the headers, if any. */ 
   export_header();

   /* Scan it. */
   while ((result = cursor->c_get(cursor, &key_dbt, &val_dbt, DB_NEXT)) == 0) {

      /* Make a sanity check on what we got. */
      checksize("key", sizeof(provdb_key), key_dbt.size);
      if (val_dbt.size < sizeof(provdb_val)) {
	 errx(1, "runt value (found %zu; expected at least %zu)",
	      (size_t)val_dbt.size, sizeof(provdb_val));
      }

      /* Apply types. Copy the value for safekeeping. */
      key = key_dbt.data;
      val = domalloc(val_dbt.size);
      memcpy(val, val_dbt.data, val_dbt.size);

      /* Get the left and center components of the triple. */
      triple_init(&t);
      t.pv = *key;
      t.isancestry = (val->pdb_flags & PROVDB_ANCESTRY) != 0;
      t.attrib = dostrdup(PROVDB_VAL_ATTR(val));

      if (g_lowercase) {
	 strlower(t.attrib);
      }

      /* Get right component of the triple based on the value type. */

      data = PROVDB_VAL_VALUE(val);
      datalen = PROVDB_VAL_VALUELEN(val);
      tokenized = (val->pdb_flags & PROVDB_TOKENIZED) != 0;

      switch (PROVDB_VAL_VALUETYPE(val)) {

       case PROV_TYPE_NIL:
	t.valtype = V_NIL;
	/* nothing else */
	break;

       case PROV_TYPE_STRING:
	t.valtype = V_STRING;
	if (PROVDB_VAL_ISTOKENIZED(val)) {
	   checksize("token value", sizeof(tnum_t), datalen);
	   gettoken(*(tnum_t *) data, &t.val_string);
	}
	else {
	   makestring(data, datalen, &t.val_string);
	}
	break;

       case PROV_TYPE_MULTISTRING:
	if (PROVDB_VAL_ISTOKENIZED(val)) {
	   tnum_t *toks;
	   unsigned i, ntoks;

	   /* check size */
	   if (datalen % sizeof(tnum_t) != 0) {
	      err(1, "tokenized multistring value has unaligned size %zu",
		  datalen);
	   }

	   /* copy the tokens for safekeeping */
	   ntoks = datalen / sizeof(tnum_t);
	   toks = domalloc(datalen);
	   memcpy(toks, data, datalen);

	   /* fetch the strings for the tokens */
	   triple_set_multistring(&t, ntoks);
	   for (i=0; i<ntoks; i++) {
	       gettoken(toks[i], &t.val_multistring.v[i]);
	   }

	   /* tidy up */
	   free(toks);
	}
	else {
	   size_t x, y;
	   unsigned pos;

	   /* count how many strings we have */
	   pos = 0;
	   for (x = 0; x < datalen; x += y+1) {
	      y = wdb_strnlen(data+x, datalen-x);
	      pos++;
	   }

	   triple_set_multistring(&t, pos);

	   pos = 0;
	   for (x = 0; x < datalen; x += y+1) {
	      y = wdb_strnlen(data+x, datalen-x);
	      makestring(data+x, y, &t.val_multistring.v[pos++]);
	   }
	}
	break;

       case PROV_TYPE_INT:
	checksize("int value", sizeof(int32_t), datalen);
	t.valtype = V_INT;
	t.val_int = *(int32_t *)data;
	break;

       case PROV_TYPE_REAL:
	checksize("real value", sizeof(double), datalen);
	t.valtype = V_REAL;
	t.val_real = *(double *)data;
	break;

       case PROV_TYPE_TIMESTAMP:
	checksize("timestamp value", sizeof(struct prov_timestamp), datalen);
	t.valtype = V_TIMESTAMP;
	t.val_timestamp = *(struct prov_timestamp *)data;
	break;

       case PROV_TYPE_INODE:
	checksize("inode value", sizeof(uint32_t), datalen);
	t.valtype = V_INODE;
	t.val_inode = *(uint32_t *)data;
	break;

       case PROV_TYPE_PNODE:
	checksize("pnode value", sizeof(pnode_t), datalen);
	t.valtype = V_PNODE;
	t.val_pnode = *(pnode_t *)data;
	break;

       case PROV_TYPE_PNODEVERSION:
	checksize("pnode-version value",
		  sizeof(struct pnode_version), datalen);
	t.valtype = V_PNODEVERSION;
	t.val_pnodeversion = *(struct pnode_version *)data;
	break;

       case PROV_TYPE_OBJECT:
       case PROV_TYPE_OBJECTVERSION:
       default:
	errx(1, "illegal value type %u", PROVDB_VAL_VALUETYPE(val));
	break;
      }

      export_triple(&t);

      triple_cleanup(&t);
      free(val);
   }
   if (result != DB_NOTFOUND) {
      wdb_check_err(result);
   }

   /* close the cursor */
   /* XXX */
}

////////////////////////////////////////////////////////////
// main

static void usage(const char *name) {
   fprintf(stderr, "Usage: %s [options] [path]\n", name);
   fprintf(stderr, "   -p           preserve case of edge names\n");
   fprintf(stderr, "   -t <fmt>     choose output format\n");
   fprintf(stderr, "      text      plain text dump (default)\n");
   fprintf(stderr, "      passrdf   RDF with PASS edge naming\n");
   fprintf(stderr, "   -v           suppress versions in output\n");
   exit(1);
}

int main(int argc, char *argv[]) {
   const char *path;
   int ch;

   /* establish defaults */
   path = ".";
   g_format = F_TEXT;
   g_noversions = false;
   g_lowercase = true;

   /* handle args */
   while ((ch = getopt(argc, argv, "pt:v")) != -1) {
      switch (ch) {
       case 'p':
	g_lowercase = false;
	break;
       case 't':
	/* should probably make a table for this */
	if (!strcmp(optarg, "text")) {
	   g_format = F_TEXT;
	}
	else if (!strcmp(optarg, "passrdf")) {
	   g_format = F_PASSRDF;
	}
	else {
	   errx(1, "Unknown output format %s", optarg);
	}
	break;
       case 'v':
	g_noversions = true;
	break;
       default:
	usage(argv[0]);
	break;
      }
   }
   if (optind < argc) {
      path = argv[optind++];
   }
   if (optind < argc) {
      usage(argv[0]);
   }

   /* open the database */
   if (wdb_startup(path, WDB_O_RDONLY) < 0) {
      err(1, "wdb_startup: %s", path);
   }

   /* do it */
   export();

   /* close database */
   wdb_shutdown();

   return 0;
}
