/*
 * Copyright 2009, 2010
 *	The President and Fellows of Harvard College.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Simple optimizations.
 *
 * This module should contain only simple transforms that require
 * little or no analysis, context tracking, or non-local knowledge,
 * and do only one pass over the tree.
 *
 * It should also not contain any transforms that are subtle in either
 * semantics or implementation, because when we're looking at dumps we
 * really want to be able to read the post-baseopt dump instead of the
 * raw output from tuplify, and we want to be able to do so without
 * worrying excessively over whether it's correct.
 */

#include <stdlib.h> /* for qsort */
#include <string.h> /* for strcmp */

#include "datatype.h"
#include "pqlvalue.h"
#include "tcalc.h"
#include "tcutil.h"
#include "passes.h"

struct baseopt {
   struct pqlcontext *pql;
};

static void tcexpr_replacevar(struct pqlcontext *ctx,
			      struct tcexpr *te,
			      struct tcvar *oldvar,
			      struct tcvar *newvar);

static struct tcexpr *tcexpr_baseopt(struct baseopt *ctx, struct tcexpr *te);

////////////////////////////////////////////////////////////
// context management

static void baseopt_init(struct baseopt *ctx, struct pqlcontext *pql) {
   ctx->pql = pql;
}

static void baseopt_cleanup(struct baseopt *ctx) {
   (void)ctx;
}

////////////////////////////////////////////////////////////
// small tools/utils

/*
 * Given:
 *    lambda V1: P1
 *    lambda V2: P2
 * produce
 *    lambda V3: P3
 * where
 *    P3 === P1 and P2
 *
 * Pretend NULL expressions are "true".
 *
 * The inputs are consumed; replacements are produced.
 *
 * XXX this should be shared with the similar code in stepjoins.c.
 */
static void combine_predicates(struct baseopt *ctx,
			       struct tcvar *v1,
			       struct tcexpr *p1,
			       struct tcvar *v2,
			       struct tcexpr *p2,
			       struct tcvar **vret,
			       struct tcexpr **pret) {
   if (p1 == NULL) {
      PQLASSERT(v1 == NULL);
      *vret = v2;
      *pret = p2;
      return;
   }
   if (p2 == NULL) {
      PQLASSERT(v2 == NULL);
      *vret = v1;
      *pret = p1;
      return;
   }

   if (v1 != NULL && v2 != NULL) {
      tcexpr_replacevar(ctx->pql, p2, v2, v1);
      tcvar_decref(ctx->pql, v2);
      *vret = v1;
   }
   else if (v1 != NULL) {
      *vret = v1;
   }
   else if (v2 != NULL) {
      *vret = v2;
   }
   else {
      *vret = NULL;
   }

   *pret = mktcexpr_bop(ctx->pql, p1, F_AND, p2);
   (*pret)->datatype = p1->datatype;
}

/*
 * Check if TE is a constant true/false value.
 */

static bool tcexpr_istrue(struct tcexpr *te) {
   return (te->type == TCE_VALUE && pqlvalue_isbool(te->value)
	   && pqlvalue_bool_get(te->value));
}

static bool tcexpr_isfalse(struct tcexpr *te) {
   return (te->type == TCE_VALUE && pqlvalue_isbool(te->value)
	   && !pqlvalue_bool_get(te->value));
}

static bool columnarray_contains(const struct columnarray *arr,
				 const struct column *col) {
   unsigned i, num;

   num = columnarray_num(arr);
   for (i=0; i<num; i++) {
      if (columnarray_get(arr, i) == col) {
	 return true;
      }
   }
   return false;
}

static bool columnarray_contains_any2(const struct columnarray *arr,
				      const struct column *col1,
				      const struct column *col2) {
   unsigned i, num;

   num = columnarray_num(arr);
   for (i=0; i<num; i++) {
      if (columnarray_get(arr, i) == col1 ||
	  columnarray_get(arr, i) == col2) {
	 return true;
      }
   }
   return false;
}

static bool columnarray_contains_any3(const struct columnarray *arr,
				      const struct column *col1,
				      const struct column *col2,
				      const struct column *col3) {
   unsigned i, num;

   num = columnarray_num(arr);
   for (i=0; i<num; i++) {
      if (columnarray_get(arr, i) == col1 ||
	  columnarray_get(arr, i) == col2 ||
	  columnarray_get(arr, i) == col3) {
	 return true;
      }
   }
   return false;
}

static bool columnarray_contains_any4(const struct columnarray *arr,
				      const struct column *col1,
				      const struct column *col2,
				      const struct column *col3,
				      const struct column *col4) {
   unsigned i, num;

   num = columnarray_num(arr);
   for (i=0; i<num; i++) {
      if (columnarray_get(arr, i) == col1 ||
	  columnarray_get(arr, i) == col2 ||
	  columnarray_get(arr, i) == col3 ||
	  columnarray_get(arr, i) == col4) {
	 return true;
      }
   }
   return false;
}

static bool columnarray_contains_any6(const struct columnarray *arr,
				      const struct column *col1,
				      const struct column *col2,
				      const struct column *col3,
				      const struct column *col4,
				      const struct column *col5,
				      const struct column *col6) {
   unsigned i, num;

   num = columnarray_num(arr);
   for (i=0; i<num; i++) {
      if (columnarray_get(arr, i) == col1 ||
	  columnarray_get(arr, i) == col2 ||
	  columnarray_get(arr, i) == col3 ||
	  columnarray_get(arr, i) == col4 ||
	  columnarray_get(arr, i) == col5 ||
	  columnarray_get(arr, i) == col6) {
	 return true;
      }
   }
   return false;
}

static bool columnarray_contains_any(const struct columnarray *arr1,
				     const struct columnarray *arr2) {
   struct column *col1, *col2;
   unsigned i, num1, j, num2;

   num1 = columnarray_num(arr1);
   num2 = columnarray_num(arr2);
   for (i=0; i<num1; i++) {
      col1 = columnarray_get(arr1, i);
      for (j=0; j<num2; j++) {
	 col2 = columnarray_get(arr2, j);
	 if (col1 == col2) {
	    return true;
	 }
      }
   }
   return false;
}

static bool columnarray_contains_all2(const struct columnarray *arr,
				      const struct column *col1,
				      const struct column *col2) {
   unsigned i, num, ct = 0;

   num = columnarray_num(arr);
   for (i=0; i<num; i++) {
      if (columnarray_get(arr, i) == col1 ||
	  columnarray_get(arr, i) == col2) {
	 ct++;
      }
   }
   return ct == 2;
}

static bool columnarray_contains_all3(const struct columnarray *arr,
				      const struct column *col1,
				      const struct column *col2,
				      const struct column *col3) {
   unsigned i, num, ct = 0;

   num = columnarray_num(arr);
   for (i=0; i<num; i++) {
      if (columnarray_get(arr, i) == col1 ||
	  columnarray_get(arr, i) == col2 ||
	  columnarray_get(arr, i) == col3) {
	 ct++;
      }
   }
   return ct == 3;
}

static bool columnarray_contains_all4(const struct columnarray *arr,
				      const struct column *col1,
				      const struct column *col2,
				      const struct column *col3,
				      const struct column *col4) {
   unsigned i, num, ct = 0;

   num = columnarray_num(arr);
   for (i=0; i<num; i++) {
      if (columnarray_get(arr, i) == col1 ||
	  columnarray_get(arr, i) == col2 ||
	  columnarray_get(arr, i) == col3 ||
	  columnarray_get(arr, i) == col4) {
	 ct++;
      }
   }
   return ct == 4;
}

static bool columnarray_contains_all6(const struct columnarray *arr,
				      const struct column *col1,
				      const struct column *col2,
				      const struct column *col3,
				      const struct column *col4,
				      const struct column *col5,
				      const struct column *col6) {
   unsigned i, num, ct = 0;

   num = columnarray_num(arr);
   for (i=0; i<num; i++) {
      if (columnarray_get(arr, i) == col1 ||
	  columnarray_get(arr, i) == col2 ||
	  columnarray_get(arr, i) == col3 ||
	  columnarray_get(arr, i) == col4 ||
	  columnarray_get(arr, i) == col5 ||
	  columnarray_get(arr, i) == col6) {
	 ct++;
      }
   }
   return ct == 6;
}

/*
 * Check if arr1 is a superset of arr2.
 */
static bool columnarray_contains_all(const struct columnarray *arr1,
				     const struct columnarray *arr2) {
   struct column *col1, *col2;
   unsigned i, num1, j, num2, ct = 0;

   num1 = columnarray_num(arr1);
   num2 = columnarray_num(arr2);
   for (i=0; i<num1; i++) {
      col1 = columnarray_get(arr1, i);
      for (j=0; j<num2; j++) {
	 col2 = columnarray_get(arr2, j);
	 if (col1 == col2) {
	    ct++;
	 }
      }
   }
   return ct == num2;
}

static int columnarray_find(const struct columnarray *arr,
			    const struct column *col,
			    unsigned *index_ret) {
   unsigned i, num;

   num = columnarray_num(arr);
   for (i=0; i<num; i++) {
      if (columnarray_get(arr, i) == col) {
	 *index_ret = i;
	 return 0;
      }
   }
   return -1;
}

#if 0
static void columnarray_removeptr(struct columnarray *arr,
				  struct column *col) {
   unsigned index;
   int ret;

   ret = columnarray_find(arr, col, &index);
   PQLASSERT(ret == 0);
   columnarray_remove(arr, index);
}
#endif

static void columnarray_mayberemoveptr(struct columnarray *arr,
				       struct column *col) {
   unsigned index;
   int ret;

   ret = columnarray_find(arr, col, &index);
   if (ret == 0) {
      columnarray_remove(arr, index);
   }
}

static bool columnarray_eq(const struct columnarray *a,
			   const struct columnarray *b) {
   unsigned i, num;

   num = columnarray_num(a);
   if (num != columnarray_num(b)) {
      return false;
   }
   for (i=0; i<num; i++) {
      if (columnarray_get(a, i) != columnarray_get(b, i)) {
	 return false;
      }
   }
   return true;
}

#if 0
/*
 * Move columns from FROM to TO.
 */
static void columnarray_moveappend(struct pqlcontext *pql,
				   struct columnarray *to,
				   struct columnarray *from) {
   unsigned i, num;
   struct column *col;

   num = columnarray_num(from);
   for (i=0; i<num; i++) {
      col = columnarray_get(from, i);
      columnarray_add(pql, to, col, NULL);
   }
   columnarray_setsize(pql, from, 0);
}
#endif

static bool datatype_contains_column(struct datatype *t, struct column *col) {
   return datatype_subprojectscalar(t, col) != NULL;
}

#if 0
/*
 * Fill a columnarray with all toplevel column names in type T.
 */
static void gettopcols(struct pqlcontext *pql, struct columnarray *fill,
		       struct datatype *t, bool doincref) {
   struct column *col;
   unsigned i, num;

   if (datatype_isset(t)) {
      t = datatype_set_member(t);
   }
   else if (datatype_issequence(t)) {
      t = datatype_sequence_member(t);
   }
   PQLASSERT(datatype_istuple(t));

   num = datatype_arity(t);
   for (i=0; i<num; i++) {
      col = datatype_tuple_getcolumn(t, i);
      if (doincref) {
	 column_incref(col);
      }
      columnarray_add(pql, fill, col, NULL);
   }
}
#endif

/*
 * Return true if SUB, which is an adjoin, adjoins an exact copy of 
 * some other column, and if so return that column in FROMCOL_RET.
 *
 * Does not return a new reference in FROMCOL_RET - caller should
 * incref if the column is affixed anywhere persistent.
 */
static bool adjoin_copies_column(struct baseopt *ctx, struct tcexpr *sub,
				 struct column **fromcol_ret) {
   struct tcvar *lvar;
   struct tcexpr *lexp;

   (void)ctx;

   PQLASSERT(sub->type == TCE_ADJOIN);
   lvar = sub->adjoin.var;
   lexp = sub->adjoin.expr;

   // XXX I don't think the first test is needed, or maybe even wanted
   if (lexp->type == TCE_PROJECTTUPLE &&
       lexp->projecttuple.sub->type == TCE_READVAR &&
       lexp->projecttuple.sub->readvar == lvar &&
       columnarray_num(lexp->projecttuple.cols) == 1) {
      *fromcol_ret = columnarray_get(lexp->projecttuple.cols, 0);
      return true;
   }
   if (lexp->type == TCE_PROJECTSCALAR &&
       lexp->projectscalar.sub->type == TCE_READVAR &&
       lexp->projectscalar.sub->readvar == lvar) {
      *fromcol_ret = lexp->projectscalar.col;
      return true;
   }
   return false;
}

#if 0
/*
 * Transform ARR into ~ARR, using T for the reference list of columns.
 * Only toplevel columns are handled. Refcounts columns.
 *
 * XXX this is copied from typeinf.c. Either it should be shared or it
 * should go away from here entirely (IOW, all complements wanted
 * would be taken upstream...)
 */
static void columnarray_take_complement(struct baseopt *ctx,
					struct columnarray *arr,
					struct datatype *t) {
   struct column *col;
   bool *keep;
   unsigned i, arity, j, num;

   if (datatype_isset(t)) {
      t = datatype_set_member(t);
   }
   else if (datatype_issequence(t)) {
      t = datatype_sequence_member(t);
   }
   PQLASSERT(datatype_istuple(t));

   arity = datatype_arity(t);
   num = columnarray_num(arr);
   PQLASSERT(num <= arity);

   keep = domalloc(ctx->pql, arity * sizeof(*keep));
   for (i=0; i<arity; i++) {
      keep[i] = true;
   }

   for (i=0; i<arity; i++) {
      col = datatype_tuple_getcolumn(t, i);
      for (j=0; j<num; j++) {
	 if (col == columnarray_get(arr, j)) {
	    keep[i] = false;
	 }
      }
   }

   for (j=0; j<num; j++) {
      column_decref(ctx->pql, columnarray_get(arr, j));
   }
   columnarray_setsize(ctx->pql, arr, arity - num);

   for (i=j=0; i<arity; i++) {
      if (keep[i]) {
	 col = datatype_tuple_getcolumn(t, i);
	 column_incref(col);
	 columnarray_set(arr, j++, col);
      }
   }
   PQLASSERT(j == arity - num);

   dofree(ctx->pql, keep, arity * sizeof(*keep));
}
#endif

////////////////////////////////////////////////////////////
// larger tests

/*
 * Check if TE uses any of the columns in COLS.
 */
static bool tcexpr_uses_columns(struct baseopt *ctx, struct tcexpr *te,
				struct columnarray *cols) {
   unsigned i, num;
   bool ret;

   ret = false; // gcc 4.1

   if (te == NULL) {
      return false;
   }

   switch (te->type) {
    case TCE_FILTER:
     ret = tcexpr_uses_columns(ctx, te->filter.sub, cols) ||
	tcexpr_uses_columns(ctx, te->filter.predexpr, cols);
     break;
    case TCE_PROJECTTUPLE:
     ret = tcexpr_uses_columns(ctx, te->projecttuple.sub, cols) ||
	columnarray_contains_any(te->projecttuple.cols, cols);
     break;
    case TCE_PROJECTSCALAR:
     ret = columnarray_contains(cols, te->projectscalar.col) ||
	tcexpr_uses_columns(ctx, te->projectscalar.sub, cols);
     break;
    case TCE_RENAME:
     ret = columnarray_contains_any2(cols,
				     te->rename.insidecol,
				     te->rename.outsidecol) ||
	tcexpr_uses_columns(ctx, te->rename.sub, cols);
     break;
    case TCE_JOIN:
     ret = tcexpr_uses_columns(ctx, te->join.left, cols) ||
	tcexpr_uses_columns(ctx, te->join.right, cols) ||
	tcexpr_uses_columns(ctx, te->join.predexpr, cols);
     break;
    case TCE_ORDER:
     ret = tcexpr_uses_columns(ctx, te->order.sub, cols) ||
	columnarray_contains_any(te->order.cols, cols);
     break;
    case TCE_UNIQ:
     ret = tcexpr_uses_columns(ctx, te->uniq.sub, cols) ||
	columnarray_contains_any(te->uniq.cols, cols);
     break;
    case TCE_NEST:
     ret = columnarray_contains(cols, te->nest.newcol) ||
	tcexpr_uses_columns(ctx, te->nest.sub, cols) ||
	columnarray_contains_any(te->nest.cols, cols);
     break;
    case TCE_UNNEST:
     ret = columnarray_contains(cols, te->unnest.col) ||
	tcexpr_uses_columns(ctx, te->unnest.sub, cols);
     break;
    case TCE_DISTINGUISH:
     ret = columnarray_contains(cols, te->distinguish.newcol) ||
	tcexpr_uses_columns(ctx, te->distinguish.sub, cols);
     break;
    case TCE_ADJOIN:
     ret = columnarray_contains(cols, te->adjoin.newcol) ||
	tcexpr_uses_columns(ctx, te->adjoin.left, cols) ||
	tcexpr_uses_columns(ctx, te->adjoin.expr, cols);
     break;
    case TCE_STEP:
     ret = columnarray_contains_any4(cols,
				     te->step.subcolumn,
				     te->step.leftobjcolumn,
				     te->step.edgecolumn,
				     te->step.rightobjcolumn) ||
	tcexpr_uses_columns(ctx, te->step.sub, cols) ||
	tcexpr_uses_columns(ctx, te->step.predexpr, cols);
     break;
    case TCE_REPEAT:
     ret = columnarray_contains_any6(cols,
				     te->repeat.subendcolumn,
				     te->repeat.bodystartcolumn,
				     te->repeat.bodypathcolumn,
				     te->repeat.bodyendcolumn,
				     te->repeat.repeatpathcolumn,
				     te->repeat.repeatendcolumn) ||
	tcexpr_uses_columns(ctx, te->repeat.sub, cols) ||
	tcexpr_uses_columns(ctx, te->repeat.body, cols);
     break;
    case TCE_SCAN:
     ret = columnarray_contains_any3(cols,
				     te->scan.leftobjcolumn,
				     te->scan.edgecolumn,
				     te->scan.rightobjcolumn);
     break;
    case TCE_BOP:
     ret = tcexpr_uses_columns(ctx, te->bop.left, cols) ||
	tcexpr_uses_columns(ctx, te->bop.right, cols);
     break;
    case TCE_UOP:
     ret = tcexpr_uses_columns(ctx, te->uop.sub, cols);
     break;
    case TCE_FUNC:
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	if (tcexpr_uses_columns(ctx, tcexprarray_get(&te->func.args, i),
				cols)) {
	   return true;
	}
     }
     ret = false;
     break;
    case TCE_MAP:
     ret = tcexpr_uses_columns(ctx, te->map.set, cols) ||
	tcexpr_uses_columns(ctx, te->map.result, cols);
     break;
    case TCE_LET:
     ret = tcexpr_uses_columns(ctx, te->let.value, cols) ||
	tcexpr_uses_columns(ctx, te->let.body, cols);
     break;
    case TCE_READVAR:
    case TCE_READGLOBAL:
     ret = false;
     break;
    case TCE_CREATEPATHELEMENT:
     ret = tcexpr_uses_columns(ctx, te->createpathelement, cols);
     break;
    case TCE_SPLATTER:
     ret = tcexpr_uses_columns(ctx, te->splatter.value, cols) ||
	tcexpr_uses_columns(ctx, te->splatter.name, cols);
     break;
    case TCE_TUPLE:
     num = tcexprarray_num(&te->tuple.exprs);
     for (i=0; i<num; i++) {
	if (tcexpr_uses_columns(ctx, tcexprarray_get(&te->tuple.exprs, i),
				cols)) {
	   return true;
	}
     }
     if (columnarray_contains_any(cols, te->tuple.columns)) {
	return true;
     }
     ret = false;
     break;
    case TCE_VALUE:
     ret = false;
     break;
   }
   return ret;
}

/*
 * Check if TE uses COL.
 */
static bool tcexpr_uses_column(struct baseopt *ctx, struct tcexpr *te,
			       struct column *col) {
   struct columnarray cols;
   bool ret;
   
   columnarray_init(&cols);
   columnarray_setsize(ctx->pql, &cols, 1);
   columnarray_set(&cols, 0, col);
   ret = tcexpr_uses_columns(ctx, te, &cols);
   columnarray_setsize(ctx->pql, &cols, 0);
   columnarray_cleanup(ctx->pql, &cols);
   return ret;
}

/*
 * Check if TE uses any columns other than the columns in COLS.
 */
static bool tcexpr_onlyuses_columns(struct baseopt *ctx, struct tcexpr *te,
				    struct columnarray *cols) {
   unsigned i, num;
   bool ret;

   ret = false; // gcc 4.1

   if (te == NULL) {
      return true;
   }

   switch (te->type) {
    case TCE_FILTER:
     ret = tcexpr_onlyuses_columns(ctx, te->filter.sub, cols) ||
	tcexpr_onlyuses_columns(ctx, te->filter.predexpr, cols);
     break;
    case TCE_PROJECTTUPLE:
     ret = tcexpr_onlyuses_columns(ctx, te->projecttuple.sub, cols) ||
	columnarray_contains_all(cols, te->projecttuple.cols);
     break;
    case TCE_PROJECTSCALAR:
     ret = columnarray_contains(cols, te->projectscalar.col) ||
	tcexpr_onlyuses_columns(ctx, te->projectscalar.sub, cols);
     break;
    case TCE_RENAME:
     ret = columnarray_contains_all2(cols,
				     te->rename.insidecol,
				     te->rename.outsidecol) ||
	tcexpr_onlyuses_columns(ctx, te->rename.sub, cols);
     break;
    case TCE_JOIN:
     ret = tcexpr_onlyuses_columns(ctx, te->join.left, cols) ||
	tcexpr_onlyuses_columns(ctx, te->join.right, cols) ||
	tcexpr_onlyuses_columns(ctx, te->join.predexpr, cols);
     break;
    case TCE_ORDER:
     ret = tcexpr_onlyuses_columns(ctx, te->order.sub, cols) ||
	columnarray_contains_all(cols, te->order.cols);
     break;
    case TCE_UNIQ:
     ret = tcexpr_onlyuses_columns(ctx, te->uniq.sub, cols) ||
	columnarray_contains_all(cols, te->uniq.cols);
     break;
    case TCE_NEST:
     ret = columnarray_contains(cols, te->nest.newcol) ||
	tcexpr_onlyuses_columns(ctx, te->nest.sub, cols) ||
	columnarray_contains_all(cols, te->nest.cols);
     break;
    case TCE_UNNEST:
     ret = columnarray_contains(cols, te->unnest.col) ||
	tcexpr_onlyuses_columns(ctx, te->unnest.sub, cols);
     break;
    case TCE_DISTINGUISH:
     ret = columnarray_contains(cols, te->distinguish.newcol) ||
	tcexpr_onlyuses_columns(ctx, te->distinguish.sub, cols);
     break;
    case TCE_ADJOIN:
     ret = columnarray_contains(cols, te->adjoin.newcol) ||
	tcexpr_onlyuses_columns(ctx, te->adjoin.left, cols) ||
	tcexpr_onlyuses_columns(ctx, te->adjoin.expr, cols);
     break;
    case TCE_STEP:
     ret = columnarray_contains_all4(cols,
				     te->step.subcolumn,
				     te->step.leftobjcolumn,
				     te->step.edgecolumn,
				     te->step.rightobjcolumn) ||
	tcexpr_onlyuses_columns(ctx, te->step.sub, cols) ||
	tcexpr_onlyuses_columns(ctx, te->step.predexpr, cols);
     break;
    case TCE_REPEAT:
     ret = columnarray_contains_all6(cols,
				     te->repeat.subendcolumn,
				     te->repeat.bodystartcolumn,
				     te->repeat.bodypathcolumn,
				     te->repeat.bodyendcolumn,
				     te->repeat.repeatpathcolumn,
				     te->repeat.repeatendcolumn) ||
	tcexpr_onlyuses_columns(ctx, te->repeat.sub, cols) ||
	tcexpr_onlyuses_columns(ctx, te->repeat.body, cols);
     break;
    case TCE_SCAN:
     ret = columnarray_contains_all3(cols,
				     te->scan.leftobjcolumn,
				     te->scan.edgecolumn,
				     te->scan.rightobjcolumn);
     break;
    case TCE_BOP:
     ret = tcexpr_onlyuses_columns(ctx, te->bop.left, cols) ||
	tcexpr_onlyuses_columns(ctx, te->bop.right, cols);
     break;
    case TCE_UOP:
     ret = tcexpr_onlyuses_columns(ctx, te->uop.sub, cols);
     break;
    case TCE_FUNC:
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	if (!tcexpr_onlyuses_columns(ctx, tcexprarray_get(&te->func.args, i),
				     cols)) {
	   return false;
	}
     }
     ret = true;
     break;
    case TCE_MAP:
     ret = tcexpr_onlyuses_columns(ctx, te->map.set, cols) ||
	tcexpr_onlyuses_columns(ctx, te->map.result, cols);
     break;
    case TCE_LET:
     ret = tcexpr_onlyuses_columns(ctx, te->let.value, cols) ||
	tcexpr_onlyuses_columns(ctx, te->let.body, cols);
     break;
    case TCE_READVAR:
    case TCE_READGLOBAL:
     ret = true;
     break;
    case TCE_CREATEPATHELEMENT:
     ret = tcexpr_onlyuses_columns(ctx, te->createpathelement, cols);
     break;
    case TCE_SPLATTER:
     ret = tcexpr_onlyuses_columns(ctx, te->splatter.value, cols) ||
	tcexpr_onlyuses_columns(ctx, te->splatter.name, cols);
     break;
    case TCE_TUPLE:
     num = tcexprarray_num(&te->tuple.exprs);
     for (i=0; i<num; i++) {
	if (tcexpr_onlyuses_columns(ctx, tcexprarray_get(&te->tuple.exprs, i),
				    cols)) {
	   return false;
	}
     }
     if (!columnarray_contains_all(cols, te->tuple.columns)) {
	return false;
     }
     ret = true;
     break;
    case TCE_VALUE:
     ret = true;
     break;
   }
   return ret;
}

/*
 * Check if TE generates COL.
 */
static bool tcexpr_generates_column(struct baseopt *ctx, struct tcexpr *te,
				    struct column *col) {
   unsigned i, num;
   bool ret;

   ret = false; // gcc 4.1

   if (te == NULL) {
      return false;
   }

   switch (te->type) {
    case TCE_FILTER:
     ret = tcexpr_generates_column(ctx, te->filter.sub, col);
     break;
    case TCE_PROJECTTUPLE:
     ret = tcexpr_generates_column(ctx, te->projecttuple.sub, col);
     break;
    case TCE_PROJECTSCALAR:
     ret = tcexpr_generates_column(ctx, te->projectscalar.sub, col);
     break;
    case TCE_RENAME:
     ret = (col == te->rename.outsidecol) ||
	tcexpr_generates_column(ctx, te->rename.sub, col);
     break;
    case TCE_JOIN:
     ret = tcexpr_generates_column(ctx, te->join.left, col) ||
	tcexpr_generates_column(ctx, te->join.right, col);
     break;
    case TCE_ORDER:
     ret = tcexpr_generates_column(ctx, te->order.sub, col);
     break;
    case TCE_UNIQ:
     ret = tcexpr_generates_column(ctx, te->uniq.sub, col);
     break;
    case TCE_NEST:
     ret = (col == te->nest.newcol) ||
	tcexpr_generates_column(ctx, te->nest.sub, col);
     break;
    case TCE_UNNEST:
     ret = tcexpr_generates_column(ctx, te->unnest.sub, col);
     break;
    case TCE_DISTINGUISH:
     ret = (col == te->distinguish.newcol) ||
	tcexpr_generates_column(ctx, te->distinguish.sub, col);
     break;
    case TCE_ADJOIN:
     ret = (col == te->adjoin.newcol) ||
	tcexpr_generates_column(ctx, te->adjoin.left, col) ||
	tcexpr_generates_column(ctx, te->adjoin.expr, col);
     break;
    case TCE_STEP:
     ret =
	(col == te->step.leftobjcolumn) ||
	(col == te->step.edgecolumn) ||
	(col == te->step.rightobjcolumn) ||
	tcexpr_generates_column(ctx, te->step.sub, col);
     break;
    case TCE_REPEAT:
     ret =
	(col == te->repeat.bodybodycolumn) ||
	(col == te->repeat.repeatpathcolumn) ||
	(col == te->repeat.repeatendcolumn) ||
	tcexpr_generates_column(ctx, te->repeat.sub, col) ||
	tcexpr_generates_column(ctx, te->repeat.body, col);
     break;
    case TCE_SCAN:
     ret = 
	(col == te->scan.leftobjcolumn) ||
	(col == te->scan.edgecolumn) ||
	(col == te->scan.rightobjcolumn);
     break;
    case TCE_BOP:
     ret = tcexpr_generates_column(ctx, te->bop.left, col) ||
	tcexpr_generates_column(ctx, te->bop.right, col);
     break;
    case TCE_UOP:
     ret = tcexpr_generates_column(ctx, te->uop.sub, col);
     break;
    case TCE_FUNC:
     /* XXX is this right? */
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	if (tcexpr_generates_column(ctx, tcexprarray_get(&te->func.args, i),
				col)) {
	   return true;
	}
     }
     ret = false;
     break;
    case TCE_MAP:
     ret = tcexpr_generates_column(ctx, te->map.set, col) ||
	tcexpr_generates_column(ctx, te->map.result, col);
     break;
    case TCE_LET:
     ret = tcexpr_generates_column(ctx, te->let.value, col) ||
	tcexpr_generates_column(ctx, te->let.body, col);
     break;
    case TCE_READVAR:
    case TCE_READGLOBAL:
     ret = false;
     break;
    case TCE_CREATEPATHELEMENT:
     ret = tcexpr_generates_column(ctx, te->createpathelement, col);
     break;
    case TCE_SPLATTER:
     /* XXX this is wrong */
     ret = tcexpr_generates_column(ctx, te->splatter.value, col) ||
	tcexpr_generates_column(ctx, te->splatter.name, col);
     break;
    case TCE_TUPLE:
     num = tcexprarray_num(&te->tuple.exprs);
     for (i=0; i<num; i++) {
	if (tcexpr_generates_column(ctx, tcexprarray_get(&te->tuple.exprs, i),
				    col)) {
	   return true;
	}
     }
     if (columnarray_contains(te->tuple.columns, col)) {
	return true;
     }
     ret = false;
     break;
    case TCE_VALUE:
     ret = false;
     break;
   }
   return ret;
}

/*
 * Figure out which columns an expression uses -- but only the ones it
 * actually reads from, not any others it may just pass along. List
 * each one only once.
 */

static void column_getusedcolumns(struct baseopt *ctx, struct column *col,
				  struct columnarray *fill) {
   if (!columnarray_contains(fill, col)) {
      columnarray_add(ctx->pql, fill, col, NULL);
   }
}

static void columnarray_getusedcolumns(struct baseopt *ctx,
				       const struct columnarray *cols,
				       struct columnarray *fill) {
   unsigned i, num;

   num = columnarray_num(cols);
   for (i=0; i<num; i++) {
      column_getusedcolumns(ctx, columnarray_get(cols, i), fill);
   }
}


/*
 * XXX remove this bogus decl (and the one in joinorder) and move this
 * code to a more suitable place.
 */
void tcexpr_getusedcolumns(struct baseopt *ctx, struct tcexpr *te,
			   struct columnarray *fill);

/*static*/ void tcexpr_getusedcolumns(struct baseopt *ctx, struct tcexpr *te,
				  struct columnarray *fill) {
   unsigned i, num;

   if (te == NULL) {
      return;
   }

   switch (te->type) {
    case TCE_FILTER:
     tcexpr_getusedcolumns(ctx, te->filter.predexpr, fill);
     tcexpr_getusedcolumns(ctx, te->filter.sub, fill);
     break;
    case TCE_PROJECTTUPLE:
     /*columnarray_getusedcolumns(ctx, te->projecttuple.cols, fill);*/
     tcexpr_getusedcolumns(ctx, te->projecttuple.sub, fill);
     break;
    case TCE_PROJECTSCALAR:
     column_getusedcolumns(ctx, te->projectscalar.col, fill);
     tcexpr_getusedcolumns(ctx, te->projectscalar.sub, fill);
     break;
    case TCE_RENAME:
     /* If we're using the column on the outside, switch to the inside name. */
     {
	unsigned ix;

	if (columnarray_find(fill, te->rename.outsidecol, &ix) == 0) {
	   columnarray_set(fill, ix, te->rename.insidecol);
	}
     }
     tcexpr_getusedcolumns(ctx, te->rename.sub, fill);
     break;
    case TCE_JOIN:
     tcexpr_getusedcolumns(ctx, te->join.predexpr, fill);
     /*
      * Properly we should check each side independently and take the
      * union, because columns used on one side shouldn't affect the
      * logic on the other. But I don't think there are any cases
      * where this matters. (XXX?)
      */
     tcexpr_getusedcolumns(ctx, te->join.right, fill);
     tcexpr_getusedcolumns(ctx, te->join.left, fill);
     break;
    case TCE_ORDER:
     columnarray_getusedcolumns(ctx, te->order.cols, fill);
     tcexpr_getusedcolumns(ctx, te->order.sub, fill);
     break;
    case TCE_UNIQ:
     columnarray_getusedcolumns(ctx, te->uniq.cols, fill);
     tcexpr_getusedcolumns(ctx, te->uniq.sub, fill);
     break;
    case TCE_NEST:
     if (columnarray_contains(fill, te->nest.newcol)) {
	columnarray_mayberemoveptr(fill, te->nest.newcol);
	columnarray_getusedcolumns(ctx, te->nest.cols, fill);
     }
     tcexpr_getusedcolumns(ctx, te->nest.sub, fill);
     break;
    case TCE_UNNEST:
     column_getusedcolumns(ctx, te->unnest.col, fill);
     tcexpr_getusedcolumns(ctx, te->unnest.sub, fill);
     break;
    case TCE_DISTINGUISH:
     columnarray_mayberemoveptr(fill, te->distinguish.newcol);
     tcexpr_getusedcolumns(ctx, te->distinguish.sub, fill);
     break;
    case TCE_ADJOIN:
     if (columnarray_contains(fill, te->adjoin.newcol)) {
	columnarray_mayberemoveptr(fill, te->adjoin.newcol);
	tcexpr_getusedcolumns(ctx, te->adjoin.expr, fill);
     }
     tcexpr_getusedcolumns(ctx, te->adjoin.left, fill);
     break;
    case TCE_STEP:
     column_getusedcolumns(ctx, te->step.subcolumn, fill);
     tcexpr_getusedcolumns(ctx, te->step.predexpr, fill);
     tcexpr_getusedcolumns(ctx, te->step.sub, fill);
     break;
    case TCE_REPEAT:
     /* remove columns that are produced from the body and returned */
     columnarray_mayberemoveptr(fill, te->repeat.bodybodycolumn);
     columnarray_mayberemoveptr(fill, te->repeat.repeatendcolumn);
     if (te->repeat.repeatpathcolumn != NULL) {
	columnarray_mayberemoveptr(fill, te->repeat.repeatpathcolumn);
     }

     /* add columns we use that the body should produce */
     column_getusedcolumns(ctx, te->repeat.bodyendcolumn, fill);
     if (te->repeat.bodypathcolumn != NULL) {
	column_getusedcolumns(ctx, te->repeat.bodypathcolumn, fill);
     }

     /* scan the body */
     tcexpr_getusedcolumns(ctx, te->repeat.body, fill);

     /*
      * Do not remove the columns the body should produce. If it
      * didn't produce them but they're generated by project from
      * a variable or something, they're legitimately "used" by 
      * the overall repeat.
      */
     //columnarray_mayberemoveptr(fill, te->repeat.bodyendcolumn);
     //if (te->repeat.bodypathcolumn != NULL) {
     //   columnarray_mayberemoveptr(fill, te->repeat.bodypathcolumn);
     //}

     /* Remove the column we produce for the body. */
     columnarray_mayberemoveptr(fill, te->repeat.bodystartcolumn);

     /* Add the column we use from the left subexpr. */
     column_getusedcolumns(ctx, te->repeat.subendcolumn, fill);

     /* Scan the left subexpr. */
     tcexpr_getusedcolumns(ctx, te->repeat.sub, fill);

     /* done */
     break;

    case TCE_SCAN:
     columnarray_mayberemoveptr(fill, te->scan.leftobjcolumn);
     columnarray_mayberemoveptr(fill, te->scan.edgecolumn);
     columnarray_mayberemoveptr(fill, te->scan.rightobjcolumn);
     break;
    case TCE_BOP:
     /* like join, this should probably split left and right (XXX?) */
     tcexpr_getusedcolumns(ctx, te->bop.left, fill);
     tcexpr_getusedcolumns(ctx, te->bop.right, fill);
     break;
    case TCE_UOP:
     tcexpr_getusedcolumns(ctx, te->uop.sub, fill);
     break;
    case TCE_FUNC:
     /* like join, this should probably split the args (XXX?) */
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	tcexpr_getusedcolumns(ctx, tcexprarray_get(&te->func.args, i), fill);
     }
     break;
    case TCE_MAP:
     tcexpr_getusedcolumns(ctx, te->map.result, fill);
     tcexpr_getusedcolumns(ctx, te->map.set, fill);
     break;
    case TCE_LET:
     tcexpr_getusedcolumns(ctx, te->let.body, fill);
     tcexpr_getusedcolumns(ctx, te->let.value, fill);
     break;
    case TCE_READVAR:
     break;
    case TCE_READGLOBAL:
     break;
    case TCE_CREATEPATHELEMENT:
     {
	struct datatype *t;

	t = te->createpathelement->datatype;
	PQLASSERT(datatype_istuple(t));
	PQLASSERT(datatype_arity(t) == 3);
	for (i=0; i<3; i++) {
	   column_getusedcolumns(ctx, datatype_tuple_getcolumn(t, i), fill);
	}
     }
     tcexpr_getusedcolumns(ctx, te->createpathelement, fill);
     break;
    case TCE_SPLATTER:
     tcexpr_getusedcolumns(ctx, te->splatter.name, fill);
     tcexpr_getusedcolumns(ctx, te->splatter.value, fill);
     break;
    case TCE_TUPLE:
     /* like join, this should probably split the elements (XXX?) */
     num = tcexprarray_num(&te->tuple.exprs);
     for (i=0; i<num; i++) {
	struct column *col;

	col = columnarray_get(te->tuple.columns, i);
	if (columnarray_contains(te->tuple.columns, col)) {
	   columnarray_mayberemoveptr(fill, col);
	   tcexpr_getusedcolumns(ctx, tcexprarray_get(&te->tuple.exprs, i),
				 fill);
	}
     }
     break;
    case TCE_VALUE:
     break;
   }
}

/*
 * Check quickly if TCA and TCB are the same. This is not meant to be
 * real CSE, just a simple way to scoop up obvious multiple references
 * to the same variable.
 */
static bool tcexpr_simple_same(struct tcexpr *tca, struct tcexpr *tcb) {
   if (tca->type != tcb->type) {
      return false;
   }
   switch (tca->type) {
    case TCE_PROJECTTUPLE:
     if (!tcexpr_simple_same(tca->projecttuple.sub, tcb->projecttuple.sub)) {
	return false;
     }
     return columnarray_eq(tca->projecttuple.cols, tcb->projecttuple.cols);
    case TCE_PROJECTSCALAR:
     if (!tcexpr_simple_same(tca->projectscalar.sub, tcb->projectscalar.sub)) {
	return false;
     }
     return tca->projectscalar.col == tcb->projectscalar.col;
    case TCE_BOP:
     if (tca->bop.op != tcb->bop.op) {
	return false;
     }
     if (tcexpr_simple_same(tca->bop.left, tcb->bop.left) &&
	 tcexpr_simple_same(tca->bop.right, tcb->bop.right)) {
	return true;
     }
     if (function_commutes(tca->bop.op) &&
	 tcexpr_simple_same(tca->bop.left, tcb->bop.right) &&
	 tcexpr_simple_same(tca->bop.right, tcb->bop.left)) {
	return true;
     }
     return false;
    case TCE_UOP:
     if (tca->uop.op != tcb->uop.op) {
	return false;
     }
     return tcexpr_simple_same(tca->uop.sub, tcb->uop.sub);

    case TCE_READVAR:
     return tca->readvar == tcb->readvar;

    case TCE_READGLOBAL:
     return tca->readglobal == tcb->readglobal;

    case TCE_VALUE:
     return pqlvalue_identical(tca->value, tcb->value);

    default:
     break;
   }
   return false;
}

/*
 * Create a copy of TMPL using SUB as its subexpression.
 * Only works for certain cases.
 */
static struct tcexpr *makewrapper(struct pqlcontext *ctx,
				  const struct tcexpr *tmpl,
				  struct tcexpr *sub) {
   struct pqlcontext *pql;
   struct tcexpr *ret;
   struct column *col;
   unsigned i, num;

   /* in case we want to stick a real context in here again */
   pql = ctx /*->pql*/;

   switch (tmpl->type) {
    case TCE_PROJECTTUPLE:
     if (tmpl->projecttuple.sub != NULL) {
	sub = makewrapper(ctx, tmpl->projecttuple.sub, sub);
     }
     ret = mktcexpr_projecttuple(pql, sub);
     num = columnarray_num(tmpl->projecttuple.cols);
     for (i=0; i<num; i++) {
	col = columnarray_get(tmpl->projecttuple.cols, i);
	column_incref(col);
	columnarray_add(pql, ret->projecttuple.cols, col, NULL);
     }
     ret->datatype = datatype_subprojecttuple(ret->projecttuple.sub->datatype,
					      ret->projecttuple.cols);
     break;
    case TCE_PROJECTSCALAR:
     if (tmpl->projectscalar.sub != NULL) {
	sub = makewrapper(ctx, tmpl->projectscalar.sub, sub);
     }
     column_incref(tmpl->projectscalar.col);
     ret = mktcexpr_projectscalar(pql, sub, tmpl->projectscalar.col);
     ret->datatype =datatype_subprojectscalar(ret->projectscalar.sub->datatype,
					      ret->projectscalar.col);
     break;
    case TCE_DISTINGUISH:
     if (tmpl->distinguish.sub != NULL) {
	sub = makewrapper(ctx, tmpl->distinguish.sub, sub);
     }
     column_incref(tmpl->distinguish.newcol);
     ret = mktcexpr_distinguish(pql, sub, tmpl->distinguish.newcol);
     ret->datatype = datatype_distinguish(ret->distinguish.sub->datatype,
					  ret->distinguish.newcol);
     break;
    case TCE_ADJOIN:
     {
	struct tcexpr *newexpr;
	struct tcvar *newvar;

	if (tmpl->adjoin.left != NULL) {
	   sub = makewrapper(ctx, tmpl->adjoin.left, sub);
	}
	newexpr = tcexpr_clone(pql, tmpl->adjoin.expr);
	if (tmpl->adjoin.var != NULL) {
	   newvar = mktcvar_fresh(pql);
	   newvar->datatype = tmpl->adjoin.var->datatype;
	   tcexpr_replacevar(pql, newexpr, tmpl->adjoin.var, newvar);
	}
	else {
	   newvar = NULL;
	}
	column_incref(tmpl->adjoin.newcol);
	ret = mktcexpr_adjoin(pql, sub, newvar, newexpr, tmpl->adjoin.newcol);
	ret->datatype = datatype_adjoin(ret->adjoin.left->datatype,
					ret->adjoin.expr->datatype,
					ret->adjoin.newcol);
	PQLASSERT(ret->datatype != NULL);
     }
     break;
    default:
     PQLASSERT(0);
     return NULL;
   }

   /* if this fails, most probable cause is bad projection columns */
   PQLASSERT(ret->datatype != NULL);
   return ret;
}

/*
 * Wrap any uses of VAR in/under TE in a copy of TMPL. Assume this is
 * because the type of VAR has been changed and TMPL inverts that
 * change.
 */
struct tcexpr *wrapvar(struct pqlcontext *ctx, struct tcexpr *te,
			      struct tcvar *var,
			      const struct tcexpr *tmpl) {
   struct pqlcontext *pql;
   struct tcexpr *subexpr;
   unsigned i, num;

   /* in case we want to stick a real context in here again */
   pql = ctx /*->pql*/;

   switch (te->type) {
    case TCE_FILTER:
     te->filter.sub = wrapvar(ctx, te->filter.sub, var, tmpl);
     te->filter.predexpr = wrapvar(ctx, te->filter.predexpr, var, tmpl);
     break;
    case TCE_PROJECTTUPLE:
     te->projecttuple.sub = wrapvar(ctx, te->projecttuple.sub, var, tmpl);
     break;
    case TCE_PROJECTSCALAR:
     te->projectscalar.sub = wrapvar(ctx, te->projectscalar.sub, var, tmpl);
     break;
    case TCE_RENAME:
     te->rename.sub = wrapvar(ctx, te->rename.sub, var, tmpl);
     break;
    case TCE_JOIN:
     te->join.left = wrapvar(ctx, te->join.left, var, tmpl);
     te->join.right = wrapvar(ctx, te->join.right, var, tmpl);
     if (te->join.predexpr != NULL) {
	te->join.predexpr = wrapvar(ctx, te->join.predexpr, var, tmpl);
     }
     break;
    case TCE_ORDER:
     te->order.sub = wrapvar(ctx, te->order.sub, var, tmpl);
     break;
    case TCE_UNIQ:
     te->uniq.sub = wrapvar(ctx, te->uniq.sub, var, tmpl);
     break;
    case TCE_NEST:
     te->nest.sub = wrapvar(ctx, te->nest.sub, var, tmpl);
     break;
    case TCE_UNNEST:
     te->unnest.sub = wrapvar(ctx, te->unnest.sub, var, tmpl);
     break;
    case TCE_DISTINGUISH:
     te->distinguish.sub = wrapvar(ctx, te->distinguish.sub, var, tmpl);
     break;
    case TCE_ADJOIN:
     te->adjoin.left = wrapvar(ctx, te->adjoin.left, var, tmpl);
     te->adjoin.expr = wrapvar(ctx, te->adjoin.expr, var, tmpl);
     break;
    case TCE_STEP:
     te->join.left = wrapvar(ctx, te->step.sub, var, tmpl);
     if (te->step.predexpr != NULL) {
	te->step.predexpr = wrapvar(ctx, te->step.predexpr, var, tmpl);
     }
     break;
    case TCE_REPEAT:
     te->repeat.sub = wrapvar(ctx, te->repeat.sub, var, tmpl);
     te->repeat.body = wrapvar(ctx, te->repeat.body, var, tmpl);
     break;
    case TCE_SCAN:
     /* nothing */
     break;
    case TCE_BOP:
     te->bop.left = wrapvar(ctx, te->bop.left, var, tmpl);
     te->bop.right = wrapvar(ctx, te->bop.right, var, tmpl);
     break;
    case TCE_UOP:
     te->uop.sub = wrapvar(ctx, te->uop.sub, var, tmpl);
     break;
    case TCE_FUNC:
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	subexpr = tcexprarray_get(&te->func.args, i);
	subexpr = wrapvar(ctx, subexpr, var, tmpl);
	tcexprarray_set(&te->func.args, i, subexpr);
     }
     break;
    case TCE_MAP:
     te->map.set = wrapvar(ctx, te->map.set, var, tmpl);
     te->map.result = wrapvar(ctx, te->map.result, var, tmpl);
     break;
    case TCE_LET:
     te->let.value = wrapvar(ctx, te->let.value, var, tmpl);
     te->let.body = wrapvar(ctx, te->let.body, var, tmpl);
     break;
    case TCE_READVAR:
     if (te->readvar == var) {
	struct datatype *oldtype;

	oldtype = te->datatype;
	te->datatype = var->datatype;
	te = makewrapper(ctx, tmpl, te);
	PQLASSERT(datatype_eq_withnames(te->datatype, oldtype));
     }
     break;
    case TCE_READGLOBAL:
     break;
    case TCE_CREATEPATHELEMENT:
     te->createpathelement = wrapvar(ctx, te->createpathelement, var, tmpl);
     break;
    case TCE_SPLATTER:
     te->splatter.value = wrapvar(ctx, te->splatter.value, var, tmpl);
     te->splatter.name = wrapvar(ctx, te->splatter.name, var, tmpl);
     break;
    case TCE_TUPLE:
     num = tcexprarray_num(&te->tuple.exprs);
     for (i=0; i<num; i++) {
	subexpr = tcexprarray_get(&te->tuple.exprs, i);
	subexpr = wrapvar(ctx, subexpr, var, tmpl);
	tcexprarray_set(&te->tuple.exprs, i, subexpr);
     }
     break;
    case TCE_VALUE:
     break;
   }
   return te;
}

////////////////////////////////////////////////////////////
// larger operations

/*
 * change a column reference to a variable reference
 */
static struct tcexpr *tcexpr_column2var(struct baseopt *ctx,
					struct tcexpr *te,
					struct column *oldcol,
					struct tcvar *newvar) {
   unsigned i, num;

   if (te == NULL) {
      return NULL;
   }
   switch (te->type) {
    case TCE_FILTER:
     te->filter.sub = tcexpr_column2var(ctx, te->filter.sub, oldcol, newvar);
     te->filter.predexpr = 
	tcexpr_column2var(ctx, te->filter.predexpr, oldcol, newvar);
     break;

    case TCE_PROJECTTUPLE:
     if (columnarray_contains(te->projecttuple.cols, oldcol)) {
	if (columnarray_num(te->projecttuple.cols) == 1) {
	   /* kill off the whole works, including te->projecttuple.sub */
	   tcexpr_destroy(ctx->pql, te);
	   tcvar_incref(newvar);
	   te = mktcexpr_readvar(ctx->pql, newvar);
	   te->datatype = newvar->datatype;
	}
	else {
	   /* yuk, let's hope we don't need to implement this */
	   PQLASSERT(0); // XXX
	}
     }
     else {
	te->projecttuple.sub = tcexpr_column2var(ctx, te->projecttuple.sub,
						 oldcol, newvar);
     }
     break;

    case TCE_PROJECTSCALAR:
     if (te->projectscalar.col == oldcol) {
	/* kill off the whole works, including te->projectscalar.sub */
	tcexpr_destroy(ctx->pql, te);
	tcvar_incref(newvar);
	te = mktcexpr_readvar(ctx->pql, newvar);
	te->datatype = newvar->datatype;
     }
     else {
	te->projectscalar.sub = tcexpr_column2var(ctx, te->projectscalar.sub,
						  oldcol, newvar);
     }
     break;

    case TCE_RENAME:
     if (te->rename.outsidecol == oldcol) {
	struct tcexpr *tmp;

	tmp = tcexpr_column2var(ctx, te->rename.sub,
				te->rename.insidecol, newvar);
	te->rename.sub = NULL;
	tcexpr_destroy(ctx->pql, te);
	te = tmp;
     }
     else {
	te->rename.sub = tcexpr_column2var(ctx, te->rename.sub,
					   oldcol, newvar);
     }
     break;

    case TCE_JOIN:
     te->join.left = tcexpr_column2var(ctx, te->join.left, oldcol, newvar);
     te->join.right = tcexpr_column2var(ctx, te->join.right, oldcol, newvar);
     te->join.predexpr = tcexpr_column2var(ctx, te->join.predexpr,
					   oldcol, newvar);
     break;

    case TCE_ORDER:
    case TCE_UNIQ:
     /* should not happen */
     PQLASSERT(0);
     break;

    case TCE_NEST:
    case TCE_UNNEST:
     /* cannot handle, but should not happen, I hope */
     PQLASSERT(0); // XXX?
     break;

    case TCE_DISTINGUISH:
     /* the column came from outside, so we can't have created it */
     PQLASSERT(oldcol != te->distinguish.newcol);
     te->distinguish.sub = tcexpr_column2var(ctx, te->distinguish.sub,
					     oldcol, newvar);
     break;

    case TCE_ADJOIN:
     /* same deal */
     PQLASSERT(oldcol != te->adjoin.newcol);
     te->adjoin.left = tcexpr_column2var(ctx, te->adjoin.left, oldcol, newvar);
     te->adjoin.expr = tcexpr_column2var(ctx, te->adjoin.expr, oldcol, newvar);
     break;

    case TCE_STEP:
     /* this one may not work */
     PQLASSERT(oldcol != te->step.subcolumn);
     /* same deal again */
     PQLASSERT(oldcol != te->step.leftobjcolumn);
     PQLASSERT(oldcol != te->step.edgecolumn);
     PQLASSERT(oldcol != te->step.rightobjcolumn);
     te->step.sub = tcexpr_column2var(ctx, te->step.sub,
				      oldcol, newvar);
     te->step.predexpr = tcexpr_column2var(ctx, te->step.predexpr,
					   oldcol, newvar);
     break;
     
    case TCE_REPEAT:
     /* this one may not work */
     PQLASSERT(oldcol != te->repeat.subendcolumn);
     /* same deal again */
     PQLASSERT(oldcol != te->repeat.bodystartcolumn);
     PQLASSERT(oldcol != te->repeat.bodypathcolumn);
     PQLASSERT(oldcol != te->repeat.bodyendcolumn);
     PQLASSERT(oldcol != te->repeat.repeatpathcolumn);
     PQLASSERT(oldcol != te->repeat.repeatendcolumn);
     te->repeat.sub = tcexpr_column2var(ctx, te->repeat.sub,
					oldcol, newvar);
     te->repeat.body = tcexpr_column2var(ctx, te->repeat.body,
					 oldcol, newvar);
     break;

    case TCE_SCAN:
     /* same deal again */
     PQLASSERT(oldcol != te->scan.leftobjcolumn);
     PQLASSERT(oldcol != te->scan.edgecolumn);
     PQLASSERT(oldcol != te->scan.rightobjcolumn);
     break;

    case TCE_BOP:
     te->bop.left = tcexpr_column2var(ctx, te->bop.left, oldcol, newvar);
     te->bop.right = tcexpr_column2var(ctx, te->bop.right, oldcol, newvar);
     break;

    case TCE_UOP:
     te->uop.sub = tcexpr_column2var(ctx, te->uop.sub, oldcol, newvar);
     break;

    case TCE_FUNC:
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	struct tcexpr *sub;

	sub = tcexprarray_get(&te->func.args, i);
	sub = tcexpr_column2var(ctx, sub, oldcol, newvar);
	tcexprarray_set(&te->func.args, i, sub);
     }
     break;

    case TCE_MAP:
     te->map.set = tcexpr_column2var(ctx, te->map.set, oldcol, newvar);
     te->map.result = tcexpr_column2var(ctx, te->map.result, oldcol, newvar);
     break;

    case TCE_LET:
     te->let.value = tcexpr_column2var(ctx, te->let.value, oldcol, newvar);
     te->let.body = tcexpr_column2var(ctx, te->let.body, oldcol, newvar);
     break;

    case TCE_READVAR:
    case TCE_READGLOBAL:
     break;
     
    case TCE_CREATEPATHELEMENT:
     te->createpathelement = tcexpr_column2var(ctx, te->createpathelement,
					       oldcol, newvar);
     break;

    case TCE_SPLATTER:
     te->splatter.value = tcexpr_column2var(ctx, te->splatter.value,
					  oldcol, newvar);
     te->splatter.name = tcexpr_column2var(ctx, te->splatter.name,
				       oldcol, newvar);
     break;

    case TCE_TUPLE:
     num = tcexprarray_num(&te->tuple.exprs);
     for (i=0; i<num; i++) {
	struct tcexpr *sub;

	sub = tcexprarray_get(&te->tuple.exprs, i);
	sub = tcexpr_column2var(ctx, sub, oldcol, newvar);
	tcexprarray_set(&te->tuple.exprs, i, sub);
     }
     break;

    case TCE_VALUE:
     break;
   }
   return te;
}

/*
 * Replace uses of OLDVAR (readvar nodes, not bindings) with NEWVAR.
 */
static void tcexpr_replacevar(struct pqlcontext *ctx,
			      struct tcexpr *te,
			      struct tcvar *oldvar,
			      struct tcvar *newvar) {
   unsigned i, num;

   if (te == NULL) {
      return;
   }

   switch (te->type) {
    case TCE_FILTER:
     tcexpr_replacevar(ctx, te->filter.sub, oldvar, newvar);
     tcexpr_replacevar(ctx, te->filter.predexpr, oldvar, newvar);
     break;

    case TCE_PROJECTTUPLE:
     tcexpr_replacevar(ctx, te->projecttuple.sub, oldvar, newvar);
     break;

    case TCE_PROJECTSCALAR:
     tcexpr_replacevar(ctx, te->projectscalar.sub, oldvar, newvar);
     break;

    case TCE_RENAME:
     tcexpr_replacevar(ctx, te->rename.sub, oldvar, newvar);
     break;

    case TCE_JOIN:
     tcexpr_replacevar(ctx, te->join.left, oldvar, newvar);
     tcexpr_replacevar(ctx, te->join.right, oldvar, newvar);
     tcexpr_replacevar(ctx, te->join.predexpr, oldvar, newvar);
     break;

    case TCE_ORDER:
     tcexpr_replacevar(ctx, te->order.sub, oldvar, newvar);
     break;

    case TCE_UNIQ:
     tcexpr_replacevar(ctx, te->uniq.sub, oldvar, newvar);
     break;

    case TCE_NEST:
     tcexpr_replacevar(ctx, te->nest.sub, oldvar, newvar);
     break;

    case TCE_UNNEST:
     tcexpr_replacevar(ctx, te->unnest.sub, oldvar, newvar);
     break;

    case TCE_DISTINGUISH:
     tcexpr_replacevar(ctx, te->distinguish.sub, oldvar, newvar);
     break;

    case TCE_ADJOIN:
     tcexpr_replacevar(ctx, te->adjoin.left, oldvar, newvar);
     tcexpr_replacevar(ctx, te->adjoin.expr, oldvar, newvar);
     break;

    case TCE_STEP:
     tcexpr_replacevar(ctx, te->step.sub, oldvar, newvar);
     tcexpr_replacevar(ctx, te->step.predexpr, oldvar, newvar);
     break;

    case TCE_REPEAT:
     tcexpr_replacevar(ctx, te->repeat.sub, oldvar, newvar);
     tcexpr_replacevar(ctx, te->repeat.body, oldvar, newvar);
     break;

    case TCE_SCAN:
     /* nothing */
     break;

    case TCE_BOP:
     tcexpr_replacevar(ctx, te->bop.left, oldvar, newvar);
     tcexpr_replacevar(ctx, te->bop.right, oldvar, newvar);
     break;

    case TCE_UOP:
     tcexpr_replacevar(ctx, te->uop.sub, oldvar, newvar);
     break;

    case TCE_FUNC:
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	tcexpr_replacevar(ctx, tcexprarray_get(&te->func.args, i),
			  oldvar, newvar);
     }
     break;

    case TCE_MAP:
     tcexpr_replacevar(ctx, te->map.set, oldvar, newvar);
     tcexpr_replacevar(ctx, te->map.result, oldvar, newvar);
     break;

    case TCE_LET:
     tcexpr_replacevar(ctx, te->let.value, oldvar, newvar);
     tcexpr_replacevar(ctx, te->let.body, oldvar, newvar);
     break;

    case TCE_READVAR:
     if (te->readvar == oldvar) {
	tcvar_decref(ctx /*->pql*/, te->readvar);
	tcvar_incref(newvar);
	te->readvar = newvar;
	/* types? XXX */
     }
     break;

    case TCE_READGLOBAL:
     break;

    case TCE_CREATEPATHELEMENT:
     tcexpr_replacevar(ctx, te->createpathelement, oldvar, newvar);
     break;

    case TCE_SPLATTER:
     tcexpr_replacevar(ctx, te->splatter.value, oldvar, newvar);
     tcexpr_replacevar(ctx, te->splatter.name, oldvar, newvar);
     break;

    case TCE_TUPLE:
     num = tcexprarray_num(&te->tuple.exprs);
     for (i=0; i<num; i++) {
	tcexpr_replacevar(ctx, tcexprarray_get(&te->tuple.exprs, i),
			  oldvar, newvar);
     }
     break;

    case TCE_VALUE:
     break;
   }
}

////////////////////////////////////////////////////////////
// predicate splitting

/*
 * Given an expression TE, and two tuple-set types LT and RT, divide
 * TE into three pieces: LEFT, which makes use only of the columns in
 * LT, RIGHT, which makes use only of the columns of RT, and CENTER,
 * which can use columns from both.
 *
 * TE must be a lambda-expression; the type of TE's argument should be
 * a single element of LT x RT.
 *
 * Specifically, given:
 *
 *    (A x B) where TE
 *
 * we want an equivalent expression of the form
 *
 *    ((A where LEFT) x (B where RIGHT)) where CENTER
 *
 * where A has type LT and B has type RT.
 *
 * The general case of this splitting can get arbitrarily complicated;
 * for example, imagine something like
 *
 *    select ... from ... as A, ... as B
 *    where count((select N1 from A.name{N1})
 *                union
 *                (select N2 from B.name{N2})) > 3
 *
 * where each half of the union could be pushed inside the main A x B
 * join. This kind of thing is not the purpose of the code here; if we
 * want to do that, and we probably do for some things, it should have
 * its own standalone pass.
 *
 * All we're going to do here is take advantage of the equivalence
 *
 *    (A x B) where (lambda X: C(X) and D(X) and E(X))
 *
 *               <=>
 *
 *    (A where lambda Y: C(Y)) x (B where lambda Z: D(Z)) where lambda X: E(X)
 *
 * that applies when C touches only columns from A and D touches only
 * columns from B.
 *
 * To this end we're only going to recursively process boolean
 * expressions. (While in theory one might see something degenerate
 * like "(A and B ? 1 : 3) > 2" that could still be handled, in
 * practice we'll assume such crap will be mopped up elsewhere first.)
 *
 * At the bottom of each branch of the boolean logic, we ask the
 * subexpression where it can be evaluated; this can be LEFT, CENTER,
 * RIGHT, or ANY. Then we do the following transforms:
 *
 *   e:LEFT          e:CENTER        e:RIGHT
 *   ------------    ------------    ------------
 *   e  =>  e|-|-    e  =>  -|e|-    e  =>  -|-|e
 *
 * Note that e1|e2|e3 should be read "e1 is true on the left side, AND
 * e3 is true on the right side, AND e2 is true on the cartesian
 * product". (- = true.) Based on this we observe that:
 *
 *   (A and B and C) or (D and E and F)
 *  <=>
 *   (A or D) and (A or E) and (A or F) and
 *   (B or D) and (B or E) and (B or F) and
 *   (C or D) and (C or E) and (C or F)
 *  <=>
 *   (A or D) and
 *   (A or (E and F)) and (B or (D and E and F)) and
 *   (C or (D and E)) and
 *   (C or F)
 *
 * and write the following reduction rules:
 *
 *   ------------------------
 *   e1|e2|e3 and e4|e5|e6  =>  e1 and e4|e2 and e5|e3 and e6
 *
 *   ------------------------
 *   e1|e2|e3 or e4|e5|e6  =>
 *      e1 or e4|((e1 or (e5 and e6)) and (e2 or (e4 and e5 and e6))
 *        and (e3 or (e4 and e5)))|e3 or e6
 *
 *   ------------------------
 *   not e1|e2|e3  =>  -|(not e1) or (not e2) or (not e3)|-
 *
 *   ------------------------
 *   not e|-|-  =>  (not e)|-|-
 *
 *   ------------------------
 *   not -|e|-  =>  -|(not e)|-
 *
 *   ------------------------
 *   not -|-|e  =>  -|-|(not e)
 *
 * The NOT cases follow from the expansion of not(A and B and C)
 * combined with there being no need to move an expression to the
 * center unless it has to interact with another one already in
 * the center or from the other side.
 *
 * Note that the mess for OR can be shown to reduce to the easily
 * understood common cases
 *
 *   e1|-|-  or -|-|e6   =>  e1|e1 or e6|e6
 *   e1|-|e3 or e4|-|e6  =>  e1 or e4|(e1 or e6) and (e3 or e4)|e3 or e6
 *
 * However, this is not sufficient; we also need to deal with other
 * things that can generate booleans that can't or won't be removed
 * elsewhere easily and that might appear in or near the top of a
 * predicate. That in particular means MAP and LET.
 *
 * - If the bound variable of either MAP or LET is a boolean, recurse
 * on the expression that generates it and treat it as above when
 * handling the body.
 *
 * - If the bound variable is not a boolean, recurse on it only to
 * figure out what side it belongs to and retain that for handling
 * the body.
 *
 * Note that the side for a tuple or tuple set is one primitive side
 * value (LEFT/CENTER/RIGHT/ANY) for each column.
 *
 * In the code, a NULL tcexpr in LEFT or RIGHT indicates an unknown
 * value, that is, something that can only be computed on the other
 * side; in CENTER it indicates an unneeded value that can be computed
 * purely on the leaves.
 */

/*
 * Side value.
 *
 * This is a form of abstract interpretation; all scalars are mapped
 * to LEFT, CENTER, RIGHT, or ANY, but the structure of sets and
 * tuples is preserved.
 *
 * It's not clear that we need the structure of the sets (only the
 * tuples) but trying to do without the set wrapping can lead to
 * infinite numbers of extremely subtle problems, and has in other
 * contexts, so let's not.
 *
 * Note that a tuple or tuple set that's all LEFT can and should be
 * just handled as PS_LEFT.
 */
enum psside_types {
   PS_LEFT,
   PS_CENTER,
   PS_RIGHT,
   PS_ANY,
   PS_SET,
   PS_TUPLE,
};
struct psside {
   enum psside_types type;
   union {
      struct psside *set;
      struct pssidearray *tuple;
   };
};
DECLARRAY(psside);
DEFARRAY(psside, );

/*
 * There's enough crap here that we'll make a new context structure.
 *
 * struct psctx contains the invariant and shared stuff; struct psret
 * contains the results, which will be different at every level.
 *
 * The returned expressions in LEFT, CENTER, and RIGHT must not alter
 * or reuse TE - if something at a higher level discovers that
 * splitting is impossible, it may be necessary to throw away what's
 * been done so far and go back to the original form, and that cannot
 * easily be done in general if the original form has been destroyed.
 * (TE is marked const to facilitate this but because const in C isn't
 * transitive that's a long way from foolproof.)
 */

struct psctx {
   struct pqlcontext *pql;
   struct tcvararray sidevars;
   struct pssidearray sidevals;
};

struct psret {
   struct tcexpr *left;
   struct tcexpr *center;
   struct tcexpr *right;
   struct tcexpr *any;
};

//////////////////////////////
// psside functions

/*
 * create one of the base types
 */
static struct psside *mkpsside_base(struct psctx *ctx, enum psside_types type){
   struct psside *ps;

   ps = domalloc(ctx->pql, sizeof(*ps));
   ps->type = type;

   return ps;
}

/*
 * create a set
 */
static struct psside *mkpsside_set(struct psctx *ctx, struct psside *member) {
   struct psside *ps;

   ps = domalloc(ctx->pql, sizeof(*ps));
   ps->type = PS_SET;
   ps->set = member;

   return ps;
}

/*
 * create a tuple; set the size but don't fill it in (caller does that)
 */
static struct psside *mkpsside_tuple(struct psctx *ctx, unsigned arity) {
   struct psside *ps;
   unsigned i;

   ps = domalloc(ctx->pql, sizeof(*ps));
   ps->type = PS_TUPLE;
   ps->tuple = pssidearray_create(ctx->pql);
   pssidearray_setsize(ctx->pql, ps->tuple, arity);
   for (i=0; i<arity; i++) {
      pssidearray_set(ps->tuple, i, NULL);
   }

   return ps;
}

/*
 * copy
 */
static struct psside *psside_clone(struct psctx *ctx, const struct psside *o) {
   struct psside *ps, *sub;
   unsigned i, arity;

   ps = domalloc(ctx->pql, sizeof(*ps));
   ps->type = o->type;
   switch (ps->type) {
    case PS_LEFT:
    case PS_CENTER:
    case PS_RIGHT:
    case PS_ANY:
     break;
    case PS_SET:
     ps->set = psside_clone(ctx, o->set);
     break;
    case PS_TUPLE:
     ps->tuple = pssidearray_create(ctx->pql);
     arity = pssidearray_num(o->tuple);
     pssidearray_setsize(ctx->pql, ps->tuple, arity);
     for (i=0; i<arity; i++) {
	sub = psside_clone(ctx, pssidearray_get(o->tuple, i));
	pssidearray_set(ps->tuple, i, sub);
     }
     break;
   }
   return ps;
}

/*
 * destroy
 */
static void psside_destroy(struct psctx *ctx, struct psside *ps) {
   unsigned i, arity;

   if (ps == NULL) {
      return;
   }

   switch (ps->type) {
    case PS_LEFT:
    case PS_CENTER:
    case PS_RIGHT:
    case PS_ANY:
     break;
    case PS_SET:
     psside_destroy(ctx, ps->set);
     break;
    case PS_TUPLE:
     arity = pssidearray_num(ps->tuple);
     for (i=0; i<arity; i++) {
	psside_destroy(ctx, pssidearray_get(ps->tuple, i));
     }
     pssidearray_setsize(ctx->pql, ps->tuple, 0);
     pssidearray_destroy(ctx->pql, ps->tuple);
     break;
   }
   dofree(ctx->pql, ps, sizeof(*ps));
}

/*
 * return true if PS is one of the base values
 */
static bool psside_isbase(const struct psside *ps) {
   switch (ps->type) {
    case PS_LEFT:
    case PS_CENTER:
    case PS_RIGHT:
    case PS_ANY:
     return true;
    case PS_SET:
    case PS_TUPLE:
     return false;
   }
   PQLASSERT(0);
   return false;
}

/*
 * if PS is a set, unwrap it. if it's not a set, assert.
 */
static struct psside *psside_unset(struct psctx *ctx, struct psside *ps) {
   struct psside *ret;

   PQLASSERT(ps->type == PS_SET);
   ret = ps->set;
   ps->set = NULL;
   psside_destroy(ctx, ps);
   return ret;
}

/*
 * Find the toplevel index of COL in T. If it's not a toplevel column,
 * set SUB to true. Returns -1 if the column isn't there at all, or 0
 * if it is.
 *
 * XXX something like this should be a standard datatype operator.
 */
static int psside_findcolumnintype(struct datatype *t,
				   struct column *col,
				   unsigned *index_ret,
				   bool *sub_ret) {
   unsigned i, arity;

   if (datatype_isset(t)) {
      return psside_findcolumnintype(datatype_set_member(t), col,
				     index_ret, sub_ret);
   }
   else if (datatype_issequence(t)) {
      return psside_findcolumnintype(datatype_sequence_member(t), col,
				     index_ret, sub_ret);
   }
   else if (!datatype_istuple(t)) {
      return -1;
   }
   else {
      arity = datatype_arity(t);
      for (i=0; i<arity; i++) {
	 if (datatype_tuple_getcolumn(t, i) == col) {
	    *index_ret = i;
	    *sub_ret = false;
	    return 0;
	 }
      }
      for (i=0; i<arity; i++) {
	 if (psside_findcolumnintype(datatype_getnth(t, i), col,
				     index_ret, sub_ret) == 0) {
	    *index_ret = i;
	    *sub_ret = true;
	    return 0;
	 }
      }
      return -1;
   }
}

/*
 * Project one column from PS. Consumes PS and returns a replacement.
 *
 * Asserts if the column is missing.
 */
static struct psside *psside_projectscalar(struct psctx *ctx,
					   struct psside *ps,
					   struct datatype *t,
					   struct column *col) {
   struct psside *ret;
   int result;
   unsigned ix;
   bool sub;

   switch (ps->type) {
    case PS_LEFT:
    case PS_CENTER:
    case PS_RIGHT:
    case PS_ANY:
     break;
    case PS_SET:
     PQLASSERT(datatype_isset(t) || datatype_issequence(t));
     if (datatype_isset(t)) {
	t = datatype_set_member(t);
     }
     else {
	t = datatype_sequence_member(t);
     }
     ps->set = psside_projectscalar(ctx, ps->set, t, col);
     break;
    case PS_TUPLE:
     /* only bother with toplevel columns */
     PQLASSERT(datatype_istuple(t));
     PQLASSERT(pssidearray_num(ps->tuple) == datatype_arity(t));
     result = psside_findcolumnintype(t, col, &ix, &sub);
     PQLASSERT(result == 0);
     ret = psside_clone(ctx, pssidearray_get(ps->tuple, ix));
     if (sub && !psside_isbase(ret)) {
	ret = psside_projectscalar(ctx, ret, datatype_getnth(t, ix), col);
     }
     psside_destroy(ctx, ps);
     return ret;
   }
   return ps;
}

/*
 * Project some columns from PS. Consumes PS and returns a replacement.
 *
 * Asserts on missing columns.
 *
 * XXX: if we have a type like
 *   (A, B, { (C, D) })
 * and we project C, A, B from it, will this generate a result whose
 * columns match the datatype projection? What about B, C, D?
 */
static struct psside *psside_projecttuple(struct psctx *ctx,
					  struct psside *ps,
					  struct datatype *t,
					  struct columnarray *cols,
					  struct datatype *check) {
   struct column *col;
   struct psside *ret, *subside;
   unsigned num, i;
   int result;
   unsigned ix;
   bool sub;

   switch (ps->type) {
    case PS_LEFT:
    case PS_CENTER:
    case PS_RIGHT:
    case PS_ANY:
     break;
    case PS_SET:
     PQLASSERT(datatype_isset(t) || datatype_issequence(t));
     if (datatype_isset(t)) {
	t = datatype_set_member(t);
	check = datatype_set_member(check);
     }
     else {
	t = datatype_sequence_member(t);
	check = datatype_sequence_member(check);
     }
     ps->set = psside_projecttuple(ctx, ps->set, t, cols, check);
     break;
    case PS_TUPLE:
     PQLASSERT(datatype_istuple(t));
     PQLASSERT(pssidearray_num(ps->tuple) == datatype_arity(t));
     num = columnarray_num(cols);
     ret = mkpsside_tuple(ctx, num);
     for (i=0; i<num; i++) {
	col = columnarray_get(cols, i);
	result = psside_findcolumnintype(t, col, &ix, &sub);
	PQLASSERT(result == 0);
	subside = psside_clone(ctx, pssidearray_get(ps->tuple, ix));
	if (sub && !psside_isbase(subside)) {
	   subside = psside_projectscalar(ctx, subside,
					  datatype_getnth(t, ix),
					  col);
	}
	pssidearray_set(ret->tuple, i, subside);
     }
     psside_destroy(ctx, ps);
     PQLASSERT(pssidearray_num(ret->tuple) == datatype_arity(check));
     return ret;
   }
   return ps;
}

/*
 * Compute the results of combining A and B, for base types only.
 */
static enum psside_types pssidetypes_uniform(enum psside_types a,
					     enum psside_types b) {
   PQLASSERT(a != PS_SET && a != PS_TUPLE);
   PQLASSERT(b != PS_SET && b != PS_TUPLE);
   if (a == b) {
      return a;
   }
   if (a == PS_ANY) {
      return b;
   }
   if (b == PS_ANY) {
      return a;
   }
   /* all remaining cases (_ x C, C x _, L x R) yield C */
   return PS_CENTER;
}

/*
 * Figure out what side results if A and B must be together.
 * Does *not* collapse partitioned tuples.
 *
 * Consumes *both* arguments and returns a replacement value.
 */
static struct psside *psside_uniform(struct psctx *ctx,
				     struct psside *a,
				     struct psside *b) {
   unsigned i, num;
   struct psside *sub;

   if (a->type == PS_SET && b->type == PS_SET) {
      a->set = psside_uniform(ctx, a->set, psside_unset(ctx, b));
      return a;
   }
   /* XXX should we be stricter about the sets here? */
   if (a->type == PS_SET) {
      a = psside_unset(ctx, a);
      return psside_uniform(ctx, a, b);
   }
   if (b->type == PS_SET) {
      b = psside_unset(ctx, b);
      return psside_uniform(ctx, a, b);
   }

   if (a->type == PS_TUPLE && b->type == PS_TUPLE) {
      num = pssidearray_num(a->tuple);
      /* tuples had better match */
      PQLASSERT(num == pssidearray_num(b->tuple));
      for (i=0; i<num; i++) {
	 sub = pssidearray_get(a->tuple, i);
	 sub = psside_uniform(ctx, sub,
			      pssidearray_get(b->tuple, i));
	 pssidearray_set(a->tuple, i, sub);
      }
      pssidearray_setsize(ctx->pql, b->tuple, 0);
      psside_destroy(ctx, b);
      return a;
   }
   if (a->type == PS_TUPLE) {
      /* XXX do we really need this case? */

      num = pssidearray_num(a->tuple);
      for (i=0; i<num; i++) {
	 sub = pssidearray_get(a->tuple, i);
	 sub = psside_uniform(ctx, sub, psside_clone(ctx, b));
	 pssidearray_set(a->tuple, i, sub);
      }
      psside_destroy(ctx, b);
      return a;
   }
   if (b->type == PS_TUPLE) {
      return psside_uniform(ctx, b, a);
   }

   a->type = pssidetypes_uniform(a->type, b->type);
   psside_destroy(ctx, b);
   return a;
}

/*
 * require that PS be on only one side at a time (and not be a 
 * tuple partly on one side and partly on the other). destroys
 * argument and returns replacement value.
 */
static struct psside *psside_onesideonly(struct psctx *ctx, struct psside *ps){
   enum psside_types rt;
   unsigned i, num;
   struct psside *sub;

   switch (ps->type) {
    case PS_LEFT:
    case PS_CENTER:
    case PS_RIGHT:
    case PS_ANY:
     break;
    case PS_SET:
     return psside_onesideonly(ctx, psside_unset(ctx, ps));
    case PS_TUPLE:
     rt = PS_ANY;
     num = pssidearray_num(ps->tuple);
     for (i=0; i<num; i++) {
	sub = pssidearray_get(ps->tuple, i);
	sub = psside_onesideonly(ctx, sub);
	PQLASSERT(psside_isbase(sub));
	rt = pssidetypes_uniform(rt, sub->type);
	psside_destroy(ctx, sub);
     }
     pssidearray_setsize(ctx->pql, ps->tuple, 0);
     psside_destroy(ctx, ps);
     return mkpsside_base(ctx, rt);
   }
   return ps;
}

/*
 * Adjoin a column.
 */
static struct psside *psside_adjoin(struct psctx *ctx,
				    struct psside *ps,
				    struct psside *column) {
   switch (ps->type) {
    case PS_LEFT:
    case PS_CENTER:
    case PS_RIGHT:
    case PS_ANY:
     return psside_uniform(ctx, ps, column);
    case PS_SET:
     ps->set = psside_adjoin(ctx, ps->set, column);
     break;
    case PS_TUPLE:
     pssidearray_add(ctx->pql, ps->tuple, column, NULL);
     break;
   }
   return ps;
}

//////////////////////////////
// consing helpers

/*
 * Make an AND expression, treating a NULL subexpression as "true".
 */
static struct tcexpr *psand(struct psctx *ctx,
			    struct tcexpr *a, struct tcexpr *b) {
   struct tcexpr *ret;

   if (a != NULL && b != NULL) {
      ret = mktcexpr_bop(ctx->pql, a, F_AND, b);
      ret->datatype = datatype_bool(ctx->pql);
      return ret;
   }
   else if (a != NULL) {
      return a;
   }
   else {
      return b;
   }
}

/*
 * Make an OR expression, treating a NULL subexpression as "true".
 */
static struct tcexpr *psor(struct psctx *ctx,
			   struct tcexpr *a, struct tcexpr *b) {
   struct tcexpr *ret;

   if (a != NULL && b != NULL) {
      ret = mktcexpr_bop(ctx->pql, a, F_OR, b);
      ret->datatype = datatype_bool(ctx->pql);
      return ret;
   }
   else {
      /* either null -> true || x -> true */
      tcexpr_destroy(ctx->pql, a);
      tcexpr_destroy(ctx->pql, b);
      return NULL;
   }
}

/*
 * Make a NOT expression, treating a NULL subexpression as "true".
 */
static struct tcexpr *psnot(struct psctx *ctx, struct tcexpr *a) {
   struct tcexpr *ret;

   if (a != NULL) {
      ret = mktcexpr_uop(ctx->pql, F_NOT, a);
   }
   else {
      ret = mktcexpr_value(ctx->pql, pqlvalue_bool(ctx->pql, false));
   }
   ret->datatype = datatype_bool(ctx->pql);
   return ret;
}

/*
 * similarly but with 3 args
 */
static struct tcexpr *psand3(struct psctx *ctx, struct tcexpr *a, 
			     struct tcexpr *b, struct tcexpr *c) {
   return psand(ctx, psand(ctx, a, b), c);
}

static struct tcexpr *psor3(struct psctx *ctx, struct tcexpr *a, 
			    struct tcexpr *b, struct tcexpr *c) {
   if (c == NULL) {
      /* shortcut this */
      tcexpr_destroy(ctx->pql, a);
      tcexpr_destroy(ctx->pql, b);
      return NULL;
   }      
   return psor(ctx, psor(ctx, a, b), c);
}


//////////////////////////////
// operations on psret

/*
 * zero out a psret
 */
static void psret_zero(struct psret *ret) {
   ret->left = NULL;
   ret->center = NULL;
   ret->right = NULL;
   ret->any = NULL;
}

#if 0 /* not currently used */
/*
 * destroy the exprs in a psret
 */
static void psret_destroyexprs(struct psctx *ctx, struct psret *ret) {
   tcexpr_destroy(ctx->pql, ret->left);
   tcexpr_destroy(ctx->pql, ret->center);
   tcexpr_destroy(ctx->pql, ret->right);
   tcexpr_destroy(ctx->pql, ret->any);
}
#endif

/*
 * count how many non-null expressions are in a psret
 */
static unsigned psret_numexprs(const struct psret *psr) {
   unsigned num;

   num = 0;
   if (psr->left != NULL) {
      num++;
   }
   if (psr->center != NULL) {
      num++;
   }
   if (psr->right != NULL) {
      num++;
   }
   if (psr->any != NULL) {
      num++;
   }
   /* if ANY is non-null, the others should be null */
   PQLASSERT(num <= 3);
   return num;
}

/*
 * If A is ANY, update it to match B.
 *
 * XXX if A is expensive we probably shouldn't replicate it to
 * multiple sides but instead force it and B to the center.
 */
static void psret_matchany(struct psctx *ctx,
			   struct psret *a, const struct psret *b) {
   bool used;

   if (a->any != NULL) {
      PQLASSERT(a->left == NULL);
      PQLASSERT(a->center == NULL);
      PQLASSERT(a->right == NULL);

      used = false;
      if (b->left != NULL) {
	 a->left = used ? tcexpr_clone(ctx->pql, a->any) : a->any;
	 used = true;
      }
      if (b->center != NULL) {
	 a->center = used ? tcexpr_clone(ctx->pql, a->any) : a->any;
	 used = true;
      }
      if (b->right != NULL) {
	 a->right = used ? tcexpr_clone(ctx->pql, a->any) : a->any;
	 used = true;
      }
      if (used) {
	 a->any = NULL;
      }
      else {
	 /* leave a->any alone */
      }
   }
}

//////////////////////////////
// side analysis

static struct psside *psgetside(struct psctx *ctx, const struct tcexpr *te);

/*
 * stick a new variable in the binding list
 */
static void psside_bindvar(struct psctx *ctx,
			   /*const*/ struct tcvar *var,
			   struct psside *val) {
   unsigned num;

   num = tcvararray_num(&ctx->sidevars);
   PQLASSERT(num == pssidearray_num(&ctx->sidevals));

   tcvararray_add(ctx->pql, &ctx->sidevars, var, NULL);
   pssidearray_add(ctx->pql, &ctx->sidevals, val, NULL);
}

/*
 * remove the last variable from the binding list
 */
static void psside_unbindvar(struct psctx *ctx) {
   unsigned num;

   num = tcvararray_num(&ctx->sidevars);
   PQLASSERT(num == pssidearray_num(&ctx->sidevals));

   PQLASSERT(num > 0);

   tcvararray_setsize(ctx->pql, &ctx->sidevars, num - 1);
   pssidearray_setsize(ctx->pql, &ctx->sidevals, num - 1);
}

/*
 * look up the side value for a bound variable
 * (if not found, return NULL)
 */
static struct psside *psside_findvar(struct psctx *ctx, struct tcvar *var) {
   unsigned i, num;
   
   num = tcvararray_num(&ctx->sidevars);
   for (i=0; i<num; i++) {
      if (var == tcvararray_get(&ctx->sidevars, i)) {
	 return pssidearray_get(&ctx->sidevals, i);
      }
   }
   return NULL;
}

/*
 * find the side(s) TE can be evaluated on
 */
static struct psside *psgetside(struct psctx *ctx, const struct tcexpr *te) {
   struct psside *sub, *sub2, *sub3;
   struct psside *ret;
   unsigned i, num;

   /* demanded by gcc 4.1.3; does not seem to be needed */
   ret = NULL;

   switch (te->type) {
    case TCE_FILTER:
     /* even if things partition, can't filter independently on both sides */
     sub = psgetside(ctx, te->filter.sub);
     sub = psside_onesideonly(ctx, sub);
     PQLASSERT(psside_isbase(sub));

     psside_bindvar(ctx, te->filter.predvar, sub);
     sub2 = psgetside(ctx, te->filter.predexpr);
     psside_unbindvar(ctx);

     sub2 = psside_onesideonly(ctx, sub2);
     ret = psside_uniform(ctx, sub, sub2);
     break;
    case TCE_PROJECTTUPLE:
     sub = psgetside(ctx, te->projecttuple.sub);
     ret = psside_projecttuple(ctx, sub, te->projecttuple.sub->datatype,
			       te->projecttuple.cols,
			       te->datatype);
     break;
    case TCE_PROJECTSCALAR:
     sub = psgetside(ctx, te->projectscalar.sub);
     ret = psside_projectscalar(ctx, sub, te->projectscalar.sub->datatype,
				te->projectscalar.col);
     break;
    case TCE_RENAME:
     /* we only track columns by index so this doesn't need to do anything */
     ret = psgetside(ctx, te->rename.sub);
     break;
    case TCE_JOIN:
     /* even if we could split it, we shouldn't */
     sub = psgetside(ctx, te->join.left);
     sub = psside_onesideonly(ctx, sub);
     sub2 = psgetside(ctx, te->join.right);
     sub2 = psside_onesideonly(ctx, sub2);
     ret = psside_uniform(ctx, sub, sub2);
     if (te->join.predvar != NULL) {
	PQLASSERT(te->join.predexpr != NULL);
	PQLASSERT(psside_isbase(ret));
	psside_bindvar(ctx, te->join.predvar, ret);
	sub3 = psgetside(ctx, te->join.predexpr);
	psside_unbindvar(ctx);
	sub3 = psside_onesideonly(ctx, sub3);
	ret = psside_uniform(ctx, ret, sub3);
     }
     else {
	/* does not depend on input! */
	sub3 = psgetside(ctx, te->join.predexpr);
	PQLASSERT(sub3->type == PS_ANY);
     }
     break;
    case TCE_ORDER:
     sub = psgetside(ctx, te->order.sub);
     ret = psside_onesideonly(ctx, sub);
     break;
    case TCE_UNIQ:
     sub = psgetside(ctx, te->uniq.sub);
     ret = psside_onesideonly(ctx, sub);
     break;
    case TCE_NEST:
     sub = psgetside(ctx, te->nest.sub);
     ret = psside_onesideonly(ctx, sub);
     break;
    case TCE_UNNEST:
     sub = psgetside(ctx, te->unnest.sub);
     ret = psside_onesideonly(ctx, sub);
     break;
    case TCE_DISTINGUISH:
     sub = psgetside(ctx, te->distinguish.sub);
     ret = psside_onesideonly(ctx, sub);
     break;
    case TCE_ADJOIN:
     /*
      * This could be less conservative; specifically, if the adjoin
      * function uses only columns from one side it can be computed on
      * that side regardless of what else the subexpression may have
      * on the other side. However, we don't have the necessary dorsal
      * muscles here to set up the resulting transformation.
      */
     sub = psgetside(ctx, te->adjoin.left);

     psside_bindvar(ctx, te->adjoin.var, sub);
     sub2 = psgetside(ctx, te->adjoin.expr);
     psside_unbindvar(ctx);

     sub2 = psside_onesideonly(ctx, sub2);
     ret = psside_adjoin(ctx, sub, sub2);
     break;
    case TCE_STEP:
     sub = psgetside(ctx, te->step.sub);
     ret = psside_onesideonly(ctx, sub);
     if (te->step.predvar != NULL) {
	PQLASSERT(te->step.predexpr != NULL);
	PQLASSERT(psside_isbase(ret));
	psside_bindvar(ctx, te->step.predvar, ret);
	sub2 = psgetside(ctx, te->step.predexpr);
	psside_unbindvar(ctx);
	ret = psside_uniform(ctx, ret, sub2);
     }
     else {
	/* does not depend on input! */
	sub3 = psgetside(ctx, te->step.predexpr);
	PQLASSERT(sub3->type == PS_ANY);
     }
     break;
    case TCE_REPEAT:
     sub = psgetside(ctx, te->repeat.sub);
     sub = psside_onesideonly(ctx, sub);
     sub2 = psgetside(ctx, te->repeat.body);
     sub2 = psside_onesideonly(ctx, sub2);
     ret = psside_uniform(ctx, sub, sub2);
     break;
    case TCE_SCAN:
     ret = mkpsside_base(ctx, PS_ANY);
     break;
    case TCE_BOP:
     sub = psgetside(ctx, te->bop.left);
     sub2 = psgetside(ctx, te->bop.right);
     switch (te->bop.op) {
      case F_UNIONALL:
      case F_CONCAT:
       /* these involve no comparisons and thus can be done partitioned */
       ret = psside_uniform(ctx, sub, sub2);
       break;
      case F_UNION:
      case F_INTERSECT:
      case F_EXCEPT:
      case F_INTERSECTALL:
      case F_EXCEPTALL:
      case F_SUBSETOF:
      case F_ELEMENTOF:
      case F_EQ:
      case F_NOTEQ:
       /* these can't be done partitioned */
       sub = psside_onesideonly(ctx, sub);
       sub2 = psside_onesideonly(ctx, sub2);
       ret = psside_uniform(ctx, sub, sub2);
       break;

      case F_CHOOSE:
       /* one side is nil => always PS_ANY */
       if (sub->type == PS_ANY) {
	  psside_destroy(ctx, sub);
	  ret = sub2;
       }
       else {
	  PQLASSERT(sub2->type == PS_ANY);
	  psside_destroy(ctx, sub2);
	  ret = sub;
       }
       break;

      case F_AND:
      case F_OR:
       /*
	* While these two don't technically need to be uniform, in
	* that LEFT and RIGHT should really produce LEFT and RIGHT,
	* not CENTER, in this analysis we can't represent that.
	* However, we also shouldn't come here for anything that
	* matters.
	*/
       /* FALLTHROUGH */
      case F_STRFTIME:
      case F_LT:
      case F_GT:
      case F_LTEQ:
      case F_GTEQ:
      case F_LIKE:
      case F_GLOB:
      case F_GREP:
      case F_SOUNDEX:
      case F_ADD:
      case F_SUB:
      case F_MUL:
      case F_DIV:
      case F_MOD:
       /* arguments are scalars */
       PQLASSERT(psside_isbase(sub));
       PQLASSERT(psside_isbase(sub2));
       ret = psside_uniform(ctx, sub, sub2);
       break;

      case F_NONEMPTY:
      case F_MAKESET:
      case F_GETELEMENT:
      case F_MAYBEMAKESET:
      case F_COUNT:
      case F_SUM:
      case F_AVG:
      case F_MIN:
      case F_MAX:
      case F_ALLTRUE:
      case F_ANYTRUE:
      case F_NOT:
      case F_NEW:
      case F_NOW:
      case F_TOSTRING:
      case F_LEFT:
      case F_EDGE:
      case F_RIGHT:
      case F_NEG:
      case F_ABS:
      case F_IDENTITY:
       /* not binary */
       PQLASSERT(0);
       ret = NULL;
       break;
     }
     break;
    case TCE_UOP:
     sub = psgetside(ctx, te->uop.sub);
     switch (te->uop.op) {
      case F_NOT:
       /* similar to AND and OR above */
       /* FALLTHROUGH */
      case F_NONEMPTY:
      case F_COUNT:
      case F_SUM:
      case F_AVG:
      case F_MIN:
      case F_MAX:
      case F_ALLTRUE:
      case F_ANYTRUE:
      case F_TOSTRING:
      case F_NEG:
      case F_ABS:
      case F_LEFT:
      case F_EDGE:
      case F_RIGHT:
       ret = psside_onesideonly(ctx, sub);
       break;
      case F_MAKESET:
       ret = mkpsside_set(ctx, sub);
       break;
      case F_GETELEMENT:
       if (sub->type == PS_SET) {
	  sub = psside_unset(ctx, sub);
       }
       ret = sub;
       break;
      case F_MAYBEMAKESET:
      case F_IDENTITY:
       /* not allowed here */
       PQLASSERT(0);
       ret = NULL;
       break;
      case F_UNION:
      case F_INTERSECT:
      case F_EXCEPT:
      case F_UNIONALL:
      case F_INTERSECTALL:
      case F_EXCEPTALL:
      case F_SUBSETOF:
      case F_ELEMENTOF:
      case F_AND:
      case F_OR:
      case F_NEW:
      case F_STRFTIME:
      case F_NOW:
      case F_EQ:
      case F_NOTEQ:
      case F_LT:
      case F_GT:
      case F_LTEQ:
      case F_GTEQ:
      case F_LIKE:
      case F_GLOB:
      case F_GREP:
      case F_SOUNDEX:
      case F_CONCAT:
      case F_CHOOSE:
      case F_ADD:
      case F_SUB:
      case F_MUL:
      case F_DIV:
      case F_MOD:
       /* not unary */
       PQLASSERT(0);
       ret = NULL;
       break;
     }
     break;
    case TCE_FUNC:
     ret = mkpsside_base(ctx, PS_ANY);
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	sub = psgetside(ctx, tcexprarray_get(&te->func.args, i));
	ret = psside_uniform(ctx, ret, sub);
     }
     break;
    case TCE_MAP:
     sub = psgetside(ctx, te->map.set);
     if (!psside_isbase(sub)) {
	sub = psside_unset(ctx, sub);
     }
     psside_bindvar(ctx, te->map.var, sub);
     ret = psgetside(ctx, te->map.result);
     psside_unbindvar(ctx);
     psside_destroy(ctx, sub);
     break;
    case TCE_LET:
     sub = psgetside(ctx, te->let.value);
     psside_bindvar(ctx, te->let.var, sub);
     ret = psgetside(ctx, te->let.body);
     psside_unbindvar(ctx);
     psside_destroy(ctx, sub);
     break;
    case TCE_READVAR:
     ret = psside_findvar(ctx, te->readvar);
     if (ret == NULL) {
	ret = mkpsside_base(ctx, PS_ANY);
     }
     else {
	ret = psside_clone(ctx, ret);
     }
     break;
    case TCE_READGLOBAL:
     ret = mkpsside_base(ctx, PS_ANY);
     break;
    case TCE_CREATEPATHELEMENT:
     ret = psgetside(ctx, te->createpathelement);
     ret = psside_onesideonly(ctx, ret);
     break;
    case TCE_SPLATTER:
     sub = psgetside(ctx, te->splatter.value);
     sub2 = psgetside(ctx, te->splatter.name);
     ret = psside_uniform(ctx, sub, sub2);
     break;
    case TCE_TUPLE:
     num = tcexprarray_num(&te->tuple.exprs);
     ret = mkpsside_tuple(ctx, num);
     for (i=0; i<num; i++) {
	sub = psgetside(ctx, tcexprarray_get(&te->tuple.exprs, i));
	pssidearray_set(ret->tuple, i, sub);
     }
     break;
    case TCE_VALUE:
     ret = mkpsside_base(ctx, PS_ANY);
     break;
   }
   return ret;
}

//////////////////////////////
// the actual splitting

static bool psplit_isleaf(const struct tcexpr *te) {
   PQLASSERT(datatype_isbool(te->datatype));

   switch (te->type) {
    case TCE_FILTER:
    case TCE_PROJECTTUPLE:
    case TCE_RENAME:
    case TCE_JOIN:
    case TCE_ORDER:
    case TCE_UNIQ:
    case TCE_NEST:
    case TCE_UNNEST:
    case TCE_DISTINGUISH:
    case TCE_ADJOIN:
    case TCE_STEP:
    case TCE_REPEAT:
    case TCE_SCAN:
    case TCE_FUNC:
    case TCE_CREATEPATHELEMENT:
    case TCE_SPLATTER:
    case TCE_TUPLE:
     /* these cannot produce a boolean */
     PQLASSERT(0);
     break;

    case TCE_PROJECTSCALAR:
    case TCE_READGLOBAL:
    case TCE_READVAR:
    case TCE_VALUE:
     /* these are leaves */
     return true;

    case TCE_BOP:
    case TCE_UOP:
     switch (te->type == TCE_BOP ? te->bop.op : te->uop.op) {
      case F_IDENTITY:
       /* this should not appear */
       PQLASSERT(0);
       break;

      case F_UNION:
      case F_INTERSECT:
      case F_EXCEPT:
      case F_UNIONALL:
      case F_INTERSECTALL:
      case F_EXCEPTALL:
      case F_MAKESET:
      case F_MAYBEMAKESET:
      case F_COUNT:
      case F_SUM:
      case F_AVG:
      case F_MIN:
      case F_MAX:
      case F_NEW:
      case F_STRFTIME:
      case F_NOW:
      case F_TOSTRING:
      case F_CONCAT:
      case F_ADD:
      case F_SUB:
      case F_MUL:
      case F_DIV:
      case F_MOD:
      case F_NEG:
      case F_ABS:
       /* these cannot produce a boolean */
       PQLASSERT(0);
       break;

      case F_SUBSETOF:
      case F_ELEMENTOF:
      case F_NONEMPTY:
      case F_GETELEMENT:
      case F_ALLTRUE:
      case F_ANYTRUE:
      case F_LT:
      case F_GT:
      case F_LTEQ:
      case F_GTEQ:
      case F_LIKE:
      case F_GLOB:
      case F_GREP:
      case F_SOUNDEX:
       /* these produce booleans from nonbooleans and are leaves */
       return true;

      case F_EQ:
      case F_NOTEQ:
      case F_CHOOSE:
       /* these could be handled but aren't yet (FUTURE), so treat as leaves */
       return true;

      case F_AND:
      case F_OR:
      case F_NOT:
       /* these are not leaves */
       return false;
     }
     break;

    case TCE_MAP:
    case TCE_LET:
     /* these are not leaves (at least if we get to them to ask) */
     return false;
   }

   /* NOTREACHED */
   /* if something unexpected happens, true is better than false */
   return true;
}

static void psplit(struct psctx *ctx,
		   struct tcexpr *te,
		   struct psret *ret) {
   struct psside *ps;
   struct psret sub1, sub2;
   unsigned n;

   if (te == NULL) {
      /* null on *input* means "nothing", so return "nothing" */
      ret->left = ret->center = ret->right = NULL;
      return;
   }

   PQLASSERT(datatype_isbool(te->datatype));

   if (psplit_isleaf(te)) {
      /*
       * Leaf expression (e.g. a comparison) that generates a boolean.
       * Figure out which side(s) it can be evaluated on. Since it's a
       * boolean, it must be one of the base cases and not a column
       * partitioning.
       */
      ps = psgetside(ctx, te);
      PQLASSERT(psside_isbase(ps));
      psret_zero(ret);
      switch (ps->type) {
       case PS_LEFT: ret->left = te; break;
       case PS_CENTER: ret->center = te; break;
       case PS_RIGHT: ret->right = te; break;
       case PS_ANY: ret->any = te; break;
       case PS_SET:
       case PS_TUPLE:
	PQLASSERT(0);
	break;
      }
      psside_destroy(ctx, ps);
      return;
   }

   /*
    * The only things we expect to handle here directly are AND, OR,
    * NOT, LET, and MAP. Everything else either can't produce a
    * boolean, produces a boolean but is a leaf for this analysis,
    * or shouldn't appear at all.
    */

   if (te->type == TCE_BOP && te->bop.op == F_AND) {
      psplit(ctx, te->bop.left, &sub1);
      psplit(ctx, te->bop.right, &sub2);

      psret_matchany(ctx, &sub1, &sub2);
      psret_matchany(ctx, &sub2, &sub1);

      ret->left = psand(ctx, sub1.left, sub2.left);
      ret->center = psand(ctx, sub1.center, sub2.center);
      ret->right = psand(ctx, sub1.right, sub2.right);
      ret->any = psand(ctx, sub1.any, sub2.any);

      te->bop.left = te->bop.right = NULL;
      tcexpr_destroy(ctx->pql, te);
   }
   else if (te->type == TCE_BOP && te->bop.op == F_OR) {
      psplit(ctx, te->bop.left, &sub1);
      psplit(ctx, te->bop.right, &sub2);

      PQLASSERT(psret_numexprs(&sub1) > 0);
      PQLASSERT(psret_numexprs(&sub2) > 0);

      psret_matchany(ctx, &sub1, &sub2);
      psret_matchany(ctx, &sub2, &sub1);

      /* do this so it matches the description above */
      struct tcexpr *e1, *e2, *e3, *e4, *e5, *e6;
      struct tcexpr *e1a, *e3a, *e4a, *e4b, *e5a, *e5b, *e6a, *e6b;

      e1 = sub1.left;
      e2 = sub1.center;
      e3 = sub1.right;
      e4 = sub2.left;
      e5 = sub2.center;
      e6 = sub2.right;

      /*
       * XXX we should let-bind any of these that are "expensive".
       * (Or maybe just let-bind them all and let someone else sort it
       * out...)
       */
      e1a = tcexpr_clone(ctx->pql, e1);
      e3a = tcexpr_clone(ctx->pql, e3);
      e4a = tcexpr_clone(ctx->pql, e4);
      e4b = tcexpr_clone(ctx->pql, e4);
      e5a = tcexpr_clone(ctx->pql, e5);
      e5b = tcexpr_clone(ctx->pql, e5);
      e6a = tcexpr_clone(ctx->pql, e6);
      e6b = tcexpr_clone(ctx->pql, e6);

      PQLASSERT(sub1.any == NULL && sub2.any == NULL);

      ret->left = psor(ctx, e1a, e4a);
      ret->center = psand3(ctx,
			   psor(ctx, e1, psand(ctx, e5a, e6)),
			   psor(ctx, e2, psand3(ctx, e4, e5, e6b)),
			   psor(ctx, e3, psand(ctx, e4b, e5b)));
      ret->right = psor(ctx, e3a, e6a);
      ret->any = psor(ctx, sub1.any, sub2.any);

      te->bop.left = te->bop.right = NULL;
      tcexpr_destroy(ctx->pql, te);
   }
   else if (te->type == TCE_UOP && te->uop.op == F_NOT) {
      psplit(ctx, te->uop.sub, &sub1);

      n = psret_numexprs(&sub1);

      psret_zero(ret);

      if (sub1.any != NULL) {
	 PQLASSERT(n==1);
	 ret->any = psnot(ctx, sub1.any);
      }
      else if (sub1.left != NULL && n == 1) {
	 ret->left = psnot(ctx, sub1.left);
      }
      else if (sub1.center != NULL && n == 1) {
	 ret->center = psnot(ctx, sub1.center);
      }
      else if (sub1.right != NULL && n == 1) {
	 ret->right = psnot(ctx, sub1.right);
      }
      else {
	 ret->center = psor3(ctx,
			     psnot(ctx, sub1.left),
			     psnot(ctx, sub1.center),
			     psnot(ctx, sub1.right));
      }

      te->uop.sub = NULL;
      tcexpr_destroy(ctx->pql, te);
   }
   else if (te->type == TCE_MAP) {
      /*
       * The set to map over can't be a bool, so just do the side analysis
       * for the variable and bind it for use by the body.
       */
      PQLASSERT(!datatype_isbool(te->map.set->datatype));
      ps = psgetside(ctx, te->map.set);
      if (!psside_isbase(ps)) {
	 ps = psside_unset(ctx, ps);
      }
      psside_bindvar(ctx, te->map.var, ps);
      psplit(ctx, te->map.result, ret);
      psside_unbindvar(ctx);

      /* XXX this won't always be true */
      PQLASSERT(psret_numexprs(ret) == 1);
      if (ret->left != NULL) {
	 te->map.result = ret->left;
	 ret->left = te;
      }
      else if (ret->center != NULL) {
	 te->map.result = ret->center;
	 ret->center = te;
      }
      else if (ret->right != NULL) {
	 te->map.result = ret->right;
	 ret->right = te;
      }
      else /*if (ret->any != NULL)*/ {
	 te->map.result = ret->any;
	 ret->any = te;
      }

      psside_destroy(ctx, ps);
   }
   else if (te->type == TCE_LET) {
      /*
       * The let-value *can* be a bool. If it is, we could recurse and
       * split it. However, the logic we have so far doesn't allow for
       * splitting variables, only expressions; TCE_READVAR is a leaf
       * and thus must be computed on only one side. Therefore, all we
       * can do is run the side analysis on it to figure out what side
       * that will be, and bind it for use by the body like in map.
       * This is also what we'd do if it wasn't a bool, so just do that
       * unconditionally.
       *
       * At some point we might want to improve this to be able to
       * split boolean variables, but not now. (FUTURE)
       */
      ps = psgetside(ctx, te->let.value);

      /*
       * XXX: ps cannot be a boolean, but it can be a column partitioning;
       * for example, if we've got let X = Y |+| K and K depends only on
       * one half of Y. We ought to figure out how to split the variable
       * (at least if we get a split result back from the body expression)
       * but it's not entirely trivial to do so. So, for now, force the
       * variable to be on one side only.
       */
      ps = psside_onesideonly(ctx, ps);

      psside_bindvar(ctx, te->let.var, ps);
      psplit(ctx, te->let.body, ret);
      psside_unbindvar(ctx);

      /* XXX this won't always be true */
      PQLASSERT(psret_numexprs(ret) == 1);
      if (ret->left != NULL) {
	 te->let.body = ret->left;
	 ret->left = te;
      }
      else if (ret->center != NULL) {
	 te->let.body = ret->center;
	 ret->center = te;
      }
      else if (ret->right != NULL) {
	 te->let.body = ret->right;
	 ret->right = te;
      }
      else /*if (ret->any != NULL)*/ {
	 te->let.body = ret->any;
	 ret->any = te;
      }

      psside_destroy(ctx, ps);
   }
   else {
      PQLASSERT(0);
   }
}

/*
 * Create a wrapvar() template to adjoin more columns, taking the
 * columns from the type T.
 *
 * FUTURE: It would perhaps be better to make a single tcexpr node
 * that adjoins a pile of null columns, since this isn't the only
 * place we do that; or, better still, figure out how to adjust the
 * datatypes in place instead of inserting all these wrappers that'll
 * only get removed later. (Recall that the whole point of this code
 * is that the columns we split off and need to adjoin back are
 * unused.)
 */
struct tcexpr *psplit_adjoinwrapper(struct pqlcontext *ctx,
					   struct tcexpr *subtmpl,
					   struct datatype *starttype,
					   struct datatype *t) {
   struct pqlcontext *pql = ctx /*->pql*/;
   struct tcexpr *tmpl, *subexpr;
   struct tcvar *subvar;
   struct datatype *curtype, *subtype;
   struct column *subcol;
   unsigned i, num;

   PQLASSERT(datatype_istuple(t));

   curtype = starttype;

   tmpl = subtmpl;
   num = datatype_arity(t);
   for (i=0; i<num; i++) {
      subtype = datatype_getnth(t, i);
      subcol = datatype_tuple_getcolumn(t, i);

      subexpr = mktcexpr_value(pql, pqlvalue_nil(pql));
      subexpr->datatype = subtype;
      subvar = mktcvar_fresh(pql);
      subvar->datatype = curtype;
      column_incref(subcol);
      tmpl = mktcexpr_adjoin(pql, tmpl, subvar, subexpr, subcol);
      tmpl->datatype = datatype_adjoin(curtype, subtype, subcol);
      PQLASSERT(tmpl->datatype != NULL);
      curtype = tmpl->datatype;
   }

   return tmpl;
}

/*
 * Create a wrapvar() template to adjoin one more column.
 */
static struct tcexpr *psplit_adjoinonewrapper(struct pqlcontext *ctx,
					      struct tcexpr *subtmpl,
					      struct datatype *starttype,
					      struct datatype *addtype,
					      struct column *addcol) {
   struct pqlcontext *pql = ctx /*->pql*/;
   struct tcexpr *tmpl, *subexpr;
   struct tcvar *subvar;
   struct datatype *curtype, *subtype;
   struct column *subcol;

   curtype = starttype;

   tmpl = subtmpl;
   subtype = addtype;
   subcol = addcol;

   subexpr = mktcexpr_value(pql, pqlvalue_nil(pql));
   subexpr->datatype = subtype;
   subvar = mktcvar_fresh(pql);
   subvar->datatype = curtype;
   column_incref(subcol);
   tmpl = mktcexpr_adjoin(pql, tmpl, subvar, subexpr, subcol);
   tmpl->datatype = datatype_adjoin(curtype, subtype, subcol);
   PQLASSERT(tmpl->datatype != NULL);
   curtype = tmpl->datatype;

   return tmpl;
}

/*
 * Create a wrapvar() template to do a projection of the columns in LT
 * then RT, the idea being to reorder columns to the left after
 * adjoining on the right. (Maybe we should have an adjoin-on-left
 * operation...)
 *
 * XXX: This should be called just projectwrapper and it should
 * probably take columns from only one type. (Easy enough to use
 * datatype_tuple_concat before calling it if one needs two.)
 */
struct tcexpr *psplit_reorderwrapper(struct pqlcontext *ctx,
					    struct tcexpr *subtmpl,
					    struct datatype *lt,
					    struct datatype *rt) {
   struct pqlcontext *pql = ctx /*->pql*/;
   struct tcexpr *tmpl;
   struct column *col;
   unsigned leftarity, rightarity, i;

   PQLASSERT(datatype_istuple(lt));
   PQLASSERT(datatype_istuple(rt));

   leftarity = datatype_arity(lt);
   rightarity = datatype_arity(rt);

   tmpl = mktcexpr_projecttuple(pql, subtmpl);
   columnarray_setsize(pql, tmpl->projecttuple.cols,
		       leftarity + rightarity);
   for (i=0; i<leftarity; i++) {
      col = datatype_tuple_getcolumn(lt, i);
      column_incref(col);
      columnarray_set(tmpl->projecttuple.cols, i, col);
   }
   for (i=0; i<rightarity; i++) {
      col = datatype_tuple_getcolumn(rt, i);
      column_incref(col);
      columnarray_set(tmpl->projecttuple.cols, leftarity + i, col);
   }
   tmpl->datatype = datatype_tuple_concat(pql, lt, rt);

   return tmpl;
}

/*
 * Assign variables for RET. Reuse VAR, the variable from the original
 * predicate, to avoid unnecessary consing and to avoid unnecessary
 * variable substitutions, because that causes confusion when
 * inspecting the output. In particular we don't want this to allocate
 * a new variable ID every time it's run even if the psplit logic
 * didn't accomplish anything.
 *
 * LT and RT are the new types for the variables on the left and right
 * after the splitting.
 */
static void psplit_makevars(struct psctx *ctx, struct tcvar *origvar,
			    struct datatype *lt, struct datatype *rt,
			    struct psret *ret,
			    struct tcvar **leftvar_ret,
			    struct tcvar **centervar_ret,
			    struct tcvar **rightvar_ret) {
   struct tcvar *leftvar, *centervar, *rightvar;
   struct tcexpr *wraptemplate;
   PQLASSERT(ret->any == NULL);

   leftvar = centervar = rightvar = NULL;

   /* prefer center so the variable id stays put if possible */
   if (ret->center != NULL) {
      centervar = origvar;
   }
   else if (ret->left != NULL) {
      leftvar = origvar;
   }
   else if (ret->right != NULL) {
      rightvar = origvar;
   }

   if (ret->left != NULL) {
      if (leftvar == NULL) {
	 leftvar = mktcvar_fresh(ctx->pql);
	 tcexpr_replacevar(ctx->pql, ret->left, origvar, leftvar);
      }
      leftvar->datatype = lt;

      wraptemplate = psplit_adjoinwrapper(ctx->pql, NULL, lt, rt);
      ret->left = wrapvar(ctx->pql, ret->left, leftvar, wraptemplate);
      tcexpr_destroy(ctx->pql, wraptemplate);
   }
   if (ret->right != NULL) {
      if (rightvar == NULL) {
	 rightvar = mktcvar_fresh(ctx->pql);
	 rightvar->datatype = origvar->datatype;
	 tcexpr_replacevar(ctx->pql, ret->right, origvar, rightvar);
      }
      rightvar->datatype = rt;

      wraptemplate = psplit_adjoinwrapper(ctx->pql, NULL, rt, lt);
      wraptemplate = psplit_reorderwrapper(ctx->pql, wraptemplate, lt, rt);
      ret->right = wrapvar(ctx->pql, ret->right, rightvar, wraptemplate);
      tcexpr_destroy(ctx->pql, wraptemplate);
   }

   *leftvar_ret = leftvar;
   *centervar_ret = centervar;
   *rightvar_ret = rightvar;
}

/*
 * Split predicate VAR:EXPR into left, center, and right
 * halves based on the column partitioning given by the types LT and
 * RT.
 */
static void predicate_split(struct baseopt *ctx,
			    struct tcvar *var,
			    struct tcexpr *expr,
			    struct datatype *lt,
			    struct datatype *rt,
			    struct tcvar **leftvar_ret,
			    struct tcexpr **left_ret,
			    struct tcvar **centervar_ret,
			    struct tcexpr **center_ret,
			    struct tcvar **rightvar_ret,
			    struct tcexpr **right_ret) {
   struct psctx pctx;
   struct psret ret;
   struct datatype *arg, *res, *lm, *rm, *xm;
   struct psside *ps;
   unsigned leftarity, rightarity, i;

   if (var == NULL) {
      /* independent of data; always PS_ANY, so no need to split it */
      /* XXX should probably clone to left and right, as below */
      *leftvar_ret = *centervar_ret = *rightvar_ret = NULL;
      *left_ret = expr;
      *center_ret = NULL;
      *right_ret = NULL;
      return;
   }

   PQLASSERT(datatype_isset(lt));
   PQLASSERT(datatype_isset(rt));

   arg = var->datatype;
   res = expr->datatype;
   PQLASSERT(datatype_isbool(res));

   lm = datatype_set_member(lt);
   rm = datatype_set_member(rt);
   PQLASSERT(datatype_istuple(lm));
   PQLASSERT(datatype_istuple(rm));

   xm = datatype_tuple_concat(ctx->pql, lm, rm);
   PQLASSERT(datatype_eq_withnames(arg, xm));

   pctx.pql = ctx->pql;
   tcvararray_init(&pctx.sidevars);
   pssidearray_init(&pctx.sidevals);

   /*
    * Create a psside binding for the lambda variable.
    */
   leftarity = datatype_arity(lm);
   rightarity = datatype_arity(rm);
   ps = mkpsside_tuple(&pctx, leftarity + rightarity);
   for (i=0; i<leftarity; i++) {
      pssidearray_set(ps->tuple, i, mkpsside_base(&pctx, PS_LEFT));
   }
   for (i=0; i<rightarity; i++) {
      pssidearray_set(ps->tuple, leftarity+i, mkpsside_base(&pctx, PS_RIGHT));
   }
   psside_bindvar(&pctx, var, ps);

   /*
    * Do it.
    */
   psplit(&pctx, expr, &ret);

   psside_unbindvar(&pctx);
   psside_destroy(&pctx, ps);

   if (ret.any != NULL) {
      PQLASSERT(ret.left == NULL);
      PQLASSERT(ret.center == NULL);
      PQLASSERT(ret.right == NULL);
      /* XXX should probably copy it to both sides! (unless it's "big") */
      ret.left = ret.any;
      ret.any = NULL;
   }

   psplit_makevars(&pctx, var, lm, rm, &ret,
		   leftvar_ret, centervar_ret, rightvar_ret);

   tcvararray_cleanup(pctx.pql, &pctx.sidevars);
   pssidearray_cleanup(pctx.pql, &pctx.sidevals);

   *left_ret = ret.left;
   *center_ret = ret.center;
   *right_ret = ret.right;
}


////////////////////////////////////////////////////////////
// optimizations

/*
 * This module works as follows:
 *
 * 1. Process leaves first. This means that when working at a
 * particular point, all the subnodes are already optimized as far as
 * we can take them, so except in a few special cases there's no need
 * to reprocess whole subtrees again after working.
 *
 * 1a. After an optimization changes the node type in a particular
 * place in the tree, rerun any_baseopt() on it but not its unchanged
 * descendents.
 *
 * 2. Try combining first. If we can't combine, then try moving past.
 *
 * 3. Separate the transformation logic (that munges the tree) from
 * the decision logic that matches the tree and chooses what munging
 * to do. This makes it possible to review the logic independently of
 * mechanism, and is (I think) worthwhile even though most of the
 * transformations will be used exactly once.
 *
 * 4. The decision logic should repeat as long as the node it's
 * looking at remains the same type. If the node changes type it
 * should make a tail-call to any_baseopt().
 */

static struct tcexpr *any_baseopt(struct baseopt *ctx, struct tcexpr *te);

////////////////////////////////////////////////////////////
// destruction transforms

/*
 * Destroy TE and return its principal subexpression to replace it.
 *
 * For nodes that have multiple subexpressions, such as let bindings,
 * the node that's returned is the one we still need after the
 * optimization that calls here. Nodes where this isn't clear or isn't
 * uniquely determined will cause assertion failure.
 */
static struct tcexpr *tcexpr_destruct(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *ret;
   struct tcexpr **slot;

   /* demanded by gcc 4.1.3; does not seem to be needed */
   slot = NULL;

   switch (te->type) {
    case TCE_FILTER:      slot = &te->filter.sub; break;
    case TCE_PROJECTTUPLE: slot = &te->projecttuple.sub; break;
    case TCE_PROJECTSCALAR: slot = &te->projectscalar.sub; break;
    case TCE_RENAME:      slot = &te->rename.sub; break;
    case TCE_ORDER:       slot = &te->order.sub; break;
    case TCE_UNIQ:        slot = &te->uniq.sub; break;
    case TCE_NEST:        slot = &te->nest.sub; break;
    case TCE_UNNEST:      slot = &te->unnest.sub; break;
    case TCE_DISTINGUISH: slot = &te->distinguish.sub; break;
    case TCE_ADJOIN:      slot = &te->adjoin.left; break;
    case TCE_STEP:        slot = &te->step.sub; break;
    case TCE_UOP:         slot = &te->uop.sub; break;
    case TCE_MAP:         slot = &te->map.set; break;
    case TCE_LET:         slot = &te->let.value; break;
    case TCE_SPLATTER:    slot = &te->splatter.value; break;

    case TCE_JOIN:
    case TCE_REPEAT:
    case TCE_BOP:
    case TCE_FUNC:
    case TCE_CREATEPATHELEMENT:
    case TCE_TUPLE:
     /* two or more subexpressions of near-equal weight */
     PQLASSERT(0);
     return te;

    case TCE_SCAN:
    case TCE_READVAR:
    case TCE_READGLOBAL:
    case TCE_VALUE:
     /* no subexpression */
     PQLASSERT(0);
     return te;
   }

   ret = *slot;
   *slot = NULL;
   tcexpr_destroy(ctx->pql, te);
   return ret;
}

static struct tcexpr *bop_destruct_keepleft(struct baseopt *ctx,
					    struct tcexpr *te) {
   struct tcexpr *ret;

   PQLASSERT(te->type == TCE_BOP);
   ret = te->bop.left;
   te->bop.left = NULL;
   tcexpr_destroy(ctx->pql, te);
   return ret;
}

static struct tcexpr *bop_destruct_keepright(struct baseopt *ctx,
					     struct tcexpr *te) {
   struct tcexpr *ret;

   PQLASSERT(te->type == TCE_BOP);
   ret = te->bop.right;
   te->bop.right = NULL;
   tcexpr_destroy(ctx->pql, te);
   return ret;
}

////////////////////////////////////////////////////////////
// combining transforms

//////////////////////////////
// generic

/*
 * foo => constant
 * (consumes VAL)
 */
static struct tcexpr *comb_toconstant(struct baseopt *ctx,
				      struct tcexpr *te,
				      struct pqlvalue *val) {
   struct datatype *t;

   /* Keep the type, destroy the rest */
   t = te->datatype;
   te->datatype = NULL;;
   tcexpr_destroy(ctx->pql, te);

   /*
    * FUTURE: should be an op in tcalc.c to clean just the union part
    * of a tcexpr, so we don't have to allocate a new one here.
    */

   te = mktcexpr_value(ctx->pql, val);
   te->datatype = t;
   return te;
}

/*
 * uop(X x Y) => uop(X) newbop uop(Y)
 */
static struct tcexpr *comb_distribute_uop_over_join(struct baseopt *ctx,
						    struct tcexpr *te,
						    enum functions newbop) {
   struct tcexpr *sub;
   enum functions uop;

   PQLASSERT(te->type == TCE_UOP);

   uop = te->uop.op;
   sub = te->uop.sub;
   te->uop.sub = NULL;

   PQLASSERT(sub->type == TCE_JOIN);
   PQLASSERT(sub->join.predvar == NULL);
   PQLASSERT(sub->join.predexpr == NULL);

   te->type = TCE_BOP;
   te->bop.op = newbop;
   te->bop.left = mktcexpr_uop(ctx->pql, uop, sub->join.left);
   te->bop.right = mktcexpr_uop(ctx->pql, uop, sub->join.right);

   te->bop.left->datatype = te->datatype;
   te->bop.right->datatype = te->datatype;
   /* keep te->datatype unchanged */

   sub->join.left = sub->join.right = NULL;
   tcexpr_destroy(ctx->pql, sub);

   return te;
}

/*
 * uop(X bop Y) => uop(X) newbop uop(Y)
 */
static struct tcexpr *comb_distribute_uop_over_bop(struct baseopt *ctx,
						   struct tcexpr *te,
						   enum functions newbop) {
   struct tcexpr *sub;
   enum functions uop;

   PQLASSERT(te->type == TCE_UOP);

   uop = te->uop.op;
   sub = te->uop.sub;
   te->uop.sub = NULL;

   PQLASSERT(sub->type == TCE_BOP);

   te->type = TCE_BOP;
   te->bop.op = newbop;
   te->bop.left = mktcexpr_uop(ctx->pql, uop, sub->bop.left);
   te->bop.right = mktcexpr_uop(ctx->pql, uop, sub->bop.right);

   te->bop.left->datatype = te->datatype;
   te->bop.right->datatype = te->datatype;
   /* keep te->datatype unchanged */

   sub->bop.left = sub->bop.right = NULL;
   tcexpr_destroy(ctx->pql, sub);

   return te;
}

//////////////////////////////
// set

/*
 * X in (A union B) => X in A or X in B
 * X in (A intersect B) => X in A and X in B
 * X in (A except B) => X in A and not X in B
 */
static struct tcexpr *comb_element_setop(struct baseopt *ctx,
					 struct tcexpr *te,
					 enum functions newop,
					 bool addnot) {
   struct tcexpr *a, *b, *x;
   struct tcexpr *let, *newte;
   struct tcvar *letvar;

   PQLASSERT(te->type == TCE_BOP && te->bop.op == F_ELEMENTOF);
   PQLASSERT(te->bop.right->type == TCE_BOP);

   /* we will let-bind the element, to avoid having to do CSE on it later */

   letvar = mktcvar_fresh(ctx->pql);
   letvar->datatype = te->bop.left->datatype;

   a = te->bop.right->bop.left;
   b = te->bop.right->bop.right;
   te->bop.right->bop.left = NULL;
   te->bop.right->bop.right = NULL;

   x = mktcexpr_readvar(ctx->pql, letvar);
   x->datatype = letvar->datatype;

   a = mktcexpr_bop(ctx->pql, x, F_ELEMENTOF, a);
   a->datatype = datatype_bool(ctx->pql);

   x = mktcexpr_readvar(ctx->pql, letvar);
   x->datatype = letvar->datatype;

   b = mktcexpr_bop(ctx->pql, x, F_ELEMENTOF, b);
   b->datatype = datatype_bool(ctx->pql);

   if (addnot) {
      b = mktcexpr_uop(ctx->pql, F_NOT, b);
      b->datatype = b->uop.sub->datatype;
   }

   newte = mktcexpr_bop(ctx->pql, a, newop, b);
   newte->datatype = te->datatype;
   te->datatype = NULL;

   let = mktcexpr_let(ctx->pql, letvar, te->bop.left, newte);
   te->bop.left = NULL;

   tcexpr_destroy(ctx->pql, te);
   return let;
}
 
//////////////////////////////
// aggregate

/*
 * count(foo X) => count(X)
 */
static void comb_count_destroysub(struct baseopt *ctx, struct tcexpr *te) {
   PQLASSERT(te->type == TCE_UOP && te->uop.op == F_COUNT);
   te->uop.sub = tcexpr_destruct(ctx, te->uop.sub);
}

/*
 * count(unnest_a X) => sum(map X0 in (project_a X): count(X0))
 * sum(unnest_a X) => sum(map X0 in (project_a X): sum(X0))
 * max(unnest_a X) => max(map X0 in (project_a X): max(X0))
 * min(unnest_a X) => min(map X0 in (project_a X): min(X0))
 * anytrue(unnest_a X) => anytrue(map X0 in (project_a X): anytrue(X0))
 * alltrue(unnest_a X) => alltrue(map X0 in (project_a X): alltrue(X0))
 *
 * note: won't work for avg() directly
 */
static struct tcexpr *comb_agg_unnest(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *inner, *proj, *count, *map;
   struct tcvar *itervar;
   struct column *col;
   enum functions startop, combineop;

   PQLASSERT(te->type == TCE_UOP);
   PQLASSERT(te->uop.op == F_COUNT ||
	     te->uop.op == F_SUM ||
	     te->uop.op == F_MAX ||
	     te->uop.op == F_MIN ||
	     te->uop.op == F_ANYTRUE ||
	     te->uop.op == F_ALLTRUE);
   PQLASSERT(te->uop.sub->type == TCE_UNNEST);

   startop = te->uop.op;
   if (startop == F_COUNT) {
      combineop = F_SUM;
   }
   else {
      combineop = startop;
   }

   /* pull out the subexpression and the unnest column */
   inner = te->uop.sub->unnest.sub;
   col = te->uop.sub->unnest.col;
   column_incref(col);

   /* reuse the unnest expression node for the projection */

   proj = te->uop.sub;
   te->uop.sub = NULL;

   proj->type = TCE_PROJECTSCALAR;
   proj->projectscalar.sub = inner;
   proj->projectscalar.col = col;
   proj->datatype = datatype_subprojectscalar(inner->datatype, col);

   /* cons up the rest */

   itervar = mktcvar_fresh(ctx->pql);
   tcvar_incref(itervar);
   itervar->datatype = datatype_set_member(proj->datatype);

   count = mktcexpr_uop(ctx->pql, startop,
			mktcexpr_readvar(ctx->pql,itervar));
   count->datatype = datatype_int(ctx->pql);

   map = mktcexpr_map(ctx->pql, itervar, proj, count);
   map->datatype = datatype_set(ctx->pql, count->datatype);

   te->uop.op = combineop;
   te->uop.sub = map;

   return te;
}

/*
 * count(A ++ B) => count(A) + count(B)
 */
static struct tcexpr *comb_count_concat(struct baseopt *ctx,struct tcexpr *te){
   struct tcexpr *sub;

   PQLASSERT(te->type == TCE_UOP && te->uop.op == F_COUNT);
   PQLASSERT(te->uop.sub->type == TCE_BOP);

   sub = te->uop.sub;

   te->type = TCE_BOP;
   te->bop.op = F_ADD;
   te->bop.left = mktcexpr_uop(ctx->pql, F_COUNT, sub->bop.left);
   te->bop.right = mktcexpr_uop(ctx->pql, F_COUNT, sub->bop.right);

   te->bop.left->datatype = te->datatype;
   te->bop.right->datatype = te->datatype;

   sub->bop.left = NULL;
   sub->bop.right = NULL;
   tcexpr_destroy(ctx->pql, sub);

   return te;
}

////////////////////////////////////////////////////////////
// code motion transforms

/*
 * X op map A in B: C  =>  aggop(map A in B: X newop C)
 */
static struct tcexpr *move_bop_into_map_on_right(struct baseopt *ctx,
						 struct tcexpr *te,
						 enum functions aggop,
						 struct datatype *intermed,
						 enum functions newop) {
   struct tcexpr *map, *body, *agg;

   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.right->type == TCE_MAP);

   map = te->bop.right;
   body = map->map.result;

   te->bop.right = body;
   te->bop.op = newop;
   map->map.result = te;

   agg = mktcexpr_uop(ctx->pql, aggop, map);

   agg->datatype = te->datatype;

   map->datatype = datatype_set(ctx->pql, intermed);

   te->datatype = intermed;

   return agg;
}

/*
 * X op let A = B: C  =>  let A = B: X op C
 */
static struct tcexpr *move_bop_into_let_on_right(struct baseopt *ctx,
						 struct tcexpr *te) {
   struct tcexpr *let, *body;

   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.right->type == TCE_LET);

   (void)ctx;

   let = te->bop.right;
   body = let->let.body;

   te->bop.right = body;
   let->let.body = te;

   let->datatype = te->datatype;

   return let;
}

/*
 * op let A = B: C  =>  let A = B: op C
 */
static struct tcexpr *move_uop_into_let(struct baseopt *ctx,
					struct tcexpr *te) {
   struct tcexpr *let, *body;

   PQLASSERT(te->type == TCE_UOP);
   PQLASSERT(te->uop.sub->type == TCE_LET);

   (void)ctx;

   let = te->uop.sub;
   body = let->let.body;

   te->uop.sub = body;
   let->let.body = te;

   let->datatype = te->datatype;

   return let;
}

////////////////////////////////////////////////////////////
// optimization matching for functions

//////////////////////////////
// set

static struct tcexpr *union_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_UNION);

   (void)ctx;

   /*
    * A union (B union C) <=> (A union B) union C
    *
    * This requires data about estimated sizes, which we don't have
    * here, so this optimization will be done in a later pass,
    * probably as a side effect of join planning.
    *
    * FUTURE: we should consider transforming trees of unions into a
    * single multiunion(A, B, C) operation, because that will make it
    * easier to choose a processing order downstream. Alternatively it
    * allows processing them all at once, as large unions are best
    * done as basically a merge sort.
    */

   /*
    * A union B => (uniq A) union (uniq B)
    *
    * This has to wait until we figure out how to represent
    * sortedness. XXX
    */

   /*
    * (A intersect B) union (A intersect C) => A intersect (B union C)
    * (A except B) union (A except C) => A except (B intersect C)
    * (A except B) union (A intersect C) => A except (B except C)
    * etc.
    *
    * this is a pain to test, since there are four ways two of the
    * subexpressions could be the same. future/XXX
    */


   /* A union A => A */
   if (tcexpr_simple_same(te->bop.left, te->bop.right)) {
      return bop_destruct_keepleft(ctx, te);
   }

   if (te->bop.right->type == TCE_VALUE) {
      /* A union {} => A */
      if (pqlvalue_isset(te->bop.right->value) &&
	  pqlvalue_set_getnum(te->bop.right->value) == 0) {
	 return bop_destruct_keepleft(ctx, te);
      }
   }

   if (te->bop.left->type == TCE_VALUE) {
      /* {} union B => B */
      if (pqlvalue_isset(te->bop.left->value) &&
	  pqlvalue_set_getnum(te->bop.left->value) == 0) {
	 return bop_destruct_keepright(ctx, te);
      }
   }

   if (te->bop.left->type == TCE_VALUE &&
       te->bop.right->type == TCE_VALUE &&
       pqlvalue_isset(te->bop.left->value) &&
       pqlvalue_isset(te->bop.right->value) &&
       pqlvalue_set_getnum(te->bop.left->value) == 1 &&
       pqlvalue_set_getnum(te->bop.right->value) == 1) {
      /* catch the (set(1) union set(2)) case */
      const struct pqlvalue *val;

      // XXX this is wrong if the two sets contain the same element */

      val = pqlvalue_set_get(te->bop.right->value, 0);
      pqlvalue_set_add(te->bop.left->value, pqlvalue_clone(ctx->pql, val)); 
      return bop_destruct_keepleft(ctx, te);
   }


   return te;
}

static struct tcexpr *intersect_baseopt(struct baseopt *ctx,struct tcexpr *te){
   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_INTERSECT);

   (void)ctx;

   /*
    * A intersect (B intersect C) <=> (A intersect B) intersect C
    * (same as unions; see above)
    */

   /*
    * A intersect B => (uniq A) intersect (uniq B)
    * A intersect B => (uniq A) intersectall (uniq B)
    *
    * This has to wait until we figure out how to represent
    * sortedness. XXX
    */

   /*
    * (A union B) intersect (A union C) => A union (B intersect C)
    *
    * this is a pain to test, since there are four ways two of the
    * subexpressions could be the same. future/XXX
    */


   /* A intersect A => A */
   if (tcexpr_simple_same(te->bop.left, te->bop.right)) {
      return bop_destruct_keepleft(ctx, te);
   }

   if (te->bop.right->type == TCE_VALUE) {
      /* A intersect {} => {} */
      if (pqlvalue_isset(te->bop.right->value) &&
	  pqlvalue_set_getnum(te->bop.right->value) == 0) {
	 return bop_destruct_keepright(ctx, te);
      }
   }

   if (te->bop.left->type == TCE_VALUE) {
      /* {} intersect B => {} */
      if (pqlvalue_isset(te->bop.left->value) &&
	  pqlvalue_set_getnum(te->bop.left->value) == 0) {
	 return bop_destruct_keepleft(ctx, te);
      }
   }

   return te;
}

static struct tcexpr *except_baseopt(struct baseopt *ctx,struct tcexpr *te){
   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_EXCEPT);

   /*
    * A except B => (uniq A) except (uniq B)
    * A except B => (uniq A) exceptall (uniq B)
    *
    * This has to wait until we figure out how to represent
    * sortedness. XXX
    */

   /*
    * (A union B) except (C except A) => A union (B except C)
    * this is a pain to test, as above. XXX
    */


   /* A except A => {} */
   if (tcexpr_simple_same(te->bop.left, te->bop.right)) {
      return comb_toconstant(ctx, te, pqlvalue_emptyset(ctx->pql));
   }

   if (te->bop.right->type == TCE_VALUE) {
      /* A except {} => A */
      if (pqlvalue_isset(te->bop.right->value) &&
	  pqlvalue_set_getnum(te->bop.right->value) == 0) {
	 return bop_destruct_keepleft(ctx, te);
      }
   }

   if (te->bop.left->type == TCE_VALUE) {
      /* {} except B => {} */
      if (pqlvalue_isset(te->bop.left->value) &&
	  pqlvalue_set_getnum(te->bop.left->value) == 0) {
	 return bop_destruct_keepleft(ctx, te);
      }
   }

   return te;
}

static struct tcexpr *unionall_baseopt(struct baseopt *ctx, struct tcexpr *te){
   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_UNIONALL);

   (void)ctx;

   /*
    * A unionall (B unionall C) <=> (A unionall B) unionall C
    *
    * Even this requires data about estimated sizes (it is or can be
    * marginally more efficient to start with the largest set and add
    * the others to it) so it won't be done here.
    */

   if (te->bop.right->type == TCE_VALUE) {
      /* A unionall {} => A */
      if (pqlvalue_isset(te->bop.right->value) &&
	  pqlvalue_set_getnum(te->bop.right->value) == 0) {
	 return bop_destruct_keepleft(ctx, te);
      }
   }

   if (te->bop.left->type == TCE_VALUE) {
      /* {} unionall B => B */
      if (pqlvalue_isset(te->bop.left->value) &&
	  pqlvalue_set_getnum(te->bop.left->value) == 0) {
	 return bop_destruct_keepright(ctx, te);
      }
   }

   if (te->bop.left->type == TCE_VALUE &&
       te->bop.right->type == TCE_VALUE &&
       pqlvalue_isset(te->bop.left->value) &&
       pqlvalue_isset(te->bop.right->value) &&
       pqlvalue_set_getnum(te->bop.left->value) == 1 &&
       pqlvalue_set_getnum(te->bop.right->value) == 1) {
      /* catch the (set(1) unionall set(2)) case */
      const struct pqlvalue *val;

      val = pqlvalue_set_get(te->bop.right->value, 0);
      pqlvalue_set_add(te->bop.left->value, pqlvalue_clone(ctx->pql, val)); 
      return bop_destruct_keepleft(ctx, te);
   }

   return te;
}

static struct tcexpr *intersectall_baseopt(struct baseopt *ctx,
					   struct tcexpr *te) {
   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_INTERSECTALL);

   (void)ctx;

   /*
    * A intersectall (B intersectall C) <=> (A intersectall B) intersectall C
    * (same as unions; see above)
    */

   /* A intersectall A => A */
   if (tcexpr_simple_same(te->bop.left, te->bop.right)) {
      return bop_destruct_keepleft(ctx, te);
   }

   if (te->bop.right->type == TCE_VALUE) {
      /* A intersectall {} => {} */
      if (pqlvalue_isset(te->bop.right->value) &&
	  pqlvalue_set_getnum(te->bop.right->value) == 0) {
	 return bop_destruct_keepright(ctx, te);
      }
   }

   if (te->bop.left->type == TCE_VALUE) {
      /* {} intersectall B => {} */
      if (pqlvalue_isset(te->bop.left->value) &&
	  pqlvalue_set_getnum(te->bop.left->value) == 0) {
	 return bop_destruct_keepleft(ctx, te);
      }
   }

   return te;
}

static struct tcexpr *exceptall_baseopt(struct baseopt *ctx,struct tcexpr *te){
   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_EXCEPTALL);

   /* A exceptall A => {} */
   if (tcexpr_simple_same(te->bop.left, te->bop.right)) {
      return comb_toconstant(ctx, te, pqlvalue_emptyset(ctx->pql));
   }

   if (te->bop.right->type == TCE_VALUE) {
      /* A exceptall {} => A */
      if (pqlvalue_isset(te->bop.right->value) &&
	  pqlvalue_set_getnum(te->bop.right->value) == 0) {
	 return bop_destruct_keepleft(ctx, te);
      }
   }

   if (te->bop.left->type == TCE_VALUE) {
      /* {} exceptall B => {} */
      if (pqlvalue_isset(te->bop.left->value) &&
	  pqlvalue_set_getnum(te->bop.left->value) == 0) {
	 return bop_destruct_keepleft(ctx, te);
      }
   }

   return te;
}

static struct tcexpr *subset_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_SUBSETOF);

   /* set(X) in set(Y)  =>  X = Y */
   if (te->bop.left->type == TCE_UOP && te->bop.left->uop.op == F_MAKESET &&
       te->bop.right->type == TCE_UOP && te->bop.right->uop.op == F_MAKESET) {
      te->bop.left = tcexpr_destruct(ctx, te->bop.left);
      te->bop.right = tcexpr_destruct(ctx, te->bop.right);
      te->bop.op = F_EQ;
      return any_baseopt(ctx, te);
   }

   return te;
}

static struct tcexpr *element_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_ELEMENTOF);

   /* X in set(Y)  =>  X = Y */
   if (te->bop.right->type == TCE_UOP && te->bop.right->uop.op == F_MAKESET) {
      te->bop.right = tcexpr_destruct(ctx, te->bop.right);
      te->bop.op = F_EQ;
      return any_baseopt(ctx, te);
   }

   /*
    * X in (A union B) => let T = X: T in A or T in B
    * X in (A intersect B) => let T = X: T in A and T in B
    * X in (A except B) => let T = X: T in A and not T in B
    * X in (A ++ B) => let T = X: T in A or T in B
    */
   if (te->bop.right->type == TCE_BOP && 
       (te->bop.right->bop.op == F_UNION ||
	te->bop.right->bop.op == F_INTERSECT ||
	te->bop.right->bop.op == F_EXCEPT ||
	te->bop.right->bop.op == F_CONCAT)) {
      enum functions newop;
      bool addnot;

      switch (te->bop.right->bop.op) {
       case F_UNION:     newop = F_OR;  addnot = false; break;
       case F_INTERSECT: newop = F_AND; addnot = false; break;
       case F_EXCEPT:    newop = F_AND; addnot = true;  break;
       case F_CONCAT:    newop = F_OR;  addnot = false; break;
       default: PQLASSERT(0); break;
      }
      te = comb_element_setop(ctx, te, newop, addnot);
      PQLASSERT(te->type == TCE_LET);
      PQLASSERT(te->let.body->type == TCE_BOP);
      if (addnot) {
	 PQLASSERT(te->let.body->bop.right->type == TCE_UOP);
	 te->let.body->bop.right->uop.sub =
	    any_baseopt(ctx, te->let.body->bop.right->uop.sub);
      }
      te->let.body->bop.right = any_baseopt(ctx, te->let.body->bop.right);
      te->let.body->bop.left = any_baseopt(ctx, te->let.body->bop.left);
      te->let.body = any_baseopt(ctx, te->let.body);
      return any_baseopt(ctx, te);
   }

   /* X in (A where \K -> F(K))  =>  let T = X: T in A and F(T) */
   /* XXX notyet */

   /* X in A x B  =>  X1 in A and X2 in B, where X = X1 x X2 */
   /* XXX notyet */

   /*
    * X in (A order by ...) => X in A
    * X in (uniq A) => X in A
    */
   if (te->bop.right->type == TCE_ORDER ||
       te->bop.right->type == TCE_UNIQ) {
      te->bop.right = tcexpr_destruct(ctx, te->bop.right);
      /* do some more, by tail call */
      return element_baseopt(ctx, te);
   }

   /*
    * X in A |+| \K -> F(K) as C  =>  strip_C(X) in A and project_C(X) = F(X)
    */
   /* XXX notyet */

   /*
    * X in map A in B: C  =>  anytrue(map A in B: X = C)
    */
   if (te->bop.right->type == TCE_MAP) {
      te = move_bop_into_map_on_right(ctx, te, F_ANYTRUE,
				      datatype_bool(ctx->pql), F_EQ);
      PQLASSERT(te->type == TCE_UOP);
      PQLASSERT(te->uop.sub->type == TCE_MAP);
      te->uop.sub->map.result = any_baseopt(ctx, te->uop.sub->map.result);
      te->uop.sub = any_baseopt(ctx, te->uop.sub);
      return any_baseopt(ctx, te);
   }

   /*
    * X in let A = B: C  =>  let A = B: X in C
    */
   if (te->bop.right->type == TCE_LET) {
      te = move_bop_into_let_on_right(ctx, te);
      PQLASSERT(te->type == TCE_LET);
      te->let.body = any_baseopt(ctx, te->let.body);
      return any_baseopt(ctx, te);
   }

   /*
    * X in {} -> false
    */
   if (te->bop.right->type == TCE_VALUE &&
       ((pqlvalue_isset(te->bop.right->value) &&
	 pqlvalue_set_getnum(te->bop.right->value) == 0) ||
	(pqlvalue_issequence(te->bop.right->value) &&
	 pqlvalue_sequence_getnum(te->bop.right->value) == 0))) {
      return comb_toconstant(ctx, te, pqlvalue_bool(ctx->pql, false));
   }

   return te;
}

static struct tcexpr *nonempty_baseopt(struct baseopt *ctx, struct tcexpr *te){
   PQLASSERT(te->type == TCE_UOP);
   PQLASSERT(te->uop.op == F_NONEMPTY);

   /*
    * nonempty(foo X) => X
    */
   if (te->uop.sub->type == TCE_PROJECTTUPLE ||
       te->uop.sub->type == TCE_PROJECTSCALAR ||
       te->uop.sub->type == TCE_RENAME ||
       te->uop.sub->type == TCE_ORDER ||
       te->uop.sub->type == TCE_UNIQ ||
       te->uop.sub->type == TCE_NEST ||
       te->uop.sub->type == TCE_UNNEST ||
       te->uop.sub->type == TCE_DISTINGUISH ||
       te->uop.sub->type == TCE_ADJOIN ||
       te->uop.sub->type == TCE_MAP) {
      te->uop.sub = tcexpr_destruct(ctx, te->uop.sub);
      return nonempty_baseopt(ctx, te);
   }

   /*
    * nonempty(A x B) => nonempty(A) and nonempty(B)
    */
   if (te->uop.sub->type == TCE_JOIN &&
       te->uop.sub->join.predvar == NULL &&
       te->uop.sub->join.predexpr == NULL) {
      te = comb_distribute_uop_over_join(ctx, te, F_AND);
      PQLASSERT(te->type == TCE_BOP);
      te->bop.left = any_baseopt(ctx, te->bop.left);
      te->bop.right = any_baseopt(ctx, te->bop.right);
      return any_baseopt(ctx, te);
   }

   if (te->uop.sub->type == TCE_BOP &&
       (te->bop.right->bop.op == F_UNION ||
	te->bop.right->bop.op == F_INTERSECT ||
	te->bop.right->bop.op == F_CONCAT)) {
      enum functions newop;

      switch (te->uop.sub->bop.op) {
       case F_UNION:     newop = F_AND; break;
       case F_INTERSECT: newop = F_OR;  break;
       case F_CONCAT:    newop = F_AND; break;
       default: PQLASSERT(0); break;
      }
      te = comb_distribute_uop_over_bop(ctx, te, newop);
      PQLASSERT(te->type == TCE_BOP);
      te->bop.right = any_baseopt(ctx, te->bop.right);
      te->bop.left = any_baseopt(ctx, te->bop.left);
      return any_baseopt(ctx, te);
   }

   if (te->uop.sub->type == TCE_UOP && te->uop.sub->uop.op == F_MAKESET) {
      return comb_toconstant(ctx, te, pqlvalue_bool(ctx->pql, true));
   }

   if (te->uop.sub->type == TCE_LET) {
      te = move_uop_into_let(ctx, te);
      PQLASSERT(te->type == TCE_LET);
      te->let.body = any_baseopt(ctx, te->let.body);
      return any_baseopt(ctx, te);
   }

   if (te->uop.sub->type == TCE_VALUE && pqlvalue_isset(te->uop.sub->value)) {
      struct pqlvalue *val;

      val = pqlvalue_bool(ctx->pql,
			  pqlvalue_set_getnum(te->uop.sub->value) > 0);
      return comb_toconstant(ctx, te, val);
   }

   return te;
}

static struct tcexpr *makeset_baseopt(struct baseopt *ctx,
				      struct tcexpr *te) {
   PQLASSERT(te->type == TCE_UOP);
   PQLASSERT(te->uop.op == F_MAKESET);

   if (te->uop.sub->type == TCE_VALUE) {
      struct pqlvalue *set;

      set = pqlvalue_emptyset(ctx->pql);
      pqlvalue_set_add(set, te->uop.sub->value);
      te->uop.sub->value = NULL;
      return comb_toconstant(ctx, te, set);
   }

   return te;
}

static struct tcexpr *getelement_baseopt(struct baseopt *ctx,
					 struct tcexpr *te) {
   PQLASSERT(te->type == TCE_UOP);
   PQLASSERT(te->uop.op == F_GETELEMENT);

   if (te->uop.sub->type == TCE_UOP && te->uop.sub->uop.op == F_MAKESET) {
      te->uop.sub = tcexpr_destruct(ctx, te->uop.sub);
      return tcexpr_destruct(ctx, te);
   }

   /* constant folding */
   if (te->uop.sub->type == TCE_VALUE) {
      /*
       * Do this only for sets of one; for larger sets we need an
       * error value (not nil) and we don't have that yet. (XXX)
       */
      if (pqlvalue_isset(te->uop.sub->value) &&
	  pqlvalue_set_getnum(te->uop.sub->value) == 1) {
	 const struct pqlvalue *val;

	 val = pqlvalue_set_get(te->uop.sub->value, 0);
	 return comb_toconstant(ctx, te, pqlvalue_clone(ctx->pql, val));
      }
      if (pqlvalue_issequence(te->uop.sub->value) &&
	  pqlvalue_sequence_getnum(te->uop.sub->value) == 1) {
	 const struct pqlvalue *val;

	 val = pqlvalue_sequence_get(te->uop.sub->value, 0);
	 return comb_toconstant(ctx, te, pqlvalue_clone(ctx->pql, val));
      }
   }

   return te;
}

//////////////////////////////
// aggregator

static struct tcexpr *count_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *sub;

   while (1) {
      PQLASSERT(te->type == TCE_UOP);
      PQLASSERT(te->uop.op == F_COUNT);
      sub = te->uop.sub;

      /*
       * count(projectscalar_* X) => count(X)
       *
       * ...provided that the projection is operating on a set. If it
       * isn't, instead presumably it's projecting a set column out of
       * a single tuple and we should leave it alone.
       */
      if (sub->type == TCE_PROJECTSCALAR) {
	 if (datatype_isset(sub->projectscalar.sub->datatype)) {
	    comb_count_destroysub(ctx, te);
	    continue;
	 }
	 else {
	    /* nothing */
	 }
      }

      /*
       * count(projecttuple_* X) => count(X)
       * count(rename_* X) => count(X)
       * count(A order-by ...) => count(A)
       * count(distinguish(X)) => count(X)
       * count(adjoin X ...) => count(X)
       * count(map X0 in X: f(X)) => count(X)
       *
       * These are always true.
       */
      if (sub->type == TCE_PROJECTTUPLE ||
	  sub->type == TCE_RENAME ||
	  sub->type == TCE_ORDER ||
	  sub->type == TCE_DISTINGUISH ||
	  sub->type == TCE_ADJOIN ||
	  sub->type == TCE_MAP) {
	 comb_count_destroysub(ctx, te);
	 continue;
      }

      /*
       * count(X x Y) => count(X) + count(Y)
       */
      if (sub->type == TCE_JOIN &&
	  sub->join.predvar == NULL &&
	  sub->join.predexpr == NULL) {
	 te = comb_distribute_uop_over_join(ctx, te, F_ADD);
	 PQLASSERT(te->type == TCE_BOP);
	 te->bop.left = any_baseopt(ctx, te->bop.left);
	 te->bop.right = any_baseopt(ctx, te->bop.right);
	 return any_baseopt(ctx, te);
      }

      /*
       * count(unnest_a X) => sum(map X0 in (project_a X): count(X0))
       */
      if (sub->type == TCE_UNNEST) {
	 te = comb_agg_unnest(ctx, te);
	 PQLASSERT(te->type == TCE_UOP);
	 PQLASSERT(te->uop.sub->type == TCE_MAP);
	 te->uop.sub->map.set = any_baseopt(ctx, te->uop.sub->map.set);
	 te->uop.sub->map.result = any_baseopt(ctx, te->uop.sub->map.result);
	 te->uop.sub = any_baseopt(ctx, te->uop.sub);
	 return any_baseopt(ctx, te);
      }

      /* count(A union-all B) => count(A) + count(B) */
      if (sub->type == TCE_BOP) {
	 /* notyet */
      }

      /* count(A ++ B) => count(A) + count(B), if A/B are sequences */
      if (sub->type == TCE_BOP && sub->bop.op == F_CONCAT &&
	  datatype_issequence(sub->datatype)) {
	 te = comb_count_concat(ctx, te);
	 //te = comb_distribute_uop_over_bop(ctx, te, F_ADD); // XXX notyet
	 PQLASSERT(te->type == TCE_BOP);
	 te->bop.left = any_baseopt(ctx, te->bop.left);
	 te->bop.right = any_baseopt(ctx, te->bop.right);
	 return any_baseopt(ctx, te);
      }

      /* count(set(X)) => 1 */
      if (sub->type == TCE_UOP && sub->uop.op == F_MAKESET) {
	 return comb_toconstant(ctx, te, pqlvalue_int(ctx->pql, 1));
      }

      /* constant folding */
      if (sub->type == TCE_VALUE) {
	 unsigned val;

	 if (pqlvalue_isset(sub->value)) {
	    val = pqlvalue_set_getnum(sub->value);
	 }
	 else {
	    PQLASSERT(pqlvalue_issequence(sub->value));
	    val = pqlvalue_sequence_getnum(sub->value);
	 }
	 return comb_toconstant(ctx, te, pqlvalue_int(ctx->pql, val));
      }

      /* Nothing matched; give up. */
      break;
   }

   return te;
}

static struct tcexpr *sum_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *sub;

   while (1) {
      PQLASSERT(te->type == TCE_UOP);
      PQLASSERT(te->uop.op == F_SUM);
      sub = te->uop.sub;

      /*
       * sum(rename_* X) => count(X)
       * sum(A order-by ...) => count(A)
       */
      if (sub->type == TCE_RENAME ||
	  sub->type == TCE_ORDER) {
	 te->uop.sub = tcexpr_destruct(ctx, te->uop.sub);
	 continue;
      }

      /*
       * sum(A union B) => sum(A) + sum(B) - sum(A intersect B)
       *  ... not worthwhile
       */

      /*
       * sum(map X0 in X: X0 * K) => K * sum(X)
       */
      /* XXX notyet */

      /*
       * sum(unnest_a X) => sum(map X0 in (project_a X): sum(X0))
       */
      if (sub->type == TCE_UNNEST) {
	 te = comb_agg_unnest(ctx, te);
	 PQLASSERT(te->type == TCE_UOP);
	 PQLASSERT(te->uop.sub->type == TCE_MAP);
	 te->uop.sub->map.set = any_baseopt(ctx, te->uop.sub->map.set);
	 te->uop.sub->map.result = any_baseopt(ctx, te->uop.sub->map.result);
	 te->uop.sub = any_baseopt(ctx, te->uop.sub);
	 return any_baseopt(ctx, te);
      }

      /* sum(A unionall B) => sum(A) + sum(B) */
      if (sub->type == TCE_BOP) {
	 /* notyet XXX */
      }

      /* sum(A ++ B) => sum(A) + sum(B), if A/B are sequences */
      if (sub->type == TCE_BOP && sub->bop.op == F_CONCAT &&
	  datatype_issequence(sub->datatype)) {
	 te = comb_distribute_uop_over_bop(ctx, te, F_ADD);
	 PQLASSERT(te->type == TCE_BOP);
	 te->bop.left = any_baseopt(ctx, te->bop.left);
	 te->bop.right = any_baseopt(ctx, te->bop.right);
	 return any_baseopt(ctx, te);
      }

      /* sum(set(X)) => X */
      if (sub->type == TCE_UOP && sub->uop.op == F_MAKESET) {
	 te->uop.sub = tcexpr_destruct(ctx, te->uop.sub);
	 return tcexpr_destruct(ctx, te);
      }

      /* constant folding */
      if (sub->type == TCE_VALUE) {
	 unsigned num, i, val;
	 const struct pqlvalue *subval;

	 val = 0;
	 if (pqlvalue_isset(sub->value)) {
	    num = pqlvalue_set_getnum(sub->value);
	    for (i=0; i<num; i++) {
	       subval = pqlvalue_set_get(sub->value, i);
	       PQLASSERT(pqlvalue_isint(subval));
	       val += pqlvalue_int_get(subval);
	    }
	 }
	 else {
	    PQLASSERT(pqlvalue_issequence(sub->value));
	    num = pqlvalue_sequence_getnum(sub->value);
	    for (i=0; i<num; i++) {
	       subval = pqlvalue_sequence_get(sub->value, i);
	       PQLASSERT(pqlvalue_isint(subval));
	       val += pqlvalue_int_get(subval);
	    }
	 }
	 return comb_toconstant(ctx, te, pqlvalue_int(ctx->pql, val));
      }

      /* Nothing matched; give up. */
      break;
   }

   return te;
}

static struct tcexpr *minmax_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *sub;
   bool ismin;

   while (1) {
      PQLASSERT(te->type == TCE_UOP);
      PQLASSERT(te->uop.op == F_MIN || te->uop.op == F_MAX);
      ismin = te->uop.op == F_MIN;
      sub = te->uop.sub;

      /*
       * min(rename_* X) => min(X)
       * min(A order-by ...) => min(A)
       */
      if (sub->type == TCE_RENAME ||
	  sub->type == TCE_ORDER) {
	 te->uop.sub = tcexpr_destruct(ctx, te->uop.sub);
	 continue;
      }

      /*
       * min(A union B) => min(min(A), min(B))
       * min(A unionall B) => min(min(A), min(B))
       * min(A ++ B) => min(min(A), min(B))
       */
      /* XXX notyet */

      /*
       * min(unnest_a X) => min(map X0 in (project_a X): min(X0))
       */
      if (sub->type == TCE_UNNEST) {
	 te = comb_agg_unnest(ctx, te);
	 PQLASSERT(te->type == TCE_UOP);
	 PQLASSERT(te->uop.sub->type == TCE_MAP);
	 te->uop.sub->map.set = any_baseopt(ctx, te->uop.sub->map.set);
	 te->uop.sub->map.result = any_baseopt(ctx, te->uop.sub->map.result);
	 te->uop.sub = any_baseopt(ctx, te->uop.sub);
	 return any_baseopt(ctx, te);
      }

      /* min(set(X)) => X */
      if (sub->type == TCE_UOP && sub->uop.op == F_MAKESET) {
	 te->uop.sub = tcexpr_destruct(ctx, te->uop.sub);
	 return tcexpr_destruct(ctx, te);
      }

      /* constant folding */
      if (sub->type == TCE_VALUE) {
	 unsigned num, i, val, x;
	 bool hasval;
	 const struct pqlvalue *subval;
	 struct pqlvalue *newval;

	 val = 0; /* gcc 4.1.3 falsely warns */

	 hasval = false;
	 if (pqlvalue_isset(sub->value)) {
	    num = pqlvalue_set_getnum(sub->value);
	    for (i=0; i<num; i++) {
	       subval = pqlvalue_set_get(sub->value, i);
	       PQLASSERT(pqlvalue_isint(subval));
	       x = pqlvalue_int_get(subval);
	       if (!hasval || (ismin ? x < val : x > val)) {
		  val = x;
		  hasval = true;
	       }
	    }
	 }
	 else {
	    PQLASSERT(pqlvalue_issequence(sub->value));
	    num = pqlvalue_sequence_getnum(sub->value);
	    for (i=0; i<num; i++) {
	       subval = pqlvalue_sequence_get(sub->value, i);
	       PQLASSERT(pqlvalue_isint(subval));
	       x = pqlvalue_int_get(subval);
	       if (!hasval || (ismin ? x < val : x > val)) {
		  val = x;
		  hasval = true;
	       }
	    }
	 }
	 if (hasval) {
	    newval = pqlvalue_int(ctx->pql, val);
	 }
	 else {
	    newval = pqlvalue_nil(ctx->pql);
	 }
	 return comb_toconstant(ctx, te, newval);
      }

      /* Nothing matched; give up. */
      break;
   }

   return te;
}

static struct tcexpr *anytrue_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *sub;

   while (1) {
      PQLASSERT(te->type == TCE_UOP);
      PQLASSERT(te->uop.op == F_ANYTRUE);
      sub = te->uop.sub;

      /*
       * anytrue(rename_* X) => anytrue(X)
       * anytrue(A order-by ...) => anytrue(A)
       */
      if (sub->type == TCE_RENAME ||
	  sub->type == TCE_ORDER) {
	 te->uop.sub = tcexpr_destruct(ctx, te->uop.sub);
	 continue;
      }

      /*
       * anytrue(A union B) => anytrue(A) or anytrue(B)
       * anytrue(A unionall B) => anytrue(A) or anytrue(B)
       * anytrue(A intersect B) => anytrue(A) and anytrue(B)
       * anytrue(A intersectall B) => anytrue(A) and anytrue(B)
       * anytrue(A except B) => anytrue(A) and not anytrue(B)
       * anytrue(A exceptall B) => [does not work]
       * anytrue(A ++ B) => anytrue(A) or anytrue(B)
       */
      if (te->uop.sub->type == TCE_BOP &&
	  (te->uop.sub->bop.op == F_UNION ||
	   te->uop.sub->bop.op == F_INTERSECT ||
	   /*te->uop.sub->bop.op == F_EXCEPT ||  (XXX can't do addnot) */
	   te->uop.sub->bop.op == F_CONCAT)) {
	 enum functions newop;
	 bool addnot;

	 switch (te->bop.right->bop.op) {
	  case F_UNION:     newop = F_OR;  addnot = false; break;
	  case F_INTERSECT: newop = F_AND; addnot = false; break;
	  case F_EXCEPT:    newop = F_AND; addnot = true;  break;
	  case F_CONCAT:    newop = F_OR;  addnot = false; break;
	  default: PQLASSERT(0); break;
	 }
	 /* XXX distribute doesn't have an addnot feature */
	 te = comb_distribute_uop_over_bop(ctx, te, newop /*, addnot*/);
	 PQLASSERT(te->type == TCE_BOP);
	 if (addnot) {
	    PQLASSERT(te->bop.right->type == TCE_UOP);
	    te->bop.right->uop.sub = any_baseopt(ctx, te->bop.right->uop.sub);
	 }
	 te->bop.right = any_baseopt(ctx, te->bop.right);
	 te->bop.left = any_baseopt(ctx, te->bop.left);
	 return any_baseopt(ctx, te);
      }

      /*
       * anytrue(unnest_a X) => anytrue(map X0 in (project_a X): anytrue(X0))
       */
      if (sub->type == TCE_UNNEST) {
	 te = comb_agg_unnest(ctx, te);
	 PQLASSERT(te->type == TCE_UOP);
	 PQLASSERT(te->uop.sub->type == TCE_MAP);
	 te->uop.sub->map.set = any_baseopt(ctx, te->uop.sub->map.set);
	 te->uop.sub->map.result = any_baseopt(ctx, te->uop.sub->map.result);
	 te->uop.sub = any_baseopt(ctx, te->uop.sub);
	 return any_baseopt(ctx, te);
      }

      /* anytrue(set(X)) => X   XXX: what about nil? */
      if (sub->type == TCE_UOP && sub->uop.op == F_MAKESET) {
	 te->uop.sub = tcexpr_destruct(ctx, te->uop.sub);
	 return tcexpr_destruct(ctx, te);
      }

      /* constant folding */
      if (sub->type == TCE_VALUE) {
	 unsigned num, i;
	 const struct pqlvalue *subval;
	 bool val;

	 val = false;
	 if (pqlvalue_isset(sub->value)) {
	    num = pqlvalue_set_getnum(sub->value);
	    for (i=0; i<num; i++) {
	       subval = pqlvalue_set_get(sub->value, i);
	       PQLASSERT(pqlvalue_isbool(subval));
	       if (pqlvalue_bool_get(subval)) {
		  val = true;
	       }
	    }
	 }
	 else {
	    PQLASSERT(pqlvalue_issequence(sub->value));
	    num = pqlvalue_sequence_getnum(sub->value);
	    for (i=0; i<num; i++) {
	       subval = pqlvalue_sequence_get(sub->value, i);
	       PQLASSERT(pqlvalue_isbool(subval));
	       if (pqlvalue_bool_get(subval)) {
		  val = true;
	       }
	    }
	 }
	 return comb_toconstant(ctx, te, pqlvalue_bool(ctx->pql, val));
      }

      /* Nothing matched; give up. */
      break;
   }

   return te;
}

static struct tcexpr *alltrue_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *sub;

   while (1) {
      PQLASSERT(te->type == TCE_UOP);
      PQLASSERT(te->uop.op == F_ALLTRUE);
      sub = te->uop.sub;

      /*
       * alltrue(rename_* X) => alltrue(X)
       * alltrue(A order-by ...) => alltrue(A)
       */
      if (sub->type == TCE_RENAME ||
	  sub->type == TCE_ORDER) {
	 te->uop.sub = tcexpr_destruct(ctx, te->uop.sub);
	 continue;
      }

      /*
       * alltrue(A union B) => alltrue(A) and alltrue(B)
       * alltrue(A unionall B) => alltrue(A) and alltrue(B)
       * alltrue(A intersect B) => alltrue(A) or alltrue(B)
       * alltrue(A intersectall B) => alltrue(A) or alltrue(B)
       * alltrue(A ++ B) => alltrue(A) and alltrue(B)
       */
      if (te->uop.sub->type == TCE_BOP &&
	  (te->uop.sub->bop.op == F_UNION ||
	   te->uop.sub->bop.op == F_INTERSECT ||
	   te->uop.sub->bop.op == F_CONCAT)) {
	 enum functions newop;

	 switch (te->bop.right->bop.op) {
	  case F_UNION:     newop = F_OR;  break;
	  case F_INTERSECT: newop = F_AND; break;
	  case F_CONCAT:    newop = F_OR;  break;
	  default: PQLASSERT(0); break;
	 }
	 te = comb_distribute_uop_over_bop(ctx, te, newop);
	 PQLASSERT(te->type == TCE_BOP);
	 te->bop.right = any_baseopt(ctx, te->bop.right);
	 te->bop.left = any_baseopt(ctx, te->bop.left);
	 return any_baseopt(ctx, te);
      }

      /*
       * alltrue(A except B) => alltrue(A) and not anytrue(B)
       * alltrue(A exceptall B) => [does not work]
       */
      /* XXX available transform doesn't support this */


      /*
       * alltrue(unnest_a X) => alltrue(map X0 in (project_a X): alltrue(X0))
       */
      if (sub->type == TCE_UNNEST) {
	 te = comb_agg_unnest(ctx, te);
	 PQLASSERT(te->type == TCE_UOP);
	 PQLASSERT(te->uop.sub->type == TCE_MAP);
	 te->uop.sub->map.set = any_baseopt(ctx, te->uop.sub->map.set);
	 te->uop.sub->map.result = any_baseopt(ctx, te->uop.sub->map.result);
	 te->uop.sub = any_baseopt(ctx, te->uop.sub);
	 return any_baseopt(ctx, te);
      }

      /* alltrue(set(X)) => X   XXX: what about nil? */
      if (sub->type == TCE_UOP && sub->uop.op == F_MAKESET) {
	 te->uop.sub = tcexpr_destruct(ctx, te->uop.sub);
	 return tcexpr_destruct(ctx, te);
      }

      /* constant folding */
      if (sub->type == TCE_VALUE) {
	 unsigned num, i;
	 const struct pqlvalue *subval;
	 bool val;

	 val = true;
	 if (pqlvalue_isset(sub->value)) {
	    num = pqlvalue_set_getnum(sub->value);
	    for (i=0; i<num; i++) {
	       subval = pqlvalue_set_get(sub->value, i);
	       PQLASSERT(pqlvalue_isbool(subval));
	       if (!pqlvalue_bool_get(subval)) {
		  val = false;
	       }
	    }
	 }
	 else {
	    PQLASSERT(pqlvalue_issequence(sub->value));
	    num = pqlvalue_sequence_getnum(sub->value);
	    for (i=0; i<num; i++) {
	       subval = pqlvalue_sequence_get(sub->value, i);
	       PQLASSERT(pqlvalue_isbool(subval));
	       if (!pqlvalue_bool_get(subval)) {
		  val = false;
	       }
	    }
	 }
	 return comb_toconstant(ctx, te, pqlvalue_bool(ctx->pql, val));
      }

      /* Nothing matched; give up. */
      break;
   }

   return te;
}

//////////////////////////////
// boolean

static struct tcexpr *and_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *ret;

   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_AND);
   (void)ctx;

   /*
    * XXX. what about nil?
    */

   /* true and X => X */
   if (tcexpr_istrue(te->bop.left)) {
      ret = te->bop.right;
      te->bop.right = NULL;
      tcexpr_destroy(ctx->pql, te);
      return ret;
   }
   if (tcexpr_istrue(te->bop.right)) {
      ret = te->bop.left;
      te->bop.left = NULL;
      tcexpr_destroy(ctx->pql, te);
      return ret;
   }

   /* false and X => false */
   if (tcexpr_isfalse(te->bop.left)) {
      ret = te->bop.left;
      te->bop.left = NULL;
      tcexpr_destroy(ctx->pql, te);
      return ret;
   }
   if (tcexpr_isfalse(te->bop.right)) {
      ret = te->bop.right;
      te->bop.right = NULL;
      tcexpr_destroy(ctx->pql, te);
      return ret;
   }

   /* X and not X => false */
   if ((te->bop.right->type == TCE_UOP && te->bop.right->uop.op == F_NOT &&
       tcexpr_simple_same(te->bop.left, te->bop.right->uop.sub)) ||
       (te->bop.left->type == TCE_UOP && te->bop.left->uop.op == F_NOT &&
	tcexpr_simple_same(te->bop.right, te->bop.left->uop.sub))) {
      tcexpr_destroy(ctx->pql, te->bop.left);
      tcexpr_destroy(ctx->pql, te->bop.right);
      te->type = TCE_VALUE;
      te->value = pqlvalue_bool(ctx->pql, false);
      return te;
   }

   return te;
}

static struct tcexpr *or_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *ret;

   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_OR);
   (void)ctx;

   /*
    * XXX. what about nil?
    */

   /* false or X => X */
   if (tcexpr_isfalse(te->bop.left)) {
      ret = te->bop.right;
      te->bop.right = NULL;
      tcexpr_destroy(ctx->pql, te);
      return ret;
   }
   if (tcexpr_isfalse(te->bop.right)) {
      ret = te->bop.left;
      te->bop.left = NULL;
      tcexpr_destroy(ctx->pql, te);
      return ret;
   }

   /* true or X => true */
   if (tcexpr_istrue(te->bop.left)) {
      ret = te->bop.left;
      te->bop.left = NULL;
      tcexpr_destroy(ctx->pql, te);
      return ret;
   }
   if (tcexpr_istrue(te->bop.right)) {
      ret = te->bop.right;
      te->bop.right = NULL;
      tcexpr_destroy(ctx->pql, te);
      return ret;
   }

   /* X or not X => true */
   if ((te->bop.right->type == TCE_UOP && te->bop.right->uop.op == F_NOT &&
       tcexpr_simple_same(te->bop.left, te->bop.right->uop.sub)) ||
       (te->bop.left->type == TCE_UOP && te->bop.left->uop.op == F_NOT &&
	tcexpr_simple_same(te->bop.right, te->bop.left->uop.sub))) {
      tcexpr_destroy(ctx->pql, te->bop.left);
      tcexpr_destroy(ctx->pql, te->bop.right);
      te->type = TCE_VALUE;
      te->value = pqlvalue_bool(ctx->pql, true);
      return te;
   }

   return te;
}

static struct tcexpr *not_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *ret;

   PQLASSERT(te->type == TCE_UOP);
   PQLASSERT(te->uop.op == F_NOT);
   (void)ctx;

   /*
    * XXX. what about nil?
    */

   /* not not X => X */
   if (te->uop.sub->type == TCE_UOP && te->uop.sub->uop.op == F_NOT) {
      ret = te->uop.sub->uop.sub;
      te->uop.sub->uop.sub = NULL;
      tcexpr_destroy(ctx->pql, te);
      return ret;
   }

   /* not A == B => A != B */
   if (te->uop.sub->type == TCE_BOP && te->uop.sub->bop.op == F_EQ) {
      te->uop.sub->bop.op = F_NOTEQ;
      ret = te->uop.sub;
      te->uop.sub = NULL;
      tcexpr_destroy(ctx->pql, te);
      return ret;
   }

   /* not A != B => A == B */
   if (te->uop.sub->type == TCE_BOP && te->uop.sub->bop.op == F_NOTEQ) {
      te->uop.sub->bop.op = F_EQ;
      ret = te->uop.sub;
      te->uop.sub = NULL;
      tcexpr_destroy(ctx->pql, te);
      return ret;
   }

   return te;
}

//////////////////////////////
// comparison

static struct tcexpr *like_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_LIKE);

   /* foo like % => true */
   if (te->bop.right->type == TCE_VALUE &&
       pqlvalue_isstring(te->bop.right->value) &&
       !strcmp(pqlvalue_string_get(te->bop.right->value), "%")) {
      tcexpr_destroy(ctx->pql, te->bop.left);
      tcexpr_destroy(ctx->pql, te->bop.right);
      te->type = TCE_VALUE;
      te->value = pqlvalue_bool(ctx->pql, true);
      return te;
   }

   return te;
}

static struct tcexpr *glob_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_GLOB);

   /* foo glob * => true */
   if (te->bop.right->type == TCE_VALUE &&
       pqlvalue_isstring(te->bop.right->value) &&
       !strcmp(pqlvalue_string_get(te->bop.right->value), "*")) {
      tcexpr_destroy(ctx->pql, te->bop.left);
      tcexpr_destroy(ctx->pql, te->bop.right);
      te->type = TCE_VALUE;
      te->value = pqlvalue_bool(ctx->pql, true);
      return te;
   }

   return te;
}

static struct tcexpr *grep_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   PQLASSERT(te->type == TCE_BOP);
   PQLASSERT(te->bop.op == F_GREP);

   /* foo grep .* => true */
   if (te->bop.right->type == TCE_VALUE &&
       pqlvalue_isstring(te->bop.right->value) &&
       !strcmp(pqlvalue_string_get(te->bop.right->value), ".*")) {
      tcexpr_destroy(ctx->pql, te->bop.left);
      tcexpr_destroy(ctx->pql, te->bop.right);
      te->type = TCE_VALUE;
      te->value = pqlvalue_bool(ctx->pql, true);
      return te;
   }

   return te;
}

//////////////////////////////
// string/sequence

//////////////////////////////
// pathelement

static struct tcexpr *left_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   // XXX
   (void)ctx;
   return te;
}

static struct tcexpr *edge_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   // XXX
   (void)ctx;
   return te;
}

static struct tcexpr *right_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   // XXX
   (void)ctx;
   return te;
}

//////////////////////////////
// nil

//////////////////////////////
// numeric

////////////////////////////////////////////////////////////
// optimization matching for expressions

//////////////////////////////
// filter

static struct tcexpr *filter_baseopt(struct baseopt *ctx, struct tcexpr *te);

/*
 * Move a filter expression under its subexpression. TE is the filter,
 * and SUBSUBP should point to the subexpression slot of te->filter.sub.
 *
 * Thus e.g. filter_move(ctx, te, &te->filter.sub->projecttuple.sub)
 * moves TE under its subexpression if that subexpression is a
 * projecttuple.
 *
 * If VARWRAPPER is not null wrapvar() will be called using it as a
 * prototype to update uses of the lambda variable of the predicate.
 */
static struct tcexpr *filter_move(struct baseopt *ctx, struct tcexpr *te,
				  struct tcexpr **subsubp,
				  struct tcexpr *varwrapper) {
   struct tcexpr *sub;
   struct tcexpr *subsub;
   struct datatype *argtype, *restype;
   
   sub = te->filter.sub;
   subsub = *subsubp;
   PQLASSERT(datatype_isset(sub->datatype));
   PQLASSERT(datatype_isset(subsub->datatype));

   /*
    * In case this is the sub-slot of VARWRAPPER (which it currently
    * is in all cases where VARWRAPPER is non-null) clear it so
    * VARWRAPPER becomes a suitable template for wrapvar(). If
    * VARWRAPPER is an independently generated expression template, or
    * it's null, this is harmless.
    */
   *subsubp = NULL;

   te->filter.sub = subsub;
   te->datatype = subsub->datatype;

   /* update the predicate types */
   if (te->filter.predvar != NULL) {
      argtype = datatype_set_member(subsub->datatype);
      te->filter.predvar->datatype = argtype;
   }
   restype = te->filter.predexpr->datatype;
   PQLASSERT(datatype_isbool(restype));

   if (varwrapper != NULL) {
      PQLASSERT(te->filter.predvar != NULL);
      te->filter.predexpr = wrapvar(ctx->pql, te->filter.predexpr,
				    te->filter.predvar, varwrapper);
      te->filter.predexpr = tcexpr_baseopt(ctx, te->filter.predexpr);
   }

   te = filter_baseopt(ctx, te);
   *subsubp = te;
   return sub;
}

/*
 * Like filter_move, but replicates the filter onto two subexpressions,
 * both of which have the same type.
 */
static struct tcexpr *filter_movetwo(struct baseopt *ctx, struct tcexpr *te,
				     struct tcexpr **subsubp1,
				     struct tcexpr **subsubp2) {
   struct tcexpr *sub, *subsub1, *subsub2, *tecopy;
   struct datatype *argtype1, *restype1;
   struct datatype *argtype2, *restype2;

   sub = te->filter.sub;
   subsub1 = *subsubp1;
   subsub2 = *subsubp2;
   PQLASSERT(datatype_isset(sub->datatype));
   PQLASSERT(datatype_isset(subsub1->datatype));
   PQLASSERT(datatype_isset(subsub2->datatype));

   /* clone the filter without the subexpression */
   te->filter.sub = NULL;
   tecopy = tcexpr_clone(ctx->pql, te);

   te->filter.sub = subsub1;
   te->datatype = subsub1->datatype;
   tecopy->filter.sub = subsub2;
   tecopy->datatype = subsub2->datatype;

   /* update the predicate types */
   if (te->filter.predvar != NULL) {
      argtype1 = datatype_set_member(subsub1->datatype);
      argtype2 = datatype_set_member(subsub2->datatype);
      te->filter.predvar->datatype = argtype1;
      tecopy->filter.predvar->datatype = argtype2;
   }
   restype1 = te->filter.predexpr->datatype;
   restype2 = tecopy->filter.predexpr->datatype;
   PQLASSERT(datatype_isbool(restype1));
   PQLASSERT(datatype_isbool(restype2));

   te = filter_baseopt(ctx, te);
   tecopy = filter_baseopt(ctx, tecopy);
   *subsubp1 = te;
   *subsubp2 = tecopy;
   return sub;
}

static struct tcexpr *filter_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *sub;

   PQLASSERT(te->type == TCE_FILTER);

   sub = te->filter.sub;

   /*
    * Combine filter conditions.
    */

   if (sub->type == TCE_FILTER) {
      /* filter twice -> combine them */
      combine_predicates(ctx, 
			 sub->filter.predvar, sub->filter.predexpr,
			 te->filter.predvar, te->filter.predexpr,
			 &sub->filter.predvar, &sub->filter.predexpr);
      sub->filter.predexpr = any_baseopt(ctx, sub->filter.predexpr);
      te->filter.predvar = NULL;
      te->filter.predexpr = NULL;
      goto destroy;
   }

   if (sub->type == TCE_PROJECTTUPLE) {
      /* filter(project_c X) -> project_c(filter X) */
      return filter_move(ctx, te, &sub->projecttuple.sub, sub);
   }

   if (sub->type == TCE_PROJECTSCALAR) {
      /* filter(project_c X) -> project_c(filter X) */
      return filter_move(ctx, te, &sub->projectscalar.sub, sub);
   }

   if (sub->type == TCE_RENAME) {
      /* norenames should prevent this from occurring */
      PQLASSERT(0);
   }

   if (sub->type == TCE_JOIN) {
      /* combine the filter with the join's internal filter */
      combine_predicates(ctx,
			 sub->join.predvar, sub->join.predexpr,
			 te->filter.predvar, te->filter.predexpr,
			 &sub->join.predvar, &sub->join.predexpr);
      sub->join.predexpr = any_baseopt(ctx, sub->join.predexpr);
      te->filter.predvar = NULL;
      te->filter.predexpr = NULL;
      goto destroy;
   }

   if (sub->type == TCE_ORDER) {
      /* filter(order X) -> order(filter X) */
      return filter_move(ctx, te, &sub->order.sub, NULL);
   }

   if (sub->type == TCE_UNIQ) {
      /* filter(uniq X) -> uniq(filter X) */
      return filter_move(ctx, te, &sub->uniq.sub, NULL);
   }

   if (sub->type == TCE_NEST || sub->type == TCE_UNNEST) {
      /* XXX can we do anything here? we should be able to */
   }

   if (sub->type == TCE_DISTINGUISH) {
      /*
       * filter(distinguish X, lambda Y: PRED(Y))
       *    -> distinguish(filter(X, lambda Y: PRED1(Y)))
       *
       * where PRED1 wraps uses of Y in distinguish.
       */

      return filter_move(ctx, te, &sub->uniq.sub, sub);
   }

   if (sub->type == TCE_ADJOIN) {
      /*
       * filter (adjoin SUBSUB foo) PRED -> adjoin (filter SUBSUB PRED) foo
       */
      
      struct tcexpr *subsub, *predexpr, *wraptemplate;
      struct tcvar *predvar;
      struct datatype *varnewtype, *coltype;

      PQLASSERT(datatype_isset(sub->datatype));

      predvar = te->filter.predvar;
      predexpr = te->filter.predexpr;
      subsub = sub->adjoin.left;

      coltype = sub->adjoin.expr->datatype;

      if (predvar != NULL) {
	 varnewtype = subsub->datatype;
	 PQLASSERT(datatype_isset(varnewtype));
	 varnewtype = datatype_set_member(varnewtype);

	 predvar->datatype = varnewtype;

	 if (tcexpr_uses_column(ctx, predexpr, sub->adjoin.newcol)) {
	    /* copy the adjoin as a template (without its left subexpr) */
	    sub->adjoin.left = NULL;
	    wraptemplate = tcexpr_clone(ctx->pql, sub);
	    sub->adjoin.left = subsub;
	 }
	 else {
	    /* adjoin nil instead */
	    wraptemplate = psplit_adjoinonewrapper(ctx->pql, NULL, varnewtype,
						  coltype, sub->adjoin.newcol);
	 }
	 predexpr = wrapvar(ctx->pql, predexpr, predvar, wraptemplate);
	 tcexpr_destroy(ctx->pql, wraptemplate);
      }

      /* recurse on the new filter expression and put it into place */
      te->filter.predexpr = tcexpr_baseopt(ctx, predexpr);

      /*
       * Move the original adjoin outside the filter. We can set
       * TYPECHANGES to false here because the type of Y (in terms of
       * which the original predicate is written) has not changed.
       */
      return filter_move(ctx, te, &sub->adjoin.left, NULL);
   }

   if (sub->type == TCE_STEP) {
      /* these don't exist yet when we run (for now) */
      PQLASSERT(0);
   }

   if (sub->type == TCE_REPEAT) {
      /* repeat is a form of join; move under where possible */
      struct tcexpr *left, *center, *right;
      struct tcvar *leftvar, *centervar, *rightvar;
      struct datatype *remaindertype;

      /* type of repeat RHS; should probably be stored in the repeat node */
      {
	 struct datatype *reptype, *repsubtype;
	 struct datatypearray types;
	 struct columnarray cols;
	 unsigned reparity, repsubarity, arity, i, j;

	 reptype = sub->datatype;
	 repsubtype = sub->repeat.sub->datatype;
	 reptype = datatype_set_member(reptype);
	 repsubtype = datatype_set_member(repsubtype);
	 PQLASSERT(datatype_istuple(reptype));
	 PQLASSERT(datatype_istuple(repsubtype));
	 reparity = datatype_arity(reptype);
	 repsubarity = datatype_arity(repsubtype);
	 PQLASSERT(reparity > repsubarity);
	 arity = reparity - repsubarity;

	 datatypearray_init(&types);
	 columnarray_init(&cols);
	 datatypearray_setsize(ctx->pql, &types, arity);
	 columnarray_setsize(ctx->pql, &cols, arity);
	 for (i=0; i<arity; i++) {
	    j = repsubarity + i;
	    datatypearray_set(&types, i, datatype_getnth(reptype, j));
	    columnarray_set(&cols, i, datatype_tuple_getcolumn(reptype, j));
	 }

	 remaindertype = datatype_tuple_specific(ctx->pql, &types, &cols);
	 remaindertype = datatype_set(ctx->pql, remaindertype);

	 datatypearray_setsize(ctx->pql, &types, 0);
	 columnarray_setsize(ctx->pql, &cols, 0);
	 datatypearray_cleanup(ctx->pql, &types);
	 columnarray_cleanup(ctx->pql, &cols);
      }

      predicate_split(ctx, te->filter.predvar, te->filter.predexpr,
		      sub->repeat.sub->datatype,
		      remaindertype,
		      &leftvar, &left, &centervar, &center, &rightvar, &right);
      combine_predicates(ctx, centervar, center, rightvar, right,
			 &te->filter.predvar, &te->filter.predexpr);
      if (left != NULL) {
	 left = mktcexpr_filter(ctx->pql, sub->repeat.sub, leftvar, left);
	 left->datatype = sub->repeat.sub->datatype;
	 left = tcexpr_baseopt(ctx, left);
	 sub->repeat.sub = left;
      }
      if (te->filter.predexpr == NULL) {
	 goto destroy;
      }
   }

   if (sub->type == TCE_SCAN) {
      /* nothing */
   }


   if (sub->type == TCE_BOP) {
      switch (sub->bop.op) {
       case F_UNION:
       case F_INTERSECT:
       case F_UNIONALL:
       case F_INTERSECTALL:
       case F_EXCEPT:
       case F_EXCEPTALL:
       case F_CONCAT:
       case F_CHOOSE:
	return filter_movetwo(ctx, te, &sub->bop.left, &sub->bop.right);
       default:
	/* either do not make sense or aren't binary operators */
	PQLASSERT(0);
	break;
      }
   }

   if (sub->type == TCE_UOP) {
      switch (sub->uop.op) {
       case F_MAKESET:
	/*
	 * filter(makeset(X), lambda Y: PRED(Y))
	 *    -> let Y = X: PRED(Y) ? makeset(Y) : value(emptyset)
	 *
	 * but for now we don't have conditionals (XXX)
	 */
	break;
       case F_MAYBEMAKESET:
	/* shouldn't exist here */
	PQLASSERT(0);
	break;
       default:
	/* do not make sense or aren't unary */
	PQLASSERT(0);
	break;
      }
   }

   if (sub->type == TCE_FUNC) {
      /* none of the non-unary and non-binary functions yield sets */
      PQLASSERT(0);
   }

   if (sub->type == TCE_MAP) {
      /*
       * filter(map A in X: B(A), lambda Y: PRED(Y))
       *    ->  map A in (filter X, lambda Z: let Y=B(Z): PRED(Y)): B(A)
       *
       * but we should only clone B() under some circumstances and not
       * others... e.g. a join would probably be bad. XXX think about this.
       */
      /* XXX notyet */
   }

   if (sub->type == TCE_LET) {
      /*
       * filter(let A = X: B(A), lambda Y: PRED(Y))
       *    -> let A = X: filter(B(A), lambda Y: PRED(Y))
       */
      return filter_move(ctx, te, &sub->let.body, NULL);
   }

   if (sub->type == TCE_READVAR) {
      /*
       * We can't do anything with this with local knowledge. If the
       * variable is let-bound and used only once, hopefully there'll
       * be another pass capable of removing it. If the variable is
       * let-bound and used more than once, there isn't much we can do
       * at all, unless CSE manages to merge the two uses.
       */
   }

   if (sub->type == TCE_READGLOBAL) {
      /* can't do anything */
   }

   if (sub->type == TCE_CREATEPATHELEMENT) {
      /* doesn't yield a set, so doesn't make sense here */
      PQLASSERT(0);
   }

   if (sub->type == TCE_SPLATTER) {
      return filter_move(ctx, te, &sub->splatter.value, NULL);
   }

   if (sub->type == TCE_TUPLE) {
      /* doesn't yield a set, so doesn't make sense here */
      PQLASSERT(0);
   }

   if (sub->type == TCE_VALUE) {
      /* constant set value */

      PQLASSERT(pqlvalue_isset(sub->value));
      if (pqlvalue_set_getnum(sub->value) == 0) {
	 /* filter(emptyset) -> emptyset */
	 goto destroy;
      }

      /* otherwise, should apply the filter to it right now */
      /* XXX notyet */
   }

   /* Nothing matched; give up. */
   return te;

 destroy:
   te->filter.sub = NULL;
   tcexpr_destroy(ctx->pql, te);
   return any_baseopt(ctx, sub);
}

//////////////////////////////
// project

static struct tcexpr *projecttuple_baseopt(struct baseopt *ctx,
					   struct tcexpr *te) {

   PQLASSERT(te->type == TCE_PROJECTTUPLE);

   while (1) {
      /*
       * project(project(x)) => project(x)
       */
      if (te->projecttuple.sub->type == TCE_PROJECTTUPLE) {
	 te->projecttuple.sub = tcexpr_destruct(ctx, te->projecttuple.sub);
	 continue;
      }

      /* would not be well typed */
      PQLASSERT(te->projecttuple.sub->type != TCE_PROJECTSCALAR);

      /*
       * Check for nops; however, if the columns are in a different
       * order, bail, as we have to assume the order matters to the
       * parent expression.
       */
      {
	 struct tcexpr *sub;
	 struct datatype *subtype;
	 unsigned i, num;

	 sub = te->projecttuple.sub;
	 subtype = sub->datatype;
	 if (datatype_isset(subtype)) {
	    subtype = datatype_set_member(subtype);
	 }
	 PQLASSERT(datatype_istuple(subtype));
	 num = columnarray_num(te->projecttuple.cols);
	 if (num == datatype_arity(subtype)) {
	    for (i=0; i<num; i++) {
	       if (columnarray_get(te->projecttuple.cols, i) !=
		   datatype_tuple_getcolumn(subtype, i)) {
		  break;
	       }
	    }
	    if (i == num) {
	       /* got through without finding one that didn't match */
	       te = tcexpr_destruct(ctx, te);
	       return any_baseopt(ctx, te);
	    }
	    else {
	       break;
	    }
	 }

      }

      /*
       * project a1..an (distinguish x as b) => project a1..an (x)
       */
      if (te->projecttuple.sub->type == TCE_DISTINGUISH) {
	 struct column *newcol;

	 newcol = te->projecttuple.sub->distinguish.newcol;
	 if (!columnarray_contains(te->projecttuple.cols, newcol)) {
	    te->projecttuple.sub = tcexpr_destruct(ctx, te->projecttuple.sub);
	    continue;
	 }
      }

      /*
       * project a1..an,b (e |+| x: project c(x) as b) => 
       *     project a1..an,b (e with c as b)
       * project a1..an,b (e1 |+| x: e2(a1..an)) => project a1..an (e1) |+| e2
       * project a1..an (e1 |+| x: e2 as b) => project a1..an (e1)
       */
      if (te->projecttuple.sub->type == TCE_ADJOIN) {
	 struct datatype *undertype, *neededtype, *coltype;
	 struct column *newcol, *copycol, *tmpcol;
	 struct columnarray usedcols;
	 unsigned i, num;

	 newcol = te->projecttuple.sub->adjoin.newcol;
	 if (!columnarray_contains(te->projecttuple.cols, newcol)) {
	    /* throwing away the newly created column; throw away the adjoin */
	    te->projecttuple.sub = tcexpr_destruct(ctx, te->projecttuple.sub);
	    continue;
	 }

	 /*
	  * If we're going to copy and then drop a column, rename it
	  * instead, and then call across to norenames to clear the
	  * rename through. This only works if the subexpression
	  * generates the column; this is usually but not always the
	  * case.
	  */
	 if (adjoin_copies_column(ctx, te->projecttuple.sub, &copycol) &&
	     !columnarray_contains(te->projecttuple.cols, copycol) &&
	     tcexpr_generates_column(ctx, te->projecttuple.sub, copycol)) {
	    struct tcexpr *rn;

	    column_incref(copycol);
	    rn = mktcexpr_rename(ctx->pql,
				 te->projecttuple.sub->adjoin.left,
				 copycol, newcol);
	    rn->datatype = datatype_rename(rn->rename.sub->datatype,
					   copycol, newcol);

	    te->projecttuple.sub->adjoin.left = NULL;
	    te->projecttuple.sub->adjoin.newcol = NULL;
	    tcexpr_destroy(ctx->pql, te->projecttuple.sub);

	    te->projecttuple.sub = norenames_expr(ctx->pql, rn);
	    te->projecttuple.sub = any_baseopt(ctx, te->projecttuple.sub);
	    continue;
	 }

	 /*
	  * Find the columns used by the adjoin expression. The only
	  * columns we need under the adjoin expression are those it
	  * uses and those the project returns. (Other than the adjoin
	  * result.)
	  */

	 columnarray_init(&usedcols);
	 num = columnarray_num(te->projecttuple.cols);
	 for (i=0; i<num; i++) {
	    tmpcol = columnarray_get(te->projecttuple.cols, i);
	    if (!columnarray_contains(&usedcols, tmpcol) &&
		tmpcol != te->projecttuple.sub->adjoin.newcol) {
	       columnarray_add(ctx->pql, &usedcols, tmpcol, NULL);
	    }
	 }
	 tcexpr_getusedcolumns(ctx, te->projecttuple.sub->adjoin.expr,
			       &usedcols);
	 undertype = te->projecttuple.sub->adjoin.left->datatype;
	 neededtype = datatype_subprojecttuple(undertype, &usedcols);
	 PQLASSERT(neededtype != NULL);

	 /*
	  * The adjoinwrapper and reorderwrapper code fails miserably
	  * if it tries to handle part of a nested tuple. Make sure
	  * that for every column in neededtype, the type of the column
	  * is the same as the type of the column of the same name in
	  * undertype.
	  *
	  * XXX sometime figure out to handle the nested cases right.
	  */
	 bool bad = false;
	 {
	    struct datatype *ut, *ust, *nt, *nst;
	    struct column *col;

	    ut = undertype;
	    if (datatype_isset(ut)) {
	       ut = datatype_set_member(ut);
	    }
	    PQLASSERT(datatype_istuple(ut));
	    nt = neededtype;
	    if (datatype_isset(nt)) {
	       nt = datatype_set_member(nt);
	    }
	    PQLASSERT(datatype_istuple(nt));

	    num = datatype_arity(nt);
	    for (i=0; i<num; i++) {
	       nst = datatype_getnth(nt, i);
	       col = datatype_tuple_getcolumn(nt, i);

	       ust = datatype_subprojectscalar(ut, col);
	       /* must be present */
	       PQLASSERT(ust != NULL);

	       if (!datatype_eq_withnames(nst, ust)) {
		  bad = true;
	       }
	    }
	 }

	 if (!datatype_eq_withnames(neededtype, undertype) && !bad) {
	    struct tcexpr *adjoin, *nproj;

	    adjoin = te->projecttuple.sub;
	    nproj = mktcexpr_projecttuple(ctx->pql, adjoin->adjoin.left);
	    num = columnarray_num(&usedcols);
	    columnarray_setsize(ctx->pql, nproj->projecttuple.cols, num);
	    for (i=0; i<num; i++) {
	       tmpcol = columnarray_get(&usedcols, i);
	       column_incref(tmpcol);
	       columnarray_set(nproj->projecttuple.cols, i, tmpcol);
	    }
	    nproj->datatype = neededtype;
	    adjoin->adjoin.left = nproj;

	    PQLASSERT(datatype_eq_withnames(adjoin->adjoin.left->datatype,
					    neededtype));

	    /* update type of adjoin */
	    coltype = adjoin->adjoin.expr->datatype;
	    if (datatype_isset(adjoin->adjoin.left->datatype)) {
	       coltype = datatype_set(ctx->pql, coltype);
	    }
	    adjoin->datatype = datatype_adjoin(neededtype, coltype,
					       adjoin->adjoin.newcol);
	    PQLASSERT(adjoin->datatype != NULL);

	    if (adjoin->adjoin.var != NULL) {
	       /* update type of adjoin variable */
	       struct tcexpr *wraptemplate;
	       struct datatype *varoldtype, *varnewtype, *faketype;

	       varoldtype = adjoin->adjoin.var->datatype;
	       varnewtype = adjoin->adjoin.left->datatype;

	       /* create a type for the columns we removed */
	       faketype = datatype_substrip(nproj->projecttuple.sub->datatype,
					    nproj->projecttuple.cols);
	       PQLASSERT(faketype != NULL);

	       if (datatype_isset(varnewtype)) {
		  varnewtype = datatype_set_member(varnewtype);
		  faketype = datatype_set_member(faketype);
	       }
	       PQLASSERT(datatype_istuple(varnewtype));
	       adjoin->adjoin.var->datatype = varnewtype;

	       wraptemplate = psplit_adjoinwrapper(ctx->pql, NULL,
						   varnewtype, faketype);
	       wraptemplate = psplit_reorderwrapper(ctx->pql, wraptemplate,
						    varoldtype,
						    datatype_unit(ctx->pql));
	       adjoin->adjoin.expr = 
		  wrapvar(ctx->pql, adjoin->adjoin.expr, adjoin->adjoin.var,
			  wraptemplate);
	       tcexpr_destroy(ctx->pql, wraptemplate);

	       adjoin->adjoin.expr = tcexpr_baseopt(ctx, adjoin->adjoin.expr);
	    }

	    adjoin->adjoin.left = projecttuple_baseopt(ctx,
						       adjoin->adjoin.left);
	    te->projecttuple.sub = any_baseopt(ctx, te->projecttuple.sub);

	    columnarray_setsize(ctx->pql, &usedcols, 0);
	    columnarray_cleanup(ctx->pql, &usedcols);
	    continue;
	 }

	 columnarray_setsize(ctx->pql, &usedcols, 0);
	 columnarray_cleanup(ctx->pql, &usedcols);
      }

      /*
       * project a (scan as a, b, c) => scan as a, _, _
       * (etc.)
       */
      if (te->projecttuple.sub->type == TCE_SCAN) {
	 struct tcexpr *sub;
	 struct datatype *t;
	 bool mod = false;

	 sub = te->projecttuple.sub;
	 if (sub->scan.leftobjcolumn != NULL &&
	     !columnarray_contains(te->projecttuple.cols,
				   sub->scan.leftobjcolumn)) {
	    column_decref(ctx->pql, sub->scan.leftobjcolumn);
	    sub->scan.leftobjcolumn = NULL;
	    mod = true;
	 }
	 if (sub->scan.edgecolumn != NULL &&
	     !columnarray_contains(te->projecttuple.cols,
				   sub->scan.edgecolumn)) {
	    column_decref(ctx->pql, sub->scan.edgecolumn);
	    sub->scan.edgecolumn = NULL;
	    mod = true;
	 }
	 if (sub->scan.rightobjcolumn != NULL &&
	     !columnarray_contains(te->projecttuple.cols,
				   sub->scan.rightobjcolumn)) {
	    column_decref(ctx->pql, sub->scan.rightobjcolumn);
	    sub->scan.rightobjcolumn = NULL;
	    mod = true;
	 }

	 if (mod) {
	    /* update type */
	    t = datatype_unit(ctx->pql);
	    if (sub->scan.leftobjcolumn != NULL) {
	       t = datatype_tuple_append(ctx->pql, t,
					 datatype_absdbobj(ctx->pql),
					 sub->scan.leftobjcolumn);
	    }
	    if (sub->scan.edgecolumn != NULL) {
	       t = datatype_tuple_append(ctx->pql, t,
					 datatype_absdbedge(ctx->pql),
					 sub->scan.edgecolumn);
	    }
	    if (sub->scan.rightobjcolumn != NULL) {
	       t = datatype_tuple_append(ctx->pql, t,
					 datatype_absdbobj(ctx->pql),
					 sub->scan.rightobjcolumn);
	    }
	    te->datatype = t;
	 }

#if 0
	 /* since there can't be any other columns, nuke the project */
	 /* XXX ordering? */
	 te->projecttuple.sub = NULL;
	 tcexpr_destroy(ctx->pql, te);
	 return sub;
#else
	 continue;
#endif
      }

      /*
       * Delete any columns from the tuple expression that aren't in
       * the projection.
       *
       * XXX: ought to reorder the subexpressions to be in the same
       * order as the projection.
       *
       * XXX: what about nested columns?
       */
      if (te->projecttuple.sub->type == TCE_TUPLE) {
	 unsigned i, num;
	 struct tcexpr *sub, *te_i;
	 struct column *col_i;

	 sub = te->projecttuple.sub;
	 num = columnarray_num(sub->tuple.columns);
	 for (i=num; i-- > 0; ) {
	    col_i = columnarray_get(sub->tuple.columns, i);
	    if (!columnarray_contains(te->projecttuple.cols, col_i)) {
	       te_i = tcexprarray_get(&sub->tuple.exprs, i);
	       tcexpr_destroy(ctx->pql, te_i);
	       column_decref(ctx->pql, col_i);
	       tcexprarray_remove(&te->projecttuple.sub->tuple.exprs, i);
	       columnarray_remove(te->projecttuple.sub->tuple.columns, i);
	    }
	 }
	 te->projecttuple.sub->datatype = te->datatype;

	 continue;
      }

      /* Nothing matched; give up. */
      break;
   }

   return te;
}

static struct tcexpr *projectscalar_baseopt(struct baseopt *ctx,
					    struct tcexpr *te) {
   PQLASSERT(te->type == TCE_PROJECTSCALAR);

   while (1) {
      if (te->projectscalar.sub->type == TCE_PROJECTTUPLE) {
	 /* project(project(x)) -> project(x) */
	 te->projectscalar.sub = tcexpr_destruct(ctx, te->projectscalar.sub);
	 continue;
      }

      if (te->projectscalar.sub->type == TCE_ADJOIN) {
	 if (te->projectscalar.col == te->projectscalar.sub->adjoin.newcol) {
	    struct tcexpr *adjoin, *funcbody;

	    adjoin = te->projectscalar.sub;
	    funcbody = adjoin->adjoin.expr;

	    if (funcbody->type == TCE_PROJECTSCALAR &&
		funcbody->projectscalar.sub->type == TCE_READVAR &&
		funcbody->projectscalar.sub->readvar == adjoin->adjoin.var) {
	       /*
		* We're projecting an adjoined column that's just a
		* projection of some other column. Take that column
		* instead...
		*/
	       column_decref(ctx->pql, te->projectscalar.col);
	       //column_incref(funcbody->projectscalar.col);
	       te->projectscalar.col = funcbody->projectscalar.col;
	       funcbody->projectscalar.col = NULL;

	       te->projectscalar.sub = adjoin->adjoin.left;
	       adjoin->adjoin.left = NULL;
	       tcexpr_destroy(ctx->pql, adjoin);
	       continue;
	    }
	 }
	 else {
	    te->projectscalar.sub = tcexpr_destruct(ctx,te->projectscalar.sub);
	    continue;
	 }
      }

      /* would not be well typed */
      PQLASSERT(te->projectscalar.sub->type != TCE_PROJECTSCALAR);

      /* Nothing matched; give up. */
      break;
   }

   return te;
}

//////////////////////////////
// strip

#if 0
static struct tcexpr *strip_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *sub;
   struct column *fromcol;

   PQLASSERT(te->type == TCE_STRIP);

   while (1) {
      sub = te->strip.sub;

      /*
       * Rather than building this check into several of the cases below,
       * let's do it in one common place.
       */
      if (columnarray_num(te->strip.cols) == 0) {
	 te->strip.sub = NULL;
	 tcexpr_destroy(ctx->pql, te);
	 return sub;
      }


      /*
       * strip_a(project_a(x)) => strip_*(x)
       * strip_a(project_ab(x)) => project_b(x)
       */
      if (sub->type == TCE_PROJECTTUPLE) {
	 unsigned *indexes;
	 unsigned i, num;
	 int result;

	 num = columnarray_num(te->strip.cols);

	 if (num == columnarray_num(sub->projecttuple.cols)) {
	    /* Result is unit. Drop the projection entirely. */
	    te->strip.sub = sub->projecttuple.sub;
	    sub->projecttuple.sub = NULL;
	    tcexpr_destroy(ctx->pql, sub);

	    /* Rebuild the strip to strip everything. */
	    for (i=0; i<num; i++) {
	       column_decref(ctx->pql, columnarray_get(te->strip.cols, i));
	    }
	    columnarray_setsize(ctx->pql, te->strip.cols, 0);
	    gettopcols(ctx->pql, te->strip.cols, te->strip.sub->datatype,
		       true);

	    /* te's datatype should be unchanged */
	    continue;
	 }

	 indexes = domalloc(ctx->pql, num * sizeof(*indexes));
	 for (i=0; i<num; i++) {
	    result = columnarray_find(sub->projecttuple.cols,
				      columnarray_get(te->strip.cols, i),
				      &indexes[i]);
	    PQLASSERT(result == 0);
	 }
	 qsort(indexes, num, sizeof(indexes[0]), unsigned_sortfunc);
	 for (i=num; i-- > 0; ) {
	    column_decref(ctx->pql,
			  columnarray_get(sub->projecttuple.cols, indexes[i]));
	    columnarray_remove(sub->projecttuple.cols, indexes[i]);
	    sub->datatype = datatype_tupleset_strip(ctx->pql, sub->datatype,
						    indexes[i]);
	 }

	 dofree(ctx->pql, indexes, num * sizeof(*indexes));

	 te->strip.sub = NULL;
	 tcexpr_destroy(ctx->pql, te);

	 return any_baseopt(ctx, sub);
      }

      /*
       * strip_a(strip_b(x)) => strip_ab(x)
       */
      if (sub->type == TCE_STRIP) {
	 columnarray_moveappend(ctx->pql, te->strip.cols, sub->strip.cols);
	 te->strip.sub = sub->strip.sub;
	 sub->strip.sub = NULL;
	 tcexpr_destroy(ctx->pql, sub);
	 continue;
      }

      /*
       * strip_a(x with b as a) => strip_b(x)
       */
      if (sub->type == TCE_RENAME) {
	 unsigned index;
	 struct column *uppername;

	 if (columnarray_find(te->strip.cols, sub->rename.outsidecol,
			      &index) == 0) {
	    /* adjust strip column list */
	    uppername = columnarray_get(te->strip.cols, index);
	    PQLASSERT(uppername == sub->rename.outsidecol);
	    column_decref(ctx->pql, uppername);
	    column_incref(sub->rename.insidecol);
	    columnarray_set(te->strip.cols, index, sub->rename.insidecol);

	    /* types don't change */

	    te->strip.sub = sub->rename.sub;
	    sub->rename.sub = NULL;
	    tcexpr_destroy(ctx->pql, sub);
	    continue;
	 }
      }

      /*
       * strip_ab(nest_c(x) as a) => strip_b(uniq(sort(strip_c(x))))
       */
      if (sub->type == TCE_NEST) {
	 /* notyet */
      }

      /*
       * Ideally, if unnest produces a column X that we're going to
       * strip, we want to go find the corresponding nest or repeat
       * that creates X and shoot X before it's generated at all.
       * However, this is nontrivial.
       *
       * Nothing else useful for unnest here.
       */

      /*
       * strip_a(distinguish x as a) => x
       */
      if (sub->type == TCE_DISTINGUISH) {
	 if (columnarray_contains(te->strip.cols, sub->distinguish.newcol)) {
	    columnarray_removeptr(te->strip.cols, sub->distinguish.newcol);
	    column_decref(ctx->pql, sub->distinguish.newcol);
	    te->strip.sub = sub->distinguish.sub;
	    sub->distinguish.sub = NULL;
	    tcexpr_destroy(ctx->pql, sub);
	    continue;
	 }
      }

      /*
       * strip_a(x |+| y as a) => x
       * strip_a(x |+| \x1 -> a as b) => reorder(x with a as b)
       */
      if (sub->type == TCE_ADJOIN) {
	 if (columnarray_contains(te->strip.cols, sub->adjoin.newcol)) {
	    columnarray_removeptr(te->strip.cols, sub->adjoin.newcol);
	    column_decref(ctx->pql, sub->adjoin.newcol);
	    te->strip.sub = sub->adjoin.left;
	    sub->adjoin.left = NULL;
	    tcexpr_destroy(ctx->pql, sub);
	    continue;
	 }
	 else if (adjoin_copies_column(ctx, sub, &fromcol) &&
		  columnarray_contains(te->strip.cols, fromcol)) {
	    /*
	     * The adjoin adds a clone of a column we're supposed to
	     * wipe out. Nuke the adjoin and replace with a rename and
	     * a project (so the order of columns is preserved) and
	     * remove the column in question from the strip.
	     */
	    struct tcexpr *rn, *proj;
	    unsigned i, num;
	    struct column *thiscol;
	    struct datatype *thiscoltype, *membertype, *newcoltype, *newtype;
	    bool isset = false, isseq = false;

	    column_incref(fromcol);
	    rn = mktcexpr_rename(ctx->pql, sub->adjoin.left,
				 fromcol, sub->adjoin.newcol);
	    sub->adjoin.left = NULL;
	    sub->adjoin.newcol = NULL;
	    te->strip.sub = NULL;
	    tcexpr_destroy(ctx->pql, sub);

	    rn->datatype = datatype_rename(rn->rename.sub->datatype,
					   rn->rename.insidecol,
					   rn->rename.outsidecol);
	    PQLASSERT(rn->datatype != NULL);

	    membertype = rn->datatype;
	    if (datatype_isset(membertype)) {
	       membertype = datatype_set_member(membertype);
	       isset = true;
	    }
	    else if (datatype_issequence(membertype)) {
	       membertype = datatype_sequence_member(membertype);
	       isseq = true;
	    }

	    PQLASSERT(datatype_istuple(membertype));

	    proj = mktcexpr_projecttuple(ctx->pql, rn);
	    newtype = datatype_unit(ctx->pql);

	    newcoltype = NULL;

	    num = datatype_arity(membertype);
	    for (i=0; i<num; i++) {
	       thiscol = datatype_tuple_getcolumn(membertype, i);
	       if (thiscol != rn->rename.outsidecol) {
		  column_incref(thiscol);
		  columnarray_add(ctx->pql, proj->projecttuple.cols, thiscol,
				  NULL);
		  thiscoltype = datatype_getnth(membertype, i);
		  newtype = datatype_tuple_append(ctx->pql, newtype,
						  thiscoltype, thiscol);
	       }
	       else {
		  newcoltype = datatype_getnth(membertype, i);
	       }
	    }
	    PQLASSERT(newcoltype != NULL);
	    column_incref(rn->rename.outsidecol);
	    columnarray_add(ctx->pql, proj->projecttuple.cols,
			    rn->rename.outsidecol, NULL);
	    newtype = datatype_tuple_append(ctx->pql, newtype,
					    newcoltype, rn->rename.outsidecol);
	    if (isset) {
	       newtype = datatype_set(ctx->pql, newtype);
	    }
	    else if (isseq) {
	       newtype = datatype_sequence(ctx->pql, newtype);
	    }

	    proj->datatype = newtype;

	    columnarray_removeptr(te->strip.cols, fromcol);
	    column_decref(ctx->pql, fromcol);

	    /* Call across to norenames to clear out the rename we added */
	    proj->projecttuple.sub = norenames(ctx->pql,
					       proj->projecttuple.sub);

	    /*
	     * Now need to rerun baseopt on the changed bits
	     */

	    /* The rename should have gone away, but baseopt what's there... */
	    proj->projecttuple.sub = any_baseopt(ctx, proj->projecttuple.sub);

	    /* Do the projection... */
	    te->strip.sub = any_baseopt(ctx, proj);

	    /* and the strip itself, if it still exists */
	    continue;
	 }
	 else if (!tcexpr_uses_columns(ctx, sub->adjoin.expr, te->strip.cols)){
	    /* move the strip under the adjoin */

	    /*
	     * XXX: properly we ought to make another strip underneath
	     * the adjoin and move to it exactly the columns that the
	     * adjoin function doesn't use. Sometime...
	     */

	    struct tcexpr *wraptemplate;
	    struct datatype *coltype, *varoldtype, *varnewtype, *faketype;

	    te->strip.sub = sub->adjoin.left;
	    sub->adjoin.left = te;

	    te->datatype = datatype_substrip(te->strip.sub->datatype,
					     te->strip.cols);

	    coltype = sub->adjoin.expr->datatype;
	    if (datatype_isset(sub->adjoin.left->datatype)) {
	       coltype = datatype_set(ctx->pql, coltype);
	    }
	    sub->datatype = datatype_adjoin(sub->adjoin.left->datatype,
					    coltype, sub->adjoin.newcol);
	    PQLASSERT(sub->datatype != NULL);

	    if (sub->adjoin.var != NULL) {
	       varoldtype = sub->adjoin.var->datatype;
	       varnewtype = sub->adjoin.left->datatype;

	       /* create a type for the columns removed by the strip */
	       faketype = datatype_subprojecttuple(te->strip.sub->datatype,
						   te->strip.cols);

	       if (datatype_isset(varnewtype)) {
		  varnewtype = datatype_set_member(varnewtype);
		  faketype = datatype_set_member(faketype);
	       }
	       PQLASSERT(datatype_istuple(varnewtype));
	       sub->adjoin.var->datatype = varnewtype;

	       wraptemplate = psplit_adjoinwrapper(ctx->pql, NULL,
						   varnewtype, faketype);
	       wraptemplate = psplit_reorderwrapper(ctx->pql, wraptemplate,
						    varoldtype,
						    datatype_unit(ctx->pql));
	       sub->adjoin.expr =
		  wrapvar(ctx->pql, sub->adjoin.expr, sub->adjoin.var,
			  wraptemplate);
	       tcexpr_destroy(ctx->pql, wraptemplate);
	    }

	    sub->adjoin.left = strip_baseopt(ctx, sub->adjoin.left);
	    sub->adjoin.expr = tcexpr_baseopt(ctx, sub->adjoin.expr);
	    return any_baseopt(ctx, sub);
	 }
      }

      /*
       * Stripping out the repeat path output should delete it from the
       * repeat.
       */
      if (sub->type == TCE_REPEAT) {
	 if (sub->repeat.repeatpathcolumn != NULL &&
	     columnarray_contains(te->strip.cols,
				  sub->repeat.repeatpathcolumn)) {
	    /* notyet */
	 }
      }

      /*
       * strip_a(scan as a, b, c) => scan as _, b, c
       * strip_b(scan as a, b, c) => scan as a, _, c
       * strip_c(scan as a, b, c) => scan as a, b, _
       */
      if (sub->type == TCE_SCAN) {
	 /* XXX this needs to update sub->datatype */
	 if (sub->scan.leftobjcolumn != NULL &&
	     columnarray_contains(te->strip.cols, sub->scan.leftobjcolumn)) {
	    columnarray_removeptr(te->strip.cols, sub->scan.leftobjcolumn);
	    column_decref(ctx->pql, sub->scan.leftobjcolumn);
	    column_decref(ctx->pql, sub->scan.leftobjcolumn);
	    sub->scan.leftobjcolumn = NULL;
	 }
	 if (sub->scan.edgecolumn != NULL &&
	     columnarray_contains(te->strip.cols, sub->scan.edgecolumn)) {
	    columnarray_removeptr(te->strip.cols, sub->scan.edgecolumn);
	    column_decref(ctx->pql, sub->scan.edgecolumn);
	    column_decref(ctx->pql, sub->scan.edgecolumn);
	    sub->scan.edgecolumn = NULL;
	 }
	 if (sub->scan.rightobjcolumn != NULL &&
	     columnarray_contains(te->strip.cols, sub->scan.rightobjcolumn)) {
	    columnarray_removeptr(te->strip.cols, sub->scan.rightobjcolumn);
	    column_decref(ctx->pql, sub->scan.rightobjcolumn);
	    column_decref(ctx->pql, sub->scan.rightobjcolumn);
	    sub->scan.rightobjcolumn = NULL;
	 }
	 /* there's nothing left to strip, so we must be done! */
	 PQLASSERT(columnarray_num(te->strip.cols) == 0);
	 te->strip.sub = NULL;
	 tcexpr_destroy(ctx->pql, te);
	 return sub;
      }

      /*
       * strip_b(x as a, y as b, z as c, ...) => (x as a, z as c, ...)
       */
      if (sub->type == TCE_TUPLE) {
	 struct tcexpr *subexpr;
	 unsigned *indexes;
	 unsigned i, num;
	 int result;

	 num = columnarray_num(te->strip.cols);
	 indexes = domalloc(ctx->pql, num * sizeof(*indexes));
	 for (i=0; i<num; i++) {
	    result = columnarray_find(sub->tuple.columns,
				      columnarray_get(te->strip.cols, i),
				      &indexes[i]);
	    PQLASSERT(result == 0);
	 }
	 qsort(indexes, num, sizeof(indexes[0]), unsigned_sortfunc);
	 for (i=num; i-- > 0; ) {
	    column_decref(ctx->pql,
			  columnarray_get(sub->tuple.columns, indexes[i]));
	    columnarray_remove(sub->tuple.columns, indexes[i]);
	    subexpr = tcexprarray_get(&sub->tuple.exprs, i);
	    tcexpr_destroy(ctx->pql, subexpr);
	    tcexprarray_remove(&sub->tuple.exprs, i);
	 }
	 dofree(ctx->pql, indexes, num * sizeof(*indexes));

	 sub->datatype = te->datatype;
	 tcexpr_destroy(ctx->pql, te);
	 return any_baseopt(ctx, sub);
      }

      /*
       * strip_a(nil) = nil
       */
      if (sub->type == TCE_VALUE) {
	 if (pqlvalue_isnil(sub->value)) {
	    te->strip.sub = NULL;
	    tcexpr_destroy(ctx->pql, te);
	    return sub;
	 }
      }

      /* Nothing matched; give up. */
      break;
   }

   return te;
}
#endif /* 0 */

//////////////////////////////
// join

static struct tcexpr *join_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   PQLASSERT(te->type == TCE_JOIN);

   /*
    * Mostly, things are combined or pushed under joins, not vice versa.
    *
    * The only thing I can think of that might be worth putting here
    * (so far) is that (A x S) where S is a singleton can be written
    * as adjoining a constant S to A instead. (And likewise for (S x
    * A), although because our adjoin works only on the right it's
    * somewhat more complicated.)
    *
    * There are also some cases where joining a relation to itself can
    * be reduced to a simpler form; e.g. if A is a column of integers
    * "foo" and B is the same but renamed "bar", then
    *
    *     A except (project foo (A x B where lambda X: X.foo > X.bar))
    * 
    * is a long form of writing min(A.foo). But these are algebraic
    * curiosities not likely to appear in real life; also, they're a
    * bit too complex for this module.
    */

   /*
    * What we DO need to do here is call predicate_split() to push as
    * much of the join condition as possible into the leaves. This
    * should only be done once, because it can legitimately leave
    * stuff in place behind it and if we keep trying on that we'll
    * loop forever.
    */
   if (te->join.predexpr != NULL) {
      struct tcexpr *left, *center, *right;
      struct tcvar *leftvar, *centervar, *rightvar;

      predicate_split(ctx, te->join.predvar, te->join.predexpr,
		      te->join.left->datatype,
		      te->join.right->datatype,
		      &leftvar, &left, &centervar, &center, &rightvar, &right);
      te->join.predvar = centervar;
      te->join.predexpr = center;
      if (left != NULL) {
	 left = mktcexpr_filter(ctx->pql, te->join.left, leftvar, left);
	 left->datatype = te->join.left->datatype;
	 left = tcexpr_baseopt(ctx, left);
	 te->join.left = left;
      }
      if (right != NULL) {
	 right = mktcexpr_filter(ctx->pql, te->join.right, rightvar, right);
	 right->datatype = te->join.right->datatype;
	 right = tcexpr_baseopt(ctx, right);
	 te->join.right = right;
      }
   }

   /*
    * Ok, since REPEAT as it's currently formulated is an expensive
    * kind of repeated join, if we have (REPEAT A B) x C and the
    * predicate does not depend on B, change it to REPEAT (A x C) B.
    *
    * This is not a substitute for a join order planner but it should
    * hit a common case.
    *
    * XXX: Note that we really ought to call predicate_split in case
    * we have a separable subpredicate that's independent of B.
    * Currently we can't do that because the interface for predicate
    * splitting assumes a strict left/right partitioning and for this
    * we need A and C on the left and B on the right. Should fix that.
    * (And I am now wondering if it makes sense to rearrange splitting
    * to do a two-way instead of three-way split, and peel off
    * different left and right sections by calling it repeatedly. Or
    * if it should be set up as an N-way split instead.)
    */
   if (te->join.left->type == TCE_REPEAT) {
      struct datatype *repeattype, *subtype;
      struct columnarray cols;
      unsigned repeatarity, subarity, i;

      /* Get the columns produced by the repeat */

      repeattype = te->join.left->datatype;
      PQLASSERT(datatype_isset(repeattype));
      repeattype = datatype_set_member(repeattype);
      PQLASSERT(datatype_istuple(repeattype));

      subtype = te->join.left->repeat.sub->datatype;
      PQLASSERT(datatype_isset(subtype));
      subtype = datatype_set_member(subtype);
      PQLASSERT(datatype_istuple(subtype));

      repeatarity = datatype_arity(repeattype);
      subarity = datatype_arity(subtype);

      /* type of SUB (the repeat's lhs) should be a prefix of REPEAT */
      PQLASSERT(subarity < repeatarity);
      for (i=0; i<subarity; i++) {
	 PQLASSERT(datatype_tuple_getcolumn(repeattype, i) ==
		   datatype_tuple_getcolumn(subtype, i));
      }
      
      /* columns in the repeat's rhs are the ones not in the lhs */
      columnarray_init(&cols);
      columnarray_setsize(ctx->pql, &cols, repeatarity - subarity);
      for (i=subarity; i<repeatarity; i++) {
	 columnarray_set(&cols, i-subarity,
			 datatype_tuple_getcolumn(repeattype, i));
      }
      
      if (!tcexpr_uses_columns(ctx, te->join.predexpr, &cols)) {
	 /*
	  * Predicate is independent of the repeat output. Move it
	  * inside.
	  */

	 struct tcexpr *join, *repeat, *left, *right, *project;
	 struct datatype *jointype, *oldtype, *lefttype, *righttype;
	 struct datatype *type, *checktype;
	 struct column *col, *checkcol;
	 unsigned joinarity, oldarity, leftarity, rightarity;
	 unsigned oldrightpos;
	 struct datatypearray types;

	 datatypearray_init(&types);

	 /* name all the expressions */
	 join = te;
	 repeat = te->join.left;
	 left = repeat->repeat.sub;
	 right = te->join.right;

	 /* get prior overall type */
	 oldtype = datatype_set_member(join->datatype);

	 /* reorganize them */
	 join->join.left = left;
	 join->join.right = right;
	 repeat->repeat.sub = join;

	 /* new types */
	 lefttype = datatype_set_member(left->datatype);
	 righttype = datatype_set_member(right->datatype);
	 jointype = datatype_tuple_concat(ctx->pql, lefttype, righttype);
	 join->datatype = datatype_set(ctx->pql, jointype);

	 /*
	  * new type for repeat result
	  */

	 oldarity = datatype_arity(oldtype);
	 leftarity = datatype_arity(lefttype);
	 rightarity = datatype_arity(righttype);
	 joinarity = datatype_arity(jointype);
	 PQLASSERT(leftarity + rightarity == joinarity);

	 datatypearray_setsize(ctx->pql, &types, oldarity);
	 columnarray_setsize(ctx->pql, &cols, oldarity);

	 for (i=0; i<leftarity; i++) {
	    type = datatype_getnth(lefttype, i);
	    col = datatype_tuple_getcolumn(lefttype, i);
	    checktype = datatype_getnth(oldtype, i);
	    checkcol = datatype_tuple_getcolumn(oldtype, i);
	    PQLASSERT(datatype_eq_withnames(type, checktype));
	    PQLASSERT(col == checkcol);
	    datatypearray_set(&types, i, type);
	    columnarray_set(&cols, i, col);
	 }
	 oldrightpos = oldarity - rightarity;
	 for (i=0; i<rightarity; i++) {
	    type = datatype_getnth(righttype, i);
	    col = datatype_tuple_getcolumn(righttype, i);
	    checktype = datatype_getnth(oldtype, i + oldrightpos);
	    checkcol = datatype_tuple_getcolumn(oldtype, i + oldrightpos);
	    PQLASSERT(datatype_eq_withnames(type, checktype));
	    PQLASSERT(col == checkcol);
	    datatypearray_set(&types, i + leftarity, type);
	    columnarray_set(&cols, i + leftarity, col);
	 }
	 for (i=leftarity; i<oldrightpos; i++) {
	    type = datatype_getnth(oldtype, i);
	    col = datatype_tuple_getcolumn(oldtype, i);
	    datatypearray_set(&types, i + rightarity, type);
	    columnarray_set(&cols, i + rightarity, col);
	 }

	 repeat->datatype = datatype_tuple_specific(ctx->pql, &types, &cols);
	 repeat->datatype = datatype_set(ctx->pql, repeat->datatype);

	 /*
	  * new type for the join predicate, if any
	  */
	 if (join->join.predexpr != NULL && join->join.predvar != NULL) {
	    /*
	     * Update the type of the variable. It was an element of
	     * A x B x C; it is now an element of A x C.
	     */
	    struct tcexpr *wraptemplate;
	    struct datatype *vartype;

	    vartype = datatype_set_member(join->datatype);
	    join->join.predvar->datatype = vartype;

	    /*
	     * Now use wrapvar to update uses of the variable; adjoin
	     * columns of nil matching the columns in B, and then
	     * reorder.
	     */
	    wraptemplate = NULL;
	    for (i=leftarity; i<oldrightpos; i++) {
	       type = datatype_getnth(oldtype, i);
	       col = datatype_tuple_getcolumn(oldtype, i);
	       wraptemplate = psplit_adjoinonewrapper(ctx->pql, wraptemplate,
						      vartype, type, col);
	    }
	    wraptemplate = psplit_reorderwrapper(ctx->pql, wraptemplate,
						 oldtype,
						 datatype_unit(ctx->pql));
	    join->join.predexpr = wrapvar(ctx->pql, join->join.predexpr,
					  join->join.predvar,
					  wraptemplate);
	    tcexpr_destroy(ctx->pql, wraptemplate);
	 }

	 /* wrap in a project to preserve column order */
	 project = mktcexpr_projecttuple(ctx->pql, repeat);
	 columnarray_setsize(ctx->pql, project->projecttuple.cols, oldarity);
	 for (i=0; i<oldarity; i++) {
	    col = datatype_tuple_getcolumn(oldtype, i);
	    column_incref(col);
	    columnarray_set(project->projecttuple.cols, i, col);
	 }
	 project->datatype = datatype_set(ctx->pql, oldtype);

	 datatypearray_setsize(ctx->pql, &types, 0);
	 datatypearray_cleanup(ctx->pql, &types);
	 columnarray_setsize(ctx->pql, &cols, 0);
	 columnarray_cleanup(ctx->pql, &cols);

	 repeat->repeat.sub = join_baseopt(ctx, repeat->repeat.sub);
	 /* could just call repeat_baseopt but there's no forward decl */
	 project->projecttuple.sub = any_baseopt(ctx,
						 project->projecttuple.sub);

	 return tcexpr_baseopt(ctx, project);
      }

      columnarray_setsize(ctx->pql, &cols, 0);
      columnarray_cleanup(ctx->pql, &cols);
   }

#if 0
   if (te->join.left->type == TCE_STRIP) {
      /*
       * Convert it to PROJECTTUPLE so the subsequent logic can be applied.
       */
      struct tcexpr *strip, *stripsub;
      struct columnarray *stripcols;

      strip = te->join.left;
      stripsub = strip->strip.sub;
      stripcols = strip->strip.cols;

      columnarray_take_complement(ctx, stripcols, stripsub->datatype);

      // XXX this should be removed. bypassing ctors is evil

      strip->type = TCE_PROJECTTUPLE;
      strip->projecttuple.sub = stripsub;
      strip->projecttuple.cols = stripcols;
      strip->projecttuple.complement_cols = false;
   }
#endif

   if (te->join.left->type == TCE_PROJECTTUPLE) {
      /*
       * Move the join inside the project so we can get at what's
       * under the project. (If the project were going to have gone
       * away by merging with it, it already would have.)
       *
       * It might be the case that the project is throwing away
       * columns on the left that are also found on the right. In that
       * case, apply renames.
       */

      struct tcexpr *sub, *subsub;
      struct datatype *subsubtype, *righttype;
      struct column *col, *newcol;
      unsigned num, rightarity, i;
      bool renames;

      sub = te->join.left;
      subsub = sub->projecttuple.sub;
      subsubtype = subsub->datatype;
      PQLASSERT(datatype_isset(subsubtype));
      subsubtype = datatype_set_member(subsubtype);
      PQLASSERT(datatype_istuple(subsubtype));

      renames = false;
      num = datatype_arity(subsubtype);
      for (i=0; i<num; i++) {
	 col = datatype_tuple_getcolumn(subsubtype, i);
	 if (columnarray_contains(sub->projecttuple.cols, col)) {
	    continue;
	 }
	 if (datatype_contains_column(te->join.right->datatype, col)) {
	    newcol = column_create(ctx->pql, NULL);
	    column_incref(col);
	    subsub = mktcexpr_rename(ctx->pql, subsub, col, newcol);
	    subsub->datatype = datatype_rename(subsub->rename.sub->datatype,
					       col, newcol);
	    renames = true;
	 }
      }
      if (renames) {
	 subsub = norenames_expr(ctx->pql, subsub);
	 subsubtype = datatype_set_member(subsub->datatype);
      }

      te->join.left = subsub;
      sub->projecttuple.sub = te;

      righttype = te->join.right->datatype;
      PQLASSERT(datatype_isset(righttype));
      righttype = datatype_set_member(righttype);
      PQLASSERT(datatype_istuple(righttype));

      rightarity = datatype_arity(righttype);
      for (i=0; i<rightarity; i++) {
	 col = datatype_tuple_getcolumn(righttype, i);
	 column_incref(col);
	 columnarray_add(ctx->pql, sub->projecttuple.cols, col, NULL);
      }

      te->datatype = datatype_join(te->join.left->datatype,
				   te->join.right->datatype);
      if (te->join.predexpr != NULL && te->join.predvar != NULL) {
	 /*
	  * Here the variable has gained new columns, specifically the
	  * ones discarded by the projection. The uses of the variable
	  * expect those to not exist, so wrap uses of the variable in
	  * the same projection.
	  */
	 struct tcexpr *wraptemplate;
	 struct datatype *vartype;

	 vartype = te->join.predvar->datatype;
	 wraptemplate = psplit_reorderwrapper(ctx->pql, NULL, vartype,
					      datatype_unit(ctx->pql));
	 vartype = datatype_set_member(te->datatype);
	 te->join.predvar->datatype = vartype;

	 te->join.predexpr = wrapvar(ctx->pql,
				     te->join.predexpr,
				     te->join.predvar,
				     wraptemplate);

	 tcexpr_destroy(ctx->pql, wraptemplate);
      }

      sub->datatype = datatype_subprojecttuple(te->datatype,
					       sub->projecttuple.cols);
      sub->projecttuple.sub = join_baseopt(ctx, sub->projecttuple.sub);
      return tcexpr_baseopt(ctx, sub);
   }

   if (te->join.left->type == TCE_ADJOIN) {
      /*
       * (A |+| B) x C => (A x C) |+| B
       *
       * This works even if the join condition depends on B, because
       * we can stuff a copy of the adjoin inside the condition.
       */

      struct tcexpr *adjoin, *left, *wraptemplate, *project;
      struct datatype *adjoincoltype, *oldtype, *vartype, *lefttype;
      struct column *col;
      unsigned oldarity, i;

      adjoin = te->join.left;
      left = adjoin->adjoin.left;

      te->join.left = left;
      adjoin->adjoin.left = te;

      adjoincoltype = adjoin->adjoin.expr->datatype;
      oldtype = te->datatype;
      oldtype = datatype_set_member(oldtype);
      PQLASSERT(datatype_istuple(oldtype));
      oldarity = datatype_arity(oldtype);

      te->datatype = datatype_join(te->join.left->datatype,
				   te->join.right->datatype);
      adjoin->datatype = datatype_adjoin(te->datatype,
					 datatype_set(ctx->pql, adjoincoltype),
					 adjoin->adjoin.newcol);
      PQLASSERT(adjoin->datatype != NULL);

      /* this is the type for both the join predicate and adjoin func vars */
      vartype = datatype_set_member(te->datatype);

      if (adjoin->adjoin.var != NULL) {
	 adjoin->adjoin.var->datatype = vartype;

	 /*
	  * Wrap uses of the adjoin function's variable in a projection
	  * to drop the columns it just gained from the join's RHS.
	  *
	  * XXX: "reorderwrapper" is misnamed.
	  */
	 lefttype = datatype_set_member(te->join.left->datatype);
	 wraptemplate = psplit_reorderwrapper(ctx->pql, NULL,
					      lefttype,
					      datatype_unit(ctx->pql));
	 adjoin->adjoin.expr = wrapvar(ctx->pql,
				       adjoin->adjoin.expr,
				       adjoin->adjoin.var,
				       wraptemplate);
	 tcexpr_destroy(ctx->pql, wraptemplate);
      }

      if (te->join.predexpr != NULL && te->join.predvar != NULL) {
	 te->join.predvar->datatype = vartype;

	 if (tcexpr_uses_column(ctx, te->join.predexpr,
				adjoin->adjoin.newcol)) {
	    adjoin->adjoin.left = NULL;
	    wraptemplate = tcexpr_clone(ctx->pql, adjoin);
	    adjoin->adjoin.left = te;
	 }
	 else {
	    wraptemplate = psplit_adjoinonewrapper(ctx->pql, NULL,
						   vartype, adjoincoltype,
						   adjoin->adjoin.newcol);
	 }
	 wraptemplate = psplit_reorderwrapper(ctx->pql, wraptemplate,
					      oldtype,
					      datatype_unit(ctx->pql));
	 te->join.predexpr = wrapvar(ctx->pql, te->join.predexpr,
				     te->join.predvar, wraptemplate);

	 tcexpr_destroy(ctx->pql, wraptemplate);
      }

      /* wrap in a project to preserve column order */
      project = mktcexpr_projecttuple(ctx->pql, adjoin);
      columnarray_setsize(ctx->pql, project->projecttuple.cols, oldarity);
      for (i=0; i<oldarity; i++) {
	 col = datatype_tuple_getcolumn(oldtype, i);
	 column_incref(col);
	 columnarray_set(project->projecttuple.cols, i, col);
      }
      project->datatype = datatype_set(ctx->pql, oldtype);

      return tcexpr_baseopt(ctx, project);
   }


   /* XXX we ought to also apply all the above logic to the right subnode... */

   return te;
}

//////////////////////////////
// order

static struct tcexpr *order_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *sub;

   PQLASSERT(te->type == TCE_ORDER);
   (void)ctx;

   sub = te->order.sub;

   /* Sets of cardinality 1 are already ordered. */
   if ((sub->type == TCE_UOP && sub->uop.op == F_MAKESET) ||
       (sub->type == TCE_FUNC && sub->func.op == F_MAKESET &&
	tcexprarray_num(&sub->func.args) == 1)) {
      te->order.sub = NULL;
      tcexpr_destroy(ctx->pql, te);
      return sub;
   }

   return te;
}

//////////////////////////////
// uniq

static struct tcexpr *uniq_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *sub;

   PQLASSERT(te->type == TCE_UNIQ);
   (void)ctx;

   sub = te->uniq.sub;

   /* Sets of cardinality 1 are already unique. */
   if ((sub->type == TCE_UOP && sub->uop.op == F_MAKESET) ||
       (sub->type == TCE_FUNC && sub->func.op == F_MAKESET &&
	tcexprarray_num(&sub->func.args) == 1)) {
      te->uniq.sub = NULL;
      tcexpr_destroy(ctx->pql, te);
      return sub;
   }

   return te;
}

//////////////////////////////
// repeat

static struct tcexpr *repeat_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   PQLASSERT(te->type == TCE_REPEAT);
   (void)ctx;
   return te;
}

//////////////////////////////
// bop

static struct tcexpr *bop_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *left, *right;

   PQLASSERT(te->type == TCE_BOP);

   left = te->bop.left;
   right = te->bop.right;

   if (tcexpr_simple_same(left, right)) {
      struct tcexpr *ret;

      switch (te->bop.op) {
       case F_UNION:
       case F_INTERSECT:
	/* A union A => A; A intersect A => A */
	ret = te->bop.left;
	te->bop.left = NULL;
	tcexpr_destroy(ctx->pql, te);
	return ret;

       case F_EXCEPT:
	/* A except A => {} */
	ret = mktcexpr_value(ctx->pql, pqlvalue_emptyset(ctx->pql));
	ret->datatype = te->datatype;
	tcexpr_destroy(ctx->pql, te);
	return ret;

       case F_SUBSETOF:
       case F_AND:
       case F_OR:
       case F_EQ:
       case F_LTEQ:
       case F_GTEQ:
       case F_LIKE:
       case F_SOUNDEX:
	/* A in A => true; A and A => true; A or A => true; etc. */
	ret = mktcexpr_value(ctx->pql, pqlvalue_bool(ctx->pql, true));
	ret->datatype = te->datatype;
	tcexpr_destroy(ctx->pql, te);
	return ret;

       case F_ELEMENTOF:
       case F_NOTEQ:
       case F_LT:
       case F_GT:
	/* A != A => false; etc. */
	ret = mktcexpr_value(ctx->pql, pqlvalue_bool(ctx->pql, false));
	ret->datatype = te->datatype;
	tcexpr_destroy(ctx->pql, te);
	return ret;

       case F_ADD:
	/* A + A => A * 2 */
	tcexpr_destroy(ctx->pql, te->bop.right);
	te->bop.right = mktcexpr_value(ctx->pql, pqlvalue_int(ctx->pql, 2));
	te->bop.right->datatype = datatype_int(ctx->pql);
	te->bop.op = F_MUL;
	break;

       case F_SUB:
	/* A - A => 0 */
	ret = mktcexpr_value(ctx->pql, pqlvalue_int(ctx->pql, 0));
	ret->datatype = te->datatype;
	tcexpr_destroy(ctx->pql, te);
	return ret;

       case F_DIV:
       case F_MOD:
	/* note: A/A => 1 and A%A => 0, but only if A != 0 */
	break;

       default:
	break;
      }
   }

   switch (te->bop.op) {
    case F_NONEMPTY:
    case F_MAKESET:
    case F_GETELEMENT:
    case F_MAYBEMAKESET:
    case F_COUNT:
    case F_SUM:
    case F_AVG:
    case F_MIN:
    case F_MAX:
    case F_ALLTRUE:
    case F_ANYTRUE:
    case F_NOT:
    case F_NEW:
    case F_NOW:
    case F_TOSTRING:
    case F_LEFT:
    case F_EDGE:
    case F_RIGHT:
    case F_NEG:
    case F_ABS:
    case F_IDENTITY:
     /* not binary */
     PQLASSERT(0);
     break;

    case F_UNION:
     te = union_baseopt(ctx, te);
     break;
    case F_INTERSECT:
     te = intersect_baseopt(ctx, te);
     break;
    case F_EXCEPT:
     te = except_baseopt(ctx, te);
     break;
    case F_UNIONALL:
     te = unionall_baseopt(ctx, te);
     break;
    case F_INTERSECTALL:
     te = intersectall_baseopt(ctx, te);
     break;
    case F_EXCEPTALL:
     te = exceptall_baseopt(ctx, te);
     break;
    case F_SUBSETOF:
     te = subset_baseopt(ctx, te);
     break;
    case F_ELEMENTOF:
     te = element_baseopt(ctx, te);
     break;
    case F_AND:
     te = and_baseopt(ctx, te);
     break;
    case F_OR:
     te = or_baseopt(ctx, te);
     break;
    case F_STRFTIME:
     break;
    case F_EQ:
    case F_NOTEQ:
    case F_LT:
    case F_GT:
    case F_LTEQ:
    case F_GTEQ:
     break;
    case F_LIKE:
     te = like_baseopt(ctx, te);
     break;
    case F_GLOB:
     te = glob_baseopt(ctx, te);
     break;
    case F_GREP:
     te = grep_baseopt(ctx, te);
     break;
    case F_SOUNDEX:
    case F_CONCAT:
    case F_CHOOSE:
    case F_ADD:
    case F_SUB:
    case F_MUL:
    case F_DIV:
    case F_MOD:
     break;
   }

   return te;
}

//////////////////////////////
// uop

static struct tcexpr *uop_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   struct tcexpr *sub;

   PQLASSERT(te->type == TCE_UOP);

   sub = te->uop.sub;

   switch (te->uop.op) {
    case F_UNION:
    case F_INTERSECT:
    case F_EXCEPT:
    case F_UNIONALL:
    case F_INTERSECTALL:
    case F_EXCEPTALL:
    case F_SUBSETOF:
    case F_ELEMENTOF:
    case F_AND:
    case F_OR:
    case F_STRFTIME:
    case F_NOW:
    case F_EQ:
    case F_NOTEQ:
    case F_LT:
    case F_GT:
    case F_LTEQ:
    case F_GTEQ:
    case F_LIKE:
    case F_GLOB:
    case F_GREP:
    case F_SOUNDEX:
    case F_CONCAT:
    case F_CHOOSE:
    case F_ADD:
    case F_SUB:
    case F_MUL:
    case F_DIV:
    case F_MOD:
     /* not unary */
     PQLASSERT(0);
     break;

    case F_NONEMPTY:
     return nonempty_baseopt(ctx, te);
     break;
    case F_MAKESET:
     return makeset_baseopt(ctx, te);
    case F_GETELEMENT:
     return getelement_baseopt(ctx, te);
    case F_MAYBEMAKESET:
     /* not allowed downstream of typeinf */
     PQLASSERT(0);
     break;
    case F_COUNT:
     return count_baseopt(ctx, te);
    case F_SUM:
     return sum_baseopt(ctx, te);
    case F_AVG:
     break;
    case F_MIN:
    case F_MAX:
     return minmax_baseopt(ctx, te);
    case F_ALLTRUE:
     return alltrue_baseopt(ctx, te);
    case F_ANYTRUE:
     return anytrue_baseopt(ctx, te);
    case F_NOT:
     return not_baseopt(ctx, te);
    case F_LEFT:
     return left_baseopt(ctx, te);
    case F_EDGE:
     return edge_baseopt(ctx, te);
    case F_RIGHT:
     return right_baseopt(ctx, te);
    case F_NEW:
    case F_TOSTRING:
    case F_NEG:
    case F_ABS:
     break;
    case F_IDENTITY:
     /* remove it unconditionally */
     te->uop.sub = NULL;
     tcexpr_destroy(ctx->pql, te);
     return sub;
   }

   return te;
}

//////////////////////////////
// other functions

static struct tcexpr *func_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   PQLASSERT(te->type == TCE_FUNC);

   /* Switch to uop/bop if possible as it reduces consing during eval */
   if (tcexprarray_num(&te->func.args) == 1) {
      enum functions op;
      struct tcexpr *sub;

      op = te->func.op;
      sub = tcexprarray_get(&te->func.args, 0);
      tcexprarray_setsize(ctx->pql, &te->func.args, 0);
      tcexprarray_cleanup(ctx->pql, &te->func.args);
      te->type = TCE_UOP;
      te->uop.op = op;
      te->uop.sub = sub;
      return uop_baseopt(ctx, te);
   }
   else if (tcexprarray_num(&te->func.args) == 2) {
      enum functions op;
      struct tcexpr *left, *right;

      op = te->func.op;
      left = tcexprarray_get(&te->func.args, 0);
      right = tcexprarray_get(&te->func.args, 1);
      tcexprarray_setsize(ctx->pql, &te->func.args, 0);
      tcexprarray_cleanup(ctx->pql, &te->func.args);
      te->type = TCE_BOP;
      te->bop.left = left;
      te->bop.op = op;
      te->bop.right = right;
      return bop_baseopt(ctx, te);
   }

   return te;
}

//////////////////////////////
// map

static struct tcexpr *map_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   PQLASSERT(te->type == TCE_MAP);

   if (te->map.result->type == TCE_PROJECTSCALAR ||
       te->map.result->type == TCE_PROJECTTUPLE) {
      struct tcexpr *proj, **subp;
      struct datatype *resulttype;

      resulttype = te->map.result->datatype;
      if (!datatype_isset(resulttype) && !datatype_issequence(resulttype)) {
	 /*
	  * map X in S: project foo(f(X)) => project foo(map X in S: f(X))
	  * (Provided that the project is applied to an element and not a
	  * set.)
	  */
	 proj = te->map.result;
	 switch (proj->type) {
	  case TCE_PROJECTSCALAR: subp = &proj->projectscalar.sub; break;
	  case TCE_PROJECTTUPLE: subp = &proj->projecttuple.sub; break;
	  default: PQLASSERT(0); break;
	 }
	 te->map.result = *subp;
	 *subp = te;

	 proj->datatype = te->datatype;
	 te->datatype = datatype_set(ctx->pql, te->map.result->datatype);

	 *subp = map_baseopt(ctx, *subp);
	 return any_baseopt(ctx, proj);
      }
   }

   /* map X in S: X  =>  S */
   if (te->map.result->type == TCE_READVAR &&
       te->map.result->readvar == te->map.var) {
      PQLASSERT(te->map.var->refcount == 2);
      return tcexpr_destruct(ctx, te);
   }

   return te;
}

//////////////////////////////
// let

static struct tcexpr *let_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   PQLASSERT(te->type == TCE_LET);

   /* let X = E: X  =>  E */
   if (te->let.body->type == TCE_READVAR &&
       te->let.body->readvar == te->let.var) {
      PQLASSERT(te->let.var->refcount == 2);
      return tcexpr_destruct(ctx, te);
   }

   return te;
}

////////////////////////////////////////////////////////////
// dispatch for specific optimizations

static struct tcexpr *any_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   switch (te->type) {
    case TCE_FILTER:
     te = filter_baseopt(ctx, te);
     break;
    case TCE_PROJECTTUPLE:
     te = projecttuple_baseopt(ctx, te);
     break;
    case TCE_PROJECTSCALAR:
     te = projectscalar_baseopt(ctx, te);
     break;
    case TCE_RENAME:
     break;
    case TCE_JOIN:
     te = join_baseopt(ctx, te);
     break;
    case TCE_ORDER:
     te = order_baseopt(ctx, te);
     break;
    case TCE_UNIQ:
     te = uniq_baseopt(ctx, te);
     break;
    case TCE_NEST:
    case TCE_UNNEST:
    case TCE_DISTINGUISH:
    case TCE_ADJOIN:
    case TCE_STEP:
     break;
    case TCE_REPEAT:
     te = repeat_baseopt(ctx, te);
     break;
    case TCE_SCAN:
     break;
    case TCE_BOP:
     te = bop_baseopt(ctx, te);
     break;
    case TCE_UOP:
     te = uop_baseopt(ctx, te);
     break;
    case TCE_FUNC:
     te = func_baseopt(ctx, te);
     break;
    case TCE_MAP:
     te = map_baseopt(ctx, te);
     break;
    case TCE_LET:
     te = let_baseopt(ctx, te);
     break;
    case TCE_READVAR:
    case TCE_READGLOBAL:
    case TCE_CREATEPATHELEMENT:
    case TCE_SPLATTER:
    case TCE_TUPLE:
    case TCE_VALUE:
     break;
   }
   return te;
}

////////////////////////////////////////////////////////////
// recursive traversal

static struct tcexpr *tcexpr_baseopt(struct baseopt *ctx, struct tcexpr *te) {
   unsigned i, num;
   struct tcexpr *subexpr;

   switch (te->type) {
    case TCE_FILTER:
     te->filter.sub = tcexpr_baseopt(ctx, te->filter.sub);
     te->filter.predexpr = tcexpr_baseopt(ctx, te->filter.predexpr);
     break;
    case TCE_PROJECTTUPLE:
     te->projecttuple.sub = tcexpr_baseopt(ctx, te->projecttuple.sub);
     break;
    case TCE_PROJECTSCALAR:
     te->projectscalar.sub = tcexpr_baseopt(ctx, te->projectscalar.sub);
     break;
    case TCE_RENAME:
     te->rename.sub = tcexpr_baseopt(ctx, te->rename.sub);
     break;
    case TCE_JOIN:
     te->join.left = tcexpr_baseopt(ctx, te->join.left);
     te->join.right = tcexpr_baseopt(ctx, te->join.right);
     if (te->join.predexpr != NULL) {
	te->join.predexpr = tcexpr_baseopt(ctx, te->join.predexpr);
     }
     break;
    case TCE_ORDER:
     te->order.sub = tcexpr_baseopt(ctx, te->order.sub);
     break;
    case TCE_UNIQ:
     te->uniq.sub = tcexpr_baseopt(ctx, te->uniq.sub);
     break;
    case TCE_NEST:
     te->nest.sub = tcexpr_baseopt(ctx, te->nest.sub);
     break;
    case TCE_UNNEST:
     te->unnest.sub = tcexpr_baseopt(ctx, te->unnest.sub);
     break;
    case TCE_DISTINGUISH:
     te->distinguish.sub = tcexpr_baseopt(ctx, te->distinguish.sub);
     break;
    case TCE_ADJOIN:
     te->adjoin.left = tcexpr_baseopt(ctx, te->adjoin.left);
     te->adjoin.expr = tcexpr_baseopt(ctx, te->adjoin.expr);
     break;
    case TCE_STEP:
     te->join.left = tcexpr_baseopt(ctx, te->step.sub);
     if (te->step.predexpr != NULL) {
	te->step.predexpr = tcexpr_baseopt(ctx, te->step.predexpr);
     }
     break;
    case TCE_REPEAT:
     te->repeat.sub = tcexpr_baseopt(ctx, te->repeat.sub);
     te->repeat.body = tcexpr_baseopt(ctx, te->repeat.body);
     break;
    case TCE_SCAN:
     /* nothing */
     break;
    case TCE_BOP:
     te->bop.left = tcexpr_baseopt(ctx, te->bop.left);
     te->bop.right = tcexpr_baseopt(ctx, te->bop.right);
     break;
    case TCE_UOP:
     te->uop.sub = tcexpr_baseopt(ctx, te->uop.sub);
     break;
    case TCE_FUNC:
     num = tcexprarray_num(&te->func.args);
     for (i=0; i<num; i++) {
	subexpr = tcexprarray_get(&te->func.args, i);
	subexpr = tcexpr_baseopt(ctx, subexpr);
	tcexprarray_set(&te->func.args, i, subexpr);
     }
     break;
    case TCE_MAP:
     te->map.set = tcexpr_baseopt(ctx, te->map.set);
     te->map.result = tcexpr_baseopt(ctx, te->map.result);
     break;
    case TCE_LET:
     te->let.value = tcexpr_baseopt(ctx, te->let.value);
     te->let.body = tcexpr_baseopt(ctx, te->let.body);
     break;
    case TCE_READVAR:
     break;
    case TCE_READGLOBAL:
     break;
    case TCE_CREATEPATHELEMENT:
     te->createpathelement = tcexpr_baseopt(ctx, te->createpathelement);
     break;
    case TCE_SPLATTER:
     te->splatter.value = tcexpr_baseopt(ctx, te->splatter.value);
     te->splatter.name = tcexpr_baseopt(ctx, te->splatter.name);
     break;
    case TCE_TUPLE:
     num = tcexprarray_num(&te->tuple.exprs);
     for (i=0; i<num; i++) {
	subexpr = tcexprarray_get(&te->tuple.exprs, i);
	subexpr = tcexpr_baseopt(ctx, subexpr);
	tcexprarray_set(&te->tuple.exprs, i, subexpr);
     }
     break;
    case TCE_VALUE:
     break;
   }
   te = any_baseopt(ctx, te);
   return te;
}

static struct tcstmt *tcstmt_baseopt(struct baseopt *ctx, struct tcstmt *ts) {
   switch (ts->type) {
    case TCS_EXPR:
     ts->expr = tcexpr_baseopt(ctx, ts->expr);
     break;
    case TCS_BEGIN:
    case TCS_ABORT:
    case TCS_COMMIT:
     break;
   }
   return ts;
}

////////////////////////////////////////////////////////////
// entry point

struct tcstmt *baseopt(struct pqlcontext *pql, struct tcstmt *ts) {
   struct baseopt ctx;

   baseopt_init(&ctx, pql);
   ts = tcstmt_baseopt(&ctx, ts);
   baseopt_cleanup(&ctx);

   return ts;
}
